/* Generated by Cython 3.0.0 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/home/runner/work/mata/mata/3rdparty/simlib/include/mata/simlib/util/binary_relation.hh",
            "/home/runner/work/mata/mata/include/mata/alphabet.hh",
            "/home/runner/work/mata/mata/include/mata/nfa/algorithms.hh",
            "/home/runner/work/mata/mata/include/mata/nfa/nfa.hh",
            "/home/runner/work/mata/mata/include/mata/nfa/plumbing.hh",
            "/home/runner/work/mata/mata/include/mata/utils/ord-vector.hh",
            "/home/runner/work/mata/mata/include/mata/utils/sparse-set.hh",
            "/home/runner/work/mata/mata/include/mata/utils/util.hh"
        ],
        "extra_compile_args": [
            "-std=c++20",
            "-DNO_THROW_DISPATCHER"
        ],
        "include_dirs": [
            "/home/runner/work/mata/mata/include",
            "/home/runner/work/mata/mata/3rdparty",
            "/home/runner/work/mata/mata/3rdparty/re2",
            "/home/runner/work/mata/mata/3rdparty/simlib/include",
            "/home/runner/work/mata/mata/3rdparty/cudd-min/include"
        ],
        "language": "c++",
        "libraries": [
            "mata",
            "re2",
            "cudd",
            "simlib"
        ],
        "library_dirs": [
            "/home/runner/work/mata/mata/build/src",
            "/home/runner/work/mata/mata/build/3rdparty/cudd",
            "/home/runner/work/mata/mata/build/3rdparty/simlib",
            "/home/runner/work/mata/mata/build/3rdparty/re2"
        ],
        "name": "libmata.nfa.nfa",
        "sources": [
            "libmata/nfa/nfa.pyx"
        ]
    },
    "module_name": "libmata.nfa.nfa"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#define CYTHON_ABI "3_0_0"
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x030000F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(CYTHON_LIMITED_API)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(const U& other) const { return *ptr == other; }
    template<typename U> bool operator !=(const U& other) const { return *ptr != other; }
    template<typename U> bool operator==(const __Pyx_FakeReference<U>& other) const { return *ptr == *other.ptr; }
    template<typename U> bool operator!=(const __Pyx_FakeReference<U>& other) const { return *ptr != *other.ptr; }
  private:
    T *ptr;
};

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject *co=NULL, *result=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(p))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto end;
        if (!(empty = PyTuple_New(0))) goto end;
        result = (PyCodeObject*) PyObject_Call(replace, empty, kwds);
    end:
        Py_XDECREF((PyObject*) co);
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE(obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#define __PYX_HAVE__libmata__nfa__nfa
#define __PYX_HAVE_API__libmata__nfa__nfa
/* Early includes */
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <utility>

    #if __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600)
    // move should be defined for these versions of MSVC, but __cplusplus isn't set usefully
    #include <type_traits>

    namespace cython_std {
    template <typename T> typename std::remove_reference<T>::type&& move(T& t) noexcept { return std::move(t); }
    template <typename T> typename std::remove_reference<T>::type&& move(T&& t) noexcept { return std::move(t); }
    }

    #endif
    
#include <set>
#include <unordered_set>
#include <unordered_map>
#include <vector>
#include <memory>
#include <string.h>
#include <string>
#include <list>
#include <stdint.h>
#include "mata/utils/sparse-set.hh"
#include "mata/utils/ord-vector.hh"
#include "mata/utils/util.hh"
#include "mata/simlib/util/binary_relation.hh"
#include "mata/alphabet.hh"
#include <iostream>
#include <fstream>
#include <sstream>
#include "mata/nfa/nfa.hh"
#include "mata/nfa/algorithms.hh"
#include "mata/nfa/plumbing.hh"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const wchar_t *u)
{
    const wchar_t *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#endif
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else  // Py < 3.12
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "<stringsource>",
  "libmata/nfa/nfa.pyx",
  "libmata/utils.pxd",
  "libmata/alphabets.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_7libmata_5utils_BinaryRelation;
struct __pyx_obj_7libmata_9alphabets_Alphabet;
struct __pyx_obj_7libmata_3nfa_3nfa_Nfa;
struct __pyx_obj_7libmata_3nfa_3nfa_Trans;
struct __pyx_obj_7libmata_3nfa_3nfa_Run;
struct __pyx_obj_7libmata_3nfa_3nfa_Move;
struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate;

/* "libmata/utils.pxd":103
 * 
 * # Forward declarations
 * cdef class BinaryRelation:             # <<<<<<<<<<<<<<
 *     """Wrapper for binary relation."""
 *     cdef CBinaryRelation *thisptr
 */
struct __pyx_obj_7libmata_5utils_BinaryRelation {
  PyObject_HEAD
  Simlib::Util::BinaryRelation *thisptr;
};


/* "libmata/alphabets.pxd":33
 * 
 * 
 * cdef class Alphabet:             # <<<<<<<<<<<<<<
 *     cdef CAlphabet* as_base(self)
 *     cdef get_symbols(self)
 */
struct __pyx_obj_7libmata_9alphabets_Alphabet {
  PyObject_HEAD
  struct __pyx_vtabstruct_7libmata_9alphabets_Alphabet *__pyx_vtab;
};


/* "libmata/nfa/nfa.pxd":212
 * #
 * # This is needed in order for these classes to be used in other packages.
 * cdef class Nfa:             # <<<<<<<<<<<<<<
 *     # TODO: Shared pointers are not ideal as they bring some overhead which could be substantial in theory. We are not
 *     #  sure whether the shared pointers will be a problem in this case, but it would be good to pay attention to this and
 */
struct __pyx_obj_7libmata_3nfa_3nfa_Nfa {
  PyObject_HEAD
  std::shared_ptr<Mata::Nfa::Nfa>  thisptr;
  PyObject *label;
};


/* "libmata/nfa/nfa.pxd":220
 *     cdef label
 * 
 * cdef class Trans:             # <<<<<<<<<<<<<<
 *     cdef CTrans* thisptr
 *     cdef copy_from(self, CTrans trans)
 */
struct __pyx_obj_7libmata_3nfa_3nfa_Trans {
  PyObject_HEAD
  struct __pyx_vtabstruct_7libmata_3nfa_3nfa_Trans *__pyx_vtab;
  Mata::Nfa::Trans *thisptr;
};


/* "libmata/nfa/nfa.pyx":33
 *     return EPSILON
 * 
 * cdef class Run:             # <<<<<<<<<<<<<<
 *     """Wrapper over the run in NFA."""
 *     cdef mata_nfa.CRun *thisptr
 */
struct __pyx_obj_7libmata_3nfa_3nfa_Run {
  PyObject_HEAD
  Mata::Nfa::Run *thisptr;
};


/* "libmata/nfa/nfa.pyx":127
 * 
 * 
 * cdef class Move:             # <<<<<<<<<<<<<<
 *     """Wrapper over pair of symbol and states for transitions"""
 *     cdef mata_nfa.CMove *thisptr
 */
struct __pyx_obj_7libmata_3nfa_3nfa_Move {
  PyObject_HEAD
  Mata::Nfa::Move *thisptr;
};


/* "libmata/nfa/nfa.pyx":396
 *         return self.thisptr.get().size()
 * 
 *     def iterate(self):             # <<<<<<<<<<<<<<
 *         """Iterates over all transitions
 * 
 */
struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate {
  PyObject_HEAD
  Mata::Nfa::Nfa::const_iterator __pyx_v_iterator;
  Mata::Nfa::Trans __pyx_v_lhs;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self;
  struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_trans;
};



/* "libmata/alphabets.pxd":33
 * 
 * 
 * cdef class Alphabet:             # <<<<<<<<<<<<<<
 *     cdef CAlphabet* as_base(self)
 *     cdef get_symbols(self)
 */

struct __pyx_vtabstruct_7libmata_9alphabets_Alphabet {
  Mata::Alphabet *(*as_base)(struct __pyx_obj_7libmata_9alphabets_Alphabet *);
  PyObject *(*get_symbols)(struct __pyx_obj_7libmata_9alphabets_Alphabet *);
};
static struct __pyx_vtabstruct_7libmata_9alphabets_Alphabet *__pyx_vtabptr_7libmata_9alphabets_Alphabet;


/* "libmata/nfa/nfa.pyx":67
 *         self.thisptr.path = value
 * 
 * cdef class Trans:             # <<<<<<<<<<<<<<
 *     """Wrapper over the transitions in NFA."""
 * 
 */

struct __pyx_vtabstruct_7libmata_3nfa_3nfa_Trans {
  PyObject *(*copy_from)(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *, Mata::Nfa::Trans);
};
static struct __pyx_vtabstruct_7libmata_3nfa_3nfa_Trans *__pyx_vtabptr_7libmata_3nfa_3nfa_Trans;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* MoveIfSupported.proto */
#if __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600)
  #include <utility>
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) std::move(x)
#else
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) x
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyObject_Str.proto */
#define __Pyx_PyObject_Str(obj)\
    (likely(PyString_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* UnicodeConcatInPlace.proto */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
    #if CYTHON_REFNANNY
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, __pyx_refnanny)
    #else
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right)
    #endif
    static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    );
#else
#define __Pyx_PyUnicode_ConcatInPlace __Pyx_PyUnicode_Concat
#endif
#define __Pyx_PyUnicode_ConcatInPlaceSafe(left, right) ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_ConcatInPlace(left, right))

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* AssertionsEnabled.proto */
#define __Pyx_init_assertions_enabled()
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define __pyx_assertions_enabled() (1)
#elif PY_VERSION_HEX < 0x03080000  ||  CYTHON_COMPILING_IN_PYPY  ||  defined(Py_LIMITED_API)
  #define __pyx_assertions_enabled() (!Py_OptimizeFlag)
#elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030900A6
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  #undef __Pyx_init_assertions_enabled
  static void __Pyx_init_assertions_enabled(void) {
    __pyx_assertions_enabled_flag = ! _PyInterpreterState_GetConfig(__Pyx_PyThreadState_Current->interp)->optimization_level;
  }
#else
  #define __pyx_assertions_enabled() (!Py_OptimizeFlag)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE int __Pyx_PyInt_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* MergeKeywords.proto */
static int __Pyx_MergeKeywords(PyObject *kwdict, PyObject *source_mapping);

/* PyIntCompare.proto */
static CYTHON_INLINE int __Pyx_PyInt_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_0
#define __PYX_HAVE_RT_ImportType_proto_3_0_0
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_0(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_0(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_0 {
   __Pyx_ImportType_CheckSize_Error_3_0_0 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_0 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_0 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_0(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_0 check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_IsCyOrPyCFunction(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* ClassMethod.proto */
#include "descrobject.h"
CYTHON_UNUSED static PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* None.proto */
#include <new>

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static PyObject *__pyx_f_7libmata_3nfa_3nfa_5Trans_copy_from(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self, Mata::Nfa::Trans __pyx_v_trans); /* proto*/

/* Module declarations from "libcpp" */

/* Module declarations from "libcpp.utility" */

/* Module declarations from "libcpp.set" */

/* Module declarations from "libcpp.unordered_set" */

/* Module declarations from "libcpp.unordered_map" */

/* Module declarations from "libcpp.vector" */

/* Module declarations from "libcpp.memory" */

/* Module declarations from "libc.string" */

/* Module declarations from "libcpp.string" */

/* Module declarations from "libcpp.list" */

/* Module declarations from "libcpp.pair" */

/* Module declarations from "libc.stdint" */

/* Module declarations from "libmata.utils" */

/* Module declarations from "libmata.alphabets" */

/* Module declarations from "libmata.nfa.nfa" */
static Mata::Symbol __pyx_v_7libmata_3nfa_3nfa_EPSILON;
static PyObject *__pyx_f_7libmata_3nfa_3nfa_subset_map_to_dictionary(std::unordered_map<Mata::Nfa::StateSet,Mata::Nfa::State> ); /*proto*/
static std::vector<Mata::Nfa::State>  __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(PyObject *); /*proto*/
static std::vector<Mata::Symbol>  __pyx_convert_vector_from_py_Mata_3a__3a_Symbol(PyObject *); /*proto*/
static PyObject *__pyx_convert_vector_to_py_Mata_3a__3a_Symbol(std::vector<Mata::Symbol>  const &); /*proto*/
static PyObject *__pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State(std::vector<Mata::Nfa::State>  const &); /*proto*/
static PyObject *__pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State(std::pair<Mata::Nfa::State,Mata::Nfa::State>  const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyObject_string_to_py_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyUnicode_string_to_py_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyStr_string_to_py_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyBytes_string_to_py_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyByteArray_string_to_py_std__in_string(std::string const &); /*proto*/
static std::string __pyx_convert_string_from_py_std__in_string(PyObject *); /*proto*/
static PyObject *__pyx_convert_pair_to_py_std_3a__3a_pair_3c_Mata_3a__3a_Nfa_3a__3a_State_2c_Mata_3a__3a_Nfa_3a__3a_State_3e_______Mata_3a__3a_Nfa_3a__3a_State(std::pair<std::pair<Mata::Nfa::State,Mata::Nfa::State> ,Mata::Nfa::State>  const &); /*proto*/
static std::unordered_map<std::string,std::string>  __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(PyObject *); /*proto*/
static std::unordered_map<std::string,Mata::Symbol>  __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_Mata_3a__3a_Symbol(PyObject *); /*proto*/
static std::vector<std::string>  __pyx_convert_vector_from_py_std_3a__3a_string(PyObject *); /*proto*/
static std::vector<uint8_t>  __pyx_convert_vector_from_py_uint8_t(PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "libmata.nfa.nfa"
extern int __pyx_module_is_main_libmata__nfa__nfa;
int __pyx_module_is_main_libmata__nfa__nfa = 0;

/* Implementation of "libmata.nfa.nfa" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_print;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_MemoryError;
/* #### Code section: string_decls ### */
static const char __pyx_k_[] = "-[";
static const char __pyx_k_G[] = "G";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_12[] = "12";
static const char __pyx_k__2[] = "]\342\206\222";
static const char __pyx_k__3[] = "{";
  static const char __pyx_k__4[] = ",";
static const char __pyx_k__5[] = "}";
static const char __pyx_k__6[] = "[";
static const char __pyx_k__8[] = "\n";
static const char __pyx_k__9[] = " ";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_it[] = "it";
static const char __pyx_k_nx[] = "nx";
static const char __pyx_k_st[] = "st";
static const char __pyx_k_tr[] = "tr";
static const char __pyx_k_0_3[] = "0.3";
static const char __pyx_k_1_5[] = "1.5";
static const char __pyx_k_Nfa[] = "Nfa";
static const char __pyx_k_Run[] = "Run";
static const char __pyx_k__10[] = "";
static const char __pyx_k__15[] = " | ";
static const char __pyx_k__16[] = "*";
static const char __pyx_k__80[] = "_";
static const char __pyx_k_aut[] = "aut";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_cmd[] = "cmd";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_err[] = "err";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_int[] = "int";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_lhs[] = "lhs";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_neq[] = "__neq__";
static const char __pyx_k_pdf[] = "pdf";
static const char __pyx_k_red[] = "red";
static const char __pyx_k_rhs[] = "rhs";
static const char __pyx_k_run[] = "run";
static const char __pyx_k_set[] = "set";
static const char __pyx_k_src[] = "src";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_tgt[] = "tgt";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_Move[] = "Move";
static const char __pyx_k_PIPE[] = "PIPE";
static const char __pyx_k__165[] = "?";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_bool[] = "bool";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_symb[] = "symb";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_trim[] = "trim";
static const char __pyx_k_true[] = "true";
static const char __pyx_k_wait[] = "wait";
static const char __pyx_k_word[] = "word";
static const char __pyx_k_Popen[] = "Popen";
static const char __pyx_k_Trans[] = "Trans";
static const char __pyx_k_clear[] = "clear";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_color[] = "color";
static const char __pyx_k_false[] = "false";
static const char __pyx_k_input[] = "input";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_label[] = "label";
static const char __pyx_k_other[] = "other";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_quiet[] = "quiet";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shell[] = "shell";
static const char __pyx_k_shlex[] = "shlex";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_stdin[] = "stdin";
static const char __pyx_k_store[] = "store";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_style[] = "style";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_trans[] = "trans";
static const char __pyx_k_union[] = "union";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_width[] = "width";
static const char __pyx_k_STDOUT[] = "STDOUT";
static const char __pyx_k_cmd_no[] = "cmd_no";
static const char __pyx_k_cmderr[] = "cmderr";
static const char __pyx_k_cmdout[] = "cmdout";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_filled[] = "filled";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_height[] = "height";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_output[] = "output";
static const char __pyx_k_pandas[] = "pandas";
static const char __pyx_k_params[] = "params";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_resize[] = "resize";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_revert[] = "revert";
static const char __pyx_k_source[] = "source";
static const char __pyx_k_states[] = "states";
static const char __pyx_k_stderr[] = "stderr";
static const char __pyx_k_stdout[] = "stdout";
static const char __pyx_k_symbol[] = "symbol";
static const char __pyx_k_target[] = "target";
static const char __pyx_k_DiGraph[] = "DiGraph";
static const char __pyx_k_aut_dot[] = "aut.dot";
static const char __pyx_k_columns[] = "columns";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_dot_O_T[] = "dot -O -T";
static const char __pyx_k_epsilon[] = "epsilon";
static const char __pyx_k_forward[] = "forward";
static const char __pyx_k_iterate[] = "iterate";
static const char __pyx_k_lhs_map[] = "lhs_map";
static const char __pyx_k_objects[] = "objects";
static const char __pyx_k_post_of[] = "post_of";
static const char __pyx_k_process[] = "process";
static const char __pyx_k_rhs_map[] = "rhs_map";
static const char __pyx_k_store_2[] = "_store";
static const char __pyx_k_symbols[] = "symbols";
static const char __pyx_k_targets[] = "targets";
static const char __pyx_k_timeout[] = "timeout";
static const char __pyx_k_Nfa_size[] = "Nfa.size";
static const char __pyx_k_Nfa_trim[] = "Nfa.trim";
static const char __pyx_k_add_edge[] = "add_edge";
static const char __pyx_k_alphabet[] = "alphabet";
static const char __pyx_k_bool_vec[] = "bool_vec";
static const char __pyx_k_c_states[] = "c_states";
static const char __pyx_k_darkblue[] = "darkblue";
static const char __pyx_k_encoding[] = "encoding";
static const char __pyx_k_fontname[] = "fontname";
static const char __pyx_k_fontsize[] = "fontsize";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_is_state[] = "is_state";
static const char __pyx_k_iterator[] = "iterator";
static const char __pyx_k_list_int[] = "list[int]";
static const char __pyx_k_minimize[] = "minimize";
static const char __pyx_k_networkx[] = "networkx";
static const char __pyx_k_nx_Graph[] = "nx.Graph";
static const char __pyx_k_penwidth[] = "penwidth";
static const char __pyx_k_reduce_2[] = "reduce";
static const char __pyx_k_relation[] = "relation";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_state_to[] = "state_to";
static const char __pyx_k_DataFrame[] = "DataFrame";
static const char __pyx_k_Move_None[] = "Move | None";
static const char __pyx_k_Nfa_clear[] = "Nfa.clear";
static const char __pyx_k_Nfa_union[] = "Nfa.union";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_add_state[] = "add_state";
static const char __pyx_k_algorithm[] = "algorithm";
static const char __pyx_k_c_lhs_map[] = "c_lhs_map";
static const char __pyx_k_c_rhs_map[] = "c_rhs_map";
static const char __pyx_k_classical[] = "classical";
static const char __pyx_k_direction[] = "direction";
static const char __pyx_k_fillcolor[] = "fillcolor";
static const char __pyx_k_fixedsize[] = "fixedsize";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_iteritems[] = "iteritems";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_state_map[] = "state_map";
static const char __pyx_k_terminate[] = "terminate";
static const char __pyx_k_Move___neq[] = "Move.__neq__";
static const char __pyx_k_Nfa_reduce[] = "Nfa.reduce";
static const char __pyx_k_Nfa_revert[] = "Nfa.revert";
static const char __pyx_k_antichains[] = "antichains";
static const char __pyx_k_c_alphabet[] = "c_alphabet";
static const char __pyx_k_c_bool_vec[] = "c_bool_vec";
static const char __pyx_k_complement[] = "complement";
static const char __pyx_k_edge_style[] = "edge_style";
static const char __pyx_k_is_epsilon[] = "is_epsilon";
static const char __pyx_k_is_in_lang[] = "is_in_lang";
static const char __pyx_k_node_style[] = "node_style";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_returncode[] = "returncode";
static const char __pyx_k_simulation[] = "simulation";
static const char __pyx_k_sink_state[] = "sink_state";
static const char __pyx_k_state_from[] = "state_from";
static const char __pyx_k_subprocess[] = "subprocess";
static const char __pyx_k_subset_map[] = "subset_map";
static const char __pyx_k_symbol_map[] = "symbol_map";
static const char __pyx_k_to_dot_str[] = "to_dot_str";
static const char __pyx_k_trim_input[] = "trim_input";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_Nfa_iterate[] = "Nfa.iterate";
static const char __pyx_k_Nfa_post_of[] = "Nfa.post_of";
static const char __pyx_k_Trans___neq[] = "Trans.__neq__";
static const char __pyx_k_clear_final[] = "clear_final";
static const char __pyx_k_communicate[] = "communicate";
static const char __pyx_k_concatenate[] = "concatenate";
static const char __pyx_k_determinize[] = "determinize";
static const char __pyx_k_encode_word[] = "encode_word";
static const char __pyx_k_get_symbols[] = "get_symbols";
static const char __pyx_k_is_complete[] = "is_complete";
static const char __pyx_k_is_included[] = "is_included";
static const char __pyx_k_output_file[] = "output_file";
static const char __pyx_k_post_map_of[] = "post_map_of";
static const char __pyx_k_to_dot_file[] = "to_dot_file";
static const char __pyx_k_transitions[] = "transitions:\n";
static const char __pyx_k_unify_final[] = "unify_final";
static const char __pyx_k_use_epsilon[] = "use_epsilon";
static const char __pyx_k_Courier_Bold[] = "Courier-Bold";
static const char __pyx_k_Nfa_is_state[] = "Nfa.is_state";
static const char __pyx_k_Nfa_minimize[] = "Nfa.minimize";
static const char __pyx_k_c_transition[] = "c_transition";
static const char __pyx_k_divisible_by[] = "divisible_by";
static const char __pyx_k_final_states[] = "final_states: {}\n";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_input_states[] = "input_states";
static const char __pyx_k_intersection[] = "intersection";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_is_universal[] = "is_universal";
static const char __pyx_k_midnightblue[] = "midnightblue";
static const char __pyx_k_remove_trans[] = "remove_trans";
static const char __pyx_k_return_value[] = "return_value";
static const char __pyx_k_state_number[] = "state_number";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_to_dataframe[] = "to_dataframe";
static const char __pyx_k_transsymbols[] = "transsymbols";
static const char __pyx_k_Nfa_add_state[] = "Nfa.add_state";
static const char __pyx_k_add_new_state[] = "add_new_state";
static const char __pyx_k_c_product_map[] = "c_product_map";
static const char __pyx_k_c_transitions[] = "c_transitions";
static const char __pyx_k_check_results[] = "check_results";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_clear_initial[] = "clear_initial";
static const char __pyx_k_is_lang_empty[] = "is_lang_empty";
static const char __pyx_k_make_complete[] = "make_complete";
static const char __pyx_k_output_format[] = "output_format";
static const char __pyx_k_output_stream[] = "output_stream";
static const char __pyx_k_piped_command[] = "piped_command";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_relation_size[] = "relation_size";
static const char __pyx_k_transitions_2[] = "transitions";
static const char __pyx_k_unify_initial[] = "unify_initial";
static const char __pyx_k_useful_states[] = "useful_states";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_Nfa_complement[] = "Nfa.complement";
static const char __pyx_k_Nfa_is_epsilon[] = "Nfa.is_epsilon";
static const char __pyx_k_Nfa_is_in_lang[] = "Nfa.is_in_lang";
static const char __pyx_k_Nfa_to_dot_str[] = "Nfa.to_dot_str";
static const char __pyx_k_TimeoutExpired[] = "TimeoutExpired";
static const char __pyx_k_add_transition[] = "add_transition";
static const char __pyx_k_has_transition[] = "has_transition";
static const char __pyx_k_initial_states[] = "initial_states: {}\n";
static const char __pyx_k_lightsteelblue[] = "lightsteelblue";
static const char __pyx_k_one_letter_aut[] = "one_letter_aut";
static const char __pyx_k_remove_epsilon[] = "remove_epsilon";
static const char __pyx_k_Nfa_clear_final[] = "Nfa.clear_final";
static const char __pyx_k_Nfa_concatenate[] = "Nfa.concatenate";
static const char __pyx_k_Nfa_determinize[] = "Nfa.determinize";
static const char __pyx_k_Nfa_encode_word[] = "Nfa.encode_word";
static const char __pyx_k_Nfa_get_symbols[] = "Nfa.get_symbols";
static const char __pyx_k_Nfa_is_complete[] = "Nfa.is_complete";
static const char __pyx_k_Nfa_is_included[] = "Nfa.is_included";
static const char __pyx_k_Nfa_post_map_of[] = "Nfa.post_map_of";
static const char __pyx_k_Nfa_to_dot_file[] = "Nfa.to_dot_file";
static const char __pyx_k_Nfa_unify_final[] = "Nfa.unify_final";
static const char __pyx_k_accepts_epsilon[] = "accepts_epsilon";
static const char __pyx_k_captured_stderr[] = "captured stderr: ";
static const char __pyx_k_captured_stdout[] = "captured stdout: ";
static const char __pyx_k_has_final_state[] = "has_final_state";
static const char __pyx_k_libmata_nfa_nfa[] = "libmata.nfa.nfa";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_Nfa_intersection[] = "Nfa.intersection";
static const char __pyx_k_Nfa_is_universal[] = "Nfa.is_universal";
static const char __pyx_k_Nfa_remove_trans[] = "Nfa.remove_trans";
static const char __pyx_k_Nfa_to_dataframe[] = "Nfa.to_dataframe";
static const char __pyx_k_compute_relation[] = "compute_relation";
static const char __pyx_k_executed_command[] = "executed_command";
static const char __pyx_k_get_num_of_trans[] = "get_num_of_trans";
static const char __pyx_k_graphviz_command[] = "graphviz_command";
static const char __pyx_k_is_deterministic[] = "is_deterministic";
static const char __pyx_k_make_final_state[] = "make_final_state";
static const char __pyx_k_pandas_DataFrame[] = "pandas.DataFrame";
static const char __pyx_k_preserve_epsilon[] = "preserve_epsilon";
static const char __pyx_k_remove_trans_raw[] = "remove_trans_raw";
static const char __pyx_k_string_to_symbol[] = "string_to_symbol";
static const char __pyx_k_transitions_list[] = "transitions_list";
static const char __pyx_k_translate_symbol[] = "translate_symbol";
static const char __pyx_k_unpiped_commands[] = "unpiped_commands";
static const char __pyx_k_Nfa_add_new_state[] = "Nfa.add_new_state";
static const char __pyx_k_Nfa_clear_initial[] = "Nfa.clear_initial";
static const char __pyx_k_Nfa_is_lang_empty[] = "Nfa.is_lang_empty";
static const char __pyx_k_Nfa_make_complete[] = "Nfa.make_complete";
static const char __pyx_k_Nfa_unify_initial[] = "Nfa.unify_initial";
static const char __pyx_k_equivalence_check[] = "equivalence_check";
static const char __pyx_k_get_useful_states[] = "get_useful_states";
static const char __pyx_k_get_word_for_path[] = "get_word_for_path";
static const char __pyx_k_has_initial_state[] = "has_initial_state";
static const char __pyx_k_is_prefix_in_lang[] = "is_prefix_in_lang";
static const char __pyx_k_make_final_states[] = "make_final_states";
static const char __pyx_k_to_networkx_graph[] = "to_networkx_graph";
static const char __pyx_k_CalledProcessError[] = "CalledProcessError";
static const char __pyx_k_Nfa_add_transition[] = "Nfa.add_transition";
static const char __pyx_k_Nfa_has_transition[] = "Nfa.has_transition";
static const char __pyx_k_Nfa_remove_epsilon[] = "Nfa.remove_epsilon";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_get_one_letter_aut[] = "get_one_letter_aut";
static const char __pyx_k_make_initial_state[] = "make_initial_state";
static const char __pyx_k_remove_final_state[] = "remove_final_state";
static const char __pyx_k_Nfa___reduce_cython[] = "Nfa.__reduce_cython__";
static const char __pyx_k_Nfa_accepts_epsilon[] = "Nfa.accepts_epsilon";
static const char __pyx_k_Nfa_has_final_state[] = "Nfa.has_final_state";
static const char __pyx_k_Run___reduce_cython[] = "Run.__reduce_cython__";
static const char __pyx_k_epsilon_transitions[] = "epsilon_transitions";
static const char __pyx_k_libmata_nfa_nfa_pyx[] = "libmata/nfa/nfa.pyx";
static const char __pyx_k_make_initial_states[] = "make_initial_states";
static const char __pyx_k_trim_with_state_map[] = "trim_with_state_map";
static const char __pyx_k_Move___reduce_cython[] = "Move.__reduce_cython__";
static const char __pyx_k_Nfa_compute_relation[] = "Nfa.compute_relation";
static const char __pyx_k_Nfa_get_num_of_trans[] = "Nfa.get_num_of_trans";
static const char __pyx_k_Nfa_is_deterministic[] = "Nfa.is_deterministic";
static const char __pyx_k_Nfa_make_final_state[] = "Nfa.make_final_state";
static const char __pyx_k_Nfa_remove_trans_raw[] = "Nfa.remove_trans_raw";
static const char __pyx_k_error_while_dot_file[] = "error while dot file: ";
static const char __pyx_k_get_alphabet_symbols[] = "get_alphabet_symbols";
static const char __pyx_k_get_reachable_states[] = "get_reachable_states";
static const char __pyx_k_is_included_with_cex[] = "is_included_with_cex";
static const char __pyx_k_remove_initial_state[] = "remove_initial_state";
static const char __pyx_k_Nfa___setstate_cython[] = "Nfa.__setstate_cython__";
static const char __pyx_k_Nfa_equivalence_check[] = "Nfa.equivalence_check";
static const char __pyx_k_Nfa_get_useful_states[] = "Nfa.get_useful_states";
static const char __pyx_k_Nfa_get_word_for_path[] = "Nfa.get_word_for_path";
static const char __pyx_k_Nfa_has_initial_state[] = "Nfa.has_initial_state";
static const char __pyx_k_Nfa_is_prefix_in_lang[] = "Nfa.is_prefix_in_lang";
static const char __pyx_k_Nfa_make_final_states[] = "Nfa.make_final_states";
static const char __pyx_k_Nfa_to_networkx_graph[] = "Nfa.to_networkx_graph";
static const char __pyx_k_Run___setstate_cython[] = "Run.__setstate_cython__";
static const char __pyx_k_Trans___reduce_cython[] = "Trans.__reduce_cython__";
static const char __pyx_k_add_transition_object[] = "add_transition_object";
static const char __pyx_k_get_trans_as_sequence[] = "get_trans_as_sequence";
static const char __pyx_k_reduce_with_state_map[] = "reduce_with_state_map";
static const char __pyx_k_Move___setstate_cython[] = "Move.__setstate_cython__";
static const char __pyx_k_Nfa_get_one_letter_aut[] = "Nfa.get_one_letter_aut";
static const char __pyx_k_Nfa_make_initial_state[] = "Nfa.make_initial_state";
static const char __pyx_k_Nfa_remove_final_state[] = "Nfa.remove_final_state";
static const char __pyx_k_get_terminating_states[] = "get_terminating_states";
static const char __pyx_k_remove_epsilon_inplace[] = "remove_epsilon_inplace";
static const char __pyx_k_useful_states_bool_vec[] = "useful_states_bool_vec";
static const char __pyx_k_Cannot_translate_symbol[] = "Cannot translate symbol '";
static const char __pyx_k_Nfa_make_initial_states[] = "Nfa.make_initial_states";
static const char __pyx_k_Nfa_trim_with_state_map[] = "Nfa.trim_with_state_map";
static const char __pyx_k_Trans___setstate_cython[] = "Trans.__setstate_cython__";
static const char __pyx_k_get_epsilon_transitions[] = "get_epsilon_transitions";
static const char __pyx_k_Nfa_get_reachable_states[] = "Nfa.get_reachable_states";
static const char __pyx_k_Nfa_is_included_with_cex[] = "Nfa.is_included_with_cex";
static const char __pyx_k_Nfa_remove_initial_state[] = "Nfa.remove_initial_state";
static const char __pyx_k_get_transitions_to_state[] = "get_transitions_to_state";
static const char __pyx_k_Nfa_add_transition_object[] = "Nfa.add_transition_object";
static const char __pyx_k_Nfa_encode_word_line_1139[] = "Nfa.encode_word (line 1139)";
static const char __pyx_k_Nfa_get_trans_as_sequence[] = "Nfa.get_trans_as_sequence";
static const char __pyx_k_Nfa_reduce_with_state_map[] = "Nfa.reduce_with_state_map";
static const char __pyx_k_Nfa_get_terminating_states[] = "Nfa.get_terminating_states";
static const char __pyx_k_Nfa_remove_epsilon_inplace[] = "Nfa.remove_epsilon_inplace";
static const char __pyx_k_c_epsilon_transitions_iter[] = "c_epsilon_transitions_iter";
static const char __pyx_k_get_elements_from_bool_vec[] = "get_elements_from_bool_vec";
static const char __pyx_k_get_transitions_from_state[] = "get_transitions_from_state";
static const char __pyx_k_without_specified_alphabet[] = "' without specified alphabet";
static const char __pyx_k_Nfa_get_epsilon_transitions[] = "Nfa.get_epsilon_transitions";
static const char __pyx_k_determinize_with_subset_map[] = "determinize_with_subset_map";
static const char __pyx_k_run_safely_external_command[] = "run_safely_external_command";
static const char __pyx_k_Nfa_get_transitions_to_state[] = "Nfa.get_transitions_to_state";
static const char __pyx_k_intersection_with_product_map[] = "intersection_with_product_map";
static const char __pyx_k_Nfa_get_transitions_from_state[] = "Nfa.get_transitions_from_state";
static const char __pyx_k_Nfa_determinize_with_subset_map[] = "Nfa.determinize_with_subset_map";
static const char __pyx_k_Nfa_get_word_for_path_line_1122[] = "Nfa.get_word_for_path (line 1122)";
static const char __pyx_k_tuple_Nfa_dict_str_str_dict_str[] = "tuple[Nfa, dict[str, str], dict[str, str]]";
static const char __pyx_k_Encodes_word_based_on_a_string_t[] = "Encodes word based on a string to symbol map\n\n        >>> mata_nfa.Nfa.encode_word({'a': 1, 'b': 2, \"c\": 0}, \"abca\")\n        [1, 2, 0, 1]\n\n        :param dict string_to_symbol: dictionary of strings to integers\n        :param word: list of strings representing a encoded word\n        :return:\n        ";
static const char __pyx_k_For_a_given_path_set_of_states_r[] = "For a given path (set of states) returns a corresponding word\n\n        >>> mata_nfa.Nfa.get_word_for_path(lhs, [0, 1, 2])\n        ([1, 1], True)\n\n        :param Nfa lhs: source automaton\n        :param list path: list of states\n        :return: pair of word (list of symbols) and true or false, whether the search was successful\n        ";
static const char __pyx_k_Nfa_concatenate_with_result_stat[] = "Nfa.concatenate_with_result_state_maps";
static const char __pyx_k_Nfa_get_trans_from_state_as_sequ[] = "Nfa.get_trans_from_state_as_sequence";
static const char __pyx_k_Nfa_intersection_with_product_ma[] = "Nfa.intersection_with_product_map";
static const char __pyx_k_concatenate_with_result_state_ma[] = "concatenate_with_result_state_maps";
static const char __pyx_k_get_trans_from_state_as_sequence[] = "get_trans_from_state_as_sequence";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_epsilon(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static int __pyx_pf_7libmata_3nfa_3nfa_3Run___cinit__(struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self); /* proto */
static void __pyx_pf_7libmata_3nfa_3nfa_3Run_2__dealloc__(struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Run_4word___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self); /* proto */
static int __pyx_pf_7libmata_3nfa_3nfa_3Run_4word_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Run_4path___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self); /* proto */
static int __pyx_pf_7libmata_3nfa_3nfa_3Run_4path_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Run_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Run_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_3src___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_4symb___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_3tgt___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self); /* proto */
static int __pyx_pf_7libmata_3nfa_3nfa_5Trans___cinit__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self, Mata::Nfa::State __pyx_v_src, Mata::Symbol __pyx_v_s, Mata::Nfa::State __pyx_v_tgt); /* proto */
static void __pyx_pf_7libmata_3nfa_3nfa_5Trans_2__dealloc__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_4__eq__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_6__neq__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_8__str__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_10__repr__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_6symbol___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self); /* proto */
static int __pyx_pf_7libmata_3nfa_3nfa_4Move_6symbol_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_7targets___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self); /* proto */
static int __pyx_pf_7libmata_3nfa_3nfa_4Move_7targets_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_7libmata_3nfa_3nfa_4Move___cinit__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, Mata::Symbol __pyx_v_symbol, std::vector<Mata::Nfa::State>  __pyx_v_states); /* proto */
static void __pyx_pf_7libmata_3nfa_3nfa_4Move_2__dealloc__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_4__lt__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_6__gt__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_8__le__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_10__ge__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_12__eq__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_14__neq__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_16__str__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_18__repr__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7libmata_3nfa_3nfa_3Nfa___cinit__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, PyObject *__pyx_v_state_number, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet, PyObject *__pyx_v_label); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_5label___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static int __pyx_pf_7libmata_3nfa_3nfa_3Nfa_5label_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_14initial_states___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static int __pyx_pf_7libmata_3nfa_3nfa_3Nfa_14initial_states_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, std::vector<Mata::Nfa::State>  __pyx_v_value); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_12final_states___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static int __pyx_pf_7libmata_3nfa_3nfa_3Nfa_12final_states_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, std::vector<Mata::Nfa::State>  __pyx_v_value); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_2is_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_4add_new_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_6add_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_8make_initial_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_10make_initial_states(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, std::vector<Mata::Nfa::State>  __pyx_v_states); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_12has_initial_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_st); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_14remove_initial_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_16clear_initial(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_18make_final_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_20make_final_states(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, std::vector<Mata::Nfa::State>  __pyx_v_states); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_22has_final_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_st); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_24remove_final_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_26clear_final(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_28unify_initial(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_30unify_final(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_32add_transition_object(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_tr); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_34add_transition(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_src, PyObject *__pyx_v_symb, Mata::Nfa::State __pyx_v_tgt, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_36remove_trans(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_tr); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_38remove_trans_raw(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_src, Mata::Symbol __pyx_v_symb, Mata::Nfa::State __pyx_v_tgt); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_40has_transition(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_src, Mata::Symbol __pyx_v_symb, Mata::Nfa::State __pyx_v_tgt); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_42get_num_of_trans(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_44clear(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_46size(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_48iterate(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_51get_transitions_from_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_53get_transitions_to_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state_to); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_55get_trans_as_sequence(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_57get_trans_from_state_as_sequence(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state_from); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_59get_useful_states(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_61get_reachable_states(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_63get_terminating_states(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_65trim(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_67trim_with_state_map(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_69get_one_letter_aut(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_71is_epsilon(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Symbol __pyx_v_symbol); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_73__str__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_75__repr__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_77to_dot_file(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, PyObject *__pyx_v_output_file, PyObject *__pyx_v_output_format); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_79to_dot_str(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, PyObject *__pyx_v_encoding); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_81to_dataframe(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_83to_networkx_graph(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_85post_map_of(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_st, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_87post_of(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, std::vector<Mata::Nfa::State>  __pyx_v_states, Mata::Symbol __pyx_v_symbol); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_89determinize_with_subset_map(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_91determinize(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_93union(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_95intersection(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, bool __pyx_v_preserve_epsilon); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_97intersection_with_product_map(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, bool __pyx_v_preserve_epsilon); /* proto */
static struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_99concatenate(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, bool __pyx_v_use_epsilon); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_101concatenate_with_result_state_maps(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, bool __pyx_v_use_epsilon); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_103complement(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet, PyObject *__pyx_v_params); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_105make_complete(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, Mata::Nfa::State __pyx_v_sink_state, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_107revert(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_109remove_epsilon(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, Mata::Symbol __pyx_v_epsilon); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_111remove_epsilon_inplace(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Symbol __pyx_v_epsilon); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_113get_epsilon_transitions(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state, Mata::Symbol __pyx_v_epsilon); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_115minimize(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_117reduce_with_state_map(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_aut, bool __pyx_v_trim_input, PyObject *__pyx_v_params); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_119reduce(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_aut, bool __pyx_v_trim_input, PyObject *__pyx_v_params); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_121compute_relation(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, PyObject *__pyx_v_params); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_123is_deterministic(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_125is_lang_empty(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_run); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_127is_universal(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet, PyObject *__pyx_v_params); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_129is_included_with_cex(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet, PyObject *__pyx_v_params); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_131is_included(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet, PyObject *__pyx_v_params); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_133equivalence_check(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet, PyObject *__pyx_v_params); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_135get_symbols(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_137is_complete(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_139is_in_lang(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, std::vector<Mata::Symbol>  __pyx_v_word); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_141is_prefix_in_lang(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, std::vector<Mata::Symbol>  __pyx_v_word); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_143accepts_epsilon(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_145get_word_for_path(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_147encode_word(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_string_to_symbol, PyObject *__pyx_v_word); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_149__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_151__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_2divisible_by(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_k); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4run_safely_external_command(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cmd, PyObject *__pyx_v_check_results, PyObject *__pyx_v_quiet, PyObject *__pyx_v_timeout, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_6get_elements_from_bool_vec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bool_vec); /* proto */
static PyObject *__pyx_pf_7libmata_3nfa_3nfa_8store(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_tp_new_7libmata_3nfa_3nfa_Nfa(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7libmata_3nfa_3nfa_Trans(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7libmata_3nfa_3nfa_Run(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7libmata_3nfa_3nfa_Move(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7libmata_3nfa_3nfa___pyx_scope_struct__iterate(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7libmata_5utils_BinaryRelation;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7libmata_9alphabets_Alphabet;
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_7libmata_3nfa_3nfa_Nfa;
  PyObject *__pyx_type_7libmata_3nfa_3nfa_Trans;
  PyObject *__pyx_type_7libmata_3nfa_3nfa_Run;
  PyObject *__pyx_type_7libmata_3nfa_3nfa_Move;
  PyObject *__pyx_type_7libmata_3nfa_3nfa___pyx_scope_struct__iterate;
  #endif
  PyTypeObject *__pyx_ptype_7libmata_3nfa_3nfa_Nfa;
  PyTypeObject *__pyx_ptype_7libmata_3nfa_3nfa_Trans;
  PyTypeObject *__pyx_ptype_7libmata_3nfa_3nfa_Run;
  PyTypeObject *__pyx_ptype_7libmata_3nfa_3nfa_Move;
  PyTypeObject *__pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate;
  PyObject *__pyx_kp_u_;
  PyObject *__pyx_kp_u_0_3;
  PyObject *__pyx_kp_u_12;
  PyObject *__pyx_kp_u_1_5;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_s_CalledProcessError;
  PyObject *__pyx_kp_u_Cannot_translate_symbol;
  PyObject *__pyx_kp_u_Courier_Bold;
  PyObject *__pyx_n_s_DataFrame;
  PyObject *__pyx_n_s_DiGraph;
  PyObject *__pyx_kp_u_Encodes_word_based_on_a_string_t;
  PyObject *__pyx_kp_u_For_a_given_path_set_of_states_r;
  PyObject *__pyx_n_s_G;
  PyObject *__pyx_n_s_MemoryError;
  PyObject *__pyx_n_s_Move;
  PyObject *__pyx_kp_s_Move_None;
  PyObject *__pyx_n_s_Move___neq;
  PyObject *__pyx_n_s_Move___reduce_cython;
  PyObject *__pyx_n_s_Move___setstate_cython;
  PyObject *__pyx_n_s_Nfa;
  PyObject *__pyx_n_s_Nfa___reduce_cython;
  PyObject *__pyx_n_s_Nfa___setstate_cython;
  PyObject *__pyx_n_s_Nfa_accepts_epsilon;
  PyObject *__pyx_n_s_Nfa_add_new_state;
  PyObject *__pyx_n_s_Nfa_add_state;
  PyObject *__pyx_n_s_Nfa_add_transition;
  PyObject *__pyx_n_s_Nfa_add_transition_object;
  PyObject *__pyx_n_s_Nfa_clear;
  PyObject *__pyx_n_s_Nfa_clear_final;
  PyObject *__pyx_n_s_Nfa_clear_initial;
  PyObject *__pyx_n_s_Nfa_complement;
  PyObject *__pyx_n_s_Nfa_compute_relation;
  PyObject *__pyx_n_s_Nfa_concatenate;
  PyObject *__pyx_n_s_Nfa_concatenate_with_result_stat;
  PyObject *__pyx_n_s_Nfa_determinize;
  PyObject *__pyx_n_s_Nfa_determinize_with_subset_map;
  PyObject *__pyx_n_s_Nfa_encode_word;
  PyObject *__pyx_kp_u_Nfa_encode_word_line_1139;
  PyObject *__pyx_n_s_Nfa_equivalence_check;
  PyObject *__pyx_n_s_Nfa_get_epsilon_transitions;
  PyObject *__pyx_n_s_Nfa_get_num_of_trans;
  PyObject *__pyx_n_s_Nfa_get_one_letter_aut;
  PyObject *__pyx_n_s_Nfa_get_reachable_states;
  PyObject *__pyx_n_s_Nfa_get_symbols;
  PyObject *__pyx_n_s_Nfa_get_terminating_states;
  PyObject *__pyx_n_s_Nfa_get_trans_as_sequence;
  PyObject *__pyx_n_s_Nfa_get_trans_from_state_as_sequ;
  PyObject *__pyx_n_s_Nfa_get_transitions_from_state;
  PyObject *__pyx_n_s_Nfa_get_transitions_to_state;
  PyObject *__pyx_n_s_Nfa_get_useful_states;
  PyObject *__pyx_n_s_Nfa_get_word_for_path;
  PyObject *__pyx_kp_u_Nfa_get_word_for_path_line_1122;
  PyObject *__pyx_n_s_Nfa_has_final_state;
  PyObject *__pyx_n_s_Nfa_has_initial_state;
  PyObject *__pyx_n_s_Nfa_has_transition;
  PyObject *__pyx_n_s_Nfa_intersection;
  PyObject *__pyx_n_s_Nfa_intersection_with_product_ma;
  PyObject *__pyx_n_s_Nfa_is_complete;
  PyObject *__pyx_n_s_Nfa_is_deterministic;
  PyObject *__pyx_n_s_Nfa_is_epsilon;
  PyObject *__pyx_n_s_Nfa_is_in_lang;
  PyObject *__pyx_n_s_Nfa_is_included;
  PyObject *__pyx_n_s_Nfa_is_included_with_cex;
  PyObject *__pyx_n_s_Nfa_is_lang_empty;
  PyObject *__pyx_n_s_Nfa_is_prefix_in_lang;
  PyObject *__pyx_n_s_Nfa_is_state;
  PyObject *__pyx_n_s_Nfa_is_universal;
  PyObject *__pyx_n_s_Nfa_iterate;
  PyObject *__pyx_n_s_Nfa_make_complete;
  PyObject *__pyx_n_s_Nfa_make_final_state;
  PyObject *__pyx_n_s_Nfa_make_final_states;
  PyObject *__pyx_n_s_Nfa_make_initial_state;
  PyObject *__pyx_n_s_Nfa_make_initial_states;
  PyObject *__pyx_n_s_Nfa_minimize;
  PyObject *__pyx_n_s_Nfa_post_map_of;
  PyObject *__pyx_n_s_Nfa_post_of;
  PyObject *__pyx_n_s_Nfa_reduce;
  PyObject *__pyx_n_s_Nfa_reduce_with_state_map;
  PyObject *__pyx_n_s_Nfa_remove_epsilon;
  PyObject *__pyx_n_s_Nfa_remove_epsilon_inplace;
  PyObject *__pyx_n_s_Nfa_remove_final_state;
  PyObject *__pyx_n_s_Nfa_remove_initial_state;
  PyObject *__pyx_n_s_Nfa_remove_trans;
  PyObject *__pyx_n_s_Nfa_remove_trans_raw;
  PyObject *__pyx_n_s_Nfa_revert;
  PyObject *__pyx_n_s_Nfa_size;
  PyObject *__pyx_n_s_Nfa_to_dataframe;
  PyObject *__pyx_n_s_Nfa_to_dot_file;
  PyObject *__pyx_n_s_Nfa_to_dot_str;
  PyObject *__pyx_n_s_Nfa_to_networkx_graph;
  PyObject *__pyx_n_s_Nfa_trim;
  PyObject *__pyx_n_s_Nfa_trim_with_state_map;
  PyObject *__pyx_n_s_Nfa_unify_final;
  PyObject *__pyx_n_s_Nfa_unify_initial;
  PyObject *__pyx_n_s_Nfa_union;
  PyObject *__pyx_n_s_PIPE;
  PyObject *__pyx_n_s_Popen;
  PyObject *__pyx_n_s_Run;
  PyObject *__pyx_n_s_Run___reduce_cython;
  PyObject *__pyx_n_s_Run___setstate_cython;
  PyObject *__pyx_n_s_STDOUT;
  PyObject *__pyx_n_s_TimeoutExpired;
  PyObject *__pyx_n_s_Trans;
  PyObject *__pyx_n_s_Trans___neq;
  PyObject *__pyx_n_s_Trans___reduce_cython;
  PyObject *__pyx_n_s_Trans___setstate_cython;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_kp_b__10;
  PyObject *__pyx_kp_u__15;
  PyObject *__pyx_n_s__16;
  PyObject *__pyx_n_s__165;
  PyObject *__pyx_kp_u__2;
  PyObject *__pyx_kp_u__3;
  PyObject *__pyx_kp_u__4;
  PyObject *__pyx_kp_u__5;
  PyObject *__pyx_kp_u__6;
  PyObject *__pyx_kp_u__8;
  PyObject *__pyx_n_s__80;
  PyObject *__pyx_kp_u__9;
  PyObject *__pyx_n_s_accepts_epsilon;
  PyObject *__pyx_n_s_add_edge;
  PyObject *__pyx_n_s_add_new_state;
  PyObject *__pyx_n_s_add_state;
  PyObject *__pyx_n_s_add_transition;
  PyObject *__pyx_n_s_add_transition_object;
  PyObject *__pyx_n_u_algorithm;
  PyObject *__pyx_n_s_alphabet;
  PyObject *__pyx_n_u_alphabet;
  PyObject *__pyx_n_u_antichains;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_aut;
  PyObject *__pyx_kp_u_aut_dot;
  PyObject *__pyx_n_s_bool;
  PyObject *__pyx_n_s_bool_vec;
  PyObject *__pyx_n_s_c_alphabet;
  PyObject *__pyx_n_s_c_bool_vec;
  PyObject *__pyx_n_s_c_epsilon_transitions_iter;
  PyObject *__pyx_n_s_c_lhs_map;
  PyObject *__pyx_n_s_c_product_map;
  PyObject *__pyx_n_s_c_rhs_map;
  PyObject *__pyx_n_s_c_states;
  PyObject *__pyx_n_s_c_transition;
  PyObject *__pyx_n_s_c_transitions;
  PyObject *__pyx_kp_u_captured_stderr;
  PyObject *__pyx_kp_u_captured_stdout;
  PyObject *__pyx_n_s_check_results;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_u_classical;
  PyObject *__pyx_n_s_clear;
  PyObject *__pyx_n_s_clear_final;
  PyObject *__pyx_n_s_clear_initial;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_s_cls;
  PyObject *__pyx_n_s_cmd;
  PyObject *__pyx_n_s_cmd_no;
  PyObject *__pyx_n_s_cmderr;
  PyObject *__pyx_n_s_cmdout;
  PyObject *__pyx_n_u_color;
  PyObject *__pyx_n_s_columns;
  PyObject *__pyx_n_s_communicate;
  PyObject *__pyx_n_s_complement;
  PyObject *__pyx_n_s_compute_relation;
  PyObject *__pyx_n_s_concatenate;
  PyObject *__pyx_n_s_concatenate_with_result_state_ma;
  PyObject *__pyx_n_u_darkblue;
  PyObject *__pyx_n_s_data;
  PyObject *__pyx_n_s_decode;
  PyObject *__pyx_n_s_determinize;
  PyObject *__pyx_n_s_determinize_with_subset_map;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_u_direction;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_divisible_by;
  PyObject *__pyx_kp_u_dot_O_T;
  PyObject *__pyx_n_u_edge_style;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_encode;
  PyObject *__pyx_n_s_encode_word;
  PyObject *__pyx_n_s_encoding;
  PyObject *__pyx_n_s_end;
  PyObject *__pyx_n_s_epsilon;
  PyObject *__pyx_n_s_epsilon_transitions;
  PyObject *__pyx_n_s_equivalence_check;
  PyObject *__pyx_n_s_err;
  PyObject *__pyx_kp_u_error_while_dot_file;
  PyObject *__pyx_n_s_executed_command;
  PyObject *__pyx_n_u_false;
  PyObject *__pyx_n_u_fillcolor;
  PyObject *__pyx_n_u_filled;
  PyObject *__pyx_kp_u_final_states;
  PyObject *__pyx_n_u_fixedsize;
  PyObject *__pyx_n_u_fontname;
  PyObject *__pyx_n_u_fontsize;
  PyObject *__pyx_n_s_format;
  PyObject *__pyx_n_u_forward;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_get;
  PyObject *__pyx_n_s_get_alphabet_symbols;
  PyObject *__pyx_n_s_get_elements_from_bool_vec;
  PyObject *__pyx_n_s_get_epsilon_transitions;
  PyObject *__pyx_n_s_get_num_of_trans;
  PyObject *__pyx_n_s_get_one_letter_aut;
  PyObject *__pyx_n_s_get_reachable_states;
  PyObject *__pyx_n_s_get_symbols;
  PyObject *__pyx_n_s_get_terminating_states;
  PyObject *__pyx_n_s_get_trans_as_sequence;
  PyObject *__pyx_n_s_get_trans_from_state_as_sequence;
  PyObject *__pyx_n_s_get_transitions_from_state;
  PyObject *__pyx_n_s_get_transitions_to_state;
  PyObject *__pyx_n_s_get_useful_states;
  PyObject *__pyx_n_s_get_word_for_path;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_s_graphviz_command;
  PyObject *__pyx_n_s_has_final_state;
  PyObject *__pyx_n_s_has_initial_state;
  PyObject *__pyx_n_s_has_transition;
  PyObject *__pyx_n_u_height;
  PyObject *__pyx_n_s_i;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_kp_u_initial_states;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_input;
  PyObject *__pyx_n_s_input_states;
  PyObject *__pyx_n_s_int;
  PyObject *__pyx_n_s_intersection;
  PyObject *__pyx_n_s_intersection_with_product_map;
  PyObject *__pyx_n_s_is_complete;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_n_s_is_deterministic;
  PyObject *__pyx_n_s_is_epsilon;
  PyObject *__pyx_n_s_is_in_lang;
  PyObject *__pyx_n_s_is_included;
  PyObject *__pyx_n_s_is_included_with_cex;
  PyObject *__pyx_n_s_is_lang_empty;
  PyObject *__pyx_n_s_is_prefix_in_lang;
  PyObject *__pyx_n_s_is_state;
  PyObject *__pyx_n_s_is_universal;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_it;
  PyObject *__pyx_n_s_items;
  PyObject *__pyx_n_s_iterate;
  PyObject *__pyx_n_s_iterator;
  PyObject *__pyx_n_s_iteritems;
  PyObject *__pyx_n_s_j;
  PyObject *__pyx_n_s_k;
  PyObject *__pyx_n_s_key;
  PyObject *__pyx_n_s_kwargs;
  PyObject *__pyx_n_s_label;
  PyObject *__pyx_n_s_lhs;
  PyObject *__pyx_n_s_lhs_map;
  PyObject *__pyx_n_s_libmata_nfa_nfa;
  PyObject *__pyx_kp_s_libmata_nfa_nfa_pyx;
  PyObject *__pyx_n_u_lightsteelblue;
  PyObject *__pyx_kp_s_list_int;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_make_complete;
  PyObject *__pyx_n_s_make_final_state;
  PyObject *__pyx_n_s_make_final_states;
  PyObject *__pyx_n_s_make_initial_state;
  PyObject *__pyx_n_s_make_initial_states;
  PyObject *__pyx_n_s_map;
  PyObject *__pyx_n_u_midnightblue;
  PyObject *__pyx_n_s_minimize;
  PyObject *__pyx_n_u_minimize;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_neq;
  PyObject *__pyx_n_s_networkx;
  PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
  PyObject *__pyx_n_u_node_style;
  PyObject *__pyx_n_s_nx;
  PyObject *__pyx_kp_s_nx_Graph;
  PyObject *__pyx_n_s_objects;
  PyObject *__pyx_n_s_one_letter_aut;
  PyObject *__pyx_n_s_other;
  PyObject *__pyx_n_s_output;
  PyObject *__pyx_n_s_output_file;
  PyObject *__pyx_n_s_output_format;
  PyObject *__pyx_n_s_output_stream;
  PyObject *__pyx_n_s_pandas;
  PyObject *__pyx_kp_s_pandas_DataFrame;
  PyObject *__pyx_n_s_params;
  PyObject *__pyx_n_s_path;
  PyObject *__pyx_n_u_pdf;
  PyObject *__pyx_n_u_penwidth;
  PyObject *__pyx_n_s_piped_command;
  PyObject *__pyx_n_s_post_map_of;
  PyObject *__pyx_n_s_post_of;
  PyObject *__pyx_n_s_preserve_epsilon;
  PyObject *__pyx_n_s_print;
  PyObject *__pyx_n_s_process;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_quiet;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_u_red;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_2;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_reduce_with_state_map;
  PyObject *__pyx_n_s_relation;
  PyObject *__pyx_n_u_relation;
  PyObject *__pyx_n_s_relation_size;
  PyObject *__pyx_n_s_remove_epsilon;
  PyObject *__pyx_n_s_remove_epsilon_inplace;
  PyObject *__pyx_n_s_remove_final_state;
  PyObject *__pyx_n_s_remove_initial_state;
  PyObject *__pyx_n_s_remove_trans;
  PyObject *__pyx_n_s_remove_trans_raw;
  PyObject *__pyx_n_s_resize;
  PyObject *__pyx_n_s_result;
  PyObject *__pyx_n_s_return;
  PyObject *__pyx_n_s_return_value;
  PyObject *__pyx_n_s_returncode;
  PyObject *__pyx_n_s_revert;
  PyObject *__pyx_n_s_rhs;
  PyObject *__pyx_n_s_rhs_map;
  PyObject *__pyx_n_s_run;
  PyObject *__pyx_n_s_run_safely_external_command;
  PyObject *__pyx_n_s_s;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_n_s_set;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_shell;
  PyObject *__pyx_n_s_shlex;
  PyObject *__pyx_n_u_simulation;
  PyObject *__pyx_n_s_sink_state;
  PyObject *__pyx_n_s_size;
  PyObject *__pyx_n_u_source;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_s_split;
  PyObject *__pyx_n_s_src;
  PyObject *__pyx_n_s_st;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_s_state_from;
  PyObject *__pyx_n_s_state_map;
  PyObject *__pyx_n_s_state_number;
  PyObject *__pyx_n_s_state_to;
  PyObject *__pyx_n_s_states;
  PyObject *__pyx_n_s_stderr;
  PyObject *__pyx_n_s_stdin;
  PyObject *__pyx_n_s_stdout;
  PyObject *__pyx_n_s_store;
  PyObject *__pyx_n_s_store_2;
  PyObject *__pyx_n_s_str;
  PyObject *__pyx_n_s_string_to_symbol;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_strip;
  PyObject *__pyx_n_u_style;
  PyObject *__pyx_n_s_subprocess;
  PyObject *__pyx_n_s_subset_map;
  PyObject *__pyx_n_s_symb;
  PyObject *__pyx_n_s_symbol;
  PyObject *__pyx_n_u_symbol;
  PyObject *__pyx_n_s_symbol_map;
  PyObject *__pyx_n_s_symbols;
  PyObject *__pyx_n_s_t;
  PyObject *__pyx_n_u_target;
  PyObject *__pyx_n_s_targets;
  PyObject *__pyx_n_s_terminate;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_tgt;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_timeout;
  PyObject *__pyx_n_s_to_dataframe;
  PyObject *__pyx_n_s_to_dot_file;
  PyObject *__pyx_n_s_to_dot_str;
  PyObject *__pyx_n_s_to_networkx_graph;
  PyObject *__pyx_n_s_tr;
  PyObject *__pyx_n_s_trans;
  PyObject *__pyx_kp_u_transitions;
  PyObject *__pyx_n_s_transitions_2;
  PyObject *__pyx_n_s_transitions_list;
  PyObject *__pyx_n_s_translate_symbol;
  PyObject *__pyx_n_s_transsymbols;
  PyObject *__pyx_n_s_trim;
  PyObject *__pyx_n_s_trim_input;
  PyObject *__pyx_n_s_trim_with_state_map;
  PyObject *__pyx_n_u_true;
  PyObject *__pyx_kp_s_tuple_Nfa_dict_str_str_dict_str;
  PyObject *__pyx_n_s_unify_final;
  PyObject *__pyx_n_s_unify_initial;
  PyObject *__pyx_n_s_union;
  PyObject *__pyx_n_s_unpiped_commands;
  PyObject *__pyx_n_s_use_epsilon;
  PyObject *__pyx_n_s_useful_states;
  PyObject *__pyx_n_s_useful_states_bool_vec;
  PyObject *__pyx_kp_u_utf_8;
  PyObject *__pyx_n_s_v;
  PyObject *__pyx_n_s_val;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_n_s_wait;
  PyObject *__pyx_n_u_width;
  PyObject *__pyx_kp_u_without_specified_alphabet;
  PyObject *__pyx_n_s_word;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  Mata::Symbol __pyx_k__11;
  Mata::Symbol __pyx_k__12;
  Mata::Symbol __pyx_k__13;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__18;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__29;
  PyObject *__pyx_tuple__34;
  PyObject *__pyx_tuple__36;
  PyObject *__pyx_tuple__47;
  PyObject *__pyx_tuple__49;
  PyObject *__pyx_tuple__51;
  PyObject *__pyx_tuple__53;
  PyObject *__pyx_tuple__59;
  PyObject *__pyx_tuple__60;
  PyObject *__pyx_tuple__62;
  PyObject *__pyx_tuple__64;
  PyObject *__pyx_tuple__66;
  PyObject *__pyx_tuple__68;
  PyObject *__pyx_tuple__70;
  PyObject *__pyx_tuple__74;
  PyObject *__pyx_tuple__76;
  PyObject *__pyx_tuple__78;
  PyObject *__pyx_tuple__81;
  PyObject *__pyx_tuple__83;
  PyObject *__pyx_tuple__84;
  PyObject *__pyx_tuple__86;
  PyObject *__pyx_tuple__87;
  PyObject *__pyx_tuple__89;
  PyObject *__pyx_tuple__91;
  PyObject *__pyx_tuple__93;
  PyObject *__pyx_tuple__95;
  PyObject *__pyx_tuple__97;
  PyObject *__pyx_tuple__99;
  PyObject *__pyx_codeobj__7;
  PyObject *__pyx_tuple__101;
  PyObject *__pyx_tuple__103;
  PyObject *__pyx_tuple__104;
  PyObject *__pyx_tuple__106;
  PyObject *__pyx_tuple__108;
  PyObject *__pyx_tuple__110;
  PyObject *__pyx_tuple__112;
  PyObject *__pyx_tuple__115;
  PyObject *__pyx_tuple__117;
  PyObject *__pyx_tuple__119;
  PyObject *__pyx_tuple__122;
  PyObject *__pyx_tuple__124;
  PyObject *__pyx_tuple__125;
  PyObject *__pyx_tuple__127;
  PyObject *__pyx_tuple__129;
  PyObject *__pyx_tuple__131;
  PyObject *__pyx_tuple__133;
  PyObject *__pyx_tuple__135;
  PyObject *__pyx_tuple__137;
  PyObject *__pyx_tuple__138;
  PyObject *__pyx_tuple__140;
  PyObject *__pyx_tuple__142;
  PyObject *__pyx_tuple__144;
  PyObject *__pyx_tuple__146;
  PyObject *__pyx_tuple__149;
  PyObject *__pyx_tuple__151;
  PyObject *__pyx_tuple__153;
  PyObject *__pyx_tuple__157;
  PyObject *__pyx_tuple__159;
  PyObject *__pyx_tuple__161;
  PyObject *__pyx_tuple__162;
  PyObject *__pyx_codeobj__17;
  PyObject *__pyx_codeobj__19;
  PyObject *__pyx_codeobj__21;
  PyObject *__pyx_codeobj__23;
  PyObject *__pyx_codeobj__24;
  PyObject *__pyx_codeobj__25;
  PyObject *__pyx_codeobj__26;
  PyObject *__pyx_codeobj__27;
  PyObject *__pyx_codeobj__28;
  PyObject *__pyx_codeobj__30;
  PyObject *__pyx_codeobj__31;
  PyObject *__pyx_codeobj__32;
  PyObject *__pyx_codeobj__33;
  PyObject *__pyx_codeobj__35;
  PyObject *__pyx_codeobj__37;
  PyObject *__pyx_codeobj__38;
  PyObject *__pyx_codeobj__39;
  PyObject *__pyx_codeobj__40;
  PyObject *__pyx_codeobj__41;
  PyObject *__pyx_codeobj__42;
  PyObject *__pyx_codeobj__43;
  PyObject *__pyx_codeobj__44;
  PyObject *__pyx_codeobj__45;
  PyObject *__pyx_codeobj__46;
  PyObject *__pyx_codeobj__48;
  PyObject *__pyx_codeobj__50;
  PyObject *__pyx_codeobj__52;
  PyObject *__pyx_codeobj__54;
  PyObject *__pyx_codeobj__55;
  PyObject *__pyx_codeobj__56;
  PyObject *__pyx_codeobj__57;
  PyObject *__pyx_codeobj__58;
  PyObject *__pyx_codeobj__61;
  PyObject *__pyx_codeobj__63;
  PyObject *__pyx_codeobj__65;
  PyObject *__pyx_codeobj__67;
  PyObject *__pyx_codeobj__69;
  PyObject *__pyx_codeobj__71;
  PyObject *__pyx_codeobj__72;
  PyObject *__pyx_codeobj__73;
  PyObject *__pyx_codeobj__75;
  PyObject *__pyx_codeobj__77;
  PyObject *__pyx_codeobj__79;
  PyObject *__pyx_codeobj__82;
  PyObject *__pyx_codeobj__85;
  PyObject *__pyx_codeobj__88;
  PyObject *__pyx_codeobj__90;
  PyObject *__pyx_codeobj__92;
  PyObject *__pyx_codeobj__94;
  PyObject *__pyx_codeobj__96;
  PyObject *__pyx_codeobj__98;
  PyObject *__pyx_codeobj__100;
  PyObject *__pyx_codeobj__102;
  PyObject *__pyx_codeobj__105;
  PyObject *__pyx_codeobj__107;
  PyObject *__pyx_codeobj__109;
  PyObject *__pyx_codeobj__111;
  PyObject *__pyx_codeobj__113;
  PyObject *__pyx_codeobj__114;
  PyObject *__pyx_codeobj__116;
  PyObject *__pyx_codeobj__118;
  PyObject *__pyx_codeobj__120;
  PyObject *__pyx_codeobj__121;
  PyObject *__pyx_codeobj__123;
  PyObject *__pyx_codeobj__126;
  PyObject *__pyx_codeobj__128;
  PyObject *__pyx_codeobj__130;
  PyObject *__pyx_codeobj__132;
  PyObject *__pyx_codeobj__134;
  PyObject *__pyx_codeobj__136;
  PyObject *__pyx_codeobj__139;
  PyObject *__pyx_codeobj__141;
  PyObject *__pyx_codeobj__143;
  PyObject *__pyx_codeobj__145;
  PyObject *__pyx_codeobj__147;
  PyObject *__pyx_codeobj__148;
  PyObject *__pyx_codeobj__150;
  PyObject *__pyx_codeobj__152;
  PyObject *__pyx_codeobj__154;
  PyObject *__pyx_codeobj__155;
  PyObject *__pyx_codeobj__156;
  PyObject *__pyx_codeobj__158;
  PyObject *__pyx_codeobj__160;
  PyObject *__pyx_codeobj__163;
  PyObject *__pyx_codeobj__164;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7libmata_5utils_BinaryRelation);
  Py_CLEAR(clear_module_state->__pyx_ptype_7libmata_9alphabets_Alphabet);
  Py_CLEAR(clear_module_state->__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  Py_CLEAR(clear_module_state->__pyx_type_7libmata_3nfa_3nfa_Nfa);
  Py_CLEAR(clear_module_state->__pyx_ptype_7libmata_3nfa_3nfa_Trans);
  Py_CLEAR(clear_module_state->__pyx_type_7libmata_3nfa_3nfa_Trans);
  Py_CLEAR(clear_module_state->__pyx_ptype_7libmata_3nfa_3nfa_Run);
  Py_CLEAR(clear_module_state->__pyx_type_7libmata_3nfa_3nfa_Run);
  Py_CLEAR(clear_module_state->__pyx_ptype_7libmata_3nfa_3nfa_Move);
  Py_CLEAR(clear_module_state->__pyx_type_7libmata_3nfa_3nfa_Move);
  Py_CLEAR(clear_module_state->__pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate);
  Py_CLEAR(clear_module_state->__pyx_type_7libmata_3nfa_3nfa___pyx_scope_struct__iterate);
  Py_CLEAR(clear_module_state->__pyx_kp_u_);
  Py_CLEAR(clear_module_state->__pyx_kp_u_0_3);
  Py_CLEAR(clear_module_state->__pyx_kp_u_12);
  Py_CLEAR(clear_module_state->__pyx_kp_u_1_5);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_CalledProcessError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_translate_symbol);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Courier_Bold);
  Py_CLEAR(clear_module_state->__pyx_n_s_DataFrame);
  Py_CLEAR(clear_module_state->__pyx_n_s_DiGraph);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Encodes_word_based_on_a_string_t);
  Py_CLEAR(clear_module_state->__pyx_kp_u_For_a_given_path_set_of_states_r);
  Py_CLEAR(clear_module_state->__pyx_n_s_G);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Move);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Move_None);
  Py_CLEAR(clear_module_state->__pyx_n_s_Move___neq);
  Py_CLEAR(clear_module_state->__pyx_n_s_Move___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Move___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_accepts_epsilon);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_add_new_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_add_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_add_transition);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_add_transition_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_clear);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_clear_final);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_clear_initial);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_complement);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_compute_relation);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_concatenate);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_concatenate_with_result_stat);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_determinize);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_determinize_with_subset_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_encode_word);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Nfa_encode_word_line_1139);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_equivalence_check);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_get_epsilon_transitions);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_get_num_of_trans);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_get_one_letter_aut);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_get_reachable_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_get_symbols);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_get_terminating_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_get_trans_as_sequence);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_get_trans_from_state_as_sequ);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_get_transitions_from_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_get_transitions_to_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_get_useful_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_get_word_for_path);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Nfa_get_word_for_path_line_1122);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_has_final_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_has_initial_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_has_transition);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_intersection);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_intersection_with_product_ma);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_is_complete);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_is_deterministic);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_is_epsilon);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_is_in_lang);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_is_included);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_is_included_with_cex);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_is_lang_empty);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_is_prefix_in_lang);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_is_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_is_universal);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_iterate);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_make_complete);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_make_final_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_make_final_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_make_initial_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_make_initial_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_minimize);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_post_map_of);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_post_of);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_reduce_with_state_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_remove_epsilon);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_remove_epsilon_inplace);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_remove_final_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_remove_initial_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_remove_trans);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_remove_trans_raw);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_revert);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_to_dataframe);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_to_dot_file);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_to_dot_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_to_networkx_graph);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_trim);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_trim_with_state_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_unify_final);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_unify_initial);
  Py_CLEAR(clear_module_state->__pyx_n_s_Nfa_union);
  Py_CLEAR(clear_module_state->__pyx_n_s_PIPE);
  Py_CLEAR(clear_module_state->__pyx_n_s_Popen);
  Py_CLEAR(clear_module_state->__pyx_n_s_Run);
  Py_CLEAR(clear_module_state->__pyx_n_s_Run___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Run___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_STDOUT);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimeoutExpired);
  Py_CLEAR(clear_module_state->__pyx_n_s_Trans);
  Py_CLEAR(clear_module_state->__pyx_n_s_Trans___neq);
  Py_CLEAR(clear_module_state->__pyx_n_s_Trans___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Trans___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_kp_b__10);
  Py_CLEAR(clear_module_state->__pyx_kp_u__15);
  Py_CLEAR(clear_module_state->__pyx_n_s__16);
  Py_CLEAR(clear_module_state->__pyx_n_s__165);
  Py_CLEAR(clear_module_state->__pyx_kp_u__2);
  Py_CLEAR(clear_module_state->__pyx_kp_u__3);
  Py_CLEAR(clear_module_state->__pyx_kp_u__4);
  Py_CLEAR(clear_module_state->__pyx_kp_u__5);
  Py_CLEAR(clear_module_state->__pyx_kp_u__6);
  Py_CLEAR(clear_module_state->__pyx_kp_u__8);
  Py_CLEAR(clear_module_state->__pyx_n_s__80);
  Py_CLEAR(clear_module_state->__pyx_kp_u__9);
  Py_CLEAR(clear_module_state->__pyx_n_s_accepts_epsilon);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_edge);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_new_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_transition);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_transition_object);
  Py_CLEAR(clear_module_state->__pyx_n_u_algorithm);
  Py_CLEAR(clear_module_state->__pyx_n_s_alphabet);
  Py_CLEAR(clear_module_state->__pyx_n_u_alphabet);
  Py_CLEAR(clear_module_state->__pyx_n_u_antichains);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_aut);
  Py_CLEAR(clear_module_state->__pyx_kp_u_aut_dot);
  Py_CLEAR(clear_module_state->__pyx_n_s_bool);
  Py_CLEAR(clear_module_state->__pyx_n_s_bool_vec);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_alphabet);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_bool_vec);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_epsilon_transitions_iter);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_lhs_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_product_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_rhs_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_transition);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_transitions);
  Py_CLEAR(clear_module_state->__pyx_kp_u_captured_stderr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_captured_stdout);
  Py_CLEAR(clear_module_state->__pyx_n_s_check_results);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_u_classical);
  Py_CLEAR(clear_module_state->__pyx_n_s_clear);
  Py_CLEAR(clear_module_state->__pyx_n_s_clear_final);
  Py_CLEAR(clear_module_state->__pyx_n_s_clear_initial);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_cls);
  Py_CLEAR(clear_module_state->__pyx_n_s_cmd);
  Py_CLEAR(clear_module_state->__pyx_n_s_cmd_no);
  Py_CLEAR(clear_module_state->__pyx_n_s_cmderr);
  Py_CLEAR(clear_module_state->__pyx_n_s_cmdout);
  Py_CLEAR(clear_module_state->__pyx_n_u_color);
  Py_CLEAR(clear_module_state->__pyx_n_s_columns);
  Py_CLEAR(clear_module_state->__pyx_n_s_communicate);
  Py_CLEAR(clear_module_state->__pyx_n_s_complement);
  Py_CLEAR(clear_module_state->__pyx_n_s_compute_relation);
  Py_CLEAR(clear_module_state->__pyx_n_s_concatenate);
  Py_CLEAR(clear_module_state->__pyx_n_s_concatenate_with_result_state_ma);
  Py_CLEAR(clear_module_state->__pyx_n_u_darkblue);
  Py_CLEAR(clear_module_state->__pyx_n_s_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_decode);
  Py_CLEAR(clear_module_state->__pyx_n_s_determinize);
  Py_CLEAR(clear_module_state->__pyx_n_s_determinize_with_subset_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_u_direction);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_divisible_by);
  Py_CLEAR(clear_module_state->__pyx_kp_u_dot_O_T);
  Py_CLEAR(clear_module_state->__pyx_n_u_edge_style);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_encode_word);
  Py_CLEAR(clear_module_state->__pyx_n_s_encoding);
  Py_CLEAR(clear_module_state->__pyx_n_s_end);
  Py_CLEAR(clear_module_state->__pyx_n_s_epsilon);
  Py_CLEAR(clear_module_state->__pyx_n_s_epsilon_transitions);
  Py_CLEAR(clear_module_state->__pyx_n_s_equivalence_check);
  Py_CLEAR(clear_module_state->__pyx_n_s_err);
  Py_CLEAR(clear_module_state->__pyx_kp_u_error_while_dot_file);
  Py_CLEAR(clear_module_state->__pyx_n_s_executed_command);
  Py_CLEAR(clear_module_state->__pyx_n_u_false);
  Py_CLEAR(clear_module_state->__pyx_n_u_fillcolor);
  Py_CLEAR(clear_module_state->__pyx_n_u_filled);
  Py_CLEAR(clear_module_state->__pyx_kp_u_final_states);
  Py_CLEAR(clear_module_state->__pyx_n_u_fixedsize);
  Py_CLEAR(clear_module_state->__pyx_n_u_fontname);
  Py_CLEAR(clear_module_state->__pyx_n_u_fontsize);
  Py_CLEAR(clear_module_state->__pyx_n_s_format);
  Py_CLEAR(clear_module_state->__pyx_n_u_forward);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_alphabet_symbols);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_elements_from_bool_vec);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_epsilon_transitions);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_num_of_trans);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_one_letter_aut);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_reachable_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_symbols);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_terminating_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_trans_as_sequence);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_trans_from_state_as_sequence);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_transitions_from_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_transitions_to_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_useful_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_word_for_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_graphviz_command);
  Py_CLEAR(clear_module_state->__pyx_n_s_has_final_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_has_initial_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_has_transition);
  Py_CLEAR(clear_module_state->__pyx_n_u_height);
  Py_CLEAR(clear_module_state->__pyx_n_s_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_kp_u_initial_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_input);
  Py_CLEAR(clear_module_state->__pyx_n_s_input_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_intersection);
  Py_CLEAR(clear_module_state->__pyx_n_s_intersection_with_product_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_complete);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_deterministic);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_epsilon);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_in_lang);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_included);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_included_with_cex);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_lang_empty);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_prefix_in_lang);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_universal);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_it);
  Py_CLEAR(clear_module_state->__pyx_n_s_items);
  Py_CLEAR(clear_module_state->__pyx_n_s_iterate);
  Py_CLEAR(clear_module_state->__pyx_n_s_iterator);
  Py_CLEAR(clear_module_state->__pyx_n_s_iteritems);
  Py_CLEAR(clear_module_state->__pyx_n_s_j);
  Py_CLEAR(clear_module_state->__pyx_n_s_k);
  Py_CLEAR(clear_module_state->__pyx_n_s_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_kwargs);
  Py_CLEAR(clear_module_state->__pyx_n_s_label);
  Py_CLEAR(clear_module_state->__pyx_n_s_lhs);
  Py_CLEAR(clear_module_state->__pyx_n_s_lhs_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_libmata_nfa_nfa);
  Py_CLEAR(clear_module_state->__pyx_kp_s_libmata_nfa_nfa_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_u_lightsteelblue);
  Py_CLEAR(clear_module_state->__pyx_kp_s_list_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_make_complete);
  Py_CLEAR(clear_module_state->__pyx_n_s_make_final_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_make_final_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_make_initial_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_make_initial_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_map);
  Py_CLEAR(clear_module_state->__pyx_n_u_midnightblue);
  Py_CLEAR(clear_module_state->__pyx_n_s_minimize);
  Py_CLEAR(clear_module_state->__pyx_n_u_minimize);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_neq);
  Py_CLEAR(clear_module_state->__pyx_n_s_networkx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_CLEAR(clear_module_state->__pyx_n_u_node_style);
  Py_CLEAR(clear_module_state->__pyx_n_s_nx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_nx_Graph);
  Py_CLEAR(clear_module_state->__pyx_n_s_objects);
  Py_CLEAR(clear_module_state->__pyx_n_s_one_letter_aut);
  Py_CLEAR(clear_module_state->__pyx_n_s_other);
  Py_CLEAR(clear_module_state->__pyx_n_s_output);
  Py_CLEAR(clear_module_state->__pyx_n_s_output_file);
  Py_CLEAR(clear_module_state->__pyx_n_s_output_format);
  Py_CLEAR(clear_module_state->__pyx_n_s_output_stream);
  Py_CLEAR(clear_module_state->__pyx_n_s_pandas);
  Py_CLEAR(clear_module_state->__pyx_kp_s_pandas_DataFrame);
  Py_CLEAR(clear_module_state->__pyx_n_s_params);
  Py_CLEAR(clear_module_state->__pyx_n_s_path);
  Py_CLEAR(clear_module_state->__pyx_n_u_pdf);
  Py_CLEAR(clear_module_state->__pyx_n_u_penwidth);
  Py_CLEAR(clear_module_state->__pyx_n_s_piped_command);
  Py_CLEAR(clear_module_state->__pyx_n_s_post_map_of);
  Py_CLEAR(clear_module_state->__pyx_n_s_post_of);
  Py_CLEAR(clear_module_state->__pyx_n_s_preserve_epsilon);
  Py_CLEAR(clear_module_state->__pyx_n_s_print);
  Py_CLEAR(clear_module_state->__pyx_n_s_process);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_quiet);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_u_red);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_with_state_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_relation);
  Py_CLEAR(clear_module_state->__pyx_n_u_relation);
  Py_CLEAR(clear_module_state->__pyx_n_s_relation_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove_epsilon);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove_epsilon_inplace);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove_final_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove_initial_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove_trans);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove_trans_raw);
  Py_CLEAR(clear_module_state->__pyx_n_s_resize);
  Py_CLEAR(clear_module_state->__pyx_n_s_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_return);
  Py_CLEAR(clear_module_state->__pyx_n_s_return_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_returncode);
  Py_CLEAR(clear_module_state->__pyx_n_s_revert);
  Py_CLEAR(clear_module_state->__pyx_n_s_rhs);
  Py_CLEAR(clear_module_state->__pyx_n_s_rhs_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_run);
  Py_CLEAR(clear_module_state->__pyx_n_s_run_safely_external_command);
  Py_CLEAR(clear_module_state->__pyx_n_s_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_set);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_shell);
  Py_CLEAR(clear_module_state->__pyx_n_s_shlex);
  Py_CLEAR(clear_module_state->__pyx_n_u_simulation);
  Py_CLEAR(clear_module_state->__pyx_n_s_sink_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_size);
  Py_CLEAR(clear_module_state->__pyx_n_u_source);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_s_split);
  Py_CLEAR(clear_module_state->__pyx_n_s_src);
  Py_CLEAR(clear_module_state->__pyx_n_s_st);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_state_from);
  Py_CLEAR(clear_module_state->__pyx_n_s_state_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_state_number);
  Py_CLEAR(clear_module_state->__pyx_n_s_state_to);
  Py_CLEAR(clear_module_state->__pyx_n_s_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_stderr);
  Py_CLEAR(clear_module_state->__pyx_n_s_stdin);
  Py_CLEAR(clear_module_state->__pyx_n_s_stdout);
  Py_CLEAR(clear_module_state->__pyx_n_s_store);
  Py_CLEAR(clear_module_state->__pyx_n_s_store_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_string_to_symbol);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_strip);
  Py_CLEAR(clear_module_state->__pyx_n_u_style);
  Py_CLEAR(clear_module_state->__pyx_n_s_subprocess);
  Py_CLEAR(clear_module_state->__pyx_n_s_subset_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_symb);
  Py_CLEAR(clear_module_state->__pyx_n_s_symbol);
  Py_CLEAR(clear_module_state->__pyx_n_u_symbol);
  Py_CLEAR(clear_module_state->__pyx_n_s_symbol_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_symbols);
  Py_CLEAR(clear_module_state->__pyx_n_s_t);
  Py_CLEAR(clear_module_state->__pyx_n_u_target);
  Py_CLEAR(clear_module_state->__pyx_n_s_targets);
  Py_CLEAR(clear_module_state->__pyx_n_s_terminate);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_tgt);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_timeout);
  Py_CLEAR(clear_module_state->__pyx_n_s_to_dataframe);
  Py_CLEAR(clear_module_state->__pyx_n_s_to_dot_file);
  Py_CLEAR(clear_module_state->__pyx_n_s_to_dot_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_to_networkx_graph);
  Py_CLEAR(clear_module_state->__pyx_n_s_tr);
  Py_CLEAR(clear_module_state->__pyx_n_s_trans);
  Py_CLEAR(clear_module_state->__pyx_kp_u_transitions);
  Py_CLEAR(clear_module_state->__pyx_n_s_transitions_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_transitions_list);
  Py_CLEAR(clear_module_state->__pyx_n_s_translate_symbol);
  Py_CLEAR(clear_module_state->__pyx_n_s_transsymbols);
  Py_CLEAR(clear_module_state->__pyx_n_s_trim);
  Py_CLEAR(clear_module_state->__pyx_n_s_trim_input);
  Py_CLEAR(clear_module_state->__pyx_n_s_trim_with_state_map);
  Py_CLEAR(clear_module_state->__pyx_n_u_true);
  Py_CLEAR(clear_module_state->__pyx_kp_s_tuple_Nfa_dict_str_str_dict_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_unify_final);
  Py_CLEAR(clear_module_state->__pyx_n_s_unify_initial);
  Py_CLEAR(clear_module_state->__pyx_n_s_union);
  Py_CLEAR(clear_module_state->__pyx_n_s_unpiped_commands);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_epsilon);
  Py_CLEAR(clear_module_state->__pyx_n_s_useful_states);
  Py_CLEAR(clear_module_state->__pyx_n_s_useful_states_bool_vec);
  Py_CLEAR(clear_module_state->__pyx_kp_u_utf_8);
  Py_CLEAR(clear_module_state->__pyx_n_s_v);
  Py_CLEAR(clear_module_state->__pyx_n_s_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_wait);
  Py_CLEAR(clear_module_state->__pyx_n_u_width);
  Py_CLEAR(clear_module_state->__pyx_kp_u_without_specified_alphabet);
  Py_CLEAR(clear_module_state->__pyx_n_s_word);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__18);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__29);
  Py_CLEAR(clear_module_state->__pyx_tuple__34);
  Py_CLEAR(clear_module_state->__pyx_tuple__36);
  Py_CLEAR(clear_module_state->__pyx_tuple__47);
  Py_CLEAR(clear_module_state->__pyx_tuple__49);
  Py_CLEAR(clear_module_state->__pyx_tuple__51);
  Py_CLEAR(clear_module_state->__pyx_tuple__53);
  Py_CLEAR(clear_module_state->__pyx_tuple__59);
  Py_CLEAR(clear_module_state->__pyx_tuple__60);
  Py_CLEAR(clear_module_state->__pyx_tuple__62);
  Py_CLEAR(clear_module_state->__pyx_tuple__64);
  Py_CLEAR(clear_module_state->__pyx_tuple__66);
  Py_CLEAR(clear_module_state->__pyx_tuple__68);
  Py_CLEAR(clear_module_state->__pyx_tuple__70);
  Py_CLEAR(clear_module_state->__pyx_tuple__74);
  Py_CLEAR(clear_module_state->__pyx_tuple__76);
  Py_CLEAR(clear_module_state->__pyx_tuple__78);
  Py_CLEAR(clear_module_state->__pyx_tuple__81);
  Py_CLEAR(clear_module_state->__pyx_tuple__83);
  Py_CLEAR(clear_module_state->__pyx_tuple__84);
  Py_CLEAR(clear_module_state->__pyx_tuple__86);
  Py_CLEAR(clear_module_state->__pyx_tuple__87);
  Py_CLEAR(clear_module_state->__pyx_tuple__89);
  Py_CLEAR(clear_module_state->__pyx_tuple__91);
  Py_CLEAR(clear_module_state->__pyx_tuple__93);
  Py_CLEAR(clear_module_state->__pyx_tuple__95);
  Py_CLEAR(clear_module_state->__pyx_tuple__97);
  Py_CLEAR(clear_module_state->__pyx_tuple__99);
  Py_CLEAR(clear_module_state->__pyx_codeobj__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__101);
  Py_CLEAR(clear_module_state->__pyx_tuple__103);
  Py_CLEAR(clear_module_state->__pyx_tuple__104);
  Py_CLEAR(clear_module_state->__pyx_tuple__106);
  Py_CLEAR(clear_module_state->__pyx_tuple__108);
  Py_CLEAR(clear_module_state->__pyx_tuple__110);
  Py_CLEAR(clear_module_state->__pyx_tuple__112);
  Py_CLEAR(clear_module_state->__pyx_tuple__115);
  Py_CLEAR(clear_module_state->__pyx_tuple__117);
  Py_CLEAR(clear_module_state->__pyx_tuple__119);
  Py_CLEAR(clear_module_state->__pyx_tuple__122);
  Py_CLEAR(clear_module_state->__pyx_tuple__124);
  Py_CLEAR(clear_module_state->__pyx_tuple__125);
  Py_CLEAR(clear_module_state->__pyx_tuple__127);
  Py_CLEAR(clear_module_state->__pyx_tuple__129);
  Py_CLEAR(clear_module_state->__pyx_tuple__131);
  Py_CLEAR(clear_module_state->__pyx_tuple__133);
  Py_CLEAR(clear_module_state->__pyx_tuple__135);
  Py_CLEAR(clear_module_state->__pyx_tuple__137);
  Py_CLEAR(clear_module_state->__pyx_tuple__138);
  Py_CLEAR(clear_module_state->__pyx_tuple__140);
  Py_CLEAR(clear_module_state->__pyx_tuple__142);
  Py_CLEAR(clear_module_state->__pyx_tuple__144);
  Py_CLEAR(clear_module_state->__pyx_tuple__146);
  Py_CLEAR(clear_module_state->__pyx_tuple__149);
  Py_CLEAR(clear_module_state->__pyx_tuple__151);
  Py_CLEAR(clear_module_state->__pyx_tuple__153);
  Py_CLEAR(clear_module_state->__pyx_tuple__157);
  Py_CLEAR(clear_module_state->__pyx_tuple__159);
  Py_CLEAR(clear_module_state->__pyx_tuple__161);
  Py_CLEAR(clear_module_state->__pyx_tuple__162);
  Py_CLEAR(clear_module_state->__pyx_codeobj__17);
  Py_CLEAR(clear_module_state->__pyx_codeobj__19);
  Py_CLEAR(clear_module_state->__pyx_codeobj__21);
  Py_CLEAR(clear_module_state->__pyx_codeobj__23);
  Py_CLEAR(clear_module_state->__pyx_codeobj__24);
  Py_CLEAR(clear_module_state->__pyx_codeobj__25);
  Py_CLEAR(clear_module_state->__pyx_codeobj__26);
  Py_CLEAR(clear_module_state->__pyx_codeobj__27);
  Py_CLEAR(clear_module_state->__pyx_codeobj__28);
  Py_CLEAR(clear_module_state->__pyx_codeobj__30);
  Py_CLEAR(clear_module_state->__pyx_codeobj__31);
  Py_CLEAR(clear_module_state->__pyx_codeobj__32);
  Py_CLEAR(clear_module_state->__pyx_codeobj__33);
  Py_CLEAR(clear_module_state->__pyx_codeobj__35);
  Py_CLEAR(clear_module_state->__pyx_codeobj__37);
  Py_CLEAR(clear_module_state->__pyx_codeobj__38);
  Py_CLEAR(clear_module_state->__pyx_codeobj__39);
  Py_CLEAR(clear_module_state->__pyx_codeobj__40);
  Py_CLEAR(clear_module_state->__pyx_codeobj__41);
  Py_CLEAR(clear_module_state->__pyx_codeobj__42);
  Py_CLEAR(clear_module_state->__pyx_codeobj__43);
  Py_CLEAR(clear_module_state->__pyx_codeobj__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__45);
  Py_CLEAR(clear_module_state->__pyx_codeobj__46);
  Py_CLEAR(clear_module_state->__pyx_codeobj__48);
  Py_CLEAR(clear_module_state->__pyx_codeobj__50);
  Py_CLEAR(clear_module_state->__pyx_codeobj__52);
  Py_CLEAR(clear_module_state->__pyx_codeobj__54);
  Py_CLEAR(clear_module_state->__pyx_codeobj__55);
  Py_CLEAR(clear_module_state->__pyx_codeobj__56);
  Py_CLEAR(clear_module_state->__pyx_codeobj__57);
  Py_CLEAR(clear_module_state->__pyx_codeobj__58);
  Py_CLEAR(clear_module_state->__pyx_codeobj__61);
  Py_CLEAR(clear_module_state->__pyx_codeobj__63);
  Py_CLEAR(clear_module_state->__pyx_codeobj__65);
  Py_CLEAR(clear_module_state->__pyx_codeobj__67);
  Py_CLEAR(clear_module_state->__pyx_codeobj__69);
  Py_CLEAR(clear_module_state->__pyx_codeobj__71);
  Py_CLEAR(clear_module_state->__pyx_codeobj__72);
  Py_CLEAR(clear_module_state->__pyx_codeobj__73);
  Py_CLEAR(clear_module_state->__pyx_codeobj__75);
  Py_CLEAR(clear_module_state->__pyx_codeobj__77);
  Py_CLEAR(clear_module_state->__pyx_codeobj__79);
  Py_CLEAR(clear_module_state->__pyx_codeobj__82);
  Py_CLEAR(clear_module_state->__pyx_codeobj__85);
  Py_CLEAR(clear_module_state->__pyx_codeobj__88);
  Py_CLEAR(clear_module_state->__pyx_codeobj__90);
  Py_CLEAR(clear_module_state->__pyx_codeobj__92);
  Py_CLEAR(clear_module_state->__pyx_codeobj__94);
  Py_CLEAR(clear_module_state->__pyx_codeobj__96);
  Py_CLEAR(clear_module_state->__pyx_codeobj__98);
  Py_CLEAR(clear_module_state->__pyx_codeobj__100);
  Py_CLEAR(clear_module_state->__pyx_codeobj__102);
  Py_CLEAR(clear_module_state->__pyx_codeobj__105);
  Py_CLEAR(clear_module_state->__pyx_codeobj__107);
  Py_CLEAR(clear_module_state->__pyx_codeobj__109);
  Py_CLEAR(clear_module_state->__pyx_codeobj__111);
  Py_CLEAR(clear_module_state->__pyx_codeobj__113);
  Py_CLEAR(clear_module_state->__pyx_codeobj__114);
  Py_CLEAR(clear_module_state->__pyx_codeobj__116);
  Py_CLEAR(clear_module_state->__pyx_codeobj__118);
  Py_CLEAR(clear_module_state->__pyx_codeobj__120);
  Py_CLEAR(clear_module_state->__pyx_codeobj__121);
  Py_CLEAR(clear_module_state->__pyx_codeobj__123);
  Py_CLEAR(clear_module_state->__pyx_codeobj__126);
  Py_CLEAR(clear_module_state->__pyx_codeobj__128);
  Py_CLEAR(clear_module_state->__pyx_codeobj__130);
  Py_CLEAR(clear_module_state->__pyx_codeobj__132);
  Py_CLEAR(clear_module_state->__pyx_codeobj__134);
  Py_CLEAR(clear_module_state->__pyx_codeobj__136);
  Py_CLEAR(clear_module_state->__pyx_codeobj__139);
  Py_CLEAR(clear_module_state->__pyx_codeobj__141);
  Py_CLEAR(clear_module_state->__pyx_codeobj__143);
  Py_CLEAR(clear_module_state->__pyx_codeobj__145);
  Py_CLEAR(clear_module_state->__pyx_codeobj__147);
  Py_CLEAR(clear_module_state->__pyx_codeobj__148);
  Py_CLEAR(clear_module_state->__pyx_codeobj__150);
  Py_CLEAR(clear_module_state->__pyx_codeobj__152);
  Py_CLEAR(clear_module_state->__pyx_codeobj__154);
  Py_CLEAR(clear_module_state->__pyx_codeobj__155);
  Py_CLEAR(clear_module_state->__pyx_codeobj__156);
  Py_CLEAR(clear_module_state->__pyx_codeobj__158);
  Py_CLEAR(clear_module_state->__pyx_codeobj__160);
  Py_CLEAR(clear_module_state->__pyx_codeobj__163);
  Py_CLEAR(clear_module_state->__pyx_codeobj__164);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7libmata_5utils_BinaryRelation);
  Py_VISIT(traverse_module_state->__pyx_ptype_7libmata_9alphabets_Alphabet);
  Py_VISIT(traverse_module_state->__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  Py_VISIT(traverse_module_state->__pyx_type_7libmata_3nfa_3nfa_Nfa);
  Py_VISIT(traverse_module_state->__pyx_ptype_7libmata_3nfa_3nfa_Trans);
  Py_VISIT(traverse_module_state->__pyx_type_7libmata_3nfa_3nfa_Trans);
  Py_VISIT(traverse_module_state->__pyx_ptype_7libmata_3nfa_3nfa_Run);
  Py_VISIT(traverse_module_state->__pyx_type_7libmata_3nfa_3nfa_Run);
  Py_VISIT(traverse_module_state->__pyx_ptype_7libmata_3nfa_3nfa_Move);
  Py_VISIT(traverse_module_state->__pyx_type_7libmata_3nfa_3nfa_Move);
  Py_VISIT(traverse_module_state->__pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate);
  Py_VISIT(traverse_module_state->__pyx_type_7libmata_3nfa_3nfa___pyx_scope_struct__iterate);
  Py_VISIT(traverse_module_state->__pyx_kp_u_);
  Py_VISIT(traverse_module_state->__pyx_kp_u_0_3);
  Py_VISIT(traverse_module_state->__pyx_kp_u_12);
  Py_VISIT(traverse_module_state->__pyx_kp_u_1_5);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_CalledProcessError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_translate_symbol);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Courier_Bold);
  Py_VISIT(traverse_module_state->__pyx_n_s_DataFrame);
  Py_VISIT(traverse_module_state->__pyx_n_s_DiGraph);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Encodes_word_based_on_a_string_t);
  Py_VISIT(traverse_module_state->__pyx_kp_u_For_a_given_path_set_of_states_r);
  Py_VISIT(traverse_module_state->__pyx_n_s_G);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Move);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Move_None);
  Py_VISIT(traverse_module_state->__pyx_n_s_Move___neq);
  Py_VISIT(traverse_module_state->__pyx_n_s_Move___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Move___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_accepts_epsilon);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_add_new_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_add_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_add_transition);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_add_transition_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_clear);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_clear_final);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_clear_initial);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_complement);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_compute_relation);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_concatenate);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_concatenate_with_result_stat);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_determinize);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_determinize_with_subset_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_encode_word);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Nfa_encode_word_line_1139);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_equivalence_check);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_get_epsilon_transitions);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_get_num_of_trans);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_get_one_letter_aut);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_get_reachable_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_get_symbols);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_get_terminating_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_get_trans_as_sequence);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_get_trans_from_state_as_sequ);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_get_transitions_from_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_get_transitions_to_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_get_useful_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_get_word_for_path);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Nfa_get_word_for_path_line_1122);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_has_final_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_has_initial_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_has_transition);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_intersection);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_intersection_with_product_ma);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_is_complete);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_is_deterministic);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_is_epsilon);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_is_in_lang);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_is_included);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_is_included_with_cex);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_is_lang_empty);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_is_prefix_in_lang);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_is_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_is_universal);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_iterate);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_make_complete);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_make_final_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_make_final_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_make_initial_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_make_initial_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_minimize);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_post_map_of);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_post_of);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_reduce_with_state_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_remove_epsilon);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_remove_epsilon_inplace);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_remove_final_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_remove_initial_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_remove_trans);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_remove_trans_raw);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_revert);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_to_dataframe);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_to_dot_file);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_to_dot_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_to_networkx_graph);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_trim);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_trim_with_state_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_unify_final);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_unify_initial);
  Py_VISIT(traverse_module_state->__pyx_n_s_Nfa_union);
  Py_VISIT(traverse_module_state->__pyx_n_s_PIPE);
  Py_VISIT(traverse_module_state->__pyx_n_s_Popen);
  Py_VISIT(traverse_module_state->__pyx_n_s_Run);
  Py_VISIT(traverse_module_state->__pyx_n_s_Run___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Run___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_STDOUT);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimeoutExpired);
  Py_VISIT(traverse_module_state->__pyx_n_s_Trans);
  Py_VISIT(traverse_module_state->__pyx_n_s_Trans___neq);
  Py_VISIT(traverse_module_state->__pyx_n_s_Trans___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Trans___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_kp_b__10);
  Py_VISIT(traverse_module_state->__pyx_kp_u__15);
  Py_VISIT(traverse_module_state->__pyx_n_s__16);
  Py_VISIT(traverse_module_state->__pyx_n_s__165);
  Py_VISIT(traverse_module_state->__pyx_kp_u__2);
  Py_VISIT(traverse_module_state->__pyx_kp_u__3);
  Py_VISIT(traverse_module_state->__pyx_kp_u__4);
  Py_VISIT(traverse_module_state->__pyx_kp_u__5);
  Py_VISIT(traverse_module_state->__pyx_kp_u__6);
  Py_VISIT(traverse_module_state->__pyx_kp_u__8);
  Py_VISIT(traverse_module_state->__pyx_n_s__80);
  Py_VISIT(traverse_module_state->__pyx_kp_u__9);
  Py_VISIT(traverse_module_state->__pyx_n_s_accepts_epsilon);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_edge);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_new_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_transition);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_transition_object);
  Py_VISIT(traverse_module_state->__pyx_n_u_algorithm);
  Py_VISIT(traverse_module_state->__pyx_n_s_alphabet);
  Py_VISIT(traverse_module_state->__pyx_n_u_alphabet);
  Py_VISIT(traverse_module_state->__pyx_n_u_antichains);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_aut);
  Py_VISIT(traverse_module_state->__pyx_kp_u_aut_dot);
  Py_VISIT(traverse_module_state->__pyx_n_s_bool);
  Py_VISIT(traverse_module_state->__pyx_n_s_bool_vec);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_alphabet);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_bool_vec);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_epsilon_transitions_iter);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_lhs_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_product_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_rhs_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_transition);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_transitions);
  Py_VISIT(traverse_module_state->__pyx_kp_u_captured_stderr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_captured_stdout);
  Py_VISIT(traverse_module_state->__pyx_n_s_check_results);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_u_classical);
  Py_VISIT(traverse_module_state->__pyx_n_s_clear);
  Py_VISIT(traverse_module_state->__pyx_n_s_clear_final);
  Py_VISIT(traverse_module_state->__pyx_n_s_clear_initial);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_cls);
  Py_VISIT(traverse_module_state->__pyx_n_s_cmd);
  Py_VISIT(traverse_module_state->__pyx_n_s_cmd_no);
  Py_VISIT(traverse_module_state->__pyx_n_s_cmderr);
  Py_VISIT(traverse_module_state->__pyx_n_s_cmdout);
  Py_VISIT(traverse_module_state->__pyx_n_u_color);
  Py_VISIT(traverse_module_state->__pyx_n_s_columns);
  Py_VISIT(traverse_module_state->__pyx_n_s_communicate);
  Py_VISIT(traverse_module_state->__pyx_n_s_complement);
  Py_VISIT(traverse_module_state->__pyx_n_s_compute_relation);
  Py_VISIT(traverse_module_state->__pyx_n_s_concatenate);
  Py_VISIT(traverse_module_state->__pyx_n_s_concatenate_with_result_state_ma);
  Py_VISIT(traverse_module_state->__pyx_n_u_darkblue);
  Py_VISIT(traverse_module_state->__pyx_n_s_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_decode);
  Py_VISIT(traverse_module_state->__pyx_n_s_determinize);
  Py_VISIT(traverse_module_state->__pyx_n_s_determinize_with_subset_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_u_direction);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_divisible_by);
  Py_VISIT(traverse_module_state->__pyx_kp_u_dot_O_T);
  Py_VISIT(traverse_module_state->__pyx_n_u_edge_style);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_encode_word);
  Py_VISIT(traverse_module_state->__pyx_n_s_encoding);
  Py_VISIT(traverse_module_state->__pyx_n_s_end);
  Py_VISIT(traverse_module_state->__pyx_n_s_epsilon);
  Py_VISIT(traverse_module_state->__pyx_n_s_epsilon_transitions);
  Py_VISIT(traverse_module_state->__pyx_n_s_equivalence_check);
  Py_VISIT(traverse_module_state->__pyx_n_s_err);
  Py_VISIT(traverse_module_state->__pyx_kp_u_error_while_dot_file);
  Py_VISIT(traverse_module_state->__pyx_n_s_executed_command);
  Py_VISIT(traverse_module_state->__pyx_n_u_false);
  Py_VISIT(traverse_module_state->__pyx_n_u_fillcolor);
  Py_VISIT(traverse_module_state->__pyx_n_u_filled);
  Py_VISIT(traverse_module_state->__pyx_kp_u_final_states);
  Py_VISIT(traverse_module_state->__pyx_n_u_fixedsize);
  Py_VISIT(traverse_module_state->__pyx_n_u_fontname);
  Py_VISIT(traverse_module_state->__pyx_n_u_fontsize);
  Py_VISIT(traverse_module_state->__pyx_n_s_format);
  Py_VISIT(traverse_module_state->__pyx_n_u_forward);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_alphabet_symbols);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_elements_from_bool_vec);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_epsilon_transitions);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_num_of_trans);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_one_letter_aut);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_reachable_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_symbols);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_terminating_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_trans_as_sequence);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_trans_from_state_as_sequence);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_transitions_from_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_transitions_to_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_useful_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_word_for_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_graphviz_command);
  Py_VISIT(traverse_module_state->__pyx_n_s_has_final_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_has_initial_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_has_transition);
  Py_VISIT(traverse_module_state->__pyx_n_u_height);
  Py_VISIT(traverse_module_state->__pyx_n_s_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_kp_u_initial_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_input);
  Py_VISIT(traverse_module_state->__pyx_n_s_input_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_intersection);
  Py_VISIT(traverse_module_state->__pyx_n_s_intersection_with_product_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_complete);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_deterministic);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_epsilon);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_in_lang);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_included);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_included_with_cex);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_lang_empty);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_prefix_in_lang);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_universal);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_it);
  Py_VISIT(traverse_module_state->__pyx_n_s_items);
  Py_VISIT(traverse_module_state->__pyx_n_s_iterate);
  Py_VISIT(traverse_module_state->__pyx_n_s_iterator);
  Py_VISIT(traverse_module_state->__pyx_n_s_iteritems);
  Py_VISIT(traverse_module_state->__pyx_n_s_j);
  Py_VISIT(traverse_module_state->__pyx_n_s_k);
  Py_VISIT(traverse_module_state->__pyx_n_s_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_kwargs);
  Py_VISIT(traverse_module_state->__pyx_n_s_label);
  Py_VISIT(traverse_module_state->__pyx_n_s_lhs);
  Py_VISIT(traverse_module_state->__pyx_n_s_lhs_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_libmata_nfa_nfa);
  Py_VISIT(traverse_module_state->__pyx_kp_s_libmata_nfa_nfa_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_u_lightsteelblue);
  Py_VISIT(traverse_module_state->__pyx_kp_s_list_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_make_complete);
  Py_VISIT(traverse_module_state->__pyx_n_s_make_final_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_make_final_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_make_initial_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_make_initial_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_map);
  Py_VISIT(traverse_module_state->__pyx_n_u_midnightblue);
  Py_VISIT(traverse_module_state->__pyx_n_s_minimize);
  Py_VISIT(traverse_module_state->__pyx_n_u_minimize);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_neq);
  Py_VISIT(traverse_module_state->__pyx_n_s_networkx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_VISIT(traverse_module_state->__pyx_n_u_node_style);
  Py_VISIT(traverse_module_state->__pyx_n_s_nx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_nx_Graph);
  Py_VISIT(traverse_module_state->__pyx_n_s_objects);
  Py_VISIT(traverse_module_state->__pyx_n_s_one_letter_aut);
  Py_VISIT(traverse_module_state->__pyx_n_s_other);
  Py_VISIT(traverse_module_state->__pyx_n_s_output);
  Py_VISIT(traverse_module_state->__pyx_n_s_output_file);
  Py_VISIT(traverse_module_state->__pyx_n_s_output_format);
  Py_VISIT(traverse_module_state->__pyx_n_s_output_stream);
  Py_VISIT(traverse_module_state->__pyx_n_s_pandas);
  Py_VISIT(traverse_module_state->__pyx_kp_s_pandas_DataFrame);
  Py_VISIT(traverse_module_state->__pyx_n_s_params);
  Py_VISIT(traverse_module_state->__pyx_n_s_path);
  Py_VISIT(traverse_module_state->__pyx_n_u_pdf);
  Py_VISIT(traverse_module_state->__pyx_n_u_penwidth);
  Py_VISIT(traverse_module_state->__pyx_n_s_piped_command);
  Py_VISIT(traverse_module_state->__pyx_n_s_post_map_of);
  Py_VISIT(traverse_module_state->__pyx_n_s_post_of);
  Py_VISIT(traverse_module_state->__pyx_n_s_preserve_epsilon);
  Py_VISIT(traverse_module_state->__pyx_n_s_print);
  Py_VISIT(traverse_module_state->__pyx_n_s_process);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_quiet);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_u_red);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_with_state_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_relation);
  Py_VISIT(traverse_module_state->__pyx_n_u_relation);
  Py_VISIT(traverse_module_state->__pyx_n_s_relation_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove_epsilon);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove_epsilon_inplace);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove_final_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove_initial_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove_trans);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove_trans_raw);
  Py_VISIT(traverse_module_state->__pyx_n_s_resize);
  Py_VISIT(traverse_module_state->__pyx_n_s_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_return);
  Py_VISIT(traverse_module_state->__pyx_n_s_return_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_returncode);
  Py_VISIT(traverse_module_state->__pyx_n_s_revert);
  Py_VISIT(traverse_module_state->__pyx_n_s_rhs);
  Py_VISIT(traverse_module_state->__pyx_n_s_rhs_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_run);
  Py_VISIT(traverse_module_state->__pyx_n_s_run_safely_external_command);
  Py_VISIT(traverse_module_state->__pyx_n_s_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_set);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_shell);
  Py_VISIT(traverse_module_state->__pyx_n_s_shlex);
  Py_VISIT(traverse_module_state->__pyx_n_u_simulation);
  Py_VISIT(traverse_module_state->__pyx_n_s_sink_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_size);
  Py_VISIT(traverse_module_state->__pyx_n_u_source);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_s_split);
  Py_VISIT(traverse_module_state->__pyx_n_s_src);
  Py_VISIT(traverse_module_state->__pyx_n_s_st);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_state_from);
  Py_VISIT(traverse_module_state->__pyx_n_s_state_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_state_number);
  Py_VISIT(traverse_module_state->__pyx_n_s_state_to);
  Py_VISIT(traverse_module_state->__pyx_n_s_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_stderr);
  Py_VISIT(traverse_module_state->__pyx_n_s_stdin);
  Py_VISIT(traverse_module_state->__pyx_n_s_stdout);
  Py_VISIT(traverse_module_state->__pyx_n_s_store);
  Py_VISIT(traverse_module_state->__pyx_n_s_store_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_string_to_symbol);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_strip);
  Py_VISIT(traverse_module_state->__pyx_n_u_style);
  Py_VISIT(traverse_module_state->__pyx_n_s_subprocess);
  Py_VISIT(traverse_module_state->__pyx_n_s_subset_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_symb);
  Py_VISIT(traverse_module_state->__pyx_n_s_symbol);
  Py_VISIT(traverse_module_state->__pyx_n_u_symbol);
  Py_VISIT(traverse_module_state->__pyx_n_s_symbol_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_symbols);
  Py_VISIT(traverse_module_state->__pyx_n_s_t);
  Py_VISIT(traverse_module_state->__pyx_n_u_target);
  Py_VISIT(traverse_module_state->__pyx_n_s_targets);
  Py_VISIT(traverse_module_state->__pyx_n_s_terminate);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_tgt);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_timeout);
  Py_VISIT(traverse_module_state->__pyx_n_s_to_dataframe);
  Py_VISIT(traverse_module_state->__pyx_n_s_to_dot_file);
  Py_VISIT(traverse_module_state->__pyx_n_s_to_dot_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_to_networkx_graph);
  Py_VISIT(traverse_module_state->__pyx_n_s_tr);
  Py_VISIT(traverse_module_state->__pyx_n_s_trans);
  Py_VISIT(traverse_module_state->__pyx_kp_u_transitions);
  Py_VISIT(traverse_module_state->__pyx_n_s_transitions_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_transitions_list);
  Py_VISIT(traverse_module_state->__pyx_n_s_translate_symbol);
  Py_VISIT(traverse_module_state->__pyx_n_s_transsymbols);
  Py_VISIT(traverse_module_state->__pyx_n_s_trim);
  Py_VISIT(traverse_module_state->__pyx_n_s_trim_input);
  Py_VISIT(traverse_module_state->__pyx_n_s_trim_with_state_map);
  Py_VISIT(traverse_module_state->__pyx_n_u_true);
  Py_VISIT(traverse_module_state->__pyx_kp_s_tuple_Nfa_dict_str_str_dict_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_unify_final);
  Py_VISIT(traverse_module_state->__pyx_n_s_unify_initial);
  Py_VISIT(traverse_module_state->__pyx_n_s_union);
  Py_VISIT(traverse_module_state->__pyx_n_s_unpiped_commands);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_epsilon);
  Py_VISIT(traverse_module_state->__pyx_n_s_useful_states);
  Py_VISIT(traverse_module_state->__pyx_n_s_useful_states_bool_vec);
  Py_VISIT(traverse_module_state->__pyx_kp_u_utf_8);
  Py_VISIT(traverse_module_state->__pyx_n_s_v);
  Py_VISIT(traverse_module_state->__pyx_n_s_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_wait);
  Py_VISIT(traverse_module_state->__pyx_n_u_width);
  Py_VISIT(traverse_module_state->__pyx_kp_u_without_specified_alphabet);
  Py_VISIT(traverse_module_state->__pyx_n_s_word);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__18);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__29);
  Py_VISIT(traverse_module_state->__pyx_tuple__34);
  Py_VISIT(traverse_module_state->__pyx_tuple__36);
  Py_VISIT(traverse_module_state->__pyx_tuple__47);
  Py_VISIT(traverse_module_state->__pyx_tuple__49);
  Py_VISIT(traverse_module_state->__pyx_tuple__51);
  Py_VISIT(traverse_module_state->__pyx_tuple__53);
  Py_VISIT(traverse_module_state->__pyx_tuple__59);
  Py_VISIT(traverse_module_state->__pyx_tuple__60);
  Py_VISIT(traverse_module_state->__pyx_tuple__62);
  Py_VISIT(traverse_module_state->__pyx_tuple__64);
  Py_VISIT(traverse_module_state->__pyx_tuple__66);
  Py_VISIT(traverse_module_state->__pyx_tuple__68);
  Py_VISIT(traverse_module_state->__pyx_tuple__70);
  Py_VISIT(traverse_module_state->__pyx_tuple__74);
  Py_VISIT(traverse_module_state->__pyx_tuple__76);
  Py_VISIT(traverse_module_state->__pyx_tuple__78);
  Py_VISIT(traverse_module_state->__pyx_tuple__81);
  Py_VISIT(traverse_module_state->__pyx_tuple__83);
  Py_VISIT(traverse_module_state->__pyx_tuple__84);
  Py_VISIT(traverse_module_state->__pyx_tuple__86);
  Py_VISIT(traverse_module_state->__pyx_tuple__87);
  Py_VISIT(traverse_module_state->__pyx_tuple__89);
  Py_VISIT(traverse_module_state->__pyx_tuple__91);
  Py_VISIT(traverse_module_state->__pyx_tuple__93);
  Py_VISIT(traverse_module_state->__pyx_tuple__95);
  Py_VISIT(traverse_module_state->__pyx_tuple__97);
  Py_VISIT(traverse_module_state->__pyx_tuple__99);
  Py_VISIT(traverse_module_state->__pyx_codeobj__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__101);
  Py_VISIT(traverse_module_state->__pyx_tuple__103);
  Py_VISIT(traverse_module_state->__pyx_tuple__104);
  Py_VISIT(traverse_module_state->__pyx_tuple__106);
  Py_VISIT(traverse_module_state->__pyx_tuple__108);
  Py_VISIT(traverse_module_state->__pyx_tuple__110);
  Py_VISIT(traverse_module_state->__pyx_tuple__112);
  Py_VISIT(traverse_module_state->__pyx_tuple__115);
  Py_VISIT(traverse_module_state->__pyx_tuple__117);
  Py_VISIT(traverse_module_state->__pyx_tuple__119);
  Py_VISIT(traverse_module_state->__pyx_tuple__122);
  Py_VISIT(traverse_module_state->__pyx_tuple__124);
  Py_VISIT(traverse_module_state->__pyx_tuple__125);
  Py_VISIT(traverse_module_state->__pyx_tuple__127);
  Py_VISIT(traverse_module_state->__pyx_tuple__129);
  Py_VISIT(traverse_module_state->__pyx_tuple__131);
  Py_VISIT(traverse_module_state->__pyx_tuple__133);
  Py_VISIT(traverse_module_state->__pyx_tuple__135);
  Py_VISIT(traverse_module_state->__pyx_tuple__137);
  Py_VISIT(traverse_module_state->__pyx_tuple__138);
  Py_VISIT(traverse_module_state->__pyx_tuple__140);
  Py_VISIT(traverse_module_state->__pyx_tuple__142);
  Py_VISIT(traverse_module_state->__pyx_tuple__144);
  Py_VISIT(traverse_module_state->__pyx_tuple__146);
  Py_VISIT(traverse_module_state->__pyx_tuple__149);
  Py_VISIT(traverse_module_state->__pyx_tuple__151);
  Py_VISIT(traverse_module_state->__pyx_tuple__153);
  Py_VISIT(traverse_module_state->__pyx_tuple__157);
  Py_VISIT(traverse_module_state->__pyx_tuple__159);
  Py_VISIT(traverse_module_state->__pyx_tuple__161);
  Py_VISIT(traverse_module_state->__pyx_tuple__162);
  Py_VISIT(traverse_module_state->__pyx_codeobj__17);
  Py_VISIT(traverse_module_state->__pyx_codeobj__19);
  Py_VISIT(traverse_module_state->__pyx_codeobj__21);
  Py_VISIT(traverse_module_state->__pyx_codeobj__23);
  Py_VISIT(traverse_module_state->__pyx_codeobj__24);
  Py_VISIT(traverse_module_state->__pyx_codeobj__25);
  Py_VISIT(traverse_module_state->__pyx_codeobj__26);
  Py_VISIT(traverse_module_state->__pyx_codeobj__27);
  Py_VISIT(traverse_module_state->__pyx_codeobj__28);
  Py_VISIT(traverse_module_state->__pyx_codeobj__30);
  Py_VISIT(traverse_module_state->__pyx_codeobj__31);
  Py_VISIT(traverse_module_state->__pyx_codeobj__32);
  Py_VISIT(traverse_module_state->__pyx_codeobj__33);
  Py_VISIT(traverse_module_state->__pyx_codeobj__35);
  Py_VISIT(traverse_module_state->__pyx_codeobj__37);
  Py_VISIT(traverse_module_state->__pyx_codeobj__38);
  Py_VISIT(traverse_module_state->__pyx_codeobj__39);
  Py_VISIT(traverse_module_state->__pyx_codeobj__40);
  Py_VISIT(traverse_module_state->__pyx_codeobj__41);
  Py_VISIT(traverse_module_state->__pyx_codeobj__42);
  Py_VISIT(traverse_module_state->__pyx_codeobj__43);
  Py_VISIT(traverse_module_state->__pyx_codeobj__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__45);
  Py_VISIT(traverse_module_state->__pyx_codeobj__46);
  Py_VISIT(traverse_module_state->__pyx_codeobj__48);
  Py_VISIT(traverse_module_state->__pyx_codeobj__50);
  Py_VISIT(traverse_module_state->__pyx_codeobj__52);
  Py_VISIT(traverse_module_state->__pyx_codeobj__54);
  Py_VISIT(traverse_module_state->__pyx_codeobj__55);
  Py_VISIT(traverse_module_state->__pyx_codeobj__56);
  Py_VISIT(traverse_module_state->__pyx_codeobj__57);
  Py_VISIT(traverse_module_state->__pyx_codeobj__58);
  Py_VISIT(traverse_module_state->__pyx_codeobj__61);
  Py_VISIT(traverse_module_state->__pyx_codeobj__63);
  Py_VISIT(traverse_module_state->__pyx_codeobj__65);
  Py_VISIT(traverse_module_state->__pyx_codeobj__67);
  Py_VISIT(traverse_module_state->__pyx_codeobj__69);
  Py_VISIT(traverse_module_state->__pyx_codeobj__71);
  Py_VISIT(traverse_module_state->__pyx_codeobj__72);
  Py_VISIT(traverse_module_state->__pyx_codeobj__73);
  Py_VISIT(traverse_module_state->__pyx_codeobj__75);
  Py_VISIT(traverse_module_state->__pyx_codeobj__77);
  Py_VISIT(traverse_module_state->__pyx_codeobj__79);
  Py_VISIT(traverse_module_state->__pyx_codeobj__82);
  Py_VISIT(traverse_module_state->__pyx_codeobj__85);
  Py_VISIT(traverse_module_state->__pyx_codeobj__88);
  Py_VISIT(traverse_module_state->__pyx_codeobj__90);
  Py_VISIT(traverse_module_state->__pyx_codeobj__92);
  Py_VISIT(traverse_module_state->__pyx_codeobj__94);
  Py_VISIT(traverse_module_state->__pyx_codeobj__96);
  Py_VISIT(traverse_module_state->__pyx_codeobj__98);
  Py_VISIT(traverse_module_state->__pyx_codeobj__100);
  Py_VISIT(traverse_module_state->__pyx_codeobj__102);
  Py_VISIT(traverse_module_state->__pyx_codeobj__105);
  Py_VISIT(traverse_module_state->__pyx_codeobj__107);
  Py_VISIT(traverse_module_state->__pyx_codeobj__109);
  Py_VISIT(traverse_module_state->__pyx_codeobj__111);
  Py_VISIT(traverse_module_state->__pyx_codeobj__113);
  Py_VISIT(traverse_module_state->__pyx_codeobj__114);
  Py_VISIT(traverse_module_state->__pyx_codeobj__116);
  Py_VISIT(traverse_module_state->__pyx_codeobj__118);
  Py_VISIT(traverse_module_state->__pyx_codeobj__120);
  Py_VISIT(traverse_module_state->__pyx_codeobj__121);
  Py_VISIT(traverse_module_state->__pyx_codeobj__123);
  Py_VISIT(traverse_module_state->__pyx_codeobj__126);
  Py_VISIT(traverse_module_state->__pyx_codeobj__128);
  Py_VISIT(traverse_module_state->__pyx_codeobj__130);
  Py_VISIT(traverse_module_state->__pyx_codeobj__132);
  Py_VISIT(traverse_module_state->__pyx_codeobj__134);
  Py_VISIT(traverse_module_state->__pyx_codeobj__136);
  Py_VISIT(traverse_module_state->__pyx_codeobj__139);
  Py_VISIT(traverse_module_state->__pyx_codeobj__141);
  Py_VISIT(traverse_module_state->__pyx_codeobj__143);
  Py_VISIT(traverse_module_state->__pyx_codeobj__145);
  Py_VISIT(traverse_module_state->__pyx_codeobj__147);
  Py_VISIT(traverse_module_state->__pyx_codeobj__148);
  Py_VISIT(traverse_module_state->__pyx_codeobj__150);
  Py_VISIT(traverse_module_state->__pyx_codeobj__152);
  Py_VISIT(traverse_module_state->__pyx_codeobj__154);
  Py_VISIT(traverse_module_state->__pyx_codeobj__155);
  Py_VISIT(traverse_module_state->__pyx_codeobj__156);
  Py_VISIT(traverse_module_state->__pyx_codeobj__158);
  Py_VISIT(traverse_module_state->__pyx_codeobj__160);
  Py_VISIT(traverse_module_state->__pyx_codeobj__163);
  Py_VISIT(traverse_module_state->__pyx_codeobj__164);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7libmata_5utils_BinaryRelation __pyx_mstate_global->__pyx_ptype_7libmata_5utils_BinaryRelation
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7libmata_9alphabets_Alphabet __pyx_mstate_global->__pyx_ptype_7libmata_9alphabets_Alphabet
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_7libmata_3nfa_3nfa_Nfa __pyx_mstate_global->__pyx_type_7libmata_3nfa_3nfa_Nfa
#define __pyx_type_7libmata_3nfa_3nfa_Trans __pyx_mstate_global->__pyx_type_7libmata_3nfa_3nfa_Trans
#define __pyx_type_7libmata_3nfa_3nfa_Run __pyx_mstate_global->__pyx_type_7libmata_3nfa_3nfa_Run
#define __pyx_type_7libmata_3nfa_3nfa_Move __pyx_mstate_global->__pyx_type_7libmata_3nfa_3nfa_Move
#define __pyx_type_7libmata_3nfa_3nfa___pyx_scope_struct__iterate __pyx_mstate_global->__pyx_type_7libmata_3nfa_3nfa___pyx_scope_struct__iterate
#endif
#define __pyx_ptype_7libmata_3nfa_3nfa_Nfa __pyx_mstate_global->__pyx_ptype_7libmata_3nfa_3nfa_Nfa
#define __pyx_ptype_7libmata_3nfa_3nfa_Trans __pyx_mstate_global->__pyx_ptype_7libmata_3nfa_3nfa_Trans
#define __pyx_ptype_7libmata_3nfa_3nfa_Run __pyx_mstate_global->__pyx_ptype_7libmata_3nfa_3nfa_Run
#define __pyx_ptype_7libmata_3nfa_3nfa_Move __pyx_mstate_global->__pyx_ptype_7libmata_3nfa_3nfa_Move
#define __pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate __pyx_mstate_global->__pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate
#define __pyx_kp_u_ __pyx_mstate_global->__pyx_kp_u_
#define __pyx_kp_u_0_3 __pyx_mstate_global->__pyx_kp_u_0_3
#define __pyx_kp_u_12 __pyx_mstate_global->__pyx_kp_u_12
#define __pyx_kp_u_1_5 __pyx_mstate_global->__pyx_kp_u_1_5
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_s_CalledProcessError __pyx_mstate_global->__pyx_n_s_CalledProcessError
#define __pyx_kp_u_Cannot_translate_symbol __pyx_mstate_global->__pyx_kp_u_Cannot_translate_symbol
#define __pyx_kp_u_Courier_Bold __pyx_mstate_global->__pyx_kp_u_Courier_Bold
#define __pyx_n_s_DataFrame __pyx_mstate_global->__pyx_n_s_DataFrame
#define __pyx_n_s_DiGraph __pyx_mstate_global->__pyx_n_s_DiGraph
#define __pyx_kp_u_Encodes_word_based_on_a_string_t __pyx_mstate_global->__pyx_kp_u_Encodes_word_based_on_a_string_t
#define __pyx_kp_u_For_a_given_path_set_of_states_r __pyx_mstate_global->__pyx_kp_u_For_a_given_path_set_of_states_r
#define __pyx_n_s_G __pyx_mstate_global->__pyx_n_s_G
#define __pyx_n_s_MemoryError __pyx_mstate_global->__pyx_n_s_MemoryError
#define __pyx_n_s_Move __pyx_mstate_global->__pyx_n_s_Move
#define __pyx_kp_s_Move_None __pyx_mstate_global->__pyx_kp_s_Move_None
#define __pyx_n_s_Move___neq __pyx_mstate_global->__pyx_n_s_Move___neq
#define __pyx_n_s_Move___reduce_cython __pyx_mstate_global->__pyx_n_s_Move___reduce_cython
#define __pyx_n_s_Move___setstate_cython __pyx_mstate_global->__pyx_n_s_Move___setstate_cython
#define __pyx_n_s_Nfa __pyx_mstate_global->__pyx_n_s_Nfa
#define __pyx_n_s_Nfa___reduce_cython __pyx_mstate_global->__pyx_n_s_Nfa___reduce_cython
#define __pyx_n_s_Nfa___setstate_cython __pyx_mstate_global->__pyx_n_s_Nfa___setstate_cython
#define __pyx_n_s_Nfa_accepts_epsilon __pyx_mstate_global->__pyx_n_s_Nfa_accepts_epsilon
#define __pyx_n_s_Nfa_add_new_state __pyx_mstate_global->__pyx_n_s_Nfa_add_new_state
#define __pyx_n_s_Nfa_add_state __pyx_mstate_global->__pyx_n_s_Nfa_add_state
#define __pyx_n_s_Nfa_add_transition __pyx_mstate_global->__pyx_n_s_Nfa_add_transition
#define __pyx_n_s_Nfa_add_transition_object __pyx_mstate_global->__pyx_n_s_Nfa_add_transition_object
#define __pyx_n_s_Nfa_clear __pyx_mstate_global->__pyx_n_s_Nfa_clear
#define __pyx_n_s_Nfa_clear_final __pyx_mstate_global->__pyx_n_s_Nfa_clear_final
#define __pyx_n_s_Nfa_clear_initial __pyx_mstate_global->__pyx_n_s_Nfa_clear_initial
#define __pyx_n_s_Nfa_complement __pyx_mstate_global->__pyx_n_s_Nfa_complement
#define __pyx_n_s_Nfa_compute_relation __pyx_mstate_global->__pyx_n_s_Nfa_compute_relation
#define __pyx_n_s_Nfa_concatenate __pyx_mstate_global->__pyx_n_s_Nfa_concatenate
#define __pyx_n_s_Nfa_concatenate_with_result_stat __pyx_mstate_global->__pyx_n_s_Nfa_concatenate_with_result_stat
#define __pyx_n_s_Nfa_determinize __pyx_mstate_global->__pyx_n_s_Nfa_determinize
#define __pyx_n_s_Nfa_determinize_with_subset_map __pyx_mstate_global->__pyx_n_s_Nfa_determinize_with_subset_map
#define __pyx_n_s_Nfa_encode_word __pyx_mstate_global->__pyx_n_s_Nfa_encode_word
#define __pyx_kp_u_Nfa_encode_word_line_1139 __pyx_mstate_global->__pyx_kp_u_Nfa_encode_word_line_1139
#define __pyx_n_s_Nfa_equivalence_check __pyx_mstate_global->__pyx_n_s_Nfa_equivalence_check
#define __pyx_n_s_Nfa_get_epsilon_transitions __pyx_mstate_global->__pyx_n_s_Nfa_get_epsilon_transitions
#define __pyx_n_s_Nfa_get_num_of_trans __pyx_mstate_global->__pyx_n_s_Nfa_get_num_of_trans
#define __pyx_n_s_Nfa_get_one_letter_aut __pyx_mstate_global->__pyx_n_s_Nfa_get_one_letter_aut
#define __pyx_n_s_Nfa_get_reachable_states __pyx_mstate_global->__pyx_n_s_Nfa_get_reachable_states
#define __pyx_n_s_Nfa_get_symbols __pyx_mstate_global->__pyx_n_s_Nfa_get_symbols
#define __pyx_n_s_Nfa_get_terminating_states __pyx_mstate_global->__pyx_n_s_Nfa_get_terminating_states
#define __pyx_n_s_Nfa_get_trans_as_sequence __pyx_mstate_global->__pyx_n_s_Nfa_get_trans_as_sequence
#define __pyx_n_s_Nfa_get_trans_from_state_as_sequ __pyx_mstate_global->__pyx_n_s_Nfa_get_trans_from_state_as_sequ
#define __pyx_n_s_Nfa_get_transitions_from_state __pyx_mstate_global->__pyx_n_s_Nfa_get_transitions_from_state
#define __pyx_n_s_Nfa_get_transitions_to_state __pyx_mstate_global->__pyx_n_s_Nfa_get_transitions_to_state
#define __pyx_n_s_Nfa_get_useful_states __pyx_mstate_global->__pyx_n_s_Nfa_get_useful_states
#define __pyx_n_s_Nfa_get_word_for_path __pyx_mstate_global->__pyx_n_s_Nfa_get_word_for_path
#define __pyx_kp_u_Nfa_get_word_for_path_line_1122 __pyx_mstate_global->__pyx_kp_u_Nfa_get_word_for_path_line_1122
#define __pyx_n_s_Nfa_has_final_state __pyx_mstate_global->__pyx_n_s_Nfa_has_final_state
#define __pyx_n_s_Nfa_has_initial_state __pyx_mstate_global->__pyx_n_s_Nfa_has_initial_state
#define __pyx_n_s_Nfa_has_transition __pyx_mstate_global->__pyx_n_s_Nfa_has_transition
#define __pyx_n_s_Nfa_intersection __pyx_mstate_global->__pyx_n_s_Nfa_intersection
#define __pyx_n_s_Nfa_intersection_with_product_ma __pyx_mstate_global->__pyx_n_s_Nfa_intersection_with_product_ma
#define __pyx_n_s_Nfa_is_complete __pyx_mstate_global->__pyx_n_s_Nfa_is_complete
#define __pyx_n_s_Nfa_is_deterministic __pyx_mstate_global->__pyx_n_s_Nfa_is_deterministic
#define __pyx_n_s_Nfa_is_epsilon __pyx_mstate_global->__pyx_n_s_Nfa_is_epsilon
#define __pyx_n_s_Nfa_is_in_lang __pyx_mstate_global->__pyx_n_s_Nfa_is_in_lang
#define __pyx_n_s_Nfa_is_included __pyx_mstate_global->__pyx_n_s_Nfa_is_included
#define __pyx_n_s_Nfa_is_included_with_cex __pyx_mstate_global->__pyx_n_s_Nfa_is_included_with_cex
#define __pyx_n_s_Nfa_is_lang_empty __pyx_mstate_global->__pyx_n_s_Nfa_is_lang_empty
#define __pyx_n_s_Nfa_is_prefix_in_lang __pyx_mstate_global->__pyx_n_s_Nfa_is_prefix_in_lang
#define __pyx_n_s_Nfa_is_state __pyx_mstate_global->__pyx_n_s_Nfa_is_state
#define __pyx_n_s_Nfa_is_universal __pyx_mstate_global->__pyx_n_s_Nfa_is_universal
#define __pyx_n_s_Nfa_iterate __pyx_mstate_global->__pyx_n_s_Nfa_iterate
#define __pyx_n_s_Nfa_make_complete __pyx_mstate_global->__pyx_n_s_Nfa_make_complete
#define __pyx_n_s_Nfa_make_final_state __pyx_mstate_global->__pyx_n_s_Nfa_make_final_state
#define __pyx_n_s_Nfa_make_final_states __pyx_mstate_global->__pyx_n_s_Nfa_make_final_states
#define __pyx_n_s_Nfa_make_initial_state __pyx_mstate_global->__pyx_n_s_Nfa_make_initial_state
#define __pyx_n_s_Nfa_make_initial_states __pyx_mstate_global->__pyx_n_s_Nfa_make_initial_states
#define __pyx_n_s_Nfa_minimize __pyx_mstate_global->__pyx_n_s_Nfa_minimize
#define __pyx_n_s_Nfa_post_map_of __pyx_mstate_global->__pyx_n_s_Nfa_post_map_of
#define __pyx_n_s_Nfa_post_of __pyx_mstate_global->__pyx_n_s_Nfa_post_of
#define __pyx_n_s_Nfa_reduce __pyx_mstate_global->__pyx_n_s_Nfa_reduce
#define __pyx_n_s_Nfa_reduce_with_state_map __pyx_mstate_global->__pyx_n_s_Nfa_reduce_with_state_map
#define __pyx_n_s_Nfa_remove_epsilon __pyx_mstate_global->__pyx_n_s_Nfa_remove_epsilon
#define __pyx_n_s_Nfa_remove_epsilon_inplace __pyx_mstate_global->__pyx_n_s_Nfa_remove_epsilon_inplace
#define __pyx_n_s_Nfa_remove_final_state __pyx_mstate_global->__pyx_n_s_Nfa_remove_final_state
#define __pyx_n_s_Nfa_remove_initial_state __pyx_mstate_global->__pyx_n_s_Nfa_remove_initial_state
#define __pyx_n_s_Nfa_remove_trans __pyx_mstate_global->__pyx_n_s_Nfa_remove_trans
#define __pyx_n_s_Nfa_remove_trans_raw __pyx_mstate_global->__pyx_n_s_Nfa_remove_trans_raw
#define __pyx_n_s_Nfa_revert __pyx_mstate_global->__pyx_n_s_Nfa_revert
#define __pyx_n_s_Nfa_size __pyx_mstate_global->__pyx_n_s_Nfa_size
#define __pyx_n_s_Nfa_to_dataframe __pyx_mstate_global->__pyx_n_s_Nfa_to_dataframe
#define __pyx_n_s_Nfa_to_dot_file __pyx_mstate_global->__pyx_n_s_Nfa_to_dot_file
#define __pyx_n_s_Nfa_to_dot_str __pyx_mstate_global->__pyx_n_s_Nfa_to_dot_str
#define __pyx_n_s_Nfa_to_networkx_graph __pyx_mstate_global->__pyx_n_s_Nfa_to_networkx_graph
#define __pyx_n_s_Nfa_trim __pyx_mstate_global->__pyx_n_s_Nfa_trim
#define __pyx_n_s_Nfa_trim_with_state_map __pyx_mstate_global->__pyx_n_s_Nfa_trim_with_state_map
#define __pyx_n_s_Nfa_unify_final __pyx_mstate_global->__pyx_n_s_Nfa_unify_final
#define __pyx_n_s_Nfa_unify_initial __pyx_mstate_global->__pyx_n_s_Nfa_unify_initial
#define __pyx_n_s_Nfa_union __pyx_mstate_global->__pyx_n_s_Nfa_union
#define __pyx_n_s_PIPE __pyx_mstate_global->__pyx_n_s_PIPE
#define __pyx_n_s_Popen __pyx_mstate_global->__pyx_n_s_Popen
#define __pyx_n_s_Run __pyx_mstate_global->__pyx_n_s_Run
#define __pyx_n_s_Run___reduce_cython __pyx_mstate_global->__pyx_n_s_Run___reduce_cython
#define __pyx_n_s_Run___setstate_cython __pyx_mstate_global->__pyx_n_s_Run___setstate_cython
#define __pyx_n_s_STDOUT __pyx_mstate_global->__pyx_n_s_STDOUT
#define __pyx_n_s_TimeoutExpired __pyx_mstate_global->__pyx_n_s_TimeoutExpired
#define __pyx_n_s_Trans __pyx_mstate_global->__pyx_n_s_Trans
#define __pyx_n_s_Trans___neq __pyx_mstate_global->__pyx_n_s_Trans___neq
#define __pyx_n_s_Trans___reduce_cython __pyx_mstate_global->__pyx_n_s_Trans___reduce_cython
#define __pyx_n_s_Trans___setstate_cython __pyx_mstate_global->__pyx_n_s_Trans___setstate_cython
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_kp_b__10 __pyx_mstate_global->__pyx_kp_b__10
#define __pyx_kp_u__15 __pyx_mstate_global->__pyx_kp_u__15
#define __pyx_n_s__16 __pyx_mstate_global->__pyx_n_s__16
#define __pyx_n_s__165 __pyx_mstate_global->__pyx_n_s__165
#define __pyx_kp_u__2 __pyx_mstate_global->__pyx_kp_u__2
#define __pyx_kp_u__3 __pyx_mstate_global->__pyx_kp_u__3
#define __pyx_kp_u__4 __pyx_mstate_global->__pyx_kp_u__4
#define __pyx_kp_u__5 __pyx_mstate_global->__pyx_kp_u__5
#define __pyx_kp_u__6 __pyx_mstate_global->__pyx_kp_u__6
#define __pyx_kp_u__8 __pyx_mstate_global->__pyx_kp_u__8
#define __pyx_n_s__80 __pyx_mstate_global->__pyx_n_s__80
#define __pyx_kp_u__9 __pyx_mstate_global->__pyx_kp_u__9
#define __pyx_n_s_accepts_epsilon __pyx_mstate_global->__pyx_n_s_accepts_epsilon
#define __pyx_n_s_add_edge __pyx_mstate_global->__pyx_n_s_add_edge
#define __pyx_n_s_add_new_state __pyx_mstate_global->__pyx_n_s_add_new_state
#define __pyx_n_s_add_state __pyx_mstate_global->__pyx_n_s_add_state
#define __pyx_n_s_add_transition __pyx_mstate_global->__pyx_n_s_add_transition
#define __pyx_n_s_add_transition_object __pyx_mstate_global->__pyx_n_s_add_transition_object
#define __pyx_n_u_algorithm __pyx_mstate_global->__pyx_n_u_algorithm
#define __pyx_n_s_alphabet __pyx_mstate_global->__pyx_n_s_alphabet
#define __pyx_n_u_alphabet __pyx_mstate_global->__pyx_n_u_alphabet
#define __pyx_n_u_antichains __pyx_mstate_global->__pyx_n_u_antichains
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_aut __pyx_mstate_global->__pyx_n_s_aut
#define __pyx_kp_u_aut_dot __pyx_mstate_global->__pyx_kp_u_aut_dot
#define __pyx_n_s_bool __pyx_mstate_global->__pyx_n_s_bool
#define __pyx_n_s_bool_vec __pyx_mstate_global->__pyx_n_s_bool_vec
#define __pyx_n_s_c_alphabet __pyx_mstate_global->__pyx_n_s_c_alphabet
#define __pyx_n_s_c_bool_vec __pyx_mstate_global->__pyx_n_s_c_bool_vec
#define __pyx_n_s_c_epsilon_transitions_iter __pyx_mstate_global->__pyx_n_s_c_epsilon_transitions_iter
#define __pyx_n_s_c_lhs_map __pyx_mstate_global->__pyx_n_s_c_lhs_map
#define __pyx_n_s_c_product_map __pyx_mstate_global->__pyx_n_s_c_product_map
#define __pyx_n_s_c_rhs_map __pyx_mstate_global->__pyx_n_s_c_rhs_map
#define __pyx_n_s_c_states __pyx_mstate_global->__pyx_n_s_c_states
#define __pyx_n_s_c_transition __pyx_mstate_global->__pyx_n_s_c_transition
#define __pyx_n_s_c_transitions __pyx_mstate_global->__pyx_n_s_c_transitions
#define __pyx_kp_u_captured_stderr __pyx_mstate_global->__pyx_kp_u_captured_stderr
#define __pyx_kp_u_captured_stdout __pyx_mstate_global->__pyx_kp_u_captured_stdout
#define __pyx_n_s_check_results __pyx_mstate_global->__pyx_n_s_check_results
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_u_classical __pyx_mstate_global->__pyx_n_u_classical
#define __pyx_n_s_clear __pyx_mstate_global->__pyx_n_s_clear
#define __pyx_n_s_clear_final __pyx_mstate_global->__pyx_n_s_clear_final
#define __pyx_n_s_clear_initial __pyx_mstate_global->__pyx_n_s_clear_initial
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_s_cls __pyx_mstate_global->__pyx_n_s_cls
#define __pyx_n_s_cmd __pyx_mstate_global->__pyx_n_s_cmd
#define __pyx_n_s_cmd_no __pyx_mstate_global->__pyx_n_s_cmd_no
#define __pyx_n_s_cmderr __pyx_mstate_global->__pyx_n_s_cmderr
#define __pyx_n_s_cmdout __pyx_mstate_global->__pyx_n_s_cmdout
#define __pyx_n_u_color __pyx_mstate_global->__pyx_n_u_color
#define __pyx_n_s_columns __pyx_mstate_global->__pyx_n_s_columns
#define __pyx_n_s_communicate __pyx_mstate_global->__pyx_n_s_communicate
#define __pyx_n_s_complement __pyx_mstate_global->__pyx_n_s_complement
#define __pyx_n_s_compute_relation __pyx_mstate_global->__pyx_n_s_compute_relation
#define __pyx_n_s_concatenate __pyx_mstate_global->__pyx_n_s_concatenate
#define __pyx_n_s_concatenate_with_result_state_ma __pyx_mstate_global->__pyx_n_s_concatenate_with_result_state_ma
#define __pyx_n_u_darkblue __pyx_mstate_global->__pyx_n_u_darkblue
#define __pyx_n_s_data __pyx_mstate_global->__pyx_n_s_data
#define __pyx_n_s_decode __pyx_mstate_global->__pyx_n_s_decode
#define __pyx_n_s_determinize __pyx_mstate_global->__pyx_n_s_determinize
#define __pyx_n_s_determinize_with_subset_map __pyx_mstate_global->__pyx_n_s_determinize_with_subset_map
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_u_direction __pyx_mstate_global->__pyx_n_u_direction
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_divisible_by __pyx_mstate_global->__pyx_n_s_divisible_by
#define __pyx_kp_u_dot_O_T __pyx_mstate_global->__pyx_kp_u_dot_O_T
#define __pyx_n_u_edge_style __pyx_mstate_global->__pyx_n_u_edge_style
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_encode __pyx_mstate_global->__pyx_n_s_encode
#define __pyx_n_s_encode_word __pyx_mstate_global->__pyx_n_s_encode_word
#define __pyx_n_s_encoding __pyx_mstate_global->__pyx_n_s_encoding
#define __pyx_n_s_end __pyx_mstate_global->__pyx_n_s_end
#define __pyx_n_s_epsilon __pyx_mstate_global->__pyx_n_s_epsilon
#define __pyx_n_s_epsilon_transitions __pyx_mstate_global->__pyx_n_s_epsilon_transitions
#define __pyx_n_s_equivalence_check __pyx_mstate_global->__pyx_n_s_equivalence_check
#define __pyx_n_s_err __pyx_mstate_global->__pyx_n_s_err
#define __pyx_kp_u_error_while_dot_file __pyx_mstate_global->__pyx_kp_u_error_while_dot_file
#define __pyx_n_s_executed_command __pyx_mstate_global->__pyx_n_s_executed_command
#define __pyx_n_u_false __pyx_mstate_global->__pyx_n_u_false
#define __pyx_n_u_fillcolor __pyx_mstate_global->__pyx_n_u_fillcolor
#define __pyx_n_u_filled __pyx_mstate_global->__pyx_n_u_filled
#define __pyx_kp_u_final_states __pyx_mstate_global->__pyx_kp_u_final_states
#define __pyx_n_u_fixedsize __pyx_mstate_global->__pyx_n_u_fixedsize
#define __pyx_n_u_fontname __pyx_mstate_global->__pyx_n_u_fontname
#define __pyx_n_u_fontsize __pyx_mstate_global->__pyx_n_u_fontsize
#define __pyx_n_s_format __pyx_mstate_global->__pyx_n_s_format
#define __pyx_n_u_forward __pyx_mstate_global->__pyx_n_u_forward
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_get __pyx_mstate_global->__pyx_n_s_get
#define __pyx_n_s_get_alphabet_symbols __pyx_mstate_global->__pyx_n_s_get_alphabet_symbols
#define __pyx_n_s_get_elements_from_bool_vec __pyx_mstate_global->__pyx_n_s_get_elements_from_bool_vec
#define __pyx_n_s_get_epsilon_transitions __pyx_mstate_global->__pyx_n_s_get_epsilon_transitions
#define __pyx_n_s_get_num_of_trans __pyx_mstate_global->__pyx_n_s_get_num_of_trans
#define __pyx_n_s_get_one_letter_aut __pyx_mstate_global->__pyx_n_s_get_one_letter_aut
#define __pyx_n_s_get_reachable_states __pyx_mstate_global->__pyx_n_s_get_reachable_states
#define __pyx_n_s_get_symbols __pyx_mstate_global->__pyx_n_s_get_symbols
#define __pyx_n_s_get_terminating_states __pyx_mstate_global->__pyx_n_s_get_terminating_states
#define __pyx_n_s_get_trans_as_sequence __pyx_mstate_global->__pyx_n_s_get_trans_as_sequence
#define __pyx_n_s_get_trans_from_state_as_sequence __pyx_mstate_global->__pyx_n_s_get_trans_from_state_as_sequence
#define __pyx_n_s_get_transitions_from_state __pyx_mstate_global->__pyx_n_s_get_transitions_from_state
#define __pyx_n_s_get_transitions_to_state __pyx_mstate_global->__pyx_n_s_get_transitions_to_state
#define __pyx_n_s_get_useful_states __pyx_mstate_global->__pyx_n_s_get_useful_states
#define __pyx_n_s_get_word_for_path __pyx_mstate_global->__pyx_n_s_get_word_for_path
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_s_graphviz_command __pyx_mstate_global->__pyx_n_s_graphviz_command
#define __pyx_n_s_has_final_state __pyx_mstate_global->__pyx_n_s_has_final_state
#define __pyx_n_s_has_initial_state __pyx_mstate_global->__pyx_n_s_has_initial_state
#define __pyx_n_s_has_transition __pyx_mstate_global->__pyx_n_s_has_transition
#define __pyx_n_u_height __pyx_mstate_global->__pyx_n_u_height
#define __pyx_n_s_i __pyx_mstate_global->__pyx_n_s_i
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_kp_u_initial_states __pyx_mstate_global->__pyx_kp_u_initial_states
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_input __pyx_mstate_global->__pyx_n_s_input
#define __pyx_n_s_input_states __pyx_mstate_global->__pyx_n_s_input_states
#define __pyx_n_s_int __pyx_mstate_global->__pyx_n_s_int
#define __pyx_n_s_intersection __pyx_mstate_global->__pyx_n_s_intersection
#define __pyx_n_s_intersection_with_product_map __pyx_mstate_global->__pyx_n_s_intersection_with_product_map
#define __pyx_n_s_is_complete __pyx_mstate_global->__pyx_n_s_is_complete
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_n_s_is_deterministic __pyx_mstate_global->__pyx_n_s_is_deterministic
#define __pyx_n_s_is_epsilon __pyx_mstate_global->__pyx_n_s_is_epsilon
#define __pyx_n_s_is_in_lang __pyx_mstate_global->__pyx_n_s_is_in_lang
#define __pyx_n_s_is_included __pyx_mstate_global->__pyx_n_s_is_included
#define __pyx_n_s_is_included_with_cex __pyx_mstate_global->__pyx_n_s_is_included_with_cex
#define __pyx_n_s_is_lang_empty __pyx_mstate_global->__pyx_n_s_is_lang_empty
#define __pyx_n_s_is_prefix_in_lang __pyx_mstate_global->__pyx_n_s_is_prefix_in_lang
#define __pyx_n_s_is_state __pyx_mstate_global->__pyx_n_s_is_state
#define __pyx_n_s_is_universal __pyx_mstate_global->__pyx_n_s_is_universal
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_it __pyx_mstate_global->__pyx_n_s_it
#define __pyx_n_s_items __pyx_mstate_global->__pyx_n_s_items
#define __pyx_n_s_iterate __pyx_mstate_global->__pyx_n_s_iterate
#define __pyx_n_s_iterator __pyx_mstate_global->__pyx_n_s_iterator
#define __pyx_n_s_iteritems __pyx_mstate_global->__pyx_n_s_iteritems
#define __pyx_n_s_j __pyx_mstate_global->__pyx_n_s_j
#define __pyx_n_s_k __pyx_mstate_global->__pyx_n_s_k
#define __pyx_n_s_key __pyx_mstate_global->__pyx_n_s_key
#define __pyx_n_s_kwargs __pyx_mstate_global->__pyx_n_s_kwargs
#define __pyx_n_s_label __pyx_mstate_global->__pyx_n_s_label
#define __pyx_n_s_lhs __pyx_mstate_global->__pyx_n_s_lhs
#define __pyx_n_s_lhs_map __pyx_mstate_global->__pyx_n_s_lhs_map
#define __pyx_n_s_libmata_nfa_nfa __pyx_mstate_global->__pyx_n_s_libmata_nfa_nfa
#define __pyx_kp_s_libmata_nfa_nfa_pyx __pyx_mstate_global->__pyx_kp_s_libmata_nfa_nfa_pyx
#define __pyx_n_u_lightsteelblue __pyx_mstate_global->__pyx_n_u_lightsteelblue
#define __pyx_kp_s_list_int __pyx_mstate_global->__pyx_kp_s_list_int
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_make_complete __pyx_mstate_global->__pyx_n_s_make_complete
#define __pyx_n_s_make_final_state __pyx_mstate_global->__pyx_n_s_make_final_state
#define __pyx_n_s_make_final_states __pyx_mstate_global->__pyx_n_s_make_final_states
#define __pyx_n_s_make_initial_state __pyx_mstate_global->__pyx_n_s_make_initial_state
#define __pyx_n_s_make_initial_states __pyx_mstate_global->__pyx_n_s_make_initial_states
#define __pyx_n_s_map __pyx_mstate_global->__pyx_n_s_map
#define __pyx_n_u_midnightblue __pyx_mstate_global->__pyx_n_u_midnightblue
#define __pyx_n_s_minimize __pyx_mstate_global->__pyx_n_s_minimize
#define __pyx_n_u_minimize __pyx_mstate_global->__pyx_n_u_minimize
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_neq __pyx_mstate_global->__pyx_n_s_neq
#define __pyx_n_s_networkx __pyx_mstate_global->__pyx_n_s_networkx
#define __pyx_kp_s_no_default___reduce___due_to_non __pyx_mstate_global->__pyx_kp_s_no_default___reduce___due_to_non
#define __pyx_n_u_node_style __pyx_mstate_global->__pyx_n_u_node_style
#define __pyx_n_s_nx __pyx_mstate_global->__pyx_n_s_nx
#define __pyx_kp_s_nx_Graph __pyx_mstate_global->__pyx_kp_s_nx_Graph
#define __pyx_n_s_objects __pyx_mstate_global->__pyx_n_s_objects
#define __pyx_n_s_one_letter_aut __pyx_mstate_global->__pyx_n_s_one_letter_aut
#define __pyx_n_s_other __pyx_mstate_global->__pyx_n_s_other
#define __pyx_n_s_output __pyx_mstate_global->__pyx_n_s_output
#define __pyx_n_s_output_file __pyx_mstate_global->__pyx_n_s_output_file
#define __pyx_n_s_output_format __pyx_mstate_global->__pyx_n_s_output_format
#define __pyx_n_s_output_stream __pyx_mstate_global->__pyx_n_s_output_stream
#define __pyx_n_s_pandas __pyx_mstate_global->__pyx_n_s_pandas
#define __pyx_kp_s_pandas_DataFrame __pyx_mstate_global->__pyx_kp_s_pandas_DataFrame
#define __pyx_n_s_params __pyx_mstate_global->__pyx_n_s_params
#define __pyx_n_s_path __pyx_mstate_global->__pyx_n_s_path
#define __pyx_n_u_pdf __pyx_mstate_global->__pyx_n_u_pdf
#define __pyx_n_u_penwidth __pyx_mstate_global->__pyx_n_u_penwidth
#define __pyx_n_s_piped_command __pyx_mstate_global->__pyx_n_s_piped_command
#define __pyx_n_s_post_map_of __pyx_mstate_global->__pyx_n_s_post_map_of
#define __pyx_n_s_post_of __pyx_mstate_global->__pyx_n_s_post_of
#define __pyx_n_s_preserve_epsilon __pyx_mstate_global->__pyx_n_s_preserve_epsilon
#define __pyx_n_s_print __pyx_mstate_global->__pyx_n_s_print
#define __pyx_n_s_process __pyx_mstate_global->__pyx_n_s_process
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_quiet __pyx_mstate_global->__pyx_n_s_quiet
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_u_red __pyx_mstate_global->__pyx_n_u_red
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_2 __pyx_mstate_global->__pyx_n_s_reduce_2
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_reduce_with_state_map __pyx_mstate_global->__pyx_n_s_reduce_with_state_map
#define __pyx_n_s_relation __pyx_mstate_global->__pyx_n_s_relation
#define __pyx_n_u_relation __pyx_mstate_global->__pyx_n_u_relation
#define __pyx_n_s_relation_size __pyx_mstate_global->__pyx_n_s_relation_size
#define __pyx_n_s_remove_epsilon __pyx_mstate_global->__pyx_n_s_remove_epsilon
#define __pyx_n_s_remove_epsilon_inplace __pyx_mstate_global->__pyx_n_s_remove_epsilon_inplace
#define __pyx_n_s_remove_final_state __pyx_mstate_global->__pyx_n_s_remove_final_state
#define __pyx_n_s_remove_initial_state __pyx_mstate_global->__pyx_n_s_remove_initial_state
#define __pyx_n_s_remove_trans __pyx_mstate_global->__pyx_n_s_remove_trans
#define __pyx_n_s_remove_trans_raw __pyx_mstate_global->__pyx_n_s_remove_trans_raw
#define __pyx_n_s_resize __pyx_mstate_global->__pyx_n_s_resize
#define __pyx_n_s_result __pyx_mstate_global->__pyx_n_s_result
#define __pyx_n_s_return __pyx_mstate_global->__pyx_n_s_return
#define __pyx_n_s_return_value __pyx_mstate_global->__pyx_n_s_return_value
#define __pyx_n_s_returncode __pyx_mstate_global->__pyx_n_s_returncode
#define __pyx_n_s_revert __pyx_mstate_global->__pyx_n_s_revert
#define __pyx_n_s_rhs __pyx_mstate_global->__pyx_n_s_rhs
#define __pyx_n_s_rhs_map __pyx_mstate_global->__pyx_n_s_rhs_map
#define __pyx_n_s_run __pyx_mstate_global->__pyx_n_s_run
#define __pyx_n_s_run_safely_external_command __pyx_mstate_global->__pyx_n_s_run_safely_external_command
#define __pyx_n_s_s __pyx_mstate_global->__pyx_n_s_s
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_n_s_set __pyx_mstate_global->__pyx_n_s_set
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_shell __pyx_mstate_global->__pyx_n_s_shell
#define __pyx_n_s_shlex __pyx_mstate_global->__pyx_n_s_shlex
#define __pyx_n_u_simulation __pyx_mstate_global->__pyx_n_u_simulation
#define __pyx_n_s_sink_state __pyx_mstate_global->__pyx_n_s_sink_state
#define __pyx_n_s_size __pyx_mstate_global->__pyx_n_s_size
#define __pyx_n_u_source __pyx_mstate_global->__pyx_n_u_source
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_s_split __pyx_mstate_global->__pyx_n_s_split
#define __pyx_n_s_src __pyx_mstate_global->__pyx_n_s_src
#define __pyx_n_s_st __pyx_mstate_global->__pyx_n_s_st
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_s_state_from __pyx_mstate_global->__pyx_n_s_state_from
#define __pyx_n_s_state_map __pyx_mstate_global->__pyx_n_s_state_map
#define __pyx_n_s_state_number __pyx_mstate_global->__pyx_n_s_state_number
#define __pyx_n_s_state_to __pyx_mstate_global->__pyx_n_s_state_to
#define __pyx_n_s_states __pyx_mstate_global->__pyx_n_s_states
#define __pyx_n_s_stderr __pyx_mstate_global->__pyx_n_s_stderr
#define __pyx_n_s_stdin __pyx_mstate_global->__pyx_n_s_stdin
#define __pyx_n_s_stdout __pyx_mstate_global->__pyx_n_s_stdout
#define __pyx_n_s_store __pyx_mstate_global->__pyx_n_s_store
#define __pyx_n_s_store_2 __pyx_mstate_global->__pyx_n_s_store_2
#define __pyx_n_s_str __pyx_mstate_global->__pyx_n_s_str
#define __pyx_n_s_string_to_symbol __pyx_mstate_global->__pyx_n_s_string_to_symbol
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_strip __pyx_mstate_global->__pyx_n_s_strip
#define __pyx_n_u_style __pyx_mstate_global->__pyx_n_u_style
#define __pyx_n_s_subprocess __pyx_mstate_global->__pyx_n_s_subprocess
#define __pyx_n_s_subset_map __pyx_mstate_global->__pyx_n_s_subset_map
#define __pyx_n_s_symb __pyx_mstate_global->__pyx_n_s_symb
#define __pyx_n_s_symbol __pyx_mstate_global->__pyx_n_s_symbol
#define __pyx_n_u_symbol __pyx_mstate_global->__pyx_n_u_symbol
#define __pyx_n_s_symbol_map __pyx_mstate_global->__pyx_n_s_symbol_map
#define __pyx_n_s_symbols __pyx_mstate_global->__pyx_n_s_symbols
#define __pyx_n_s_t __pyx_mstate_global->__pyx_n_s_t
#define __pyx_n_u_target __pyx_mstate_global->__pyx_n_u_target
#define __pyx_n_s_targets __pyx_mstate_global->__pyx_n_s_targets
#define __pyx_n_s_terminate __pyx_mstate_global->__pyx_n_s_terminate
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_tgt __pyx_mstate_global->__pyx_n_s_tgt
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_timeout __pyx_mstate_global->__pyx_n_s_timeout
#define __pyx_n_s_to_dataframe __pyx_mstate_global->__pyx_n_s_to_dataframe
#define __pyx_n_s_to_dot_file __pyx_mstate_global->__pyx_n_s_to_dot_file
#define __pyx_n_s_to_dot_str __pyx_mstate_global->__pyx_n_s_to_dot_str
#define __pyx_n_s_to_networkx_graph __pyx_mstate_global->__pyx_n_s_to_networkx_graph
#define __pyx_n_s_tr __pyx_mstate_global->__pyx_n_s_tr
#define __pyx_n_s_trans __pyx_mstate_global->__pyx_n_s_trans
#define __pyx_kp_u_transitions __pyx_mstate_global->__pyx_kp_u_transitions
#define __pyx_n_s_transitions_2 __pyx_mstate_global->__pyx_n_s_transitions_2
#define __pyx_n_s_transitions_list __pyx_mstate_global->__pyx_n_s_transitions_list
#define __pyx_n_s_translate_symbol __pyx_mstate_global->__pyx_n_s_translate_symbol
#define __pyx_n_s_transsymbols __pyx_mstate_global->__pyx_n_s_transsymbols
#define __pyx_n_s_trim __pyx_mstate_global->__pyx_n_s_trim
#define __pyx_n_s_trim_input __pyx_mstate_global->__pyx_n_s_trim_input
#define __pyx_n_s_trim_with_state_map __pyx_mstate_global->__pyx_n_s_trim_with_state_map
#define __pyx_n_u_true __pyx_mstate_global->__pyx_n_u_true
#define __pyx_kp_s_tuple_Nfa_dict_str_str_dict_str __pyx_mstate_global->__pyx_kp_s_tuple_Nfa_dict_str_str_dict_str
#define __pyx_n_s_unify_final __pyx_mstate_global->__pyx_n_s_unify_final
#define __pyx_n_s_unify_initial __pyx_mstate_global->__pyx_n_s_unify_initial
#define __pyx_n_s_union __pyx_mstate_global->__pyx_n_s_union
#define __pyx_n_s_unpiped_commands __pyx_mstate_global->__pyx_n_s_unpiped_commands
#define __pyx_n_s_use_epsilon __pyx_mstate_global->__pyx_n_s_use_epsilon
#define __pyx_n_s_useful_states __pyx_mstate_global->__pyx_n_s_useful_states
#define __pyx_n_s_useful_states_bool_vec __pyx_mstate_global->__pyx_n_s_useful_states_bool_vec
#define __pyx_kp_u_utf_8 __pyx_mstate_global->__pyx_kp_u_utf_8
#define __pyx_n_s_v __pyx_mstate_global->__pyx_n_s_v
#define __pyx_n_s_val __pyx_mstate_global->__pyx_n_s_val
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_n_s_wait __pyx_mstate_global->__pyx_n_s_wait
#define __pyx_n_u_width __pyx_mstate_global->__pyx_n_u_width
#define __pyx_kp_u_without_specified_alphabet __pyx_mstate_global->__pyx_kp_u_without_specified_alphabet
#define __pyx_n_s_word __pyx_mstate_global->__pyx_n_s_word
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_k__11 __pyx_mstate_global->__pyx_k__11
#define __pyx_k__12 __pyx_mstate_global->__pyx_k__12
#define __pyx_k__13 __pyx_mstate_global->__pyx_k__13
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__18 __pyx_mstate_global->__pyx_tuple__18
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__29 __pyx_mstate_global->__pyx_tuple__29
#define __pyx_tuple__34 __pyx_mstate_global->__pyx_tuple__34
#define __pyx_tuple__36 __pyx_mstate_global->__pyx_tuple__36
#define __pyx_tuple__47 __pyx_mstate_global->__pyx_tuple__47
#define __pyx_tuple__49 __pyx_mstate_global->__pyx_tuple__49
#define __pyx_tuple__51 __pyx_mstate_global->__pyx_tuple__51
#define __pyx_tuple__53 __pyx_mstate_global->__pyx_tuple__53
#define __pyx_tuple__59 __pyx_mstate_global->__pyx_tuple__59
#define __pyx_tuple__60 __pyx_mstate_global->__pyx_tuple__60
#define __pyx_tuple__62 __pyx_mstate_global->__pyx_tuple__62
#define __pyx_tuple__64 __pyx_mstate_global->__pyx_tuple__64
#define __pyx_tuple__66 __pyx_mstate_global->__pyx_tuple__66
#define __pyx_tuple__68 __pyx_mstate_global->__pyx_tuple__68
#define __pyx_tuple__70 __pyx_mstate_global->__pyx_tuple__70
#define __pyx_tuple__74 __pyx_mstate_global->__pyx_tuple__74
#define __pyx_tuple__76 __pyx_mstate_global->__pyx_tuple__76
#define __pyx_tuple__78 __pyx_mstate_global->__pyx_tuple__78
#define __pyx_tuple__81 __pyx_mstate_global->__pyx_tuple__81
#define __pyx_tuple__83 __pyx_mstate_global->__pyx_tuple__83
#define __pyx_tuple__84 __pyx_mstate_global->__pyx_tuple__84
#define __pyx_tuple__86 __pyx_mstate_global->__pyx_tuple__86
#define __pyx_tuple__87 __pyx_mstate_global->__pyx_tuple__87
#define __pyx_tuple__89 __pyx_mstate_global->__pyx_tuple__89
#define __pyx_tuple__91 __pyx_mstate_global->__pyx_tuple__91
#define __pyx_tuple__93 __pyx_mstate_global->__pyx_tuple__93
#define __pyx_tuple__95 __pyx_mstate_global->__pyx_tuple__95
#define __pyx_tuple__97 __pyx_mstate_global->__pyx_tuple__97
#define __pyx_tuple__99 __pyx_mstate_global->__pyx_tuple__99
#define __pyx_codeobj__7 __pyx_mstate_global->__pyx_codeobj__7
#define __pyx_tuple__101 __pyx_mstate_global->__pyx_tuple__101
#define __pyx_tuple__103 __pyx_mstate_global->__pyx_tuple__103
#define __pyx_tuple__104 __pyx_mstate_global->__pyx_tuple__104
#define __pyx_tuple__106 __pyx_mstate_global->__pyx_tuple__106
#define __pyx_tuple__108 __pyx_mstate_global->__pyx_tuple__108
#define __pyx_tuple__110 __pyx_mstate_global->__pyx_tuple__110
#define __pyx_tuple__112 __pyx_mstate_global->__pyx_tuple__112
#define __pyx_tuple__115 __pyx_mstate_global->__pyx_tuple__115
#define __pyx_tuple__117 __pyx_mstate_global->__pyx_tuple__117
#define __pyx_tuple__119 __pyx_mstate_global->__pyx_tuple__119
#define __pyx_tuple__122 __pyx_mstate_global->__pyx_tuple__122
#define __pyx_tuple__124 __pyx_mstate_global->__pyx_tuple__124
#define __pyx_tuple__125 __pyx_mstate_global->__pyx_tuple__125
#define __pyx_tuple__127 __pyx_mstate_global->__pyx_tuple__127
#define __pyx_tuple__129 __pyx_mstate_global->__pyx_tuple__129
#define __pyx_tuple__131 __pyx_mstate_global->__pyx_tuple__131
#define __pyx_tuple__133 __pyx_mstate_global->__pyx_tuple__133
#define __pyx_tuple__135 __pyx_mstate_global->__pyx_tuple__135
#define __pyx_tuple__137 __pyx_mstate_global->__pyx_tuple__137
#define __pyx_tuple__138 __pyx_mstate_global->__pyx_tuple__138
#define __pyx_tuple__140 __pyx_mstate_global->__pyx_tuple__140
#define __pyx_tuple__142 __pyx_mstate_global->__pyx_tuple__142
#define __pyx_tuple__144 __pyx_mstate_global->__pyx_tuple__144
#define __pyx_tuple__146 __pyx_mstate_global->__pyx_tuple__146
#define __pyx_tuple__149 __pyx_mstate_global->__pyx_tuple__149
#define __pyx_tuple__151 __pyx_mstate_global->__pyx_tuple__151
#define __pyx_tuple__153 __pyx_mstate_global->__pyx_tuple__153
#define __pyx_tuple__157 __pyx_mstate_global->__pyx_tuple__157
#define __pyx_tuple__159 __pyx_mstate_global->__pyx_tuple__159
#define __pyx_tuple__161 __pyx_mstate_global->__pyx_tuple__161
#define __pyx_tuple__162 __pyx_mstate_global->__pyx_tuple__162
#define __pyx_codeobj__17 __pyx_mstate_global->__pyx_codeobj__17
#define __pyx_codeobj__19 __pyx_mstate_global->__pyx_codeobj__19
#define __pyx_codeobj__21 __pyx_mstate_global->__pyx_codeobj__21
#define __pyx_codeobj__23 __pyx_mstate_global->__pyx_codeobj__23
#define __pyx_codeobj__24 __pyx_mstate_global->__pyx_codeobj__24
#define __pyx_codeobj__25 __pyx_mstate_global->__pyx_codeobj__25
#define __pyx_codeobj__26 __pyx_mstate_global->__pyx_codeobj__26
#define __pyx_codeobj__27 __pyx_mstate_global->__pyx_codeobj__27
#define __pyx_codeobj__28 __pyx_mstate_global->__pyx_codeobj__28
#define __pyx_codeobj__30 __pyx_mstate_global->__pyx_codeobj__30
#define __pyx_codeobj__31 __pyx_mstate_global->__pyx_codeobj__31
#define __pyx_codeobj__32 __pyx_mstate_global->__pyx_codeobj__32
#define __pyx_codeobj__33 __pyx_mstate_global->__pyx_codeobj__33
#define __pyx_codeobj__35 __pyx_mstate_global->__pyx_codeobj__35
#define __pyx_codeobj__37 __pyx_mstate_global->__pyx_codeobj__37
#define __pyx_codeobj__38 __pyx_mstate_global->__pyx_codeobj__38
#define __pyx_codeobj__39 __pyx_mstate_global->__pyx_codeobj__39
#define __pyx_codeobj__40 __pyx_mstate_global->__pyx_codeobj__40
#define __pyx_codeobj__41 __pyx_mstate_global->__pyx_codeobj__41
#define __pyx_codeobj__42 __pyx_mstate_global->__pyx_codeobj__42
#define __pyx_codeobj__43 __pyx_mstate_global->__pyx_codeobj__43
#define __pyx_codeobj__44 __pyx_mstate_global->__pyx_codeobj__44
#define __pyx_codeobj__45 __pyx_mstate_global->__pyx_codeobj__45
#define __pyx_codeobj__46 __pyx_mstate_global->__pyx_codeobj__46
#define __pyx_codeobj__48 __pyx_mstate_global->__pyx_codeobj__48
#define __pyx_codeobj__50 __pyx_mstate_global->__pyx_codeobj__50
#define __pyx_codeobj__52 __pyx_mstate_global->__pyx_codeobj__52
#define __pyx_codeobj__54 __pyx_mstate_global->__pyx_codeobj__54
#define __pyx_codeobj__55 __pyx_mstate_global->__pyx_codeobj__55
#define __pyx_codeobj__56 __pyx_mstate_global->__pyx_codeobj__56
#define __pyx_codeobj__57 __pyx_mstate_global->__pyx_codeobj__57
#define __pyx_codeobj__58 __pyx_mstate_global->__pyx_codeobj__58
#define __pyx_codeobj__61 __pyx_mstate_global->__pyx_codeobj__61
#define __pyx_codeobj__63 __pyx_mstate_global->__pyx_codeobj__63
#define __pyx_codeobj__65 __pyx_mstate_global->__pyx_codeobj__65
#define __pyx_codeobj__67 __pyx_mstate_global->__pyx_codeobj__67
#define __pyx_codeobj__69 __pyx_mstate_global->__pyx_codeobj__69
#define __pyx_codeobj__71 __pyx_mstate_global->__pyx_codeobj__71
#define __pyx_codeobj__72 __pyx_mstate_global->__pyx_codeobj__72
#define __pyx_codeobj__73 __pyx_mstate_global->__pyx_codeobj__73
#define __pyx_codeobj__75 __pyx_mstate_global->__pyx_codeobj__75
#define __pyx_codeobj__77 __pyx_mstate_global->__pyx_codeobj__77
#define __pyx_codeobj__79 __pyx_mstate_global->__pyx_codeobj__79
#define __pyx_codeobj__82 __pyx_mstate_global->__pyx_codeobj__82
#define __pyx_codeobj__85 __pyx_mstate_global->__pyx_codeobj__85
#define __pyx_codeobj__88 __pyx_mstate_global->__pyx_codeobj__88
#define __pyx_codeobj__90 __pyx_mstate_global->__pyx_codeobj__90
#define __pyx_codeobj__92 __pyx_mstate_global->__pyx_codeobj__92
#define __pyx_codeobj__94 __pyx_mstate_global->__pyx_codeobj__94
#define __pyx_codeobj__96 __pyx_mstate_global->__pyx_codeobj__96
#define __pyx_codeobj__98 __pyx_mstate_global->__pyx_codeobj__98
#define __pyx_codeobj__100 __pyx_mstate_global->__pyx_codeobj__100
#define __pyx_codeobj__102 __pyx_mstate_global->__pyx_codeobj__102
#define __pyx_codeobj__105 __pyx_mstate_global->__pyx_codeobj__105
#define __pyx_codeobj__107 __pyx_mstate_global->__pyx_codeobj__107
#define __pyx_codeobj__109 __pyx_mstate_global->__pyx_codeobj__109
#define __pyx_codeobj__111 __pyx_mstate_global->__pyx_codeobj__111
#define __pyx_codeobj__113 __pyx_mstate_global->__pyx_codeobj__113
#define __pyx_codeobj__114 __pyx_mstate_global->__pyx_codeobj__114
#define __pyx_codeobj__116 __pyx_mstate_global->__pyx_codeobj__116
#define __pyx_codeobj__118 __pyx_mstate_global->__pyx_codeobj__118
#define __pyx_codeobj__120 __pyx_mstate_global->__pyx_codeobj__120
#define __pyx_codeobj__121 __pyx_mstate_global->__pyx_codeobj__121
#define __pyx_codeobj__123 __pyx_mstate_global->__pyx_codeobj__123
#define __pyx_codeobj__126 __pyx_mstate_global->__pyx_codeobj__126
#define __pyx_codeobj__128 __pyx_mstate_global->__pyx_codeobj__128
#define __pyx_codeobj__130 __pyx_mstate_global->__pyx_codeobj__130
#define __pyx_codeobj__132 __pyx_mstate_global->__pyx_codeobj__132
#define __pyx_codeobj__134 __pyx_mstate_global->__pyx_codeobj__134
#define __pyx_codeobj__136 __pyx_mstate_global->__pyx_codeobj__136
#define __pyx_codeobj__139 __pyx_mstate_global->__pyx_codeobj__139
#define __pyx_codeobj__141 __pyx_mstate_global->__pyx_codeobj__141
#define __pyx_codeobj__143 __pyx_mstate_global->__pyx_codeobj__143
#define __pyx_codeobj__145 __pyx_mstate_global->__pyx_codeobj__145
#define __pyx_codeobj__147 __pyx_mstate_global->__pyx_codeobj__147
#define __pyx_codeobj__148 __pyx_mstate_global->__pyx_codeobj__148
#define __pyx_codeobj__150 __pyx_mstate_global->__pyx_codeobj__150
#define __pyx_codeobj__152 __pyx_mstate_global->__pyx_codeobj__152
#define __pyx_codeobj__154 __pyx_mstate_global->__pyx_codeobj__154
#define __pyx_codeobj__155 __pyx_mstate_global->__pyx_codeobj__155
#define __pyx_codeobj__156 __pyx_mstate_global->__pyx_codeobj__156
#define __pyx_codeobj__158 __pyx_mstate_global->__pyx_codeobj__158
#define __pyx_codeobj__160 __pyx_mstate_global->__pyx_codeobj__160
#define __pyx_codeobj__163 __pyx_mstate_global->__pyx_codeobj__163
#define __pyx_codeobj__164 __pyx_mstate_global->__pyx_codeobj__164
/* #### Code section: module_code ### */

/* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State")
 * cdef vector[X] __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

static std::vector<Mata::Nfa::State>  __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(PyObject *__pyx_v_o) {
  std::vector<Mata::Nfa::State>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<Mata::Nfa::State>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  Mata::Nfa::State __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_v_item); if (unlikely((__pyx_t_5 == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(0, 48, __pyx_L1_error)
    try {
      __pyx_v_v.push_back(((Mata::Nfa::State)__pyx_t_5));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 48, __pyx_L1_error)
    }

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State")
 * cdef vector[X] __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<Mata::Symbol>  __pyx_convert_vector_from_py_Mata_3a__3a_Symbol(PyObject *__pyx_v_o) {
  std::vector<Mata::Symbol>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<Mata::Symbol>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  Mata::Symbol __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_Mata_3a__3a_Symbol", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_Mata_3a__3a_Symbol(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_v_item); if (unlikely((__pyx_t_5 == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(0, 48, __pyx_L1_error)
    try {
      __pyx_v_v.push_back(((Mata::Symbol)__pyx_t_5));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 48, __pyx_L1_error)
    }

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_Mata_3a__3a_Symbol(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_Mata_3a__3a_Symbol")
 * cdef vector[X] __pyx_convert_vector_from_py_Mata_3a__3a_Symbol(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_Mata_3a__3a_Symbol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.to_py":66
 * 
 * @cname("__pyx_convert_vector_to_py_Mata_3a__3a_Symbol")
 * cdef object __pyx_convert_vector_to_py_Mata_3a__3a_Symbol(const vector[X]& v):             # <<<<<<<<<<<<<<
 *     if v.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()
 */

static PyObject *__pyx_convert_vector_to_py_Mata_3a__3a_Symbol(std::vector<Mata::Symbol>  const &__pyx_v_v) {
  Py_ssize_t __pyx_v_v_size_signed;
  PyObject *__pyx_v_o = NULL;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_v_item = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_Mata_3a__3a_Symbol", 0);

  /* "vector.to_py":67
 * @cname("__pyx_convert_vector_to_py_Mata_3a__3a_Symbol")
 * cdef object __pyx_convert_vector_to_py_Mata_3a__3a_Symbol(const vector[X]& v):
 *     if v.size() > <size_t> PY_SSIZE_T_MAX:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     v_size_signed = <Py_ssize_t> v.size()
 */
  __pyx_t_1 = (__pyx_v_v.size() > ((size_t)PY_SSIZE_T_MAX));
  if (unlikely(__pyx_t_1)) {

    /* "vector.to_py":68
 * cdef object __pyx_convert_vector_to_py_Mata_3a__3a_Symbol(const vector[X]& v):
 *     if v.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()             # <<<<<<<<<<<<<<
 *     v_size_signed = <Py_ssize_t> v.size()
 * 
 */
    PyErr_NoMemory(); __PYX_ERR(0, 68, __pyx_L1_error)

    /* "vector.to_py":67
 * @cname("__pyx_convert_vector_to_py_Mata_3a__3a_Symbol")
 * cdef object __pyx_convert_vector_to_py_Mata_3a__3a_Symbol(const vector[X]& v):
 *     if v.size() > <size_t> PY_SSIZE_T_MAX:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     v_size_signed = <Py_ssize_t> v.size()
 */
  }

  /* "vector.to_py":69
 *     if v.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()
 *     v_size_signed = <Py_ssize_t> v.size()             # <<<<<<<<<<<<<<
 * 
 *     o = PyList_New(v_size_signed)
 */
  __pyx_v_v_size_signed = ((Py_ssize_t)__pyx_v_v.size());

  /* "vector.to_py":71
 *     v_size_signed = <Py_ssize_t> v.size()
 * 
 *     o = PyList_New(v_size_signed)             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t i
 */
  __pyx_t_2 = PyList_New(__pyx_v_v_size_signed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_o = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "vector.to_py":76
 *     cdef object item
 * 
 *     for i in range(v_size_signed):             # <<<<<<<<<<<<<<
 *         item = v[i]
 *         Py_INCREF(item)
 */
  __pyx_t_3 = __pyx_v_v_size_signed;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "vector.to_py":77
 * 
 *     for i in range(v_size_signed):
 *         item = v[i]             # <<<<<<<<<<<<<<
 *         Py_INCREF(item)
 *         PyList_SET_ITEM(o, i, item)
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "vector.to_py":78
 *     for i in range(v_size_signed):
 *         item = v[i]
 *         Py_INCREF(item)             # <<<<<<<<<<<<<<
 *         PyList_SET_ITEM(o, i, item)
 * 
 */
    Py_INCREF(__pyx_v_item);

    /* "vector.to_py":79
 *         item = v[i]
 *         Py_INCREF(item)
 *         PyList_SET_ITEM(o, i, item)             # <<<<<<<<<<<<<<
 * 
 *     return o
 */
    PyList_SET_ITEM(__pyx_v_o, __pyx_v_i, __pyx_v_item);
  }

  /* "vector.to_py":81
 *         PyList_SET_ITEM(o, i, item)
 * 
 *     return o             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_o);
  __pyx_r = __pyx_v_o;
  goto __pyx_L0;

  /* "vector.to_py":66
 * 
 * @cname("__pyx_convert_vector_to_py_Mata_3a__3a_Symbol")
 * cdef object __pyx_convert_vector_to_py_Mata_3a__3a_Symbol(const vector[X]& v):             # <<<<<<<<<<<<<<
 *     if v.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_Mata_3a__3a_Symbol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State(std::vector<Mata::Nfa::State>  const &__pyx_v_v) {
  Py_ssize_t __pyx_v_v_size_signed;
  PyObject *__pyx_v_o = NULL;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_v_item = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State", 0);

  /* "vector.to_py":67
 * @cname("__pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State")
 * cdef object __pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State(const vector[X]& v):
 *     if v.size() > <size_t> PY_SSIZE_T_MAX:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     v_size_signed = <Py_ssize_t> v.size()
 */
  __pyx_t_1 = (__pyx_v_v.size() > ((size_t)PY_SSIZE_T_MAX));
  if (unlikely(__pyx_t_1)) {

    /* "vector.to_py":68
 * cdef object __pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State(const vector[X]& v):
 *     if v.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()             # <<<<<<<<<<<<<<
 *     v_size_signed = <Py_ssize_t> v.size()
 * 
 */
    PyErr_NoMemory(); __PYX_ERR(0, 68, __pyx_L1_error)

    /* "vector.to_py":67
 * @cname("__pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State")
 * cdef object __pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State(const vector[X]& v):
 *     if v.size() > <size_t> PY_SSIZE_T_MAX:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     v_size_signed = <Py_ssize_t> v.size()
 */
  }

  /* "vector.to_py":69
 *     if v.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()
 *     v_size_signed = <Py_ssize_t> v.size()             # <<<<<<<<<<<<<<
 * 
 *     o = PyList_New(v_size_signed)
 */
  __pyx_v_v_size_signed = ((Py_ssize_t)__pyx_v_v.size());

  /* "vector.to_py":71
 *     v_size_signed = <Py_ssize_t> v.size()
 * 
 *     o = PyList_New(v_size_signed)             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t i
 */
  __pyx_t_2 = PyList_New(__pyx_v_v_size_signed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_o = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "vector.to_py":76
 *     cdef object item
 * 
 *     for i in range(v_size_signed):             # <<<<<<<<<<<<<<
 *         item = v[i]
 *         Py_INCREF(item)
 */
  __pyx_t_3 = __pyx_v_v_size_signed;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "vector.to_py":77
 * 
 *     for i in range(v_size_signed):
 *         item = v[i]             # <<<<<<<<<<<<<<
 *         Py_INCREF(item)
 *         PyList_SET_ITEM(o, i, item)
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "vector.to_py":78
 *     for i in range(v_size_signed):
 *         item = v[i]
 *         Py_INCREF(item)             # <<<<<<<<<<<<<<
 *         PyList_SET_ITEM(o, i, item)
 * 
 */
    Py_INCREF(__pyx_v_item);

    /* "vector.to_py":79
 *         item = v[i]
 *         Py_INCREF(item)
 *         PyList_SET_ITEM(o, i, item)             # <<<<<<<<<<<<<<
 * 
 *     return o
 */
    PyList_SET_ITEM(__pyx_v_o, __pyx_v_i, __pyx_v_item);
  }

  /* "vector.to_py":81
 *         PyList_SET_ITEM(o, i, item)
 * 
 *     return o             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_o);
  __pyx_r = __pyx_v_o;
  goto __pyx_L0;

  /* "vector.to_py":66
 * 
 * @cname("__pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State")
 * cdef object __pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State(const vector[X]& v):             # <<<<<<<<<<<<<<
 *     if v.size() > <size_t> PY_SSIZE_T_MAX:
 *         raise MemoryError()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pair.to_py":190
 * 
 * @cname("__pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State")
 * cdef object __pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State(const pair[X,Y]& p):             # <<<<<<<<<<<<<<
 *     return p.first, p.second
 * 
 */

static PyObject *__pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State(std::pair<Mata::Nfa::State,Mata::Nfa::State>  const &__pyx_v_p) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State", 0);

  /* "pair.to_py":191
 * @cname("__pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State")
 * cdef object __pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State(const pair[X,Y]& p):
 *     return p.first, p.second             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_p.first); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_p.second); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pair.to_py":190
 * 
 * @cname("__pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State")
 * cdef object __pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State(const pair[X,Y]& p):             # <<<<<<<<<<<<<<
 *     return p.first, p.second
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pair.to_py.__pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":31
 * 
 * @cname("__pyx_convert_PyObject_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyObject_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyObject_string_to_py_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyObject_string_to_py_std__in_string", 0);

  /* "string.to_py":32
 * @cname("__pyx_convert_PyObject_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyObject_string_to_py_std__in_string(const string& s):
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyUnicode_FromStringAndSize(const char*, size_t)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":31
 * 
 * @cname("__pyx_convert_PyObject_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyObject_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyObject_string_to_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":37
 * 
 * @cname("__pyx_convert_PyUnicode_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyUnicode_string_to_py_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyUnicode_string_to_py_std__in_string", 0);

  /* "string.to_py":38
 * @cname("__pyx_convert_PyUnicode_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_std__in_string(const string& s):
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyStr_FromStringAndSize(const char*, size_t)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyUnicode_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":37
 * 
 * @cname("__pyx_convert_PyUnicode_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyUnicode_string_to_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":43
 * 
 * @cname("__pyx_convert_PyStr_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyStr_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyStr_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyStr_string_to_py_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyStr_string_to_py_std__in_string", 0);

  /* "string.to_py":44
 * @cname("__pyx_convert_PyStr_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyStr_string_to_py_std__in_string(const string& s):
 *     return __Pyx_PyStr_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyBytes_FromStringAndSize(const char*, size_t)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyStr_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":43
 * 
 * @cname("__pyx_convert_PyStr_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyStr_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyStr_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyStr_string_to_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":49
 * 
 * @cname("__pyx_convert_PyBytes_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyBytes_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyBytes_string_to_py_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyBytes_string_to_py_std__in_string", 0);

  /* "string.to_py":50
 * @cname("__pyx_convert_PyBytes_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyBytes_string_to_py_std__in_string(const string& s):
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyByteArray_FromStringAndSize(const char*, size_t)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":49
 * 
 * @cname("__pyx_convert_PyBytes_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyBytes_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyBytes_string_to_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":55
 * 
 * @cname("__pyx_convert_PyByteArray_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), s.size())
 * 
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyByteArray_string_to_py_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyByteArray_string_to_py_std__in_string", 0);

  /* "string.to_py":56
 * @cname("__pyx_convert_PyByteArray_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_std__in_string(const string& s):
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyByteArray_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":55
 * 
 * @cname("__pyx_convert_PyByteArray_string_to_py_std__in_string")
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), s.size())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyByteArray_string_to_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.from_py":13
 * 
 * @cname("__pyx_convert_string_from_py_std__in_string")
 * cdef string __pyx_convert_string_from_py_std__in_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 */

static std::string __pyx_convert_string_from_py_std__in_string(PyObject *__pyx_v_o) {
  Py_ssize_t __pyx_v_length;
  char const *__pyx_v_data;
  std::string __pyx_r;
  __Pyx_RefNannyDeclarations
  char const *__pyx_t_1;
  std::string __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_string_from_py_std__in_string", 0);

  /* "string.from_py":14
 * @cname("__pyx_convert_string_from_py_std__in_string")
 * cdef string __pyx_convert_string_from_py_std__in_string(object o) except *:
 *     cdef Py_ssize_t length = 0             # <<<<<<<<<<<<<<
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 *     return string(data, length)
 */
  __pyx_v_length = 0;

  /* "string.from_py":15
 * cdef string __pyx_convert_string_from_py_std__in_string(object o) except *:
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)             # <<<<<<<<<<<<<<
 *     return string(data, length)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_AsStringAndSize(__pyx_v_o, (&__pyx_v_length)); if (unlikely(__pyx_t_1 == ((char const *)NULL))) __PYX_ERR(0, 15, __pyx_L1_error)
  __pyx_v_data = __pyx_t_1;

  /* "string.from_py":16
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 *     return string(data, length)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  try {
    __pyx_t_2 = std::string(__pyx_v_data, __pyx_v_length);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 16, __pyx_L1_error)
  }
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "string.from_py":13
 * 
 * @cname("__pyx_convert_string_from_py_std__in_string")
 * cdef string __pyx_convert_string_from_py_std__in_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("string.from_py.__pyx_convert_string_from_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pair.to_py":190
 * 
 * @cname("__pyx_convert_pair_to_py_std_3a__3a_pair_3c_Mata_3a__3a_Nfa_3a__3a_State_2c_Mata_3a__3a_Nfa_3a__3a_State_3e_______Mata_3a__3a_Nfa_3a__3a_State")
 * cdef object __pyx_convert_pair_to_py_std_3a__3a_pair_3c_Mata_3a__3a_Nfa_3a__3a_State_2c_Mata_3a__3a_Nfa_3a__3a_State_3e_______Mata_3a__3a_Nfa_3a__3a_State(const pair[X,Y]& p):             # <<<<<<<<<<<<<<
 *     return p.first, p.second
 * 
 */

static PyObject *__pyx_convert_pair_to_py_std_3a__3a_pair_3c_Mata_3a__3a_Nfa_3a__3a_State_2c_Mata_3a__3a_Nfa_3a__3a_State_3e_______Mata_3a__3a_Nfa_3a__3a_State(std::pair<std::pair<Mata::Nfa::State,Mata::Nfa::State> ,Mata::Nfa::State>  const &__pyx_v_p) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_pair_to_py_std_3a__3a_pair_3c_Mata_3a__3a_Nfa_3a__3a_State_2c_Mata_3a__3a_Nfa_3a__3a_State_3e_______Mata_3a__3a_Nfa_3a__3a_State", 0);

  /* "pair.to_py":191
 * @cname("__pyx_convert_pair_to_py_std_3a__3a_pair_3c_Mata_3a__3a_Nfa_3a__3a_State_2c_Mata_3a__3a_Nfa_3a__3a_State_3e_______Mata_3a__3a_Nfa_3a__3a_State")
 * cdef object __pyx_convert_pair_to_py_std_3a__3a_pair_3c_Mata_3a__3a_Nfa_3a__3a_State_2c_Mata_3a__3a_Nfa_3a__3a_State_3e_______Mata_3a__3a_Nfa_3a__3a_State(const pair[X,Y]& p):
 *     return p.first, p.second             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State(__pyx_v_p.first); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_p.second); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pair.to_py":190
 * 
 * @cname("__pyx_convert_pair_to_py_std_3a__3a_pair_3c_Mata_3a__3a_Nfa_3a__3a_State_2c_Mata_3a__3a_Nfa_3a__3a_State_3e_______Mata_3a__3a_Nfa_3a__3a_State")
 * cdef object __pyx_convert_pair_to_py_std_3a__3a_pair_3c_Mata_3a__3a_Nfa_3a__3a_State_2c_Mata_3a__3a_Nfa_3a__3a_State_3e_______Mata_3a__3a_Nfa_3a__3a_State(const pair[X,Y]& p):             # <<<<<<<<<<<<<<
 *     return p.first, p.second
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pair.to_py.__pyx_convert_pair_to_py_std_3a__3a_pair_3c_Mata_3a__3a_Nfa_3a__3a_State_2c_Mata_3a__3a_Nfa_3a__3a_State_3e_______Mata_3a__3a_Nfa_3a__3a_State", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "map.from_py":207
 * 
 * @cname("__pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string")
 * cdef map[X,Y] __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef map[X,Y] m
 *     if PY_MAJOR_VERSION < 3:
 */

static std::unordered_map<std::string,std::string>  __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(PyObject *__pyx_v_o) {
  std::unordered_map<std::string,std::string>  __pyx_v_m;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  std::unordered_map<std::string,std::string>  __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  std::string __pyx_t_9;
  std::string __pyx_t_10;
  std::pair<std::string,std::string>  __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string", 0);

  /* "map.from_py":209
 * cdef map[X,Y] __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(object o) except *:
 *     cdef map[X,Y] m
 *     if PY_MAJOR_VERSION < 3:             # <<<<<<<<<<<<<<
 *         for key, value in o.iteritems():
 *             m.insert(pair[X,Y](<X>key, <Y>value))
 */
  __pyx_t_1 = (PY_MAJOR_VERSION < 3);
  if (__pyx_t_1) {

    /* "map.from_py":210
 *     cdef map[X,Y] m
 *     if PY_MAJOR_VERSION < 3:
 *         for key, value in o.iteritems():             # <<<<<<<<<<<<<<
 *             m.insert(pair[X,Y](<X>key, <Y>value))
 *     else:
 */
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_o == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "iteritems");
      __PYX_ERR(0, 210, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_o, 0, __pyx_n_s_iteritems, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "map.from_py":211
 *     if PY_MAJOR_VERSION < 3:
 *         for key, value in o.iteritems():
 *             m.insert(pair[X,Y](<X>key, <Y>value))             # <<<<<<<<<<<<<<
 *     else:
 *         for key, value in o.items():
 */
      __pyx_t_9 = __pyx_convert_string_from_py_std__in_string(__pyx_v_key); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 211, __pyx_L1_error)
      __pyx_t_10 = __pyx_convert_string_from_py_std__in_string(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 211, __pyx_L1_error)
      try {
        __pyx_t_11 = std::pair<std::string,std::string> (((std::string)__pyx_t_9), ((std::string)__pyx_t_10));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 211, __pyx_L1_error)
      }
      try {
        __pyx_v_m.insert(__pyx_t_11);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 211, __pyx_L1_error)
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "map.from_py":209
 * cdef map[X,Y] __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(object o) except *:
 *     cdef map[X,Y] m
 *     if PY_MAJOR_VERSION < 3:             # <<<<<<<<<<<<<<
 *         for key, value in o.iteritems():
 *             m.insert(pair[X,Y](<X>key, <Y>value))
 */
    goto __pyx_L3;
  }

  /* "map.from_py":213
 *             m.insert(pair[X,Y](<X>key, <Y>value))
 *     else:
 *         for key, value in o.items():             # <<<<<<<<<<<<<<
 *             m.insert(pair[X,Y](<X>key, <Y>value))
 *     return m
 */
  /*else*/ {
    __pyx_t_4 = 0;
    if (unlikely(__pyx_v_o == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 213, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_v_o, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_5)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_3, &__pyx_t_4, &__pyx_t_7, &__pyx_t_6, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "map.from_py":214
 *     else:
 *         for key, value in o.items():
 *             m.insert(pair[X,Y](<X>key, <Y>value))             # <<<<<<<<<<<<<<
 *     return m
 * 
 */
      __pyx_t_10 = __pyx_convert_string_from_py_std__in_string(__pyx_v_key); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L1_error)
      __pyx_t_9 = __pyx_convert_string_from_py_std__in_string(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L1_error)
      try {
        __pyx_t_11 = std::pair<std::string,std::string> (((std::string)__pyx_t_10), ((std::string)__pyx_t_9));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 214, __pyx_L1_error)
      }
      try {
        __pyx_v_m.insert(__pyx_t_11);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 214, __pyx_L1_error)
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "map.from_py":215
 *         for key, value in o.items():
 *             m.insert(pair[X,Y](<X>key, <Y>value))
 *     return m             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_m;
  goto __pyx_L0;

  /* "map.from_py":207
 * 
 * @cname("__pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string")
 * cdef map[X,Y] __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef map[X,Y] m
 *     if PY_MAJOR_VERSION < 3:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("map.from_py.__pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::unordered_map<std::string,Mata::Symbol>  __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_Mata_3a__3a_Symbol(PyObject *__pyx_v_o) {
  std::unordered_map<std::string,Mata::Symbol>  __pyx_v_m;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  std::unordered_map<std::string,Mata::Symbol>  __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  std::string __pyx_t_9;
  Mata::Symbol __pyx_t_10;
  std::pair<std::string,Mata::Symbol>  __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_unordered_map_from_py_std_3a__3a_string__and_Mata_3a__3a_Symbol", 0);

  /* "map.from_py":209
 * cdef map[X,Y] __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_Mata_3a__3a_Symbol(object o) except *:
 *     cdef map[X,Y] m
 *     if PY_MAJOR_VERSION < 3:             # <<<<<<<<<<<<<<
 *         for key, value in o.iteritems():
 *             m.insert(pair[X,Y](<X>key, <Y>value))
 */
  __pyx_t_1 = (PY_MAJOR_VERSION < 3);
  if (__pyx_t_1) {

    /* "map.from_py":210
 *     cdef map[X,Y] m
 *     if PY_MAJOR_VERSION < 3:
 *         for key, value in o.iteritems():             # <<<<<<<<<<<<<<
 *             m.insert(pair[X,Y](<X>key, <Y>value))
 *     else:
 */
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_o == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "iteritems");
      __PYX_ERR(0, 210, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_o, 0, __pyx_n_s_iteritems, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "map.from_py":211
 *     if PY_MAJOR_VERSION < 3:
 *         for key, value in o.iteritems():
 *             m.insert(pair[X,Y](<X>key, <Y>value))             # <<<<<<<<<<<<<<
 *     else:
 *         for key, value in o.items():
 */
      __pyx_t_9 = __pyx_convert_string_from_py_std__in_string(__pyx_v_key); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 211, __pyx_L1_error)
      __pyx_t_10 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_10 == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(0, 211, __pyx_L1_error)
      try {
        __pyx_t_11 = std::pair<std::string,Mata::Symbol> (((std::string)__pyx_t_9), ((Mata::Symbol)__pyx_t_10));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 211, __pyx_L1_error)
      }
      try {
        __pyx_v_m.insert(__pyx_t_11);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 211, __pyx_L1_error)
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "map.from_py":209
 * cdef map[X,Y] __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_Mata_3a__3a_Symbol(object o) except *:
 *     cdef map[X,Y] m
 *     if PY_MAJOR_VERSION < 3:             # <<<<<<<<<<<<<<
 *         for key, value in o.iteritems():
 *             m.insert(pair[X,Y](<X>key, <Y>value))
 */
    goto __pyx_L3;
  }

  /* "map.from_py":213
 *             m.insert(pair[X,Y](<X>key, <Y>value))
 *     else:
 *         for key, value in o.items():             # <<<<<<<<<<<<<<
 *             m.insert(pair[X,Y](<X>key, <Y>value))
 *     return m
 */
  /*else*/ {
    __pyx_t_4 = 0;
    if (unlikely(__pyx_v_o == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 213, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_v_o, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_5)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_3, &__pyx_t_4, &__pyx_t_7, &__pyx_t_6, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "map.from_py":214
 *     else:
 *         for key, value in o.items():
 *             m.insert(pair[X,Y](<X>key, <Y>value))             # <<<<<<<<<<<<<<
 *     return m
 * 
 */
      __pyx_t_9 = __pyx_convert_string_from_py_std__in_string(__pyx_v_key); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L1_error)
      __pyx_t_10 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_10 == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L1_error)
      try {
        __pyx_t_11 = std::pair<std::string,Mata::Symbol> (((std::string)__pyx_t_9), ((Mata::Symbol)__pyx_t_10));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 214, __pyx_L1_error)
      }
      try {
        __pyx_v_m.insert(__pyx_t_11);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 214, __pyx_L1_error)
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "map.from_py":215
 *         for key, value in o.items():
 *             m.insert(pair[X,Y](<X>key, <Y>value))
 *     return m             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_m;
  goto __pyx_L0;

  /* "map.from_py":207
 * 
 * @cname("__pyx_convert_unordered_map_from_py_std_3a__3a_string__and_Mata_3a__3a_Symbol")
 * cdef map[X,Y] __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_Mata_3a__3a_Symbol(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef map[X,Y] m
 *     if PY_MAJOR_VERSION < 3:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("map.from_py.__pyx_convert_unordered_map_from_py_std_3a__3a_string__and_Mata_3a__3a_Symbol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_std_3a__3a_string")
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

static std::vector<std::string>  __pyx_convert_vector_from_py_std_3a__3a_string(PyObject *__pyx_v_o) {
  std::vector<std::string>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<std::string>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  std::string __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_std_3a__3a_string", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_string(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __pyx_convert_string_from_py_std__in_string(__pyx_v_item); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 48, __pyx_L1_error)
    try {
      __pyx_v_v.push_back(((std::string)__pyx_t_5));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 48, __pyx_L1_error)
    }

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_string(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_std_3a__3a_string")
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_std_3a__3a_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<uint8_t>  __pyx_convert_vector_from_py_uint8_t(PyObject *__pyx_v_o) {
  std::vector<uint8_t>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<uint8_t>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_uint8_t", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_uint8_t(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_v_item); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 48, __pyx_L1_error)
    try {
      __pyx_v_v.push_back(((uint8_t)__pyx_t_5));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 48, __pyx_L1_error)
    }

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_uint8_t(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_uint8_t")
 * cdef vector[X] __pyx_convert_vector_from_py_uint8_t(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_uint8_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":30
 * cdef Symbol EPSILON = CEPSILON
 * 
 * def epsilon():             # <<<<<<<<<<<<<<
 *     return EPSILON
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_1epsilon(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_1epsilon = {"epsilon", (PyCFunction)__pyx_pw_7libmata_3nfa_3nfa_1epsilon, METH_NOARGS, 0};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_1epsilon(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("epsilon (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_epsilon(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_epsilon(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("epsilon", 0);

  /* "libmata/nfa/nfa.pyx":31
 * 
 * def epsilon():
 *     return EPSILON             # <<<<<<<<<<<<<<
 * 
 * cdef class Run:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_7libmata_3nfa_3nfa_EPSILON); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":30
 * cdef Symbol EPSILON = CEPSILON
 * 
 * def epsilon():             # <<<<<<<<<<<<<<
 *     return EPSILON
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.epsilon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":37
 *     cdef mata_nfa.CRun *thisptr
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         """Constructor of the transition
 * 
 */

/* Python wrapper */
static int __pyx_pw_7libmata_3nfa_3nfa_3Run_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7libmata_3nfa_3nfa_3Run_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Run___cinit__(((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7libmata_3nfa_3nfa_3Run___cinit__(struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Mata::Nfa::Run *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "libmata/nfa/nfa.pyx":44
 *         :param State tgt: target state
 *         """
 *         self.thisptr = new mata_nfa.CRun()             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  try {
    __pyx_t_1 = new Mata::Nfa::Run();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 44, __pyx_L1_error)
  }
  __pyx_v_self->thisptr = __pyx_t_1;

  /* "libmata/nfa/nfa.pyx":37
 *     cdef mata_nfa.CRun *thisptr
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         """Constructor of the transition
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Run.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":46
 *         self.thisptr = new mata_nfa.CRun()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         """Destructor"""
 *         if self.thisptr != NULL:
 */

/* Python wrapper */
static void __pyx_pw_7libmata_3nfa_3nfa_3Run_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7libmata_3nfa_3nfa_3Run_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7libmata_3nfa_3nfa_3Run_2__dealloc__(((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7libmata_3nfa_3nfa_3Run_2__dealloc__(struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "libmata/nfa/nfa.pyx":48
 *     def __dealloc__(self):
 *         """Destructor"""
 *         if self.thisptr != NULL:             # <<<<<<<<<<<<<<
 *             del self.thisptr
 * 
 */
  __pyx_t_1 = (__pyx_v_self->thisptr != NULL);
  if (__pyx_t_1) {

    /* "libmata/nfa/nfa.pyx":49
 *         """Destructor"""
 *         if self.thisptr != NULL:
 *             del self.thisptr             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
    delete __pyx_v_self->thisptr;

    /* "libmata/nfa/nfa.pyx":48
 *     def __dealloc__(self):
 *         """Destructor"""
 *         if self.thisptr != NULL:             # <<<<<<<<<<<<<<
 *             del self.thisptr
 * 
 */
  }

  /* "libmata/nfa/nfa.pyx":46
 *         self.thisptr = new mata_nfa.CRun()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         """Destructor"""
 *         if self.thisptr != NULL:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "libmata/nfa/nfa.pyx":51
 *             del self.thisptr
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def word(self):
 *         return self.thisptr.word
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Run_4word_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Run_4word_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Run_4word___get__(((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Run_4word___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "libmata/nfa/nfa.pyx":53
 *     @property
 *     def word(self):
 *         return self.thisptr.word             # <<<<<<<<<<<<<<
 * 
 *     @word.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_Mata_3a__3a_Symbol(__pyx_v_self->thisptr->word); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":51
 *             del self.thisptr
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def word(self):
 *         return self.thisptr.word
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Run.word.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":55
 *         return self.thisptr.word
 * 
 *     @word.setter             # <<<<<<<<<<<<<<
 *     def word(self, value):
 *         self.thisptr.word = value
 */

/* Python wrapper */
static int __pyx_pw_7libmata_3nfa_3nfa_3Run_4word_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7libmata_3nfa_3nfa_3Run_4word_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Run_4word_2__set__(((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7libmata_3nfa_3nfa_3Run_4word_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  std::vector<Mata::Symbol>  __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "libmata/nfa/nfa.pyx":57
 *     @word.setter
 *     def word(self, value):
 *         self.thisptr.word = value             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __pyx_convert_vector_from_py_Mata_3a__3a_Symbol(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 57, __pyx_L1_error)
  __pyx_v_self->thisptr->word = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "libmata/nfa/nfa.pyx":55
 *         return self.thisptr.word
 * 
 *     @word.setter             # <<<<<<<<<<<<<<
 *     def word(self, value):
 *         self.thisptr.word = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Run.word.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":59
 *         self.thisptr.word = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def path(self):
 *         return self.thisptr.path
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Run_4path_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Run_4path_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Run_4path___get__(((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Run_4path___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "libmata/nfa/nfa.pyx":61
 *     @property
 *     def path(self):
 *         return self.thisptr.path             # <<<<<<<<<<<<<<
 * 
 *     @path.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State(__pyx_v_self->thisptr->path); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":59
 *         self.thisptr.word = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def path(self):
 *         return self.thisptr.path
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Run.path.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":63
 *         return self.thisptr.path
 * 
 *     @path.setter             # <<<<<<<<<<<<<<
 *     def path(self, value):
 *         self.thisptr.path = value
 */

/* Python wrapper */
static int __pyx_pw_7libmata_3nfa_3nfa_3Run_4path_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7libmata_3nfa_3nfa_3Run_4path_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Run_4path_2__set__(((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7libmata_3nfa_3nfa_3Run_4path_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  std::vector<Mata::Nfa::State>  __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "libmata/nfa/nfa.pyx":65
 *     @path.setter
 *     def path(self, value):
 *         self.thisptr.path = value             # <<<<<<<<<<<<<<
 * 
 * cdef class Trans:
 */
  __pyx_t_1 = __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_v_self->thisptr->path = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "libmata/nfa/nfa.pyx":63
 *         return self.thisptr.path
 * 
 *     @path.setter             # <<<<<<<<<<<<<<
 *     def path(self, value):
 *         self.thisptr.path = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Run.path.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Run_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Run_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Run_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Run_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Run_4__reduce_cython__(((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Run_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Run.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Run_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Run_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Run_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Run_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Run.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Run_6__setstate_cython__(((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Run_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Run.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":70
 *     """Wrapper over the transitions in NFA."""
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def src(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_3src_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_3src_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_5Trans_3src___get__(((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_3src___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "libmata/nfa/nfa.pyx":75
 *         :return: source state of the transition
 *         """
 *         return self.thisptr.src             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->thisptr->src); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":70
 *     """Wrapper over the transitions in NFA."""
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def src(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.src.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":77
 *         return self.thisptr.src
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def symb(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_4symb_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_4symb_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_5Trans_4symb___get__(((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_4symb___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "libmata/nfa/nfa.pyx":82
 *         :return: symbol for the transition
 *         """
 *         return self.thisptr.symb             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->thisptr->symb); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":77
 *         return self.thisptr.src
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def symb(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.symb.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":84
 *         return self.thisptr.symb
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tgt(self):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_3tgt_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_3tgt_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_5Trans_3tgt___get__(((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_3tgt___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "libmata/nfa/nfa.pyx":89
 *         :return: target state of the transition
 *         """
 *         return self.thisptr.tgt             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, State src=0, Symbol s=0, State tgt=0):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->thisptr->tgt); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":84
 *         return self.thisptr.symb
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tgt(self):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.tgt.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":91
 *         return self.thisptr.tgt
 * 
 *     def __cinit__(self, State src=0, Symbol s=0, State tgt=0):             # <<<<<<<<<<<<<<
 *         """Constructor of the transition
 * 
 */

/* Python wrapper */
static int __pyx_pw_7libmata_3nfa_3nfa_5Trans_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7libmata_3nfa_3nfa_5Trans_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Mata::Nfa::State __pyx_v_src;
  Mata::Symbol __pyx_v_s;
  Mata::Nfa::State __pyx_v_tgt;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_src,&__pyx_n_s_s,&__pyx_n_s_tgt,0};
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_src);
          if (value) { values[0] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 91, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_s);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 91, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tgt);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 91, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 91, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_src = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_src == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 91, __pyx_L3_error)
    } else {
      __pyx_v_src = ((Mata::Nfa::State)0);
    }
    if (values[1]) {
      __pyx_v_s = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_s == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 91, __pyx_L3_error)
    } else {
      __pyx_v_s = ((Mata::Symbol)0);
    }
    if (values[2]) {
      __pyx_v_tgt = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_tgt == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 91, __pyx_L3_error)
    } else {
      __pyx_v_tgt = ((Mata::Nfa::State)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 3, __pyx_nargs); __PYX_ERR(1, 91, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_5Trans___cinit__(((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_v_self), __pyx_v_src, __pyx_v_s, __pyx_v_tgt);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7libmata_3nfa_3nfa_5Trans___cinit__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self, Mata::Nfa::State __pyx_v_src, Mata::Symbol __pyx_v_s, Mata::Nfa::State __pyx_v_tgt) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Mata::Nfa::Trans *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "libmata/nfa/nfa.pyx":98
 *         :param State tgt: target state
 *         """
 *         self.thisptr = new mata_nfa.CTrans(src, s, tgt)             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  try {
    __pyx_t_1 = new Mata::Nfa::Trans(__pyx_v_src, __pyx_v_s, __pyx_v_tgt);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 98, __pyx_L1_error)
  }
  __pyx_v_self->thisptr = __pyx_t_1;

  /* "libmata/nfa/nfa.pyx":91
 *         return self.thisptr.tgt
 * 
 *     def __cinit__(self, State src=0, Symbol s=0, State tgt=0):             # <<<<<<<<<<<<<<
 *         """Constructor of the transition
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":100
 *         self.thisptr = new mata_nfa.CTrans(src, s, tgt)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         """Destructor"""
 *         if self.thisptr != NULL:
 */

/* Python wrapper */
static void __pyx_pw_7libmata_3nfa_3nfa_5Trans_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7libmata_3nfa_3nfa_5Trans_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7libmata_3nfa_3nfa_5Trans_2__dealloc__(((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7libmata_3nfa_3nfa_5Trans_2__dealloc__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "libmata/nfa/nfa.pyx":102
 *     def __dealloc__(self):
 *         """Destructor"""
 *         if self.thisptr != NULL:             # <<<<<<<<<<<<<<
 *             del self.thisptr
 * 
 */
  __pyx_t_1 = (__pyx_v_self->thisptr != NULL);
  if (__pyx_t_1) {

    /* "libmata/nfa/nfa.pyx":103
 *         """Destructor"""
 *         if self.thisptr != NULL:
 *             del self.thisptr             # <<<<<<<<<<<<<<
 * 
 *     cdef copy_from(self, CTrans trans):
 */
    delete __pyx_v_self->thisptr;

    /* "libmata/nfa/nfa.pyx":102
 *     def __dealloc__(self):
 *         """Destructor"""
 *         if self.thisptr != NULL:             # <<<<<<<<<<<<<<
 *             del self.thisptr
 * 
 */
  }

  /* "libmata/nfa/nfa.pyx":100
 *         self.thisptr = new mata_nfa.CTrans(src, s, tgt)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         """Destructor"""
 *         if self.thisptr != NULL:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "libmata/nfa/nfa.pyx":105
 *             del self.thisptr
 * 
 *     cdef copy_from(self, CTrans trans):             # <<<<<<<<<<<<<<
 *         """Copies the internals of trans into the wrapped pointer
 * 
 */

static PyObject *__pyx_f_7libmata_3nfa_3nfa_5Trans_copy_from(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self, Mata::Nfa::Trans __pyx_v_trans) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Mata::Nfa::State __pyx_t_1;
  Mata::Symbol __pyx_t_2;
  __Pyx_RefNannySetupContext("copy_from", 0);

  /* "libmata/nfa/nfa.pyx":110
 *         :param CTrans trans: copied transition
 *         """
 *         self.thisptr.src = trans.src             # <<<<<<<<<<<<<<
 *         self.thisptr.symb = trans.symb
 *         self.thisptr.tgt = trans.tgt
 */
  __pyx_t_1 = __pyx_v_trans.src;
  __pyx_v_self->thisptr->src = __pyx_t_1;

  /* "libmata/nfa/nfa.pyx":111
 *         """
 *         self.thisptr.src = trans.src
 *         self.thisptr.symb = trans.symb             # <<<<<<<<<<<<<<
 *         self.thisptr.tgt = trans.tgt
 * 
 */
  __pyx_t_2 = __pyx_v_trans.symb;
  __pyx_v_self->thisptr->symb = __pyx_t_2;

  /* "libmata/nfa/nfa.pyx":112
 *         self.thisptr.src = trans.src
 *         self.thisptr.symb = trans.symb
 *         self.thisptr.tgt = trans.tgt             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, Trans other):
 */
  __pyx_t_1 = __pyx_v_trans.tgt;
  __pyx_v_self->thisptr->tgt = __pyx_t_1;

  /* "libmata/nfa/nfa.pyx":105
 *             del self.thisptr
 * 
 *     cdef copy_from(self, CTrans trans):             # <<<<<<<<<<<<<<
 *         """Copies the internals of trans into the wrapped pointer
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":114
 *         self.thisptr.tgt = trans.tgt
 * 
 *     def __eq__(self, Trans other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) == dereference(other.thisptr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_5__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_5__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7libmata_3nfa_3nfa_Trans, 1, "other", 0))) __PYX_ERR(1, 114, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_5Trans_4__eq__(((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_v_self), ((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_4__eq__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "libmata/nfa/nfa.pyx":115
 * 
 *     def __eq__(self, Trans other):
 *         return dereference(self.thisptr) == dereference(other.thisptr)             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, Trans other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(((*__pyx_v_self->thisptr) == (*__pyx_v_other->thisptr))); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":114
 *         self.thisptr.tgt = trans.tgt
 * 
 *     def __eq__(self, Trans other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) == dereference(other.thisptr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":117
 *         return dereference(self.thisptr) == dereference(other.thisptr)
 * 
 *     def __neq__(self, Trans other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) != dereference(other.thisptr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_7__neq__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_5Trans_7__neq__ = {"__neq__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_5Trans_7__neq__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_7__neq__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neq__ (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 117, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__neq__") < 0)) __PYX_ERR(1, 117, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__neq__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 117, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.__neq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7libmata_3nfa_3nfa_Trans, 1, "other", 0))) __PYX_ERR(1, 117, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_5Trans_6__neq__(((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_6__neq__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__neq__", 0);

  /* "libmata/nfa/nfa.pyx":118
 * 
 *     def __neq__(self, Trans other):
 *         return dereference(self.thisptr) != dereference(other.thisptr)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(((*__pyx_v_self->thisptr) != (*__pyx_v_other->thisptr))); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":117
 *         return dereference(self.thisptr) == dereference(other.thisptr)
 * 
 *     def __neq__(self, Trans other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) != dereference(other.thisptr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.__neq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":120
 *         return dereference(self.thisptr) != dereference(other.thisptr)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f"{self.thisptr.src}-[{self.thisptr.symb}]\u2192{self.thisptr.tgt}"
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_9__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_9__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_5Trans_8__str__(((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_8__str__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "libmata/nfa/nfa.pyx":121
 * 
 *     def __str__(self):
 *         return f"{self.thisptr.src}-[{self.thisptr.symb}]\u2192{self.thisptr.tgt}"             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_self->thisptr->src); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_);
  __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_self->thisptr->symb); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_3 = (65535 > __pyx_t_3) ? 65535 : __pyx_t_3;
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_self->thisptr->tgt); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":120
 *         return dereference(self.thisptr) != dereference(other.thisptr)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return f"{self.thisptr.src}-[{self.thisptr.symb}]\u2192{self.thisptr.tgt}"
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":123
 *         return f"{self.thisptr.src}-[{self.thisptr.symb}]\u2192{self.thisptr.tgt}"
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_11__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_11__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_5Trans_10__repr__(((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_10__repr__(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "libmata/nfa/nfa.pyx":124
 * 
 *     def __repr__(self):
 *         return str(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Str(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":123
 *         return f"{self.thisptr.src}-[{self.thisptr.symb}]\u2192{self.thisptr.tgt}"
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_5Trans_13__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_5Trans_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_5Trans_12__reduce_cython__(((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_5Trans_15__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_5Trans_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5Trans_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_5Trans_14__setstate_cython__(((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_5Trans_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Trans.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":131
 *     cdef mata_nfa.CMove *thisptr
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def symbol(self):
 *         return self.thisptr.symbol
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_6symbol_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_6symbol_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_6symbol___get__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_6symbol___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "libmata/nfa/nfa.pyx":133
 *     @property
 *     def symbol(self):
 *         return self.thisptr.symbol             # <<<<<<<<<<<<<<
 * 
 *     @symbol.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->thisptr->symbol); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":131
 *     cdef mata_nfa.CMove *thisptr
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def symbol(self):
 *         return self.thisptr.symbol
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.symbol.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":135
 *         return self.thisptr.symbol
 * 
 *     @symbol.setter             # <<<<<<<<<<<<<<
 *     def symbol(self, value):
 *         self.thisptr.symbol = value
 */

/* Python wrapper */
static int __pyx_pw_7libmata_3nfa_3nfa_4Move_6symbol_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7libmata_3nfa_3nfa_4Move_6symbol_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_6symbol_2__set__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7libmata_3nfa_3nfa_4Move_6symbol_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Mata::Symbol __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "libmata/nfa/nfa.pyx":137
 *     @symbol.setter
 *     def symbol(self, value):
 *         self.thisptr.symbol = value             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 137, __pyx_L1_error)
  __pyx_v_self->thisptr->symbol = __pyx_t_1;

  /* "libmata/nfa/nfa.pyx":135
 *         return self.thisptr.symbol
 * 
 *     @symbol.setter             # <<<<<<<<<<<<<<
 *     def symbol(self, value):
 *         self.thisptr.symbol = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.symbol.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":139
 *         self.thisptr.symbol = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def targets(self):
 *         cdef vector[State] states_as_vector = self.thisptr.targets.ToVector()
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_7targets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_7targets_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_7targets___get__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_7targets___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self) {
  std::vector<Mata::Nfa::State>  __pyx_v_states_as_vector;
  Mata::Nfa::State __pyx_7genexpr__pyx_v_s;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::vector<Mata::Nfa::State> ::iterator __pyx_t_2;
  Mata::Nfa::State __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "libmata/nfa/nfa.pyx":141
 *     @property
 *     def targets(self):
 *         cdef vector[State] states_as_vector = self.thisptr.targets.ToVector()             # <<<<<<<<<<<<<<
 *         return [s for s in states_as_vector]
 * 
 */
  __pyx_v_states_as_vector = ((std::vector<Mata::Nfa::State> )__pyx_v_self->thisptr->targets.ToVector());

  /* "libmata/nfa/nfa.pyx":142
 *     def targets(self):
 *         cdef vector[State] states_as_vector = self.thisptr.targets.ToVector()
 *         return [s for s in states_as_vector]             # <<<<<<<<<<<<<<
 * 
 *     @targets.setter
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_states_as_vector.begin();
    for (;;) {
      if (!(__pyx_t_2 != __pyx_v_states_as_vector.end())) break;
      __pyx_t_3 = *__pyx_t_2;
      ++__pyx_t_2;
      __pyx_7genexpr__pyx_v_s = __pyx_t_3;
      __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_7genexpr__pyx_v_s); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 142, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(1, 142, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":139
 *         self.thisptr.symbol = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def targets(self):
 *         cdef vector[State] states_as_vector = self.thisptr.targets.ToVector()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.targets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":144
 *         return [s for s in states_as_vector]
 * 
 *     @targets.setter             # <<<<<<<<<<<<<<
 *     def targets(self, value):
 *         cdef StateSet targets = StateSet(value)
 */

/* Python wrapper */
static int __pyx_pw_7libmata_3nfa_3nfa_4Move_7targets_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7libmata_3nfa_3nfa_4Move_7targets_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_7targets_2__set__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7libmata_3nfa_3nfa_4Move_7targets_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, PyObject *__pyx_v_value) {
  Mata::Nfa::StateSet __pyx_v_targets;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  std::vector<Mata::Nfa::State>  __pyx_t_1;
  Mata::Nfa::StateSet __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "libmata/nfa/nfa.pyx":146
 *     @targets.setter
 *     def targets(self, value):
 *         cdef StateSet targets = StateSet(value)             # <<<<<<<<<<<<<<
 *         self.thisptr.targets = targets
 * 
 */
  __pyx_t_1 = __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 146, __pyx_L1_error)
  try {
    __pyx_t_2 = Mata::Nfa::StateSet(__pyx_t_1);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 146, __pyx_L1_error)
  }
  __pyx_v_targets = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_2);

  /* "libmata/nfa/nfa.pyx":147
 *     def targets(self, value):
 *         cdef StateSet targets = StateSet(value)
 *         self.thisptr.targets = targets             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, Symbol symbol, vector[State] states):
 */
  __pyx_v_self->thisptr->targets = __pyx_v_targets;

  /* "libmata/nfa/nfa.pyx":144
 *         return [s for s in states_as_vector]
 * 
 *     @targets.setter             # <<<<<<<<<<<<<<
 *     def targets(self, value):
 *         cdef StateSet targets = StateSet(value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.targets.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":149
 *         self.thisptr.targets = targets
 * 
 *     def __cinit__(self, Symbol symbol, vector[State] states):             # <<<<<<<<<<<<<<
 *         cdef StateSet targets = StateSet(states)
 *         self.thisptr = new mata_nfa.CMove(symbol, targets)
 */

/* Python wrapper */
static int __pyx_pw_7libmata_3nfa_3nfa_4Move_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7libmata_3nfa_3nfa_4Move_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Mata::Symbol __pyx_v_symbol;
  std::vector<Mata::Nfa::State>  __pyx_v_states;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_symbol,&__pyx_n_s_states,0};
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_symbol)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 149, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_states)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 149, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(1, 149, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 149, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_symbol = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_symbol == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 149, __pyx_L3_error)
    __pyx_v_states = __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(values[1]); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 149, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, __pyx_nargs); __PYX_ERR(1, 149, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move___cinit__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self), __pyx_v_symbol, __PYX_STD_MOVE_IF_SUPPORTED(__pyx_v_states));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7libmata_3nfa_3nfa_4Move___cinit__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, Mata::Symbol __pyx_v_symbol, std::vector<Mata::Nfa::State>  __pyx_v_states) {
  Mata::Nfa::StateSet __pyx_v_targets;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Mata::Nfa::StateSet __pyx_t_1;
  Mata::Nfa::Move *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "libmata/nfa/nfa.pyx":150
 * 
 *     def __cinit__(self, Symbol symbol, vector[State] states):
 *         cdef StateSet targets = StateSet(states)             # <<<<<<<<<<<<<<
 *         self.thisptr = new mata_nfa.CMove(symbol, targets)
 * 
 */
  try {
    __pyx_t_1 = Mata::Nfa::StateSet(((std::vector<Mata::Nfa::State>  &)__pyx_v_states));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 150, __pyx_L1_error)
  }
  __pyx_v_targets = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "libmata/nfa/nfa.pyx":151
 *     def __cinit__(self, Symbol symbol, vector[State] states):
 *         cdef StateSet targets = StateSet(states)
 *         self.thisptr = new mata_nfa.CMove(symbol, targets)             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  try {
    __pyx_t_2 = new Mata::Nfa::Move(__pyx_v_symbol, __pyx_v_targets);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 151, __pyx_L1_error)
  }
  __pyx_v_self->thisptr = __pyx_t_2;

  /* "libmata/nfa/nfa.pyx":149
 *         self.thisptr.targets = targets
 * 
 *     def __cinit__(self, Symbol symbol, vector[State] states):             # <<<<<<<<<<<<<<
 *         cdef StateSet targets = StateSet(states)
 *         self.thisptr = new mata_nfa.CMove(symbol, targets)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":153
 *         self.thisptr = new mata_nfa.CMove(symbol, targets)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.thisptr != NULL:
 *             del self.thisptr
 */

/* Python wrapper */
static void __pyx_pw_7libmata_3nfa_3nfa_4Move_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7libmata_3nfa_3nfa_4Move_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7libmata_3nfa_3nfa_4Move_2__dealloc__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7libmata_3nfa_3nfa_4Move_2__dealloc__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "libmata/nfa/nfa.pyx":154
 * 
 *     def __dealloc__(self):
 *         if self.thisptr != NULL:             # <<<<<<<<<<<<<<
 *             del self.thisptr
 * 
 */
  __pyx_t_1 = (__pyx_v_self->thisptr != NULL);
  if (__pyx_t_1) {

    /* "libmata/nfa/nfa.pyx":155
 *     def __dealloc__(self):
 *         if self.thisptr != NULL:
 *             del self.thisptr             # <<<<<<<<<<<<<<
 * 
 *     def __lt__(self, Move other):
 */
    delete __pyx_v_self->thisptr;

    /* "libmata/nfa/nfa.pyx":154
 * 
 *     def __dealloc__(self):
 *         if self.thisptr != NULL:             # <<<<<<<<<<<<<<
 *             del self.thisptr
 * 
 */
  }

  /* "libmata/nfa/nfa.pyx":153
 *         self.thisptr = new mata_nfa.CMove(symbol, targets)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.thisptr != NULL:
 *             del self.thisptr
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "libmata/nfa/nfa.pyx":157
 *             del self.thisptr
 * 
 *     def __lt__(self, Move other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) < dereference(other.thisptr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_5__lt__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_5__lt__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__lt__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7libmata_3nfa_3nfa_Move, 1, "other", 0))) __PYX_ERR(1, 157, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_4__lt__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self), ((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_4__lt__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__lt__", 0);

  /* "libmata/nfa/nfa.pyx":158
 * 
 *     def __lt__(self, Move other):
 *         return dereference(self.thisptr) < dereference(other.thisptr)             # <<<<<<<<<<<<<<
 * 
 *     def __gt__(self, Move other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(((*__pyx_v_self->thisptr) < (*__pyx_v_other->thisptr))); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":157
 *             del self.thisptr
 * 
 *     def __lt__(self, Move other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) < dereference(other.thisptr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__lt__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":160
 *         return dereference(self.thisptr) < dereference(other.thisptr)
 * 
 *     def __gt__(self, Move other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) > dereference(other.thisptr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_7__gt__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_7__gt__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__gt__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7libmata_3nfa_3nfa_Move, 1, "other", 0))) __PYX_ERR(1, 160, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_6__gt__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self), ((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_6__gt__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__gt__", 0);

  /* "libmata/nfa/nfa.pyx":161
 * 
 *     def __gt__(self, Move other):
 *         return dereference(self.thisptr) > dereference(other.thisptr)             # <<<<<<<<<<<<<<
 * 
 *     def __le__(self, Move other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(((*__pyx_v_self->thisptr) > (*__pyx_v_other->thisptr))); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":160
 *         return dereference(self.thisptr) < dereference(other.thisptr)
 * 
 *     def __gt__(self, Move other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) > dereference(other.thisptr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__gt__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":163
 *         return dereference(self.thisptr) > dereference(other.thisptr)
 * 
 *     def __le__(self, Move other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) <= dereference(other.thisptr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_9__le__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_9__le__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__le__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7libmata_3nfa_3nfa_Move, 1, "other", 0))) __PYX_ERR(1, 163, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_8__le__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self), ((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_8__le__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__le__", 0);

  /* "libmata/nfa/nfa.pyx":164
 * 
 *     def __le__(self, Move other):
 *         return dereference(self.thisptr) <= dereference(other.thisptr)             # <<<<<<<<<<<<<<
 * 
 *     def __ge__(self, Move other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(((*__pyx_v_self->thisptr) <= (*__pyx_v_other->thisptr))); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":163
 *         return dereference(self.thisptr) > dereference(other.thisptr)
 * 
 *     def __le__(self, Move other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) <= dereference(other.thisptr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__le__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":166
 *         return dereference(self.thisptr) <= dereference(other.thisptr)
 * 
 *     def __ge__(self, Move other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) >= dereference(other.thisptr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_11__ge__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_11__ge__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ge__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7libmata_3nfa_3nfa_Move, 1, "other", 0))) __PYX_ERR(1, 166, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_10__ge__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self), ((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_10__ge__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ge__", 0);

  /* "libmata/nfa/nfa.pyx":167
 * 
 *     def __ge__(self, Move other):
 *         return dereference(self.thisptr) >= dereference(other.thisptr)             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, Move other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(((*__pyx_v_self->thisptr) >= (*__pyx_v_other->thisptr))); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":166
 *         return dereference(self.thisptr) <= dereference(other.thisptr)
 * 
 *     def __ge__(self, Move other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) >= dereference(other.thisptr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__ge__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":169
 *         return dereference(self.thisptr) >= dereference(other.thisptr)
 * 
 *     def __eq__(self, Move other):             # <<<<<<<<<<<<<<
 *         return self.symbol == other.symbol and self.targets == other.targets
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_13__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_13__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7libmata_3nfa_3nfa_Move, 1, "other", 0))) __PYX_ERR(1, 169, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_12__eq__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self), ((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_12__eq__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "libmata/nfa/nfa.pyx":170
 * 
 *     def __eq__(self, Move other):
 *         return self.symbol == other.symbol and self.targets == other.targets             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, Move other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_symbol); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_symbol); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(1, 170, __pyx_L1_error)
  if (__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_targets); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_targets); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":169
 *         return dereference(self.thisptr) >= dereference(other.thisptr)
 * 
 *     def __eq__(self, Move other):             # <<<<<<<<<<<<<<
 *         return self.symbol == other.symbol and self.targets == other.targets
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":172
 *         return self.symbol == other.symbol and self.targets == other.targets
 * 
 *     def __neq__(self, Move other):             # <<<<<<<<<<<<<<
 *         return self.symbol != other.symbol or self.targets != other.targets
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_15__neq__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_4Move_15__neq__ = {"__neq__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_4Move_15__neq__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_15__neq__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neq__ (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 172, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__neq__") < 0)) __PYX_ERR(1, 172, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__neq__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 172, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__neq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7libmata_3nfa_3nfa_Move, 1, "other", 0))) __PYX_ERR(1, 172, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_14__neq__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_14__neq__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__neq__", 0);

  /* "libmata/nfa/nfa.pyx":173
 * 
 *     def __neq__(self, Move other):
 *         return self.symbol != other.symbol or self.targets != other.targets             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_symbol); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_symbol); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(1, 173, __pyx_L1_error)
  if (!__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_targets); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_targets); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":172
 *         return self.symbol == other.symbol and self.targets == other.targets
 * 
 *     def __neq__(self, Move other):             # <<<<<<<<<<<<<<
 *         return self.symbol != other.symbol or self.targets != other.targets
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__neq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":175
 *         return self.symbol != other.symbol or self.targets != other.targets
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         trans = "{" + ",".join(map(str, self.targets)) + "}"
 *         return f"[{self.symbol}]\u2192{trans}"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_17__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_17__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_16__str__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_16__str__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self) {
  PyObject *__pyx_v_trans = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "libmata/nfa/nfa.pyx":176
 * 
 *     def __str__(self):
 *         trans = "{" + ",".join(map(str, self.targets)) + "}"             # <<<<<<<<<<<<<<
 *         return f"[{self.symbol}]\u2192{trans}"
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_targets); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)(&PyUnicode_Type));
  __Pyx_GIVEREF((PyObject *)(&PyUnicode_Type));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&PyUnicode_Type)));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyUnicode_Join(__pyx_kp_u__4, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_kp_u__3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_1, __pyx_kp_u__5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_trans = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":177
 *     def __str__(self):
 *         trans = "{" + ",".join(map(str, self.targets)) + "}"
 *         return f"[{self.symbol}]\u2192{trans}"             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 127;
  __Pyx_INCREF(__pyx_kp_u__6);
  __pyx_t_3 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__6);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u__6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_symbol); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
  __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_4 = (65535 > __pyx_t_4) ? 65535 : __pyx_t_4;
  __pyx_t_3 += 2;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u__2);
  __Pyx_INCREF(__pyx_v_trans);
  __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_trans) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_trans) : __pyx_t_4;
  __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_v_trans);
  __Pyx_GIVEREF(__pyx_v_trans);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_v_trans);
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 4, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":175
 *         return self.symbol != other.symbol or self.targets != other.targets
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         trans = "{" + ",".join(map(str, self.targets)) + "}"
 *         return f"[{self.symbol}]\u2192{trans}"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_trans);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":179
 *         return f"[{self.symbol}]\u2192{trans}"
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_19__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_19__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_18__repr__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_18__repr__(struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "libmata/nfa/nfa.pyx":180
 * 
 *     def __repr__(self):
 *         return str(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Str(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":179
 *         return f"[{self.symbol}]\u2192{trans}"
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_21__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_4Move_21__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_4Move_21__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_21__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_20__reduce_cython__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_23__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_4Move_23__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_4Move_23__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_4Move_23__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4Move_22__setstate_cython__(((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4Move_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Move.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":189
 *     """
 * 
 *     def __cinit__(self, state_number = 0, alph.Alphabet alphabet = None, label=None):             # <<<<<<<<<<<<<<
 *         """Constructor of the NFA.
 * 
 */

/* Python wrapper */
static int __pyx_pw_7libmata_3nfa_3nfa_3Nfa_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7libmata_3nfa_3nfa_3Nfa_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_state_number = 0;
  struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet = 0;
  PyObject *__pyx_v_label = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state_number,&__pyx_n_s_alphabet,&__pyx_n_s_label,0};
    PyObject* values[3] = {0,0,0};
    values[0] = ((PyObject *)__pyx_int_0);
    values[1] = (PyObject *)((struct __pyx_obj_7libmata_9alphabets_Alphabet *)Py_None);
    values[2] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state_number);
          if (value) { values[0] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 189, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_alphabet);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 189, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_label);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 189, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 189, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_state_number = values[0];
    __pyx_v_alphabet = ((struct __pyx_obj_7libmata_9alphabets_Alphabet *)values[1]);
    __pyx_v_label = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 3, __pyx_nargs); __PYX_ERR(1, 189, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_alphabet), __pyx_ptype_7libmata_9alphabets_Alphabet, 1, "alphabet", 0))) __PYX_ERR(1, 189, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa___cinit__(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_state_number, __pyx_v_alphabet, __pyx_v_label);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7libmata_3nfa_3nfa_3Nfa___cinit__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, PyObject *__pyx_v_state_number, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet, PyObject *__pyx_v_label) {
  Mata::Alphabet *__pyx_v_c_alphabet;
  Mata::Nfa::StateSet __pyx_v_empty_default_state_set;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Mata::Alphabet *__pyx_t_2;
  unsigned long __pyx_t_3;
  std::shared_ptr<Mata::Nfa::Nfa>  __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "libmata/nfa/nfa.pyx":195
 *         :param alph.Alphabet alphabet: alphabet corresponding to the automaton
 *         """
 *         cdef CAlphabet* c_alphabet = NULL             # <<<<<<<<<<<<<<
 *         cdef StateSet empty_default_state_set
 *         if alphabet:
 */
  __pyx_v_c_alphabet = NULL;

  /* "libmata/nfa/nfa.pyx":197
 *         cdef CAlphabet* c_alphabet = NULL
 *         cdef StateSet empty_default_state_set
 *         if alphabet:             # <<<<<<<<<<<<<<
 *             c_alphabet = alphabet.as_base()
 *         self.thisptr = make_shared[CNfa](
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_alphabet)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 197, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "libmata/nfa/nfa.pyx":198
 *         cdef StateSet empty_default_state_set
 *         if alphabet:
 *             c_alphabet = alphabet.as_base()             # <<<<<<<<<<<<<<
 *         self.thisptr = make_shared[CNfa](
 *             mata_nfa.CNfa(state_number, empty_default_state_set, empty_default_state_set, c_alphabet)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_7libmata_9alphabets_Alphabet *)__pyx_v_alphabet->__pyx_vtab)->as_base(__pyx_v_alphabet); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 198, __pyx_L1_error)
    __pyx_v_c_alphabet = __pyx_t_2;

    /* "libmata/nfa/nfa.pyx":197
 *         cdef CAlphabet* c_alphabet = NULL
 *         cdef StateSet empty_default_state_set
 *         if alphabet:             # <<<<<<<<<<<<<<
 *             c_alphabet = alphabet.as_base()
 *         self.thisptr = make_shared[CNfa](
 */
  }

  /* "libmata/nfa/nfa.pyx":200
 *             c_alphabet = alphabet.as_base()
 *         self.thisptr = make_shared[CNfa](
 *             mata_nfa.CNfa(state_number, empty_default_state_set, empty_default_state_set, c_alphabet)             # <<<<<<<<<<<<<<
 *         )
 *         self.label = label
 */
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_long(__pyx_v_state_number); if (unlikely((__pyx_t_3 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 200, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":199
 *         if alphabet:
 *             c_alphabet = alphabet.as_base()
 *         self.thisptr = make_shared[CNfa](             # <<<<<<<<<<<<<<
 *             mata_nfa.CNfa(state_number, empty_default_state_set, empty_default_state_set, c_alphabet)
 *         )
 */
  try {
    __pyx_t_4 = std::make_shared<Mata::Nfa::Nfa>(Mata::Nfa::Nfa(__pyx_t_3, __pyx_v_empty_default_state_set, __pyx_v_empty_default_state_set, __pyx_v_c_alphabet));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 199, __pyx_L1_error)
  }
  __pyx_v_self->thisptr = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4);

  /* "libmata/nfa/nfa.pyx":202
 *             mata_nfa.CNfa(state_number, empty_default_state_set, empty_default_state_set, c_alphabet)
 *         )
 *         self.label = label             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF(__pyx_v_label);
  __Pyx_GIVEREF(__pyx_v_label);
  __Pyx_GOTREF(__pyx_v_self->label);
  __Pyx_DECREF(__pyx_v_self->label);
  __pyx_v_self->label = __pyx_v_label;

  /* "libmata/nfa/nfa.pyx":189
 *     """
 * 
 *     def __cinit__(self, state_number = 0, alph.Alphabet alphabet = None, label=None):             # <<<<<<<<<<<<<<
 *         """Constructor of the NFA.
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":204
 *         self.label = label
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def label(self):
 *         return self.label
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_5label_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_5label_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_5label___get__(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_5label___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "libmata/nfa/nfa.pyx":206
 *     @property
 *     def label(self):
 *         return self.label             # <<<<<<<<<<<<<<
 * 
 *     @label.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->label);
  __pyx_r = __pyx_v_self->label;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":204
 *         self.label = label
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def label(self):
 *         return self.label
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":208
 *         return self.label
 * 
 *     @label.setter             # <<<<<<<<<<<<<<
 *     def label(self, value):
 *         self.label = value
 */

/* Python wrapper */
static int __pyx_pw_7libmata_3nfa_3nfa_3Nfa_5label_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7libmata_3nfa_3nfa_3Nfa_5label_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_5label_2__set__(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7libmata_3nfa_3nfa_3Nfa_5label_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "libmata/nfa/nfa.pyx":210
 *     @label.setter
 *     def label(self, value):
 *         self.label = value             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->label);
  __Pyx_DECREF(__pyx_v_self->label);
  __pyx_v_self->label = __pyx_v_value;

  /* "libmata/nfa/nfa.pyx":208
 *         return self.label
 * 
 *     @label.setter             # <<<<<<<<<<<<<<
 *     def label(self, value):
 *         self.label = value
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":212
 *         self.label = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def initial_states(self):
 *         return [initial_state for initial_state in self.thisptr.get().initial]
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_14initial_states_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_14initial_states_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_14initial_states___get__(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_14initial_states___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  Mata::Nfa::State __pyx_8genexpr1__pyx_v_initial_state;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Mata::Util::SparseSet<Mata::Nfa::State> ::iterator __pyx_t_2;
  Mata::Util::SparseSet<Mata::Nfa::State>  __pyx_t_3;
  Mata::Nfa::State __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "libmata/nfa/nfa.pyx":214
 *     @property
 *     def initial_states(self):
 *         return [initial_state for initial_state in self.thisptr.get().initial]             # <<<<<<<<<<<<<<
 * 
 *     @initial_states.setter
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_v_self->thisptr.get()->initial;
    __pyx_t_2 = __pyx_t_3.begin();
    for (;;) {
      if (!(__pyx_t_2 != __pyx_t_3.end())) break;
      __pyx_t_4 = *__pyx_t_2;
      ++__pyx_t_2;
      __pyx_8genexpr1__pyx_v_initial_state = __pyx_t_4;
      __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_8genexpr1__pyx_v_initial_state); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 214, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 214, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":212
 *         self.label = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def initial_states(self):
 *         return [initial_state for initial_state in self.thisptr.get().initial]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.initial_states.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":216
 *         return [initial_state for initial_state in self.thisptr.get().initial]
 * 
 *     @initial_states.setter             # <<<<<<<<<<<<<<
 *     def initial_states(self, vector[State] value):
 *         self.thisptr.get().initial.clear()
 */

/* Python wrapper */
static int __pyx_pw_7libmata_3nfa_3nfa_3Nfa_14initial_states_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_7libmata_3nfa_3nfa_3Nfa_14initial_states_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  std::vector<Mata::Nfa::State>  __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(__pyx_arg_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 217, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.initial_states.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_14initial_states_2__set__(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __PYX_STD_MOVE_IF_SUPPORTED(__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7libmata_3nfa_3nfa_3Nfa_14initial_states_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, std::vector<Mata::Nfa::State>  __pyx_v_value) {
  Mata::Nfa::State __pyx_v_state;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  std::vector<Mata::Nfa::State> ::iterator __pyx_t_1;
  Mata::Nfa::State __pyx_t_2;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "libmata/nfa/nfa.pyx":218
 *     @initial_states.setter
 *     def initial_states(self, vector[State] value):
 *         self.thisptr.get().initial.clear()             # <<<<<<<<<<<<<<
 *         for state in value:
 *             self.thisptr.get().initial.insert(state)
 */
  __pyx_v_self->thisptr.get()->initial.clear();

  /* "libmata/nfa/nfa.pyx":219
 *     def initial_states(self, vector[State] value):
 *         self.thisptr.get().initial.clear()
 *         for state in value:             # <<<<<<<<<<<<<<
 *             self.thisptr.get().initial.insert(state)
 * 
 */
  __pyx_t_1 = __pyx_v_value.begin();
  for (;;) {
    if (!(__pyx_t_1 != __pyx_v_value.end())) break;
    __pyx_t_2 = *__pyx_t_1;
    ++__pyx_t_1;
    __pyx_v_state = __pyx_t_2;

    /* "libmata/nfa/nfa.pyx":220
 *         self.thisptr.get().initial.clear()
 *         for state in value:
 *             self.thisptr.get().initial.insert(state)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
    __pyx_v_self->thisptr.get()->initial.insert(__pyx_v_state);

    /* "libmata/nfa/nfa.pyx":219
 *     def initial_states(self, vector[State] value):
 *         self.thisptr.get().initial.clear()
 *         for state in value:             # <<<<<<<<<<<<<<
 *             self.thisptr.get().initial.insert(state)
 * 
 */
  }

  /* "libmata/nfa/nfa.pyx":216
 *         return [initial_state for initial_state in self.thisptr.get().initial]
 * 
 *     @initial_states.setter             # <<<<<<<<<<<<<<
 *     def initial_states(self, vector[State] value):
 *         self.thisptr.get().initial.clear()
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":222
 *             self.thisptr.get().initial.insert(state)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def final_states(self):
 *         return [final_state for final_state in self.thisptr.get().final]
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_12final_states_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_12final_states_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_12final_states___get__(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_12final_states___get__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  Mata::Nfa::State __pyx_8genexpr2__pyx_v_final_state;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Mata::Util::SparseSet<Mata::Nfa::State> ::iterator __pyx_t_2;
  Mata::Util::SparseSet<Mata::Nfa::State>  __pyx_t_3;
  Mata::Nfa::State __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "libmata/nfa/nfa.pyx":224
 *     @property
 *     def final_states(self):
 *         return [final_state for final_state in self.thisptr.get().final]             # <<<<<<<<<<<<<<
 * 
 *     @final_states.setter
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_v_self->thisptr.get()->final;
    __pyx_t_2 = __pyx_t_3.begin();
    for (;;) {
      if (!(__pyx_t_2 != __pyx_t_3.end())) break;
      __pyx_t_4 = *__pyx_t_2;
      ++__pyx_t_2;
      __pyx_8genexpr2__pyx_v_final_state = __pyx_t_4;
      __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_8genexpr2__pyx_v_final_state); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 224, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":222
 *             self.thisptr.get().initial.insert(state)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def final_states(self):
 *         return [final_state for final_state in self.thisptr.get().final]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.final_states.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":226
 *         return [final_state for final_state in self.thisptr.get().final]
 * 
 *     @final_states.setter             # <<<<<<<<<<<<<<
 *     def final_states(self, vector[State] value):
 *         self.thisptr.get().final.clear()
 */

/* Python wrapper */
static int __pyx_pw_7libmata_3nfa_3nfa_3Nfa_12final_states_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_7libmata_3nfa_3nfa_3Nfa_12final_states_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  std::vector<Mata::Nfa::State>  __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(__pyx_arg_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 227, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.final_states.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_12final_states_2__set__(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __PYX_STD_MOVE_IF_SUPPORTED(__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7libmata_3nfa_3nfa_3Nfa_12final_states_2__set__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, std::vector<Mata::Nfa::State>  __pyx_v_value) {
  Mata::Nfa::State __pyx_v_state;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  std::vector<Mata::Nfa::State> ::iterator __pyx_t_1;
  Mata::Nfa::State __pyx_t_2;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "libmata/nfa/nfa.pyx":228
 *     @final_states.setter
 *     def final_states(self, vector[State] value):
 *         self.thisptr.get().final.clear()             # <<<<<<<<<<<<<<
 *         for state in value:
 *             self.thisptr.get().final.insert(state)
 */
  __pyx_v_self->thisptr.get()->final.clear();

  /* "libmata/nfa/nfa.pyx":229
 *     def final_states(self, vector[State] value):
 *         self.thisptr.get().final.clear()
 *         for state in value:             # <<<<<<<<<<<<<<
 *             self.thisptr.get().final.insert(state)
 * 
 */
  __pyx_t_1 = __pyx_v_value.begin();
  for (;;) {
    if (!(__pyx_t_1 != __pyx_v_value.end())) break;
    __pyx_t_2 = *__pyx_t_1;
    ++__pyx_t_1;
    __pyx_v_state = __pyx_t_2;

    /* "libmata/nfa/nfa.pyx":230
 *         self.thisptr.get().final.clear()
 *         for state in value:
 *             self.thisptr.get().final.insert(state)             # <<<<<<<<<<<<<<
 * 
 *     def is_state(self, state):
 */
    __pyx_v_self->thisptr.get()->final.insert(__pyx_v_state);

    /* "libmata/nfa/nfa.pyx":229
 *     def final_states(self, vector[State] value):
 *         self.thisptr.get().final.clear()
 *         for state in value:             # <<<<<<<<<<<<<<
 *             self.thisptr.get().final.insert(state)
 * 
 */
  }

  /* "libmata/nfa/nfa.pyx":226
 *         return [final_state for final_state in self.thisptr.get().final]
 * 
 *     @final_states.setter             # <<<<<<<<<<<<<<
 *     def final_states(self, vector[State] value):
 *         self.thisptr.get().final.clear()
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":232
 *             self.thisptr.get().final.insert(state)
 * 
 *     def is_state(self, state):             # <<<<<<<<<<<<<<
 *         """Tests if state is in the automaton
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_3is_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_2is_state, "Tests if state is in the automaton\n\n        :param int state: tested state\n        :return: true if state is in the automaton\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_3is_state = {"is_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_3is_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_2is_state};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_3is_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_state (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 232, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "is_state") < 0)) __PYX_ERR(1, 232, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 232, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_2is_state(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_2is_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Mata::Nfa::State __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_state", 0);

  /* "libmata/nfa/nfa.pyx":238
 *         :return: true if state is in the automaton
 *         """
 *         return self.thisptr.get().is_state(state)             # <<<<<<<<<<<<<<
 * 
 *     def add_new_state(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_state); if (unlikely((__pyx_t_1 == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 238, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->thisptr.get()->is_state(__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":232
 *             self.thisptr.get().final.insert(state)
 * 
 *     def is_state(self, state):             # <<<<<<<<<<<<<<
 *         """Tests if state is in the automaton
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":240
 *         return self.thisptr.get().is_state(state)
 * 
 *     def add_new_state(self):             # <<<<<<<<<<<<<<
 *         """Adds new state to automaton
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_5add_new_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_4add_new_state, "Adds new state to automaton\n\n        :return: number of the state\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_5add_new_state = {"add_new_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_5add_new_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_4add_new_state};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_5add_new_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_new_state (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("add_new_state", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "add_new_state", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_4add_new_state(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_4add_new_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_new_state", 0);

  /* "libmata/nfa/nfa.pyx":245
 *         :return: number of the state
 *         """
 *         return self.thisptr.get().add_state()             # <<<<<<<<<<<<<<
 * 
 *     def add_state(self, State state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->thisptr.get()->add_state()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":240
 *         return self.thisptr.get().is_state(state)
 * 
 *     def add_new_state(self):             # <<<<<<<<<<<<<<
 *         """Adds new state to automaton
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.add_new_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":247
 *         return self.thisptr.get().add_state()
 * 
 *     def add_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Adds passed state to the automaton.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_7add_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_6add_state, "Adds passed state to the automaton.\n\n        :return: number of the state\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_7add_state = {"add_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_7add_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_6add_state};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_7add_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_state;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_state (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 247, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "add_state") < 0)) __PYX_ERR(1, 247, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_state = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_state == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 247, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 247, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.add_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_6add_state(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_6add_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_state", 0);

  /* "libmata/nfa/nfa.pyx":252
 *         :return: number of the state
 *         """
 *         return self.thisptr.get().add_state(state)             # <<<<<<<<<<<<<<
 * 
 *     def make_initial_state(self, State state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->thisptr.get()->add_state(__pyx_v_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":247
 *         return self.thisptr.get().add_state()
 * 
 *     def add_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Adds passed state to the automaton.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.add_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":254
 *         return self.thisptr.get().add_state(state)
 * 
 *     def make_initial_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Makes specified state from the automaton initial.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_9make_initial_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_8make_initial_state, "Makes specified state from the automaton initial.\n\n        :param State state: State to be made initial.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_9make_initial_state = {"make_initial_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_9make_initial_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_8make_initial_state};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_9make_initial_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_state;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_initial_state (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 254, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "make_initial_state") < 0)) __PYX_ERR(1, 254, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_state = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_state == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 254, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("make_initial_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 254, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.make_initial_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_8make_initial_state(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_8make_initial_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_initial_state", 0);

  /* "libmata/nfa/nfa.pyx":259
 *         :param State state: State to be made initial.
 *         """
 *         self.thisptr.get().initial.insert(state)             # <<<<<<<<<<<<<<
 * 
 *     def make_initial_states(self, vector[State] states):
 */
  __pyx_v_self->thisptr.get()->initial.insert(__pyx_v_state);

  /* "libmata/nfa/nfa.pyx":254
 *         return self.thisptr.get().add_state(state)
 * 
 *     def make_initial_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Makes specified state from the automaton initial.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":261
 *         self.thisptr.get().initial.insert(state)
 * 
 *     def make_initial_states(self, vector[State] states):             # <<<<<<<<<<<<<<
 *         """Makes specified states from the automaton initial.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_11make_initial_states(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_10make_initial_states, "Makes specified states from the automaton initial.\n\n        :param list states: List of states to be made initial.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_11make_initial_states = {"make_initial_states", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_11make_initial_states, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_10make_initial_states};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_11make_initial_states(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  std::vector<Mata::Nfa::State>  __pyx_v_states;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_initial_states (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_states,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_states)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 261, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "make_initial_states") < 0)) __PYX_ERR(1, 261, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_states = __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(values[0]); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 261, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("make_initial_states", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 261, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.make_initial_states", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_10make_initial_states(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __PYX_STD_MOVE_IF_SUPPORTED(__pyx_v_states));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_10make_initial_states(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, std::vector<Mata::Nfa::State>  __pyx_v_states) {
  Mata::Nfa::State __pyx_v_state;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  std::vector<Mata::Nfa::State> ::iterator __pyx_t_1;
  Mata::Nfa::State __pyx_t_2;
  __Pyx_RefNannySetupContext("make_initial_states", 0);

  /* "libmata/nfa/nfa.pyx":266
 *         :param list states: List of states to be made initial.
 *         """
 *         for state in states:             # <<<<<<<<<<<<<<
 *             self.thisptr.get().initial.insert(state)
 * 
 */
  __pyx_t_1 = __pyx_v_states.begin();
  for (;;) {
    if (!(__pyx_t_1 != __pyx_v_states.end())) break;
    __pyx_t_2 = *__pyx_t_1;
    ++__pyx_t_1;
    __pyx_v_state = __pyx_t_2;

    /* "libmata/nfa/nfa.pyx":267
 *         """
 *         for state in states:
 *             self.thisptr.get().initial.insert(state)             # <<<<<<<<<<<<<<
 * 
 *     def has_initial_state(self, State st):
 */
    __pyx_v_self->thisptr.get()->initial.insert(__pyx_v_state);

    /* "libmata/nfa/nfa.pyx":266
 *         :param list states: List of states to be made initial.
 *         """
 *         for state in states:             # <<<<<<<<<<<<<<
 *             self.thisptr.get().initial.insert(state)
 * 
 */
  }

  /* "libmata/nfa/nfa.pyx":261
 *         self.thisptr.get().initial.insert(state)
 * 
 *     def make_initial_states(self, vector[State] states):             # <<<<<<<<<<<<<<
 *         """Makes specified states from the automaton initial.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":269
 *             self.thisptr.get().initial.insert(state)
 * 
 *     def has_initial_state(self, State st):             # <<<<<<<<<<<<<<
 *         """Tests if automaton contains given state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_13has_initial_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_12has_initial_state, "Tests if automaton contains given state\n\n        :param State st: tested state\n        :return: true if automaton contains given state\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_13has_initial_state = {"has_initial_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_13has_initial_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_12has_initial_state};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_13has_initial_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_st;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_initial_state (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_st,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_st)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 269, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "has_initial_state") < 0)) __PYX_ERR(1, 269, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_st = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_st == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 269, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("has_initial_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 269, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.has_initial_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_12has_initial_state(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_st);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_12has_initial_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_st) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("has_initial_state", 0);

  /* "libmata/nfa/nfa.pyx":275
 *         :return: true if automaton contains given state
 *         """
 *         return self.thisptr.get().initial[st]             # <<<<<<<<<<<<<<
 * 
 *     def remove_initial_state(self, State state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->thisptr.get()->initial[__pyx_v_st])); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":269
 *             self.thisptr.get().initial.insert(state)
 * 
 *     def has_initial_state(self, State st):             # <<<<<<<<<<<<<<
 *         """Tests if automaton contains given state
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.has_initial_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":277
 *         return self.thisptr.get().initial[st]
 * 
 *     def remove_initial_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Removes state from initial state set of the automaton.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_15remove_initial_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_14remove_initial_state, "Removes state from initial state set of the automaton.\n\n        :param State state: State to be removed from initial states.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_15remove_initial_state = {"remove_initial_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_15remove_initial_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_14remove_initial_state};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_15remove_initial_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_state;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_initial_state (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 277, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "remove_initial_state") < 0)) __PYX_ERR(1, 277, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_state = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_state == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 277, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove_initial_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 277, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.remove_initial_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_14remove_initial_state(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_14remove_initial_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_initial_state", 0);

  /* "libmata/nfa/nfa.pyx":282
 *         :param State state: State to be removed from initial states.
 *         """
 *         self.thisptr.get().initial.erase(state)             # <<<<<<<<<<<<<<
 * 
 *     def clear_initial(self):
 */
  __pyx_v_self->thisptr.get()->initial.erase(__pyx_v_state);

  /* "libmata/nfa/nfa.pyx":277
 *         return self.thisptr.get().initial[st]
 * 
 *     def remove_initial_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Removes state from initial state set of the automaton.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":284
 *         self.thisptr.get().initial.erase(state)
 * 
 *     def clear_initial(self):             # <<<<<<<<<<<<<<
 *         """Clears initial state set of the automaton."""
 *         self.thisptr.get().initial.clear()
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_17clear_initial(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_16clear_initial, "Clears initial state set of the automaton.");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_17clear_initial = {"clear_initial", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_17clear_initial, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_16clear_initial};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_17clear_initial(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear_initial (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("clear_initial", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "clear_initial", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_16clear_initial(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_16clear_initial(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear_initial", 0);

  /* "libmata/nfa/nfa.pyx":286
 *     def clear_initial(self):
 *         """Clears initial state set of the automaton."""
 *         self.thisptr.get().initial.clear()             # <<<<<<<<<<<<<<
 * 
 *     def make_final_state(self, State state):
 */
  __pyx_v_self->thisptr.get()->initial.clear();

  /* "libmata/nfa/nfa.pyx":284
 *         self.thisptr.get().initial.erase(state)
 * 
 *     def clear_initial(self):             # <<<<<<<<<<<<<<
 *         """Clears initial state set of the automaton."""
 *         self.thisptr.get().initial.clear()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":288
 *         self.thisptr.get().initial.clear()
 * 
 *     def make_final_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Makes specified state from the automaton final.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_19make_final_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_18make_final_state, "Makes specified state from the automaton final.\n\n        :param State state: State to be made final.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_19make_final_state = {"make_final_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_19make_final_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_18make_final_state};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_19make_final_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_state;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_final_state (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 288, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "make_final_state") < 0)) __PYX_ERR(1, 288, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_state = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_state == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 288, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("make_final_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 288, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.make_final_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_18make_final_state(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_18make_final_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_final_state", 0);

  /* "libmata/nfa/nfa.pyx":293
 *         :param State state: State to be made final.
 *         """
 *         self.thisptr.get().final.insert(state)             # <<<<<<<<<<<<<<
 * 
 *     def make_final_states(self, vector[State] states):
 */
  __pyx_v_self->thisptr.get()->final.insert(__pyx_v_state);

  /* "libmata/nfa/nfa.pyx":288
 *         self.thisptr.get().initial.clear()
 * 
 *     def make_final_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Makes specified state from the automaton final.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":295
 *         self.thisptr.get().final.insert(state)
 * 
 *     def make_final_states(self, vector[State] states):             # <<<<<<<<<<<<<<
 *         """Makes specified states from the automaton final.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_21make_final_states(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_20make_final_states, "Makes specified states from the automaton final.\n\n        :param vector[State] states: List of states to be made final.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_21make_final_states = {"make_final_states", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_21make_final_states, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_20make_final_states};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_21make_final_states(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  std::vector<Mata::Nfa::State>  __pyx_v_states;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_final_states (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_states,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_states)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 295, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "make_final_states") < 0)) __PYX_ERR(1, 295, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_states = __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(values[0]); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 295, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("make_final_states", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 295, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.make_final_states", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_20make_final_states(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __PYX_STD_MOVE_IF_SUPPORTED(__pyx_v_states));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_20make_final_states(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, std::vector<Mata::Nfa::State>  __pyx_v_states) {
  Mata::Nfa::State __pyx_v_state;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  std::vector<Mata::Nfa::State> ::iterator __pyx_t_1;
  Mata::Nfa::State __pyx_t_2;
  __Pyx_RefNannySetupContext("make_final_states", 0);

  /* "libmata/nfa/nfa.pyx":300
 *         :param vector[State] states: List of states to be made final.
 *         """
 *         for state in states:             # <<<<<<<<<<<<<<
 *             self.thisptr.get().final.insert(state)
 * 
 */
  __pyx_t_1 = __pyx_v_states.begin();
  for (;;) {
    if (!(__pyx_t_1 != __pyx_v_states.end())) break;
    __pyx_t_2 = *__pyx_t_1;
    ++__pyx_t_1;
    __pyx_v_state = __pyx_t_2;

    /* "libmata/nfa/nfa.pyx":301
 *         """
 *         for state in states:
 *             self.thisptr.get().final.insert(state)             # <<<<<<<<<<<<<<
 * 
 *     def has_final_state(self, State st):
 */
    __pyx_v_self->thisptr.get()->final.insert(__pyx_v_state);

    /* "libmata/nfa/nfa.pyx":300
 *         :param vector[State] states: List of states to be made final.
 *         """
 *         for state in states:             # <<<<<<<<<<<<<<
 *             self.thisptr.get().final.insert(state)
 * 
 */
  }

  /* "libmata/nfa/nfa.pyx":295
 *         self.thisptr.get().final.insert(state)
 * 
 *     def make_final_states(self, vector[State] states):             # <<<<<<<<<<<<<<
 *         """Makes specified states from the automaton final.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":303
 *             self.thisptr.get().final.insert(state)
 * 
 *     def has_final_state(self, State st):             # <<<<<<<<<<<<<<
 *         """Tests if automaton contains given state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_23has_final_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_22has_final_state, "Tests if automaton contains given state\n\n        :param State st: tested state\n        :return: true if automaton contains given state\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_23has_final_state = {"has_final_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_23has_final_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_22has_final_state};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_23has_final_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_st;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_final_state (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_st,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_st)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 303, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "has_final_state") < 0)) __PYX_ERR(1, 303, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_st = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_st == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 303, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("has_final_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 303, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.has_final_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_22has_final_state(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_st);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_22has_final_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_st) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("has_final_state", 0);

  /* "libmata/nfa/nfa.pyx":309
 *         :return: true if automaton contains given state
 *         """
 *         return self.thisptr.get().final[st]             # <<<<<<<<<<<<<<
 * 
 *     def remove_final_state(self, State state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->thisptr.get()->final[__pyx_v_st])); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":303
 *             self.thisptr.get().final.insert(state)
 * 
 *     def has_final_state(self, State st):             # <<<<<<<<<<<<<<
 *         """Tests if automaton contains given state
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.has_final_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":311
 *         return self.thisptr.get().final[st]
 * 
 *     def remove_final_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Removes state from final state set of the automaton.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_25remove_final_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_24remove_final_state, "Removes state from final state set of the automaton.\n\n        :param State state: State to be removed from final states.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_25remove_final_state = {"remove_final_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_25remove_final_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_24remove_final_state};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_25remove_final_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_state;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_final_state (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 311, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "remove_final_state") < 0)) __PYX_ERR(1, 311, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_state = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_state == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 311, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove_final_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 311, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.remove_final_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_24remove_final_state(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_24remove_final_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_final_state", 0);

  /* "libmata/nfa/nfa.pyx":316
 *         :param State state: State to be removed from final states.
 *         """
 *         self.thisptr.get().final.erase(state)             # <<<<<<<<<<<<<<
 * 
 *     def clear_final(self):
 */
  __pyx_v_self->thisptr.get()->final.erase(__pyx_v_state);

  /* "libmata/nfa/nfa.pyx":311
 *         return self.thisptr.get().final[st]
 * 
 *     def remove_final_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Removes state from final state set of the automaton.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":318
 *         self.thisptr.get().final.erase(state)
 * 
 *     def clear_final(self):             # <<<<<<<<<<<<<<
 *         """Clears final state set of the automaton."""
 *         self.thisptr.get().final.clear()
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_27clear_final(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_26clear_final, "Clears final state set of the automaton.");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_27clear_final = {"clear_final", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_27clear_final, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_26clear_final};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_27clear_final(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear_final (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("clear_final", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "clear_final", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_26clear_final(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_26clear_final(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear_final", 0);

  /* "libmata/nfa/nfa.pyx":320
 *     def clear_final(self):
 *         """Clears final state set of the automaton."""
 *         self.thisptr.get().final.clear()             # <<<<<<<<<<<<<<
 * 
 *     def unify_initial(self):
 */
  __pyx_v_self->thisptr.get()->final.clear();

  /* "libmata/nfa/nfa.pyx":318
 *         self.thisptr.get().final.erase(state)
 * 
 *     def clear_final(self):             # <<<<<<<<<<<<<<
 *         """Clears final state set of the automaton."""
 *         self.thisptr.get().final.clear()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":322
 *         self.thisptr.get().final.clear()
 * 
 *     def unify_initial(self):             # <<<<<<<<<<<<<<
 *         """Unify initial states into a single new initial state."""
 *         self.thisptr.get().unify_initial()
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_29unify_initial(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_28unify_initial, "Unify initial states into a single new initial state.");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_29unify_initial = {"unify_initial", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_29unify_initial, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_28unify_initial};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_29unify_initial(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unify_initial (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("unify_initial", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "unify_initial", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_28unify_initial(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_28unify_initial(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unify_initial", 0);

  /* "libmata/nfa/nfa.pyx":324
 *     def unify_initial(self):
 *         """Unify initial states into a single new initial state."""
 *         self.thisptr.get().unify_initial()             # <<<<<<<<<<<<<<
 * 
 *     def unify_final(self):
 */
  __pyx_v_self->thisptr.get()->unify_initial();

  /* "libmata/nfa/nfa.pyx":322
 *         self.thisptr.get().final.clear()
 * 
 *     def unify_initial(self):             # <<<<<<<<<<<<<<
 *         """Unify initial states into a single new initial state."""
 *         self.thisptr.get().unify_initial()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":326
 *         self.thisptr.get().unify_initial()
 * 
 *     def unify_final(self):             # <<<<<<<<<<<<<<
 *         """Unify final states into a single new final state."""
 *         self.thisptr.get().unify_final()
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_31unify_final(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_30unify_final, "Unify final states into a single new final state.");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_31unify_final = {"unify_final", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_31unify_final, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_30unify_final};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_31unify_final(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unify_final (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("unify_final", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "unify_final", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_30unify_final(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_30unify_final(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("unify_final", 0);

  /* "libmata/nfa/nfa.pyx":328
 *     def unify_final(self):
 *         """Unify final states into a single new final state."""
 *         self.thisptr.get().unify_final()             # <<<<<<<<<<<<<<
 * 
 *     def add_transition_object(self, Trans tr):
 */
  __pyx_v_self->thisptr.get()->unify_final();

  /* "libmata/nfa/nfa.pyx":326
 *         self.thisptr.get().unify_initial()
 * 
 *     def unify_final(self):             # <<<<<<<<<<<<<<
 *         """Unify final states into a single new final state."""
 *         self.thisptr.get().unify_final()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":330
 *         self.thisptr.get().unify_final()
 * 
 *     def add_transition_object(self, Trans tr):             # <<<<<<<<<<<<<<
 *         """Adds transition to automaton
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_33add_transition_object(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_32add_transition_object, "Adds transition to automaton\n\n        :param Trans tr: added transition\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_33add_transition_object = {"add_transition_object", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_33add_transition_object, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_32add_transition_object};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_33add_transition_object(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_tr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_transition_object (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tr,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tr)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 330, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "add_transition_object") < 0)) __PYX_ERR(1, 330, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_tr = ((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_transition_object", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 330, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.add_transition_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tr), __pyx_ptype_7libmata_3nfa_3nfa_Trans, 1, "tr", 0))) __PYX_ERR(1, 330, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_32add_transition_object(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_tr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_32add_transition_object(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_tr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_transition_object", 0);

  /* "libmata/nfa/nfa.pyx":335
 *         :param Trans tr: added transition
 *         """
 *         self.thisptr.get().delta.add(dereference(tr.thisptr))             # <<<<<<<<<<<<<<
 * 
 *     def add_transition(self, State src, symb, State tgt, alph.Alphabet alphabet = None):
 */
  try {
    __pyx_v_self->thisptr.get()->delta.add((*__pyx_v_tr->thisptr));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 335, __pyx_L1_error)
  }

  /* "libmata/nfa/nfa.pyx":330
 *         self.thisptr.get().unify_final()
 * 
 *     def add_transition_object(self, Trans tr):             # <<<<<<<<<<<<<<
 *         """Adds transition to automaton
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.add_transition_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":337
 *         self.thisptr.get().delta.add(dereference(tr.thisptr))
 * 
 *     def add_transition(self, State src, symb, State tgt, alph.Alphabet alphabet = None):             # <<<<<<<<<<<<<<
 *         """Constructs transition and adds it to automaton
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_35add_transition(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_34add_transition, "Constructs transition and adds it to automaton\n\n        :param State src: source state\n        :param Symbol symb: symbol\n        :param State tgt: target state\n        :param alph.Alphabet alphabet: alphabet of the transition\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_35add_transition = {"add_transition", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_35add_transition, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_34add_transition};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_35add_transition(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_src;
  PyObject *__pyx_v_symb = 0;
  Mata::Nfa::State __pyx_v_tgt;
  struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_transition (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_src,&__pyx_n_s_symb,&__pyx_n_s_tgt,&__pyx_n_s_alphabet,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = (PyObject *)((struct __pyx_obj_7libmata_9alphabets_Alphabet *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_src)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 337, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_symb)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 337, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("add_transition", 0, 3, 4, 1); __PYX_ERR(1, 337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tgt)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 337, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("add_transition", 0, 3, 4, 2); __PYX_ERR(1, 337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_alphabet);
          if (value) { values[3] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 337, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "add_transition") < 0)) __PYX_ERR(1, 337, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_src = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_src == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 337, __pyx_L3_error)
    __pyx_v_symb = values[1];
    __pyx_v_tgt = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_tgt == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 337, __pyx_L3_error)
    __pyx_v_alphabet = ((struct __pyx_obj_7libmata_9alphabets_Alphabet *)values[3]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_transition", 0, 3, 4, __pyx_nargs); __PYX_ERR(1, 337, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.add_transition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_alphabet), __pyx_ptype_7libmata_9alphabets_Alphabet, 1, "alphabet", 0))) __PYX_ERR(1, 337, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_34add_transition(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_src, __pyx_v_symb, __pyx_v_tgt, __pyx_v_alphabet);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_34add_transition(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_src, PyObject *__pyx_v_symb, Mata::Nfa::State __pyx_v_tgt, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  Mata::Symbol __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_transition", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_alphabet);

  /* "libmata/nfa/nfa.pyx":345
 *         :param alph.Alphabet alphabet: alphabet of the transition
 *         """
 *         if isinstance(symb, str):             # <<<<<<<<<<<<<<
 *             alphabet = alphabet or store().get('alphabet')
 *             if not alphabet:
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_symb); 
  if (__pyx_t_1) {

    /* "libmata/nfa/nfa.pyx":346
 *         """
 *         if isinstance(symb, str):
 *             alphabet = alphabet or store().get('alphabet')             # <<<<<<<<<<<<<<
 *             if not alphabet:
 *                 raise Exception(f"Cannot translate symbol '{symb}' without specified alphabet")
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_alphabet)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 346, __pyx_L1_error)
    if (!__pyx_t_1) {
    } else {
      __Pyx_INCREF((PyObject *)__pyx_v_alphabet);
      __pyx_t_2 = ((PyObject *)__pyx_v_alphabet);
      goto __pyx_L4_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_store); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[1] = {__pyx_t_6, };
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_get); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_alphabet};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7libmata_9alphabets_Alphabet))))) __PYX_ERR(1, 346, __pyx_L1_error)
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_L4_bool_binop_done:;
    __Pyx_DECREF_SET(__pyx_v_alphabet, ((struct __pyx_obj_7libmata_9alphabets_Alphabet *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "libmata/nfa/nfa.pyx":347
 *         if isinstance(symb, str):
 *             alphabet = alphabet or store().get('alphabet')
 *             if not alphabet:             # <<<<<<<<<<<<<<
 *                 raise Exception(f"Cannot translate symbol '{symb}' without specified alphabet")
 *             self.thisptr.get().delta.add(src, alphabet.translate_symbol(symb), tgt)
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_alphabet)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 347, __pyx_L1_error)
    __pyx_t_8 = (!__pyx_t_1);
    if (unlikely(__pyx_t_8)) {

      /* "libmata/nfa/nfa.pyx":348
 *             alphabet = alphabet or store().get('alphabet')
 *             if not alphabet:
 *                 raise Exception(f"Cannot translate symbol '{symb}' without specified alphabet")             # <<<<<<<<<<<<<<
 *             self.thisptr.get().delta.add(src, alphabet.translate_symbol(symb), tgt)
 *         else:
 */
      __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;
      __Pyx_INCREF(__pyx_kp_u_Cannot_translate_symbol);
      __pyx_t_9 += 25;
      __Pyx_GIVEREF(__pyx_kp_u_Cannot_translate_symbol);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Cannot_translate_symbol);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_symb, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_INCREF(__pyx_kp_u_without_specified_alphabet);
      __pyx_t_9 += 28;
      __Pyx_GIVEREF(__pyx_kp_u_without_specified_alphabet);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_without_specified_alphabet);
      __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(1, 348, __pyx_L1_error)

      /* "libmata/nfa/nfa.pyx":347
 *         if isinstance(symb, str):
 *             alphabet = alphabet or store().get('alphabet')
 *             if not alphabet:             # <<<<<<<<<<<<<<
 *                 raise Exception(f"Cannot translate symbol '{symb}' without specified alphabet")
 *             self.thisptr.get().delta.add(src, alphabet.translate_symbol(symb), tgt)
 */
    }

    /* "libmata/nfa/nfa.pyx":349
 *             if not alphabet:
 *                 raise Exception(f"Cannot translate symbol '{symb}' without specified alphabet")
 *             self.thisptr.get().delta.add(src, alphabet.translate_symbol(symb), tgt)             # <<<<<<<<<<<<<<
 *         else:
 *             self.thisptr.get().delta.add(src, symb, tgt)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_alphabet), __pyx_n_s_translate_symbol); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_symb};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 349, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_11 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_11 == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 349, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    try {
      __pyx_v_self->thisptr.get()->delta.add(__pyx_v_src, __pyx_t_11, __pyx_v_tgt);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(1, 349, __pyx_L1_error)
    }

    /* "libmata/nfa/nfa.pyx":345
 *         :param alph.Alphabet alphabet: alphabet of the transition
 *         """
 *         if isinstance(symb, str):             # <<<<<<<<<<<<<<
 *             alphabet = alphabet or store().get('alphabet')
 *             if not alphabet:
 */
    goto __pyx_L3;
  }

  /* "libmata/nfa/nfa.pyx":351
 *             self.thisptr.get().delta.add(src, alphabet.translate_symbol(symb), tgt)
 *         else:
 *             self.thisptr.get().delta.add(src, symb, tgt)             # <<<<<<<<<<<<<<
 * 
 *     def remove_trans(self, Trans tr):
 */
  /*else*/ {
    __pyx_t_11 = __Pyx_PyInt_As_size_t(__pyx_v_symb); if (unlikely((__pyx_t_11 == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 351, __pyx_L1_error)
    try {
      __pyx_v_self->thisptr.get()->delta.add(__pyx_v_src, __pyx_t_11, __pyx_v_tgt);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(1, 351, __pyx_L1_error)
    }
  }
  __pyx_L3:;

  /* "libmata/nfa/nfa.pyx":337
 *         self.thisptr.get().delta.add(dereference(tr.thisptr))
 * 
 *     def add_transition(self, State src, symb, State tgt, alph.Alphabet alphabet = None):             # <<<<<<<<<<<<<<
 *         """Constructs transition and adds it to automaton
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.add_transition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_alphabet);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":353
 *             self.thisptr.get().delta.add(src, symb, tgt)
 * 
 *     def remove_trans(self, Trans tr):             # <<<<<<<<<<<<<<
 *         """Removes transition from the automaton.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_37remove_trans(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_36remove_trans, "Removes transition from the automaton.\n\n        :param Trans tr: Transition to be removed.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_37remove_trans = {"remove_trans", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_37remove_trans, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_36remove_trans};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_37remove_trans(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_tr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_trans (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tr,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tr)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 353, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "remove_trans") < 0)) __PYX_ERR(1, 353, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_tr = ((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove_trans", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 353, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.remove_trans", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tr), __pyx_ptype_7libmata_3nfa_3nfa_Trans, 1, "tr", 0))) __PYX_ERR(1, 353, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_36remove_trans(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_tr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_36remove_trans(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, struct __pyx_obj_7libmata_3nfa_3nfa_Trans *__pyx_v_tr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove_trans", 0);

  /* "libmata/nfa/nfa.pyx":358
 *         :param Trans tr: Transition to be removed.
 *         """
 *         self.thisptr.get().delta.remove(dereference(tr.thisptr))             # <<<<<<<<<<<<<<
 * 
 *     def remove_trans_raw(self, State src, Symbol symb, State tgt):
 */
  try {
    __pyx_v_self->thisptr.get()->delta.remove((*__pyx_v_tr->thisptr));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 358, __pyx_L1_error)
  }

  /* "libmata/nfa/nfa.pyx":353
 *             self.thisptr.get().delta.add(src, symb, tgt)
 * 
 *     def remove_trans(self, Trans tr):             # <<<<<<<<<<<<<<
 *         """Removes transition from the automaton.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.remove_trans", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":360
 *         self.thisptr.get().delta.remove(dereference(tr.thisptr))
 * 
 *     def remove_trans_raw(self, State src, Symbol symb, State tgt):             # <<<<<<<<<<<<<<
 *         """Constructs transition and removes it from the automaton.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_39remove_trans_raw(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_38remove_trans_raw, "Constructs transition and removes it from the automaton.\n\n        :param State src: Source state of the transition to be removed.\n        :param Symbol symb: Symbol of the transition to be removed.\n        :param State tgt: Target state of the transition to be removed.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_39remove_trans_raw = {"remove_trans_raw", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_39remove_trans_raw, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_38remove_trans_raw};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_39remove_trans_raw(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_src;
  Mata::Symbol __pyx_v_symb;
  Mata::Nfa::State __pyx_v_tgt;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_trans_raw (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_src,&__pyx_n_s_symb,&__pyx_n_s_tgt,0};
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_src)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 360, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_symb)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 360, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("remove_trans_raw", 1, 3, 3, 1); __PYX_ERR(1, 360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tgt)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 360, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("remove_trans_raw", 1, 3, 3, 2); __PYX_ERR(1, 360, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "remove_trans_raw") < 0)) __PYX_ERR(1, 360, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_src = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_src == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 360, __pyx_L3_error)
    __pyx_v_symb = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_symb == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 360, __pyx_L3_error)
    __pyx_v_tgt = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_tgt == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 360, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove_trans_raw", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 360, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.remove_trans_raw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_38remove_trans_raw(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_src, __pyx_v_symb, __pyx_v_tgt);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_38remove_trans_raw(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_src, Mata::Symbol __pyx_v_symb, Mata::Nfa::State __pyx_v_tgt) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove_trans_raw", 0);

  /* "libmata/nfa/nfa.pyx":367
 *         :param State tgt: Target state of the transition to be removed.
 *         """
 *         self.thisptr.get().delta.remove(src, symb, tgt)             # <<<<<<<<<<<<<<
 * 
 *     def has_transition(self, State src, Symbol symb, State tgt):
 */
  try {
    __pyx_v_self->thisptr.get()->delta.remove(__pyx_v_src, __pyx_v_symb, __pyx_v_tgt);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 367, __pyx_L1_error)
  }

  /* "libmata/nfa/nfa.pyx":360
 *         self.thisptr.get().delta.remove(dereference(tr.thisptr))
 * 
 *     def remove_trans_raw(self, State src, Symbol symb, State tgt):             # <<<<<<<<<<<<<<
 *         """Constructs transition and removes it from the automaton.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.remove_trans_raw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":369
 *         self.thisptr.get().delta.remove(src, symb, tgt)
 * 
 *     def has_transition(self, State src, Symbol symb, State tgt):             # <<<<<<<<<<<<<<
 *         """Tests if automaton contains transition
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_41has_transition(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_40has_transition, "Tests if automaton contains transition\n\n        :param State src: source state\n        :param Symbol symb: symbol\n        :param State tgt: target state\n        :return: true if automaton contains transition\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_41has_transition = {"has_transition", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_41has_transition, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_40has_transition};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_41has_transition(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_src;
  Mata::Symbol __pyx_v_symb;
  Mata::Nfa::State __pyx_v_tgt;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_transition (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_src,&__pyx_n_s_symb,&__pyx_n_s_tgt,0};
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_src)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 369, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_symb)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 369, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("has_transition", 1, 3, 3, 1); __PYX_ERR(1, 369, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tgt)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 369, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("has_transition", 1, 3, 3, 2); __PYX_ERR(1, 369, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "has_transition") < 0)) __PYX_ERR(1, 369, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_src = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_src == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 369, __pyx_L3_error)
    __pyx_v_symb = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_symb == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 369, __pyx_L3_error)
    __pyx_v_tgt = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_tgt == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 369, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("has_transition", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 369, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.has_transition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_40has_transition(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_src, __pyx_v_symb, __pyx_v_tgt);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_40has_transition(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_src, Mata::Symbol __pyx_v_symb, Mata::Nfa::State __pyx_v_tgt) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("has_transition", 0);

  /* "libmata/nfa/nfa.pyx":377
 *         :return: true if automaton contains transition
 *         """
 *         return self.thisptr.get().delta.contains(src, symb, tgt)             # <<<<<<<<<<<<<<
 * 
 *     def get_num_of_trans(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->thisptr.get()->delta.contains(__pyx_v_src, __pyx_v_symb, __pyx_v_tgt)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":369
 *         self.thisptr.get().delta.remove(src, symb, tgt)
 * 
 *     def has_transition(self, State src, Symbol symb, State tgt):             # <<<<<<<<<<<<<<
 *         """Tests if automaton contains transition
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.has_transition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":379
 *         return self.thisptr.get().delta.contains(src, symb, tgt)
 * 
 *     def get_num_of_trans(self):             # <<<<<<<<<<<<<<
 *         """Returns number of transitions in automaton
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_43get_num_of_trans(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_42get_num_of_trans, "Returns number of transitions in automaton\n\n        :return: number of transitions in automaton\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_43get_num_of_trans = {"get_num_of_trans", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_43get_num_of_trans, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_42get_num_of_trans};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_43get_num_of_trans(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_num_of_trans (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_num_of_trans", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_num_of_trans", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_42get_num_of_trans(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_42get_num_of_trans(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_num_of_trans", 0);

  /* "libmata/nfa/nfa.pyx":384
 *         :return: number of transitions in automaton
 *         """
 *         return self.thisptr.get().get_num_of_trans()             # <<<<<<<<<<<<<<
 * 
 *     def clear(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->thisptr.get()->get_num_of_trans()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":379
 *         return self.thisptr.get().delta.contains(src, symb, tgt)
 * 
 *     def get_num_of_trans(self):             # <<<<<<<<<<<<<<
 *         """Returns number of transitions in automaton
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_num_of_trans", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":386
 *         return self.thisptr.get().get_num_of_trans()
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         """Clears all of the internals in the automaton"""
 *         self.thisptr.get().clear()
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_45clear(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_44clear, "Clears all of the internals in the automaton");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_45clear = {"clear", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_45clear, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_44clear};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_45clear(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("clear", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "clear", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_44clear(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_44clear(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear", 0);

  /* "libmata/nfa/nfa.pyx":388
 *     def clear(self):
 *         """Clears all of the internals in the automaton"""
 *         self.thisptr.get().clear()             # <<<<<<<<<<<<<<
 * 
 *     def size(self) -> int:
 */
  __pyx_v_self->thisptr.get()->clear();

  /* "libmata/nfa/nfa.pyx":386
 *         return self.thisptr.get().get_num_of_trans()
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         """Clears all of the internals in the automaton"""
 *         self.thisptr.get().clear()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":390
 *         self.thisptr.get().clear()
 * 
 *     def size(self) -> int:             # <<<<<<<<<<<<<<
 *         """Get the current number of states in the whole automaton.
 *         :return: The number of states.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_47size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_46size, "Get the current number of states in the whole automaton.\n        :return: The number of states.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_47size = {"size", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_47size, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_46size};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_47size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("size (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("size", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "size", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_46size(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_46size(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("size", 0);

  /* "libmata/nfa/nfa.pyx":394
 *         :return: The number of states.
 *         """
 *         return self.thisptr.get().size()             # <<<<<<<<<<<<<<
 * 
 *     def iterate(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->thisptr.get()->size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_t_1)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_t_1))) __PYX_ERR(1, 394, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":390
 *         self.thisptr.get().clear()
 * 
 *     def size(self) -> int:             # <<<<<<<<<<<<<<
 *         """Get the current number of states in the whole automaton.
 *         :return: The number of states.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_7libmata_3nfa_3nfa_3Nfa_50generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "libmata/nfa/nfa.pyx":396
 *         return self.thisptr.get().size()
 * 
 *     def iterate(self):             # <<<<<<<<<<<<<<
 *         """Iterates over all transitions
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_49iterate(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_48iterate, "Iterates over all transitions\n\n        :return: stream of transitions\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_49iterate = {"iterate", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_49iterate, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_48iterate};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_49iterate(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iterate (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("iterate", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "iterate", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_48iterate(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_48iterate(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("iterate", 0);
  __pyx_cur_scope = (struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *)__pyx_tp_new_7libmata_3nfa_3nfa___pyx_scope_struct__iterate(__pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 396, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_7libmata_3nfa_3nfa_3Nfa_50generator, __pyx_codeobj__7, (PyObject *) __pyx_cur_scope, __pyx_n_s_iterate, __pyx_n_s_Nfa_iterate, __pyx_n_s_libmata_nfa_nfa); if (unlikely(!gen)) __PYX_ERR(1, 396, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.iterate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7libmata_3nfa_3nfa_3Nfa_50generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *__pyx_cur_scope = ((struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iterate", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(1, 396, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":401
 *         :return: stream of transitions
 *         """
 *         iterator = self.thisptr.get().begin()             # <<<<<<<<<<<<<<
 *         while iterator != self.thisptr.get().end():
 *             trans = Trans()
 */
  __pyx_cur_scope->__pyx_v_iterator = __pyx_cur_scope->__pyx_v_self->thisptr.get()->begin();

  /* "libmata/nfa/nfa.pyx":402
 *         """
 *         iterator = self.thisptr.get().begin()
 *         while iterator != self.thisptr.get().end():             # <<<<<<<<<<<<<<
 *             trans = Trans()
 *             lhs = dereference(iterator)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_iterator != __pyx_cur_scope->__pyx_v_self->thisptr.get()->end()) != 0);
    if (!__pyx_t_1) break;

    /* "libmata/nfa/nfa.pyx":403
 *         iterator = self.thisptr.get().begin()
 *         while iterator != self.thisptr.get().end():
 *             trans = Trans()             # <<<<<<<<<<<<<<
 *             lhs = dereference(iterator)
 *             trans.copy_from(lhs)
 */
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Trans)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XGOTREF((PyObject *)__pyx_cur_scope->__pyx_v_trans);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_trans, ((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)__pyx_t_2));
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "libmata/nfa/nfa.pyx":404
 *         while iterator != self.thisptr.get().end():
 *             trans = Trans()
 *             lhs = dereference(iterator)             # <<<<<<<<<<<<<<
 *             trans.copy_from(lhs)
 *             preinc(iterator)
 */
    __pyx_cur_scope->__pyx_v_lhs = (*__pyx_cur_scope->__pyx_v_iterator);

    /* "libmata/nfa/nfa.pyx":405
 *             trans = Trans()
 *             lhs = dereference(iterator)
 *             trans.copy_from(lhs)             # <<<<<<<<<<<<<<
 *             preinc(iterator)
 *             yield trans
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_7libmata_3nfa_3nfa_Trans *)__pyx_cur_scope->__pyx_v_trans->__pyx_vtab)->copy_from(__pyx_cur_scope->__pyx_v_trans, __pyx_cur_scope->__pyx_v_lhs); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "libmata/nfa/nfa.pyx":406
 *             lhs = dereference(iterator)
 *             trans.copy_from(lhs)
 *             preinc(iterator)             # <<<<<<<<<<<<<<
 *             yield trans
 * 
 */
    (void)((++__pyx_cur_scope->__pyx_v_iterator));

    /* "libmata/nfa/nfa.pyx":407
 *             trans.copy_from(lhs)
 *             preinc(iterator)
 *             yield trans             # <<<<<<<<<<<<<<
 * 
 *     def get_transitions_from_state(self, State state):
 */
    __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_trans);
    __pyx_r = ((PyObject *)__pyx_cur_scope->__pyx_v_trans);
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(1, 407, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "libmata/nfa/nfa.pyx":396
 *         return self.thisptr.get().size()
 * 
 *     def iterate(self):             # <<<<<<<<<<<<<<
 *         """Iterates over all transitions
 * 
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("iterate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":409
 *             yield trans
 * 
 *     def get_transitions_from_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Returns list of Move for the given state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_52get_transitions_from_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_51get_transitions_from_state, "Returns list of Move for the given state\n\n        :param State state: state for which we are getting the transitions\n        :return: Move\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_52get_transitions_from_state = {"get_transitions_from_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_52get_transitions_from_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_51get_transitions_from_state};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_52get_transitions_from_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_state;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_transitions_from_state (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 409, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_transitions_from_state") < 0)) __PYX_ERR(1, 409, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_state = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_state == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 409, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_transitions_from_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 409, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_transitions_from_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_51get_transitions_from_state(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_51get_transitions_from_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state) {
  Mata::Nfa::Post __pyx_v_transitions;
  std::vector<Mata::Nfa::Move>  __pyx_v_transitions_list;
  std::vector<Mata::Nfa::Move> ::iterator __pyx_v_it;
  std::vector<Mata::Nfa::Move> ::iterator __pyx_v_end;
  PyObject *__pyx_v_transsymbols = NULL;
  struct __pyx_obj_7libmata_3nfa_3nfa_Move *__pyx_v_t = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_transitions_from_state", 0);

  /* "libmata/nfa/nfa.pyx":415
 *         :return: Move
 *         """
 *         cdef mata_nfa.CPost transitions = self.thisptr.get().get_moves_from(state)             # <<<<<<<<<<<<<<
 *         cdef vector[mata_nfa.CMove] transitions_list = transitions.ToVector()
 * 
 */
  __pyx_v_transitions = __pyx_v_self->thisptr.get()->get_moves_from(__pyx_v_state);

  /* "libmata/nfa/nfa.pyx":416
 *         """
 *         cdef mata_nfa.CPost transitions = self.thisptr.get().get_moves_from(state)
 *         cdef vector[mata_nfa.CMove] transitions_list = transitions.ToVector()             # <<<<<<<<<<<<<<
 * 
 *         cdef vector[mata_nfa.CMove].iterator it = transitions_list.begin()
 */
  __pyx_v_transitions_list = __pyx_v_transitions.ToVector();

  /* "libmata/nfa/nfa.pyx":418
 *         cdef vector[mata_nfa.CMove] transitions_list = transitions.ToVector()
 * 
 *         cdef vector[mata_nfa.CMove].iterator it = transitions_list.begin()             # <<<<<<<<<<<<<<
 *         cdef vector[mata_nfa.CMove].iterator end = transitions_list.end()
 *         transsymbols = []
 */
  __pyx_v_it = __pyx_v_transitions_list.begin();

  /* "libmata/nfa/nfa.pyx":419
 * 
 *         cdef vector[mata_nfa.CMove].iterator it = transitions_list.begin()
 *         cdef vector[mata_nfa.CMove].iterator end = transitions_list.end()             # <<<<<<<<<<<<<<
 *         transsymbols = []
 *         while it != end:
 */
  __pyx_v_end = __pyx_v_transitions_list.end();

  /* "libmata/nfa/nfa.pyx":420
 *         cdef vector[mata_nfa.CMove].iterator it = transitions_list.begin()
 *         cdef vector[mata_nfa.CMove].iterator end = transitions_list.end()
 *         transsymbols = []             # <<<<<<<<<<<<<<
 *         while it != end:
 *             t = Move(
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_transsymbols = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":421
 *         cdef vector[mata_nfa.CMove].iterator end = transitions_list.end()
 *         transsymbols = []
 *         while it != end:             # <<<<<<<<<<<<<<
 *             t = Move(
 *                 dereference(it).symbol,
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_it != __pyx_v_end);
    if (!__pyx_t_2) break;

    /* "libmata/nfa/nfa.pyx":423
 *         while it != end:
 *             t = Move(
 *                 dereference(it).symbol,             # <<<<<<<<<<<<<<
 *                 dereference(it).targets.ToVector()
 *             )
 */
    __pyx_t_1 = __Pyx_PyInt_FromSize_t((*__pyx_v_it).symbol); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "libmata/nfa/nfa.pyx":424
 *             t = Move(
 *                 dereference(it).symbol,
 *                 dereference(it).targets.ToVector()             # <<<<<<<<<<<<<<
 *             )
 *             postinc(it)
 */
    __pyx_t_3 = __pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State((*__pyx_v_it).targets.ToVector()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "libmata/nfa/nfa.pyx":422
 *         transsymbols = []
 *         while it != end:
 *             t = Move(             # <<<<<<<<<<<<<<
 *                 dereference(it).symbol,
 *                 dereference(it).targets.ToVector()
 */
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Move), __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_t, ((struct __pyx_obj_7libmata_3nfa_3nfa_Move *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "libmata/nfa/nfa.pyx":426
 *                 dereference(it).targets.ToVector()
 *             )
 *             postinc(it)             # <<<<<<<<<<<<<<
 *             transsymbols.append(t)
 *         return transsymbols
 */
    (void)((__pyx_v_it++));

    /* "libmata/nfa/nfa.pyx":427
 *             )
 *             postinc(it)
 *             transsymbols.append(t)             # <<<<<<<<<<<<<<
 *         return transsymbols
 * 
 */
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_transsymbols, ((PyObject *)__pyx_v_t)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 427, __pyx_L1_error)
  }

  /* "libmata/nfa/nfa.pyx":428
 *             postinc(it)
 *             transsymbols.append(t)
 *         return transsymbols             # <<<<<<<<<<<<<<
 * 
 *     def get_transitions_to_state(self, State state_to):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_transsymbols);
  __pyx_r = __pyx_v_transsymbols;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":409
 *             yield trans
 * 
 *     def get_transitions_from_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Returns list of Move for the given state
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_transitions_from_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_transsymbols);
  __Pyx_XDECREF((PyObject *)__pyx_v_t);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":430
 *         return transsymbols
 * 
 *     def get_transitions_to_state(self, State state_to):             # <<<<<<<<<<<<<<
 *         """Get transitions leading to state_to (state_to is their target state).
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_54get_transitions_to_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_53get_transitions_to_state, "Get transitions leading to state_to (state_to is their target state).\n\n        :return: List mata_nfa.CTrans: List of transitions leading to state_to.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_54get_transitions_to_state = {"get_transitions_to_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_54get_transitions_to_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_53get_transitions_to_state};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_54get_transitions_to_state(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_state_to;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_transitions_to_state (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state_to,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state_to)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 430, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_transitions_to_state") < 0)) __PYX_ERR(1, 430, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_state_to = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_state_to == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 430, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_transitions_to_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 430, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_transitions_to_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_53get_transitions_to_state(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_state_to);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_53get_transitions_to_state(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state_to) {
  std::vector<Mata::Nfa::Trans>  __pyx_v_c_transitions;
  PyObject *__pyx_v_trans = NULL;
  Mata::Nfa::Trans __pyx_v_c_transition;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::vector<Mata::Nfa::Trans> ::iterator __pyx_t_2;
  Mata::Nfa::Trans __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_transitions_to_state", 0);

  /* "libmata/nfa/nfa.pyx":435
 *         :return: List mata_nfa.CTrans: List of transitions leading to state_to.
 *         """
 *         cdef vector[CTrans] c_transitions = self.thisptr.get().get_transitions_to(state_to)             # <<<<<<<<<<<<<<
 *         trans = []
 *         for c_transition in c_transitions:
 */
  __pyx_v_c_transitions = __pyx_v_self->thisptr.get()->get_transitions_to(__pyx_v_state_to);

  /* "libmata/nfa/nfa.pyx":436
 *         """
 *         cdef vector[CTrans] c_transitions = self.thisptr.get().get_transitions_to(state_to)
 *         trans = []             # <<<<<<<<<<<<<<
 *         for c_transition in c_transitions:
 *             trans.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_trans = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":437
 *         cdef vector[CTrans] c_transitions = self.thisptr.get().get_transitions_to(state_to)
 *         trans = []
 *         for c_transition in c_transitions:             # <<<<<<<<<<<<<<
 *             trans.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))
 *         return trans
 */
  __pyx_t_2 = __pyx_v_c_transitions.begin();
  for (;;) {
    if (!(__pyx_t_2 != __pyx_v_c_transitions.end())) break;
    __pyx_t_3 = *__pyx_t_2;
    ++__pyx_t_2;
    __pyx_v_c_transition = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_3);

    /* "libmata/nfa/nfa.pyx":438
 *         trans = []
 *         for c_transition in c_transitions:
 *             trans.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))             # <<<<<<<<<<<<<<
 *         return trans
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_c_transition.src); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_c_transition.symb); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_c_transition.tgt); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Trans), __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_trans, __pyx_t_5); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 438, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "libmata/nfa/nfa.pyx":437
 *         cdef vector[CTrans] c_transitions = self.thisptr.get().get_transitions_to(state_to)
 *         trans = []
 *         for c_transition in c_transitions:             # <<<<<<<<<<<<<<
 *             trans.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))
 *         return trans
 */
  }

  /* "libmata/nfa/nfa.pyx":439
 *         for c_transition in c_transitions:
 *             trans.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))
 *         return trans             # <<<<<<<<<<<<<<
 * 
 *     def get_trans_as_sequence(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_trans);
  __pyx_r = __pyx_v_trans;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":430
 *         return transsymbols
 * 
 *     def get_transitions_to_state(self, State state_to):             # <<<<<<<<<<<<<<
 *         """Get transitions leading to state_to (state_to is their target state).
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_transitions_to_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_trans);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":441
 *         return trans
 * 
 *     def get_trans_as_sequence(self):             # <<<<<<<<<<<<<<
 *         """Get automaton transitions as a sequence.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_56get_trans_as_sequence(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_55get_trans_as_sequence, "Get automaton transitions as a sequence.\n\n        :return: List of automaton transitions.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_56get_trans_as_sequence = {"get_trans_as_sequence", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_56get_trans_as_sequence, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_55get_trans_as_sequence};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_56get_trans_as_sequence(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_trans_as_sequence (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_trans_as_sequence", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_trans_as_sequence", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_55get_trans_as_sequence(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_55get_trans_as_sequence(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  std::vector<Mata::Nfa::Trans>  __pyx_v_c_transitions;
  PyObject *__pyx_v_transitions = NULL;
  Mata::Nfa::Trans __pyx_v_c_transition;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::vector<Mata::Nfa::Trans> ::iterator __pyx_t_2;
  Mata::Nfa::Trans __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_trans_as_sequence", 0);

  /* "libmata/nfa/nfa.pyx":446
 *         :return: List of automaton transitions.
 *         """
 *         cdef vector[CTrans] c_transitions = self.thisptr.get().get_trans_as_sequence()             # <<<<<<<<<<<<<<
 *         transitions = []
 *         for c_transition in c_transitions:
 */
  __pyx_v_c_transitions = __pyx_v_self->thisptr.get()->get_trans_as_sequence();

  /* "libmata/nfa/nfa.pyx":447
 *         """
 *         cdef vector[CTrans] c_transitions = self.thisptr.get().get_trans_as_sequence()
 *         transitions = []             # <<<<<<<<<<<<<<
 *         for c_transition in c_transitions:
 *             transitions.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_transitions = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":448
 *         cdef vector[CTrans] c_transitions = self.thisptr.get().get_trans_as_sequence()
 *         transitions = []
 *         for c_transition in c_transitions:             # <<<<<<<<<<<<<<
 *             transitions.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))
 *         return transitions
 */
  __pyx_t_2 = __pyx_v_c_transitions.begin();
  for (;;) {
    if (!(__pyx_t_2 != __pyx_v_c_transitions.end())) break;
    __pyx_t_3 = *__pyx_t_2;
    ++__pyx_t_2;
    __pyx_v_c_transition = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_3);

    /* "libmata/nfa/nfa.pyx":449
 *         transitions = []
 *         for c_transition in c_transitions:
 *             transitions.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))             # <<<<<<<<<<<<<<
 *         return transitions
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_c_transition.src); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_c_transition.symb); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_c_transition.tgt); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Trans), __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_transitions, __pyx_t_5); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 449, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "libmata/nfa/nfa.pyx":448
 *         cdef vector[CTrans] c_transitions = self.thisptr.get().get_trans_as_sequence()
 *         transitions = []
 *         for c_transition in c_transitions:             # <<<<<<<<<<<<<<
 *             transitions.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))
 *         return transitions
 */
  }

  /* "libmata/nfa/nfa.pyx":450
 *         for c_transition in c_transitions:
 *             transitions.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))
 *         return transitions             # <<<<<<<<<<<<<<
 * 
 *     def get_trans_from_state_as_sequence(self, State state_from):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_transitions);
  __pyx_r = __pyx_v_transitions;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":441
 *         return trans
 * 
 *     def get_trans_as_sequence(self):             # <<<<<<<<<<<<<<
 *         """Get automaton transitions as a sequence.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_trans_as_sequence", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_transitions);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":452
 *         return transitions
 * 
 *     def get_trans_from_state_as_sequence(self, State state_from):             # <<<<<<<<<<<<<<
 *         """Get automaton transitions from state_from as a sequence.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_58get_trans_from_state_as_sequence(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_57get_trans_from_state_as_sequence, "Get automaton transitions from state_from as a sequence.\n\n        :return: List of automaton transitions.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_58get_trans_from_state_as_sequence = {"get_trans_from_state_as_sequence", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_58get_trans_from_state_as_sequence, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_57get_trans_from_state_as_sequence};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_58get_trans_from_state_as_sequence(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_state_from;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_trans_from_state_as_sequence (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state_from,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state_from)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 452, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_trans_from_state_as_sequence") < 0)) __PYX_ERR(1, 452, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_state_from = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_state_from == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 452, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_trans_from_state_as_sequence", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 452, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_trans_from_state_as_sequence", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_57get_trans_from_state_as_sequence(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_state_from);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_57get_trans_from_state_as_sequence(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state_from) {
  std::vector<Mata::Nfa::Trans>  __pyx_v_c_transitions;
  PyObject *__pyx_v_transitions = NULL;
  Mata::Nfa::Trans __pyx_v_c_transition;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::vector<Mata::Nfa::Trans> ::iterator __pyx_t_2;
  Mata::Nfa::Trans __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_trans_from_state_as_sequence", 0);

  /* "libmata/nfa/nfa.pyx":457
 *         :return: List of automaton transitions.
 *         """
 *         cdef vector[CTrans] c_transitions = self.thisptr.get().get_trans_from_as_sequence(state_from)             # <<<<<<<<<<<<<<
 *         transitions = []
 *         for c_transition in c_transitions:
 */
  __pyx_v_c_transitions = __pyx_v_self->thisptr.get()->get_trans_from_as_sequence(__pyx_v_state_from);

  /* "libmata/nfa/nfa.pyx":458
 *         """
 *         cdef vector[CTrans] c_transitions = self.thisptr.get().get_trans_from_as_sequence(state_from)
 *         transitions = []             # <<<<<<<<<<<<<<
 *         for c_transition in c_transitions:
 *             transitions.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_transitions = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":459
 *         cdef vector[CTrans] c_transitions = self.thisptr.get().get_trans_from_as_sequence(state_from)
 *         transitions = []
 *         for c_transition in c_transitions:             # <<<<<<<<<<<<<<
 *             transitions.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))
 *         return transitions
 */
  __pyx_t_2 = __pyx_v_c_transitions.begin();
  for (;;) {
    if (!(__pyx_t_2 != __pyx_v_c_transitions.end())) break;
    __pyx_t_3 = *__pyx_t_2;
    ++__pyx_t_2;
    __pyx_v_c_transition = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_3);

    /* "libmata/nfa/nfa.pyx":460
 *         transitions = []
 *         for c_transition in c_transitions:
 *             transitions.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))             # <<<<<<<<<<<<<<
 *         return transitions
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_c_transition.src); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_c_transition.symb); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_c_transition.tgt); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Trans), __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_transitions, __pyx_t_5); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 460, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "libmata/nfa/nfa.pyx":459
 *         cdef vector[CTrans] c_transitions = self.thisptr.get().get_trans_from_as_sequence(state_from)
 *         transitions = []
 *         for c_transition in c_transitions:             # <<<<<<<<<<<<<<
 *             transitions.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))
 *         return transitions
 */
  }

  /* "libmata/nfa/nfa.pyx":461
 *         for c_transition in c_transitions:
 *             transitions.append(Trans(c_transition.src, c_transition.symb, c_transition.tgt))
 *         return transitions             # <<<<<<<<<<<<<<
 * 
 *     def get_useful_states(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_transitions);
  __pyx_r = __pyx_v_transitions;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":452
 *         return transitions
 * 
 *     def get_trans_from_state_as_sequence(self, State state_from):             # <<<<<<<<<<<<<<
 *         """Get automaton transitions from state_from as a sequence.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_trans_from_state_as_sequence", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_transitions);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":463
 *         return transitions
 * 
 *     def get_useful_states(self):             # <<<<<<<<<<<<<<
 *         """Get useful states (states which are reachable and terminating at the same time).
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_60get_useful_states(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_59get_useful_states, "Get useful states (states which are reachable and terminating at the same time).\n\n        :return: A set of useful states.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_60get_useful_states = {"get_useful_states", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_60get_useful_states, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_59get_useful_states};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_60get_useful_states(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_useful_states (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_useful_states", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_useful_states", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_59get_useful_states(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_59get_useful_states(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  Mata::BoolVector __pyx_v_useful_states_bool_vec;
  Mata::Nfa::StateSet __pyx_v_useful_states;
  Mata::Nfa::State __pyx_8genexpr3__pyx_v_state;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Mata::Util::OrdVector<Mata::Nfa::State> ::iterator __pyx_t_2;
  Mata::Nfa::State __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_useful_states", 0);

  /* "libmata/nfa/nfa.pyx":468
 *         :return: A set of useful states.
 *         """
 *         cdef CBoolVector useful_states_bool_vec = self.thisptr.get().get_useful_states()             # <<<<<<<<<<<<<<
 *         cdef StateSet useful_states
 *         mata_nfa.get_elements(&useful_states, useful_states_bool_vec)
 */
  __pyx_v_useful_states_bool_vec = __pyx_v_self->thisptr.get()->get_useful_states();

  /* "libmata/nfa/nfa.pyx":470
 *         cdef CBoolVector useful_states_bool_vec = self.thisptr.get().get_useful_states()
 *         cdef StateSet useful_states
 *         mata_nfa.get_elements(&useful_states, useful_states_bool_vec)             # <<<<<<<<<<<<<<
 *         return { state for state in useful_states }
 * 
 */
  Mata::Nfa::Plumbing::get_elements((&__pyx_v_useful_states), __pyx_v_useful_states_bool_vec);

  /* "libmata/nfa/nfa.pyx":471
 *         cdef StateSet useful_states
 *         mata_nfa.get_elements(&useful_states, useful_states_bool_vec)
 *         return { state for state in useful_states }             # <<<<<<<<<<<<<<
 * 
 *     def get_reachable_states(self):
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PySet_New(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_useful_states.begin();
    for (;;) {
      if (!(__pyx_t_2 != __pyx_v_useful_states.end())) break;
      __pyx_t_3 = *__pyx_t_2;
      ++__pyx_t_2;
      __pyx_8genexpr3__pyx_v_state = __pyx_t_3;
      __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_8genexpr3__pyx_v_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 471, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(PySet_Add(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(1, 471, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":463
 *         return transitions
 * 
 *     def get_useful_states(self):             # <<<<<<<<<<<<<<
 *         """Get useful states (states which are reachable and terminating at the same time).
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_useful_states", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":473
 *         return { state for state in useful_states }
 * 
 *     def get_reachable_states(self):             # <<<<<<<<<<<<<<
 *         """Get reachable states.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_62get_reachable_states(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_61get_reachable_states, "Get reachable states.\n\n        :return: A set of reachable states.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_62get_reachable_states = {"get_reachable_states", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_62get_reachable_states, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_61get_reachable_states};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_62get_reachable_states(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_reachable_states (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_reachable_states", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_reachable_states", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_61get_reachable_states(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_61get_reachable_states(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  std::vector<Mata::Nfa::State>  __pyx_v_return_value;
  Mata::Nfa::State __pyx_8genexpr4__pyx_v_state;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::vector<Mata::Nfa::State> ::iterator __pyx_t_2;
  Mata::Nfa::State __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_reachable_states", 0);

  /* "libmata/nfa/nfa.pyx":478
 *         :return: A set of reachable states.
 *         """
 *         cdef vector[State] return_value = self.thisptr.get().get_reachable_states().ToVector()             # <<<<<<<<<<<<<<
 *         return {state for state in return_value}
 * 
 */
  __pyx_v_return_value = ((std::vector<Mata::Nfa::State> )__pyx_v_self->thisptr.get()->get_reachable_states().ToVector());

  /* "libmata/nfa/nfa.pyx":479
 *         """
 *         cdef vector[State] return_value = self.thisptr.get().get_reachable_states().ToVector()
 *         return {state for state in return_value}             # <<<<<<<<<<<<<<
 * 
 *     def get_terminating_states(self):
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PySet_New(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 479, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_return_value.begin();
    for (;;) {
      if (!(__pyx_t_2 != __pyx_v_return_value.end())) break;
      __pyx_t_3 = *__pyx_t_2;
      ++__pyx_t_2;
      __pyx_8genexpr4__pyx_v_state = __pyx_t_3;
      __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_8genexpr4__pyx_v_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 479, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(PySet_Add(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(1, 479, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":473
 *         return { state for state in useful_states }
 * 
 *     def get_reachable_states(self):             # <<<<<<<<<<<<<<
 *         """Get reachable states.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_reachable_states", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":481
 *         return {state for state in return_value}
 * 
 *     def get_terminating_states(self):             # <<<<<<<<<<<<<<
 *         """Get terminating states (states with a path from them leading to any final state).
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_64get_terminating_states(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_63get_terminating_states, "Get terminating states (states with a path from them leading to any final state).\n\n        :return: A set of terminating states.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_64get_terminating_states = {"get_terminating_states", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_64get_terminating_states, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_63get_terminating_states};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_64get_terminating_states(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_terminating_states (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_terminating_states", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_terminating_states", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_63get_terminating_states(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_63get_terminating_states(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  std::vector<Mata::Nfa::State>  __pyx_v_return_value;
  Mata::Nfa::State __pyx_8genexpr5__pyx_v_state;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::vector<Mata::Nfa::State> ::iterator __pyx_t_2;
  Mata::Nfa::State __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_terminating_states", 0);

  /* "libmata/nfa/nfa.pyx":486
 *         :return: A set of terminating states.
 *         """
 *         cdef vector[State] return_value = self.thisptr.get().get_terminating_states().ToVector()             # <<<<<<<<<<<<<<
 *         return {state for state in return_value}
 * 
 */
  __pyx_v_return_value = ((std::vector<Mata::Nfa::State> )__pyx_v_self->thisptr.get()->get_terminating_states().ToVector());

  /* "libmata/nfa/nfa.pyx":487
 *         """
 *         cdef vector[State] return_value = self.thisptr.get().get_terminating_states().ToVector()
 *         return {state for state in return_value}             # <<<<<<<<<<<<<<
 * 
 *     def trim(self):
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PySet_New(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_return_value.begin();
    for (;;) {
      if (!(__pyx_t_2 != __pyx_v_return_value.end())) break;
      __pyx_t_3 = *__pyx_t_2;
      ++__pyx_t_2;
      __pyx_8genexpr5__pyx_v_state = __pyx_t_3;
      __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_8genexpr5__pyx_v_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 487, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(PySet_Add(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(1, 487, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":481
 *         return {state for state in return_value}
 * 
 *     def get_terminating_states(self):             # <<<<<<<<<<<<<<
 *         """Get terminating states (states with a path from them leading to any final state).
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_terminating_states", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":489
 *         return {state for state in return_value}
 * 
 *     def trim(self):             # <<<<<<<<<<<<<<
 *         """Remove inaccessible (unreachable) and not co-accessible (non-terminating) states.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_66trim(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_65trim, "Remove inaccessible (unreachable) and not co-accessible (non-terminating) states.\n\n        Remove states which are not accessible (unreachable; state is accessible when the state is the endpoint of a path\n         starting from an initial state) or not co-accessible (non-terminating; state is co-accessible when the state is\n         the starting point of a path ending in a final state).\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_66trim = {"trim", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_66trim, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_65trim};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_66trim(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("trim (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("trim", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "trim", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_65trim(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_65trim(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("trim", 0);

  /* "libmata/nfa/nfa.pyx":496
 *          the starting point of a path ending in a final state).
 *         """
 *         self.thisptr.get().trim(NULL)             # <<<<<<<<<<<<<<
 * 
 *     def trim_with_state_map(self):
 */
  __pyx_v_self->thisptr.get()->trim(NULL);

  /* "libmata/nfa/nfa.pyx":489
 *         return {state for state in return_value}
 * 
 *     def trim(self):             # <<<<<<<<<<<<<<
 *         """Remove inaccessible (unreachable) and not co-accessible (non-terminating) states.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":498
 *         self.thisptr.get().trim(NULL)
 * 
 *     def trim_with_state_map(self):             # <<<<<<<<<<<<<<
 *         """Remove inaccessible (unreachable) and not co-accessible (non-terminating) states.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_68trim_with_state_map(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_67trim_with_state_map, "Remove inaccessible (unreachable) and not co-accessible (non-terminating) states.\n\n        Remove states which are not accessible (unreachable; state is accessible when the state is the endpoint of a path\n         starting from an initial state) or not co-accessible (non-terminating; state is co-accessible when the state is\n         the starting point of a path ending in a final state).\n\n        :return: State map of original to new states.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_68trim_with_state_map = {"trim_with_state_map", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_68trim_with_state_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_67trim_with_state_map};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_68trim_with_state_map(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("trim_with_state_map (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("trim_with_state_map", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "trim_with_state_map", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_67trim_with_state_map(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_67trim_with_state_map(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  Mata::Nfa::StateToStateMap __pyx_v_state_map;
  PyObject *__pyx_8genexpr6__pyx_v_k = NULL;
  PyObject *__pyx_8genexpr6__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::unordered_map<Mata::Nfa::State,Mata::Nfa::State> ::iterator __pyx_t_2;
  std::pair<Mata::Nfa::State,Mata::Nfa::State>  __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("trim_with_state_map", 0);

  /* "libmata/nfa/nfa.pyx":508
 *         """
 *         cdef StateToStateMap state_map
 *         self.thisptr.get().trim(&state_map)             # <<<<<<<<<<<<<<
 *         return {k: v for k, v in state_map}
 * 
 */
  __pyx_v_self->thisptr.get()->trim((&__pyx_v_state_map));

  /* "libmata/nfa/nfa.pyx":509
 *         cdef StateToStateMap state_map
 *         self.thisptr.get().trim(&state_map)
 *         return {k: v for k, v in state_map}             # <<<<<<<<<<<<<<
 * 
 *     def get_one_letter_aut(self) -> Nfa:
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 509, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_state_map.begin();
    for (;;) {
      if (!(__pyx_t_2 != __pyx_v_state_map.end())) break;
      __pyx_t_3 = *__pyx_t_2;
      ++__pyx_t_2;
      __pyx_t_4 = __pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 509, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(1, 509, __pyx_L5_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 509, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 509, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 509, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
        index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(1, 509, __pyx_L5_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L9_unpacking_done;
        __pyx_L8_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(1, 509, __pyx_L5_error)
        __pyx_L9_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_k, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_v, __pyx_t_6);
      __pyx_t_6 = 0;
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr6__pyx_v_k, (PyObject*)__pyx_8genexpr6__pyx_v_v))) __PYX_ERR(1, 509, __pyx_L5_error)
    }
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_k); __pyx_8genexpr6__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_v); __pyx_8genexpr6__pyx_v_v = 0;
    goto __pyx_L11_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_k); __pyx_8genexpr6__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_v); __pyx_8genexpr6__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L11_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":498
 *         self.thisptr.get().trim(NULL)
 * 
 *     def trim_with_state_map(self):             # <<<<<<<<<<<<<<
 *         """Remove inaccessible (unreachable) and not co-accessible (non-terminating) states.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.trim_with_state_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_k);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":511
 *         return {k: v for k, v in state_map}
 * 
 *     def get_one_letter_aut(self) -> Nfa:             # <<<<<<<<<<<<<<
 *         """Unify transitions to create a directed graph with at most a single transition between two states (using only
 *          one letter for all transitions).
 */

/* Python wrapper */
static struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_70get_one_letter_aut(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_69get_one_letter_aut, "Unify transitions to create a directed graph with at most a single transition between two states (using only\n         one letter for all transitions).\n\n        :return: mata_nfa.Nfa: One letter automaton representing a directed graph.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_70get_one_letter_aut = {"get_one_letter_aut", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_70get_one_letter_aut, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_69get_one_letter_aut};
static struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_70get_one_letter_aut(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_one_letter_aut (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_one_letter_aut", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_one_letter_aut", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_69get_one_letter_aut(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_69get_one_letter_aut(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_one_letter_aut = 0;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_one_letter_aut", 0);

  /* "libmata/nfa/nfa.pyx":517
 *         :return: mata_nfa.Nfa: One letter automaton representing a directed graph.
 *         """
 *         cdef Nfa one_letter_aut = mata_nfa.Nfa()             # <<<<<<<<<<<<<<
 *         self.thisptr.get().get_one_letter_aut(dereference(one_letter_aut.thisptr.get()))
 *         return one_letter_aut
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_one_letter_aut = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":518
 *         """
 *         cdef Nfa one_letter_aut = mata_nfa.Nfa()
 *         self.thisptr.get().get_one_letter_aut(dereference(one_letter_aut.thisptr.get()))             # <<<<<<<<<<<<<<
 *         return one_letter_aut
 * 
 */
  __pyx_v_self->thisptr.get()->get_one_letter_aut((*__pyx_v_one_letter_aut->thisptr.get()));

  /* "libmata/nfa/nfa.pyx":519
 *         cdef Nfa one_letter_aut = mata_nfa.Nfa()
 *         self.thisptr.get().get_one_letter_aut(dereference(one_letter_aut.thisptr.get()))
 *         return one_letter_aut             # <<<<<<<<<<<<<<
 * 
 *     def is_epsilon(self, Symbol symbol) -> bool:
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_one_letter_aut);
  __pyx_r = __pyx_v_one_letter_aut;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":511
 *         return {k: v for k, v in state_map}
 * 
 *     def get_one_letter_aut(self) -> Nfa:             # <<<<<<<<<<<<<<
 *         """Unify transitions to create a directed graph with at most a single transition between two states (using only
 *          one letter for all transitions).
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_one_letter_aut", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_one_letter_aut);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":521
 *         return one_letter_aut
 * 
 *     def is_epsilon(self, Symbol symbol) -> bool:             # <<<<<<<<<<<<<<
 *         """Check whether passed symbol is epsilon symbol or not.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_72is_epsilon(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_71is_epsilon, "Check whether passed symbol is epsilon symbol or not.\n\n        :param Symbol symbol: The symbol to check.\n        :return: True if the passed symbol is epsilon symbol, False otherwise.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_72is_epsilon = {"is_epsilon", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_72is_epsilon, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_71is_epsilon};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_72is_epsilon(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Symbol __pyx_v_symbol;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_epsilon (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_symbol,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_symbol)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 521, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "is_epsilon") < 0)) __PYX_ERR(1, 521, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_symbol = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_symbol == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 521, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_epsilon", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 521, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_epsilon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_71is_epsilon(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_symbol);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_71is_epsilon(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Symbol __pyx_v_symbol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_epsilon", 0);

  /* "libmata/nfa/nfa.pyx":527
 *         :return: True if the passed symbol is epsilon symbol, False otherwise.
 *         """
 *         return self.thisptr.get().is_epsilon(symbol)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->thisptr.get()->is_epsilon(__pyx_v_symbol)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":521
 *         return one_letter_aut
 * 
 *     def is_epsilon(self, Symbol symbol) -> bool:             # <<<<<<<<<<<<<<
 *         """Check whether passed symbol is epsilon symbol or not.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_epsilon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":529
 *         return self.thisptr.get().is_epsilon(symbol)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         """String representation of the automaton displays states, and transitions
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_74__str__(PyObject *__pyx_v_self); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_73__str__, "String representation of the automaton displays states, and transitions\n\n        :return: string representation of the automaton\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7libmata_3nfa_3nfa_3Nfa_73__str__;
#endif
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_74__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_73__str__(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_73__str__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_trans = NULL;
  PyObject *__pyx_v_symbol = NULL;
  Mata::Nfa::State __pyx_8genexpr7__pyx_v_s;
  Mata::Nfa::State __pyx_8genexpr8__pyx_v_s;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Mata::Util::SparseSet<Mata::Nfa::State> ::iterator __pyx_t_4;
  Mata::Util::SparseSet<Mata::Nfa::State>  __pyx_t_5;
  Mata::Nfa::State __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  Mata::Symbol __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_UCS4 __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "libmata/nfa/nfa.pyx":534
 *         :return: string representation of the automaton
 *         """
 *         result = "initial_states: {}\n".format([s for s in self.thisptr.get().initial])             # <<<<<<<<<<<<<<
 *         result += "final_states: {}\n".format([s for s in self.thisptr.get().final])
 *         result += "transitions:\n"
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_initial_states, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __pyx_v_self->thisptr.get()->initial;
    __pyx_t_4 = __pyx_t_5.begin();
    for (;;) {
      if (!(__pyx_t_4 != __pyx_t_5.end())) break;
      __pyx_t_6 = *__pyx_t_4;
      ++__pyx_t_4;
      __pyx_8genexpr7__pyx_v_s = __pyx_t_6;
      __pyx_t_7 = __Pyx_PyInt_FromSize_t(__pyx_8genexpr7__pyx_v_s); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 534, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_7))) __PYX_ERR(1, 534, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
  } /* exit inner scope */
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":535
 *         """
 *         result = "initial_states: {}\n".format([s for s in self.thisptr.get().initial])
 *         result += "final_states: {}\n".format([s for s in self.thisptr.get().final])             # <<<<<<<<<<<<<<
 *         result += "transitions:\n"
 *         for trans in self.iterate():
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_final_states, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __pyx_v_self->thisptr.get()->final;
    __pyx_t_4 = __pyx_t_5.begin();
    for (;;) {
      if (!(__pyx_t_4 != __pyx_t_5.end())) break;
      __pyx_t_6 = *__pyx_t_4;
      ++__pyx_t_4;
      __pyx_8genexpr8__pyx_v_s = __pyx_t_6;
      __pyx_t_7 = __Pyx_PyInt_FromSize_t(__pyx_8genexpr8__pyx_v_s); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 535, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_7))) __PYX_ERR(1, 535, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
  } /* exit inner scope */
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_result, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":536
 *         result = "initial_states: {}\n".format([s for s in self.thisptr.get().initial])
 *         result += "final_states: {}\n".format([s for s in self.thisptr.get().final])
 *         result += "transitions:\n"             # <<<<<<<<<<<<<<
 *         for trans in self.iterate():
 *             symbol = trans.symb if self.thisptr.get().alphabet == NULL \
 */
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_result, __pyx_kp_u_transitions); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":537
 *         result += "final_states: {}\n".format([s for s in self.thisptr.get().final])
 *         result += "transitions:\n"
 *         for trans in self.iterate():             # <<<<<<<<<<<<<<
 *             symbol = trans.symb if self.thisptr.get().alphabet == NULL \
 *                 else self.thisptr.get().alphabet.reverse_translate_symbol(trans.symb)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_iterate); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_8 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 537, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
    __pyx_t_10 = NULL;
  } else {
    __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 537, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 537, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_10)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(1, 537, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 537, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(1, 537, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 537, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_10(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 537, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_trans, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "libmata/nfa/nfa.pyx":538
 *         result += "transitions:\n"
 *         for trans in self.iterate():
 *             symbol = trans.symb if self.thisptr.get().alphabet == NULL \             # <<<<<<<<<<<<<<
 *                 else self.thisptr.get().alphabet.reverse_translate_symbol(trans.symb)
 *             result += f"{trans.src}-[{symbol}]\u2192{trans.tgt}\n"
 */
    if ((__pyx_v_self->thisptr.get()->alphabet == NULL)) {
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_trans, __pyx_n_s_symb); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 538, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __pyx_t_3;
      __pyx_t_3 = 0;
    } else {

      /* "libmata/nfa/nfa.pyx":539
 *         for trans in self.iterate():
 *             symbol = trans.symb if self.thisptr.get().alphabet == NULL \
 *                 else self.thisptr.get().alphabet.reverse_translate_symbol(trans.symb)             # <<<<<<<<<<<<<<
 *             result += f"{trans.src}-[{symbol}]\u2192{trans.tgt}\n"
 *         return result
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_trans, __pyx_n_s_symb); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 539, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_11 = __Pyx_PyInt_As_size_t(__pyx_t_3); if (unlikely((__pyx_t_11 == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 539, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __pyx_convert_PyBytes_string_to_py_std__in_string(__pyx_v_self->thisptr.get()->alphabet->reverse_translate_symbol(__pyx_t_11)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 539, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __pyx_t_3;
      __pyx_t_3 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_symbol, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "libmata/nfa/nfa.pyx":540
 *             symbol = trans.symb if self.thisptr.get().alphabet == NULL \
 *                 else self.thisptr.get().alphabet.reverse_translate_symbol(trans.symb)
 *             result += f"{trans.src}-[{symbol}]\u2192{trans.tgt}\n"             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_2 = PyTuple_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = 0;
    __pyx_t_13 = 127;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_trans, __pyx_n_s_src); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_);
    __pyx_t_12 += 2;
    __Pyx_GIVEREF(__pyx_kp_u_);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_kp_u_);
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_v_symbol, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u__2);
    __pyx_t_13 = (65535 > __pyx_t_13) ? 65535 : __pyx_t_13;
    __pyx_t_12 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__2);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_kp_u__2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_trans, __pyx_n_s_tgt); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_kp_u__8);
    __pyx_t_12 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__8);
    PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_kp_u__8);
    __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_2, 6, __pyx_t_12, __pyx_t_13); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_result, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "libmata/nfa/nfa.pyx":537
 *         result += "final_states: {}\n".format([s for s in self.thisptr.get().final])
 *         result += "transitions:\n"
 *         for trans in self.iterate():             # <<<<<<<<<<<<<<
 *             symbol = trans.symb if self.thisptr.get().alphabet == NULL \
 *                 else self.thisptr.get().alphabet.reverse_translate_symbol(trans.symb)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":541
 *                 else self.thisptr.get().alphabet.reverse_translate_symbol(trans.symb)
 *             result += f"{trans.src}-[{symbol}]\u2192{trans.tgt}\n"
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":529
 *         return self.thisptr.get().is_epsilon(symbol)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         """String representation of the automaton displays states, and transitions
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_trans);
  __Pyx_XDECREF(__pyx_v_symbol);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":543
 *         return result
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_76__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_76__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_75__repr__(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_75__repr__(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "libmata/nfa/nfa.pyx":544
 * 
 *     def __repr__(self):
 *         return str(self)             # <<<<<<<<<<<<<<
 * 
 *     def to_dot_file(self, output_file='aut.dot', output_format='pdf'):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Str(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":543
 *         return result
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":546
 *         return str(self)
 * 
 *     def to_dot_file(self, output_file='aut.dot', output_format='pdf'):             # <<<<<<<<<<<<<<
 *         """Transforms the automaton to dot format.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_78to_dot_file(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_77to_dot_file, "Transforms the automaton to dot format.\n\n        By default, the result is saved to `aut.dot`, and further to `aut.dot.pdf`.\n\n        One can choose other format that is supported by graphviz format (e.g. pdf or png).\n\n        :param str output_file: name of the output file where the automaton will be stored\n        :param str output_format: format of the output file (pdf/png/etc)\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_78to_dot_file = {"to_dot_file", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_78to_dot_file, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_77to_dot_file};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_78to_dot_file(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_output_file = 0;
  PyObject *__pyx_v_output_format = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_dot_file (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_output_file,&__pyx_n_s_output_format,0};
    PyObject* values[2] = {0,0};
    values[0] = ((PyObject *)__pyx_kp_u_aut_dot);
    values[1] = ((PyObject *)__pyx_n_u_pdf);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_output_file);
          if (value) { values[0] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 546, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_output_format);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 546, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "to_dot_file") < 0)) __PYX_ERR(1, 546, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_output_file = values[0];
    __pyx_v_output_format = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_dot_file", 0, 0, 2, __pyx_nargs); __PYX_ERR(1, 546, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.to_dot_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_77to_dot_file(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_output_file, __pyx_v_output_format);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_77to_dot_file(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, PyObject *__pyx_v_output_file, PyObject *__pyx_v_output_format) {
  std::ofstream *__pyx_v_output;
  PyObject *__pyx_v_graphviz_command = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  char const *__pyx_t_5;
  std::ofstream *__pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_dot_file", 0);

  /* "libmata/nfa/nfa.pyx":557
 *         """
 *         cdef mata_nfa.ofstream* output
 *         output = new mata_nfa.ofstream(output_file.encode('utf-8'))             # <<<<<<<<<<<<<<
 *         try:
 *             self.thisptr.get().print_to_DOT(dereference(output))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_output_file, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_utf_8};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 557, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_AsString(__pyx_t_1); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(1, 557, __pyx_L1_error)
  try {
    __pyx_t_6 = new std::ofstream(__pyx_t_5);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 557, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_output = __pyx_t_6;

  /* "libmata/nfa/nfa.pyx":558
 *         cdef mata_nfa.ofstream* output
 *         output = new mata_nfa.ofstream(output_file.encode('utf-8'))
 *         try:             # <<<<<<<<<<<<<<
 *             self.thisptr.get().print_to_DOT(dereference(output))
 *         finally:
 */
  /*try:*/ {

    /* "libmata/nfa/nfa.pyx":559
 *         output = new mata_nfa.ofstream(output_file.encode('utf-8'))
 *         try:
 *             self.thisptr.get().print_to_DOT(dereference(output))             # <<<<<<<<<<<<<<
 *         finally:
 *             del output
 */
    __pyx_v_self->thisptr.get()->print_to_DOT((*__pyx_v_output));
  }

  /* "libmata/nfa/nfa.pyx":561
 *             self.thisptr.get().print_to_DOT(dereference(output))
 *         finally:
 *             del output             # <<<<<<<<<<<<<<
 * 
 *         graphviz_command = f"dot -O -T{output_format} {output_file}"
 */
  /*finally:*/ {
    /*normal exit:*/{
      delete __pyx_v_output;
      goto __pyx_L5;
    }
    __pyx_L5:;
  }

  /* "libmata/nfa/nfa.pyx":563
 *             del output
 * 
 *         graphviz_command = f"dot -O -T{output_format} {output_file}"             # <<<<<<<<<<<<<<
 *         _, err = run_safely_external_command(graphviz_command)
 *         if err:
 */
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = 0;
  __pyx_t_8 = 127;
  __Pyx_INCREF(__pyx_kp_u_dot_O_T);
  __pyx_t_7 += 9;
  __Pyx_GIVEREF(__pyx_kp_u_dot_O_T);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_dot_O_T);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_output_format, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_8;
  __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_INCREF(__pyx_kp_u__9);
  __pyx_t_7 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__9);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__9);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_output_file, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_8;
  __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_graphviz_command = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":564
 * 
 *         graphviz_command = f"dot -O -T{output_format} {output_file}"
 *         _, err = run_safely_external_command(graphviz_command)             # <<<<<<<<<<<<<<
 *         if err:
 *             print(f"error while dot file: {err}")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_run_safely_external_command); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_graphviz_command};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 564, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_9 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_9);
    index = 0; __pyx_t_1 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_1)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_3 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(1, 564, __pyx_L1_error)
    __pyx_t_10 = NULL;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(1, 564, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }
  __pyx_v__ = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_err = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "libmata/nfa/nfa.pyx":565
 *         graphviz_command = f"dot -O -T{output_format} {output_file}"
 *         _, err = run_safely_external_command(graphviz_command)
 *         if err:             # <<<<<<<<<<<<<<
 *             print(f"error while dot file: {err}")
 * 
 */
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_err); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(1, 565, __pyx_L1_error)
  if (__pyx_t_11) {

    /* "libmata/nfa/nfa.pyx":566
 *         _, err = run_safely_external_command(graphviz_command)
 *         if err:
 *             print(f"error while dot file: {err}")             # <<<<<<<<<<<<<<
 * 
 *     def to_dot_str(self, encoding='utf-8'):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_err, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u_error_while_dot_file, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "libmata/nfa/nfa.pyx":565
 *         graphviz_command = f"dot -O -T{output_format} {output_file}"
 *         _, err = run_safely_external_command(graphviz_command)
 *         if err:             # <<<<<<<<<<<<<<
 *             print(f"error while dot file: {err}")
 * 
 */
  }

  /* "libmata/nfa/nfa.pyx":546
 *         return str(self)
 * 
 *     def to_dot_file(self, output_file='aut.dot', output_format='pdf'):             # <<<<<<<<<<<<<<
 *         """Transforms the automaton to dot format.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.to_dot_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_graphviz_command);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":568
 *             print(f"error while dot file: {err}")
 * 
 *     def to_dot_str(self, encoding='utf-8'):             # <<<<<<<<<<<<<<
 *         """Transforms the automaton to dot format string
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_80to_dot_str(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_79to_dot_str, "Transforms the automaton to dot format string\n\n        :param str encoding: encoding of the dot string\n        :return: string with dot representation of the automaton\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_80to_dot_str = {"to_dot_str", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_80to_dot_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_79to_dot_str};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_80to_dot_str(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_encoding = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_dot_str (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_encoding,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_kp_u_utf_8);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_encoding);
          if (value) { values[0] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 568, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "to_dot_str") < 0)) __PYX_ERR(1, 568, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_encoding = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_dot_str", 0, 0, 1, __pyx_nargs); __PYX_ERR(1, 568, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.to_dot_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_79to_dot_str(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_encoding);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_79to_dot_str(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, PyObject *__pyx_v_encoding) {
  std::stringstream *__pyx_v_output_stream;
  std::string __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  std::string __pyx_t_1;
  std::stringstream *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_dot_str", 0);

  /* "libmata/nfa/nfa.pyx":575
 *         """
 *         cdef mata_nfa.stringstream* output_stream
 *         output_stream = new mata_nfa.stringstream("".encode('ascii'))             # <<<<<<<<<<<<<<
 *         cdef string result
 *         try:
 */
  __pyx_t_1 = __pyx_convert_string_from_py_std__in_string(__pyx_kp_b__10); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 575, __pyx_L1_error)
  try {
    __pyx_t_2 = new std::stringstream(__pyx_t_1);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 575, __pyx_L1_error)
  }
  __pyx_v_output_stream = __pyx_t_2;

  /* "libmata/nfa/nfa.pyx":577
 *         output_stream = new mata_nfa.stringstream("".encode('ascii'))
 *         cdef string result
 *         try:             # <<<<<<<<<<<<<<
 *             self.thisptr.get().print_to_DOT(dereference(output_stream))
 *             result = output_stream.str()
 */
  /*try:*/ {

    /* "libmata/nfa/nfa.pyx":578
 *         cdef string result
 *         try:
 *             self.thisptr.get().print_to_DOT(dereference(output_stream))             # <<<<<<<<<<<<<<
 *             result = output_stream.str()
 *         finally:
 */
    __pyx_v_self->thisptr.get()->print_to_DOT((*__pyx_v_output_stream));

    /* "libmata/nfa/nfa.pyx":579
 *         try:
 *             self.thisptr.get().print_to_DOT(dereference(output_stream))
 *             result = output_stream.str()             # <<<<<<<<<<<<<<
 *         finally:
 *             del output_stream
 */
    __pyx_v_result = __pyx_v_output_stream->str();
  }

  /* "libmata/nfa/nfa.pyx":581
 *             result = output_stream.str()
 *         finally:
 *             del output_stream             # <<<<<<<<<<<<<<
 *         return result.decode(encoding)
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      delete __pyx_v_output_stream;
      goto __pyx_L5;
    }
    __pyx_L5:;
  }

  /* "libmata/nfa/nfa.pyx":582
 *         finally:
 *             del output_stream
 *         return result.decode(encoding)             # <<<<<<<<<<<<<<
 * 
 *     def to_dataframe(self) -> pandas.DataFrame:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __pyx_convert_PyBytes_string_to_py_std__in_string(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_decode); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_encoding};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 582, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":568
 *             print(f"error while dot file: {err}")
 * 
 *     def to_dot_str(self, encoding='utf-8'):             # <<<<<<<<<<<<<<
 *         """Transforms the automaton to dot format string
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.to_dot_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":584
 *         return result.decode(encoding)
 * 
 *     def to_dataframe(self) -> pandas.DataFrame:             # <<<<<<<<<<<<<<
 *         """Transforms the automaton to DataFrame format.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_82to_dataframe(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_81to_dataframe, "Transforms the automaton to DataFrame format.\n\n        Transforms the automaton into pandas.DataFrame format,\n        that is suitable for further (statistical) analysis.\n        The resulting DataFrame is in tabular format, with\n        'source', 'symbol' and 'target' as the columns\n\n        :return: automaton represented as a pandas dataframe\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_82to_dataframe = {"to_dataframe", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_82to_dataframe, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_81to_dataframe};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_82to_dataframe(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_dataframe (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("to_dataframe", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "to_dataframe", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_81to_dataframe(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_81to_dataframe(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_v_columns = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_8genexpr9__pyx_v_trans = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_dataframe", 0);

  /* "libmata/nfa/nfa.pyx":594
 *         :return: automaton represented as a pandas dataframe
 *         """
 *         columns = ['source', 'symbol', 'target']             # <<<<<<<<<<<<<<
 *         data = [
 *             [trans.src, trans.symb, trans.tgt] for trans in self.iterate()
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_source);
  __Pyx_GIVEREF(__pyx_n_u_source);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_source);
  __Pyx_INCREF(__pyx_n_u_symbol);
  __Pyx_GIVEREF(__pyx_n_u_symbol);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_symbol);
  __Pyx_INCREF(__pyx_n_u_target);
  __Pyx_GIVEREF(__pyx_n_u_target);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_target);
  __pyx_v_columns = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":595
 *         """
 *         columns = ['source', 'symbol', 'target']
 *         data = [             # <<<<<<<<<<<<<<
 *             [trans.src, trans.symb, trans.tgt] for trans in self.iterate()
 *         ]
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 595, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "libmata/nfa/nfa.pyx":596
 *         columns = ['source', 'symbol', 'target']
 *         data = [
 *             [trans.src, trans.symb, trans.tgt] for trans in self.iterate()             # <<<<<<<<<<<<<<
 *         ]
 *         return pandas.DataFrame(data, columns=columns)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_iterate); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 596, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[1] = {__pyx_t_4, };
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 596, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 596, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 596, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(1, 596, __pyx_L5_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 596, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(1, 596, __pyx_L5_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 596, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_7(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(1, 596, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_trans, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr9__pyx_v_trans, __pyx_n_s_src); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 596, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr9__pyx_v_trans, __pyx_n_s_symb); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 596, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr9__pyx_v_trans, __pyx_n_s_tgt); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 596, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PyList_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 596, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_2);
      PyList_SET_ITEM(__pyx_t_9, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4);
      PyList_SET_ITEM(__pyx_t_9, 1, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_8);
      PyList_SET_ITEM(__pyx_t_9, 2, __pyx_t_8);
      __pyx_t_2 = 0;
      __pyx_t_4 = 0;
      __pyx_t_8 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_9))) __PYX_ERR(1, 595, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_trans); __pyx_8genexpr9__pyx_v_trans = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_trans); __pyx_8genexpr9__pyx_v_trans = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_v_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":598
 *             [trans.src, trans.symb, trans.tgt] for trans in self.iterate()
 *         ]
 *         return pandas.DataFrame(data, columns=columns)             # <<<<<<<<<<<<<<
 * 
 *     def to_networkx_graph(self) -> nx.Graph:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pandas); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_DataFrame); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_data);
  __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_columns, __pyx_v_columns) < 0) __PYX_ERR(1, 598, __pyx_L1_error)
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":584
 *         return result.decode(encoding)
 * 
 *     def to_dataframe(self) -> pandas.DataFrame:             # <<<<<<<<<<<<<<
 *         """Transforms the automaton to DataFrame format.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.to_dataframe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_columns);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_trans);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":600
 *         return pandas.DataFrame(data, columns=columns)
 * 
 *     def to_networkx_graph(self) -> nx.Graph:             # <<<<<<<<<<<<<<
 *         """Transforms the automaton into networkx.Graph
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_84to_networkx_graph(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_83to_networkx_graph, "Transforms the automaton into networkx.Graph\n\n        Transforms the automaton into networkx.Graph format,\n        that is represented as graph with edges (src, tgt) with\n        additional properties.\n\n        Each symbol is added as an property to each edge.\n\n        :return:\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_84to_networkx_graph = {"to_networkx_graph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_84to_networkx_graph, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_83to_networkx_graph};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_84to_networkx_graph(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_networkx_graph (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("to_networkx_graph", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "to_networkx_graph", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_83to_networkx_graph(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_83to_networkx_graph(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_v_G = NULL;
  PyObject *__pyx_v_trans = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_networkx_graph", 0);

  /* "libmata/nfa/nfa.pyx":611
 *         :return:
 *         """
 *         G = nx.DiGraph()             # <<<<<<<<<<<<<<
 *         for trans in self.iterate():
 *             G.add_edge(trans.src, trans.tgt, symbol=trans.symb)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_nx); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_DiGraph); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_2, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_G = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":612
 *         """
 *         G = nx.DiGraph()
 *         for trans in self.iterate():             # <<<<<<<<<<<<<<
 *             G.add_edge(trans.src, trans.tgt, symbol=trans.symb)
 *         return G
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_iterate); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_2, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 612, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(1, 612, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 612, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(1, 612, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 612, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 612, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_trans, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "libmata/nfa/nfa.pyx":613
 *         G = nx.DiGraph()
 *         for trans in self.iterate():
 *             G.add_edge(trans.src, trans.tgt, symbol=trans.symb)             # <<<<<<<<<<<<<<
 *         return G
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_G, __pyx_n_s_add_edge); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_trans, __pyx_n_s_src); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_trans, __pyx_n_s_tgt); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_7);
    __pyx_t_2 = 0;
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_trans, __pyx_n_s_symb); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_symbol, __pyx_t_2) < 0) __PYX_ERR(1, 613, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "libmata/nfa/nfa.pyx":612
 *         """
 *         G = nx.DiGraph()
 *         for trans in self.iterate():             # <<<<<<<<<<<<<<
 *             G.add_edge(trans.src, trans.tgt, symbol=trans.symb)
 *         return G
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "libmata/nfa/nfa.pyx":614
 *         for trans in self.iterate():
 *             G.add_edge(trans.src, trans.tgt, symbol=trans.symb)
 *         return G             # <<<<<<<<<<<<<<
 * 
 *     def post_map_of(self, State st, alph.Alphabet alphabet):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_G);
  __pyx_r = __pyx_v_G;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":600
 *         return pandas.DataFrame(data, columns=columns)
 * 
 *     def to_networkx_graph(self) -> nx.Graph:             # <<<<<<<<<<<<<<
 *         """Transforms the automaton into networkx.Graph
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.to_networkx_graph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_G);
  __Pyx_XDECREF(__pyx_v_trans);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":616
 *         return G
 * 
 *     def post_map_of(self, State st, alph.Alphabet alphabet):             # <<<<<<<<<<<<<<
 *         """Returns mapping of symbols to set of states.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_86post_map_of(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_85post_map_of, "Returns mapping of symbols to set of states.\n\n        :param State st: source state\n        :param alph.Alphabet alphabet: alphabet of the post\n        :return: dictionary mapping symbols to set of reachable states from the symbol\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_86post_map_of = {"post_map_of", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_86post_map_of, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_85post_map_of};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_86post_map_of(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_st;
  struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("post_map_of (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_st,&__pyx_n_s_alphabet,0};
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_st)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 616, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_alphabet)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 616, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("post_map_of", 1, 2, 2, 1); __PYX_ERR(1, 616, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "post_map_of") < 0)) __PYX_ERR(1, 616, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_st = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_st == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 616, __pyx_L3_error)
    __pyx_v_alphabet = ((struct __pyx_obj_7libmata_9alphabets_Alphabet *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("post_map_of", 1, 2, 2, __pyx_nargs); __PYX_ERR(1, 616, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.post_map_of", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_alphabet), __pyx_ptype_7libmata_9alphabets_Alphabet, 1, "alphabet", 0))) __PYX_ERR(1, 616, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_85post_map_of(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_st, __pyx_v_alphabet);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_85post_map_of(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_st, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet) {
  PyObject *__pyx_v_symbol_map = NULL;
  PyObject *__pyx_v_symbol = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("post_map_of", 0);

  /* "libmata/nfa/nfa.pyx":623
 *         :return: dictionary mapping symbols to set of reachable states from the symbol
 *         """
 *         symbol_map = {}             # <<<<<<<<<<<<<<
 *         for symbol in alphabet.get_alphabet_symbols():
 *             symbol_map[symbol] = self.post_of({st}, symbol)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_symbol_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":624
 *         """
 *         symbol_map = {}
 *         for symbol in alphabet.get_alphabet_symbols():             # <<<<<<<<<<<<<<
 *             symbol_map[symbol] = self.post_of({st}, symbol)
 *         return symbol_map
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_alphabet), __pyx_n_s_get_alphabet_symbols); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 624, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 624, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 624, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(1, 624, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 624, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(1, 624, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 624, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 624, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_symbol, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "libmata/nfa/nfa.pyx":625
 *         symbol_map = {}
 *         for symbol in alphabet.get_alphabet_symbols():
 *             symbol_map[symbol] = self.post_of({st}, symbol)             # <<<<<<<<<<<<<<
 *         return symbol_map
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_post_of); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_FromSize_t(__pyx_v_st); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = PySet_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (PySet_Add(__pyx_t_8, __pyx_t_7) < 0) __PYX_ERR(1, 625, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_8, __pyx_v_symbol};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 625, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    if (unlikely((PyDict_SetItem(__pyx_v_symbol_map, __pyx_v_symbol, __pyx_t_1) < 0))) __PYX_ERR(1, 625, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "libmata/nfa/nfa.pyx":624
 *         """
 *         symbol_map = {}
 *         for symbol in alphabet.get_alphabet_symbols():             # <<<<<<<<<<<<<<
 *             symbol_map[symbol] = self.post_of({st}, symbol)
 *         return symbol_map
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":626
 *         for symbol in alphabet.get_alphabet_symbols():
 *             symbol_map[symbol] = self.post_of({st}, symbol)
 *         return symbol_map             # <<<<<<<<<<<<<<
 * 
 *     def post_of(self, vector[State] states, Symbol symbol):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_symbol_map);
  __pyx_r = __pyx_v_symbol_map;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":616
 *         return G
 * 
 *     def post_map_of(self, State st, alph.Alphabet alphabet):             # <<<<<<<<<<<<<<
 *         """Returns mapping of symbols to set of states.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.post_map_of", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_symbol_map);
  __Pyx_XDECREF(__pyx_v_symbol);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":628
 *         return symbol_map
 * 
 *     def post_of(self, vector[State] states, Symbol symbol):             # <<<<<<<<<<<<<<
 *         """Returns sets of reachable states from set of states through a symbol
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_88post_of(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_87post_of, "Returns sets of reachable states from set of states through a symbol\n\n        :param StateSet states: set of states\n        :param Symbol symbol: source symbol\n        :return: set of reachable states\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_88post_of = {"post_of", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_88post_of, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_87post_of};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_88post_of(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  std::vector<Mata::Nfa::State>  __pyx_v_states;
  Mata::Symbol __pyx_v_symbol;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("post_of (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_states,&__pyx_n_s_symbol,0};
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_states)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 628, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_symbol)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 628, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("post_of", 1, 2, 2, 1); __PYX_ERR(1, 628, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "post_of") < 0)) __PYX_ERR(1, 628, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_states = __pyx_convert_vector_from_py_Mata_3a__3a_Nfa_3a__3a_State(values[0]); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 628, __pyx_L3_error)
    __pyx_v_symbol = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_symbol == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 628, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("post_of", 1, 2, 2, __pyx_nargs); __PYX_ERR(1, 628, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.post_of", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_87post_of(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __PYX_STD_MOVE_IF_SUPPORTED(__pyx_v_states), __pyx_v_symbol);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_87post_of(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, std::vector<Mata::Nfa::State>  __pyx_v_states, Mata::Symbol __pyx_v_symbol) {
  std::vector<Mata::Nfa::State>  __pyx_v_return_value;
  Mata::Nfa::StateSet __pyx_v_input_states;
  Mata::Nfa::State __pyx_9genexpr10__pyx_v_v;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Mata::Nfa::StateSet __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  std::vector<Mata::Nfa::State> ::iterator __pyx_t_3;
  Mata::Nfa::State __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("post_of", 0);

  /* "libmata/nfa/nfa.pyx":636
 *         """
 *         cdef vector[State] return_value
 *         cdef StateSet input_states = StateSet(states)             # <<<<<<<<<<<<<<
 *         return_value = self.thisptr.get().post(input_states, symbol).ToVector()
 *         return {v for v in return_value}
 */
  try {
    __pyx_t_1 = Mata::Nfa::StateSet(((std::vector<Mata::Nfa::State>  &)__pyx_v_states));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 636, __pyx_L1_error)
  }
  __pyx_v_input_states = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "libmata/nfa/nfa.pyx":637
 *         cdef vector[State] return_value
 *         cdef StateSet input_states = StateSet(states)
 *         return_value = self.thisptr.get().post(input_states, symbol).ToVector()             # <<<<<<<<<<<<<<
 *         return {v for v in return_value}
 * 
 */
  __pyx_v_return_value = ((std::vector<Mata::Nfa::State> )__pyx_v_self->thisptr.get()->post(__pyx_v_input_states, __pyx_v_symbol).ToVector());

  /* "libmata/nfa/nfa.pyx":638
 *         cdef StateSet input_states = StateSet(states)
 *         return_value = self.thisptr.get().post(input_states, symbol).ToVector()
 *         return {v for v in return_value}             # <<<<<<<<<<<<<<
 * 
 *     # Operations
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PySet_New(NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_v_return_value.begin();
    for (;;) {
      if (!(__pyx_t_3 != __pyx_v_return_value.end())) break;
      __pyx_t_4 = *__pyx_t_3;
      ++__pyx_t_3;
      __pyx_9genexpr10__pyx_v_v = __pyx_t_4;
      __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_9genexpr10__pyx_v_v); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(PySet_Add(__pyx_t_2, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 638, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":628
 *         return symbol_map
 * 
 *     def post_of(self, vector[State] states, Symbol symbol):             # <<<<<<<<<<<<<<
 *         """Returns sets of reachable states from set of states through a symbol
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.post_of", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":641
 * 
 *     # Operations
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def determinize_with_subset_map(cls, Nfa lhs):
 *         """Determinize the lhs automaton
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_90determinize_with_subset_map(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_89determinize_with_subset_map, "Determinize the lhs automaton\n\n        :param Nfa lhs: non-deterministic finite automaton\n        :return: deterministic finite automaton, subset map\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_90determinize_with_subset_map = {"determinize_with_subset_map", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_90determinize_with_subset_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_89determinize_with_subset_map};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_90determinize_with_subset_map(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("determinize_with_subset_map (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 641, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "determinize_with_subset_map") < 0)) __PYX_ERR(1, 641, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("determinize_with_subset_map", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 641, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.determinize_with_subset_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 642, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_89determinize_with_subset_map(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_89determinize_with_subset_map(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  std::unordered_map<Mata::Nfa::StateSet,Mata::Nfa::State>  __pyx_v_subset_map;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("determinize_with_subset_map", 0);

  /* "libmata/nfa/nfa.pyx":648
 *         :return: deterministic finite automaton, subset map
 *         """
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         cdef umap[StateSet, State] subset_map
 *         mata_nfa.determinize(result.thisptr.get(), dereference(lhs.thisptr.get()), &subset_map)
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":650
 *         result = Nfa()
 *         cdef umap[StateSet, State] subset_map
 *         mata_nfa.determinize(result.thisptr.get(), dereference(lhs.thisptr.get()), &subset_map)             # <<<<<<<<<<<<<<
 *         return result, subset_map_to_dictionary(subset_map)
 * 
 */
  Mata::Nfa::Plumbing::determinize(__pyx_v_result->thisptr.get(), (*__pyx_v_lhs->thisptr.get()), (&__pyx_v_subset_map));

  /* "libmata/nfa/nfa.pyx":651
 *         cdef umap[StateSet, State] subset_map
 *         mata_nfa.determinize(result.thisptr.get(), dereference(lhs.thisptr.get()), &subset_map)
 *         return result, subset_map_to_dictionary(subset_map)             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7libmata_3nfa_3nfa_subset_map_to_dictionary(__pyx_v_subset_map); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __Pyx_GIVEREF((PyObject *)__pyx_v_result);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_result));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":641
 * 
 *     # Operations
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def determinize_with_subset_map(cls, Nfa lhs):
 *         """Determinize the lhs automaton
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.determinize_with_subset_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":653
 *         return result, subset_map_to_dictionary(subset_map)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def determinize(cls, Nfa lhs):
 *         """Determinize the lhs automaton
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_92determinize(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_91determinize, "Determinize the lhs automaton\n\n        :param Nfa lhs: non-deterministic finite automaton\n        :return: deterministic finite automaton\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_92determinize = {"determinize", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_92determinize, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_91determinize};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_92determinize(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("determinize (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 653, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "determinize") < 0)) __PYX_ERR(1, 653, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("determinize", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 653, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.determinize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 654, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_91determinize(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_91determinize(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("determinize", 0);

  /* "libmata/nfa/nfa.pyx":660
 *         :return: deterministic finite automaton
 *         """
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         mata_nfa.determinize(result.thisptr.get(), dereference(lhs.thisptr.get()), NULL)
 *         return result
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":661
 *         """
 *         result = Nfa()
 *         mata_nfa.determinize(result.thisptr.get(), dereference(lhs.thisptr.get()), NULL)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  Mata::Nfa::Plumbing::determinize(__pyx_v_result->thisptr.get(), (*__pyx_v_lhs->thisptr.get()), NULL);

  /* "libmata/nfa/nfa.pyx":662
 *         result = Nfa()
 *         mata_nfa.determinize(result.thisptr.get(), dereference(lhs.thisptr.get()), NULL)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":653
 *         return result, subset_map_to_dictionary(subset_map)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def determinize(cls, Nfa lhs):
 *         """Determinize the lhs automaton
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.determinize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":664
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def union(cls, Nfa lhs, Nfa rhs):
 *         """Performs union of lhs and rhs
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_94union(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_93union, "Performs union of lhs and rhs\n\n        :param Nfa lhs: first automaton\n        :param Nfa rhs: second automaton\n        :return: union of lhs and rhs\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_94union = {"union", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_94union, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_93union};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_94union(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("union (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_rhs,0};
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 664, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 664, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("union", 1, 2, 2, 1); __PYX_ERR(1, 664, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "union") < 0)) __PYX_ERR(1, 664, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_rhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("union", 1, 2, 2, __pyx_nargs); __PYX_ERR(1, 664, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.union", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 665, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "rhs", 0))) __PYX_ERR(1, 665, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_93union(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_rhs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_93union(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("union", 0);

  /* "libmata/nfa/nfa.pyx":672
 *         :return: union of lhs and rhs
 *         """
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         mata_nfa.uni(
 *             result.thisptr.get(), dereference(lhs.thisptr.get()), dereference(rhs.thisptr.get())
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":673
 *         """
 *         result = Nfa()
 *         mata_nfa.uni(             # <<<<<<<<<<<<<<
 *             result.thisptr.get(), dereference(lhs.thisptr.get()), dereference(rhs.thisptr.get())
 *         )
 */
  Mata::Nfa::Plumbing::uni(__pyx_v_result->thisptr.get(), (*__pyx_v_lhs->thisptr.get()), (*__pyx_v_rhs->thisptr.get()));

  /* "libmata/nfa/nfa.pyx":676
 *             result.thisptr.get(), dereference(lhs.thisptr.get()), dereference(rhs.thisptr.get())
 *         )
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":664
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def union(cls, Nfa lhs, Nfa rhs):
 *         """Performs union of lhs and rhs
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.union", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":678
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def intersection(cls, Nfa lhs, Nfa rhs, preserve_epsilon: bool = False):
 *         """Performs intersection of lhs and rhs.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_96intersection(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_95intersection, "Performs intersection of lhs and rhs.\n\n        Supports epsilon symbols when preserve_epsilon is set to True.\n\n        When computing intersection preserving epsilon transitions, create product of two NFAs, where both automata can contain \316\265-transitions. The product preserves the \316\265-transitions\n         of both automata. This means that for each \316\265-transition of the form `s-\316\265->p` and each product state `(s,a)`,\n         an \316\265-transition `(s,a)-\316\265->(p,a)` is created. Furthermore, for each \316\265-transition `s-\316\265->p` and `a-\316\265->b`,\n         a product state `(s,a)-\316\265->(p,b)` is created.\n\n        Automata must share alphabets.\n\n        :param preserve_epsilon: Whether to compute intersection preserving epsilon transitions.\n        :param Nfa lhs: First automaton.\n        :param Nfa rhs: Second automaton.\n        :return: Intersection of lhs and rhs.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_96intersection = {"intersection", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_96intersection, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_95intersection};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_96intersection(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs = 0;
  bool __pyx_v_preserve_epsilon;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("intersection (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_rhs,&__pyx_n_s_preserve_epsilon,0};
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 678, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 678, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("intersection", 0, 2, 3, 1); __PYX_ERR(1, 678, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_preserve_epsilon);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 678, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "intersection") < 0)) __PYX_ERR(1, 678, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_rhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[1]);
    if (values[2]) {
      __pyx_v_preserve_epsilon = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_preserve_epsilon == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(1, 679, __pyx_L3_error)
    } else {

      /* "libmata/nfa/nfa.pyx":679
 * 
 *     @classmethod
 *     def intersection(cls, Nfa lhs, Nfa rhs, preserve_epsilon: bool = False):             # <<<<<<<<<<<<<<
 *         """Performs intersection of lhs and rhs.
 * 
 */
      __pyx_v_preserve_epsilon = ((bool)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("intersection", 0, 2, 3, __pyx_nargs); __PYX_ERR(1, 678, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.intersection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 679, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "rhs", 0))) __PYX_ERR(1, 679, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_95intersection(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_rhs, __pyx_v_preserve_epsilon);

  /* "libmata/nfa/nfa.pyx":678
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def intersection(cls, Nfa lhs, Nfa rhs, preserve_epsilon: bool = False):
 *         """Performs intersection of lhs and rhs.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_95intersection(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, bool __pyx_v_preserve_epsilon) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("intersection", 0);

  /* "libmata/nfa/nfa.pyx":696
 *         :return: Intersection of lhs and rhs.
 *         """
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         mata_nfa.intersection(
 *             result.thisptr.get(), dereference(lhs.thisptr.get()), dereference(rhs.thisptr.get()), preserve_epsilon, NULL
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":697
 *         """
 *         result = Nfa()
 *         mata_nfa.intersection(             # <<<<<<<<<<<<<<
 *             result.thisptr.get(), dereference(lhs.thisptr.get()), dereference(rhs.thisptr.get()), preserve_epsilon, NULL
 *         )
 */
  Mata::Nfa::Plumbing::intersection(__pyx_v_result->thisptr.get(), (*__pyx_v_lhs->thisptr.get()), (*__pyx_v_rhs->thisptr.get()), __pyx_v_preserve_epsilon, NULL);

  /* "libmata/nfa/nfa.pyx":700
 *             result.thisptr.get(), dereference(lhs.thisptr.get()), dereference(rhs.thisptr.get()), preserve_epsilon, NULL
 *         )
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":678
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def intersection(cls, Nfa lhs, Nfa rhs, preserve_epsilon: bool = False):
 *         """Performs intersection of lhs and rhs.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.intersection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":702
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def intersection_with_product_map(cls, Nfa lhs, Nfa rhs, preserve_epsilon: bool = False):
 *         """Performs intersection of lhs and rhs.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_98intersection_with_product_map(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_97intersection_with_product_map, "Performs intersection of lhs and rhs.\n\n        Supports epsilon symbols when preserve_epsilon is set to True.\n\n        When computing intersection preserving epsilon transitions, create product of two NFAs, where both automata can contain \316\265-transitions. The product preserves the \316\265-transitions\n         of both automata. This means that for each \316\265-transition of the form `s-\316\265->p` and each product state `(s,a)`,\n         an \316\265-transition `(s,a)-\316\265->(p,a)` is created. Furthermore, for each \316\265-transition `s-\316\265->p` and `a-\316\265->b`,\n         a product state `(s,a)-\316\265->(p,b)` is created.\n\n        Automata must share alphabets.\n\n        :param preserve_epsilon: Whether to compute intersection preserving epsilon transitions.\n        :param Nfa lhs: First automaton.\n        :param Nfa rhs: Second automaton.\n        :return: Intersection of lhs and rhs, product map of original pairs of states to new states.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_98intersection_with_product_map = {"intersection_with_product_map", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_98intersection_with_product_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_97intersection_with_product_map};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_98intersection_with_product_map(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs = 0;
  bool __pyx_v_preserve_epsilon;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("intersection_with_product_map (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_rhs,&__pyx_n_s_preserve_epsilon,0};
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 702, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 702, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("intersection_with_product_map", 0, 2, 3, 1); __PYX_ERR(1, 702, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_preserve_epsilon);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 702, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "intersection_with_product_map") < 0)) __PYX_ERR(1, 702, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_rhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[1]);
    if (values[2]) {
      __pyx_v_preserve_epsilon = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_preserve_epsilon == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(1, 703, __pyx_L3_error)
    } else {

      /* "libmata/nfa/nfa.pyx":703
 * 
 *     @classmethod
 *     def intersection_with_product_map(cls, Nfa lhs, Nfa rhs, preserve_epsilon: bool = False):             # <<<<<<<<<<<<<<
 *         """Performs intersection of lhs and rhs.
 * 
 */
      __pyx_v_preserve_epsilon = ((bool)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("intersection_with_product_map", 0, 2, 3, __pyx_nargs); __PYX_ERR(1, 702, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.intersection_with_product_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 703, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "rhs", 0))) __PYX_ERR(1, 703, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_97intersection_with_product_map(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_rhs, __pyx_v_preserve_epsilon);

  /* "libmata/nfa/nfa.pyx":702
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def intersection_with_product_map(cls, Nfa lhs, Nfa rhs, preserve_epsilon: bool = False):
 *         """Performs intersection of lhs and rhs.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_97intersection_with_product_map(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, bool __pyx_v_preserve_epsilon) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  std::unordered_map<std::pair<Mata::Nfa::State,Mata::Nfa::State> ,Mata::Nfa::State>  __pyx_v_c_product_map;
  PyObject *__pyx_9genexpr11__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr11__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::unordered_map<std::pair<Mata::Nfa::State,Mata::Nfa::State> ,Mata::Nfa::State> ::iterator __pyx_t_2;
  std::pair<std::pair<Mata::Nfa::State,Mata::Nfa::State> ,Mata::Nfa::State>  __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("intersection_with_product_map", 0);

  /* "libmata/nfa/nfa.pyx":720
 *         :return: Intersection of lhs and rhs, product map of original pairs of states to new states.
 *         """
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         cdef umap[pair[State, State], State] c_product_map
 *         mata_nfa.intersection(
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":722
 *         result = Nfa()
 *         cdef umap[pair[State, State], State] c_product_map
 *         mata_nfa.intersection(             # <<<<<<<<<<<<<<
 *             result.thisptr.get(), dereference(lhs.thisptr.get()), dereference(rhs.thisptr.get()), preserve_epsilon,
 *             &c_product_map
 */
  Mata::Nfa::Plumbing::intersection(__pyx_v_result->thisptr.get(), (*__pyx_v_lhs->thisptr.get()), (*__pyx_v_rhs->thisptr.get()), __pyx_v_preserve_epsilon, (&__pyx_v_c_product_map));

  /* "libmata/nfa/nfa.pyx":726
 *             &c_product_map
 *         )
 *         return result, {tuple(k): v for k, v in c_product_map}             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 726, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_c_product_map.begin();
    for (;;) {
      if (!(__pyx_t_2 != __pyx_v_c_product_map.end())) break;
      __pyx_t_3 = *__pyx_t_2;
      ++__pyx_t_2;
      __pyx_t_4 = __pyx_convert_pair_to_py_std_3a__3a_pair_3c_Mata_3a__3a_Nfa_3a__3a_State_2c_Mata_3a__3a_Nfa_3a__3a_State_3e_______Mata_3a__3a_Nfa_3a__3a_State(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 726, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(1, 726, __pyx_L5_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 726, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 726, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 726, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
        index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(1, 726, __pyx_L5_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L9_unpacking_done;
        __pyx_L8_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(1, 726, __pyx_L5_error)
        __pyx_L9_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr11__pyx_v_k, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr11__pyx_v_v, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_9genexpr11__pyx_v_k); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 726, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_4, (PyObject*)__pyx_9genexpr11__pyx_v_v))) __PYX_ERR(1, 726, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_k); __pyx_9genexpr11__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_v); __pyx_9genexpr11__pyx_v_v = 0;
    goto __pyx_L11_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_k); __pyx_9genexpr11__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_v); __pyx_9genexpr11__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L11_exit_scope:;
  } /* exit inner scope */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __Pyx_GIVEREF((PyObject *)__pyx_v_result);
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_result));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":702
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def intersection_with_product_map(cls, Nfa lhs, Nfa rhs, preserve_epsilon: bool = False):
 *         """Performs intersection of lhs and rhs.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.intersection_with_product_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":728
 *         return result, {tuple(k): v for k, v in c_product_map}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def concatenate(cls, Nfa lhs, Nfa rhs, use_epsilon: bool = False) -> Nfa:
 *         """Concatenate two NFAs.
 */

/* Python wrapper */
static struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_100concatenate(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_99concatenate, "Concatenate two NFAs.\n\n        Supports epsilon symbols when @p use_epsilon is set to true.\n        :param Nfa lhs: First automaton to concatenate.\n        :param Nfa rhs: Second automaton to concatenate.\n        :param use_epsilon: Whether to concatenate over an epsilon symbol.\n        :return: Nfa: Concatenated automaton.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_100concatenate = {"concatenate", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_100concatenate, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_99concatenate};
static struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_100concatenate(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs = 0;
  bool __pyx_v_use_epsilon;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("concatenate (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_rhs,&__pyx_n_s_use_epsilon,0};
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 728, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 728, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("concatenate", 0, 2, 3, 1); __PYX_ERR(1, 728, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_use_epsilon);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 728, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "concatenate") < 0)) __PYX_ERR(1, 728, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_rhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[1]);
    if (values[2]) {
      __pyx_v_use_epsilon = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_use_epsilon == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(1, 729, __pyx_L3_error)
    } else {

      /* "libmata/nfa/nfa.pyx":729
 * 
 *     @classmethod
 *     def concatenate(cls, Nfa lhs, Nfa rhs, use_epsilon: bool = False) -> Nfa:             # <<<<<<<<<<<<<<
 *         """Concatenate two NFAs.
 * 
 */
      __pyx_v_use_epsilon = ((bool)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("concatenate", 0, 2, 3, __pyx_nargs); __PYX_ERR(1, 728, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.concatenate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 729, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "rhs", 0))) __PYX_ERR(1, 729, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_99concatenate(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_rhs, __pyx_v_use_epsilon);

  /* "libmata/nfa/nfa.pyx":728
 *         return result, {tuple(k): v for k, v in c_product_map}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def concatenate(cls, Nfa lhs, Nfa rhs, use_epsilon: bool = False) -> Nfa:
 *         """Concatenate two NFAs.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_99concatenate(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, bool __pyx_v_use_epsilon) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("concatenate", 0);

  /* "libmata/nfa/nfa.pyx":738
 *         :return: Nfa: Concatenated automaton.
 *         """
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         mata_nfa.concatenate(result.thisptr.get(), dereference(lhs.thisptr.get()), dereference(rhs.thisptr.get()),
 *                          use_epsilon, NULL, NULL)
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":739
 *         """
 *         result = Nfa()
 *         mata_nfa.concatenate(result.thisptr.get(), dereference(lhs.thisptr.get()), dereference(rhs.thisptr.get()),             # <<<<<<<<<<<<<<
 *                          use_epsilon, NULL, NULL)
 *         return result
 */
  Mata::Nfa::Plumbing::concatenate(__pyx_v_result->thisptr.get(), (*__pyx_v_lhs->thisptr.get()), (*__pyx_v_rhs->thisptr.get()), __pyx_v_use_epsilon, NULL, NULL);

  /* "libmata/nfa/nfa.pyx":741
 *         mata_nfa.concatenate(result.thisptr.get(), dereference(lhs.thisptr.get()), dereference(rhs.thisptr.get()),
 *                          use_epsilon, NULL, NULL)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":728
 *         return result, {tuple(k): v for k, v in c_product_map}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def concatenate(cls, Nfa lhs, Nfa rhs, use_epsilon: bool = False) -> Nfa:
 *         """Concatenate two NFAs.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.concatenate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":743
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def concatenate_with_result_state_maps(cls, Nfa lhs, Nfa rhs, use_epsilon: bool = False) \
 *             -> tuple[Nfa, dict[str, str], dict[str, str]]:
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_102concatenate_with_result_state_maps(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_101concatenate_with_result_state_maps, "Concatenate two NFAs.\n\n        :param Nfa lhs: First automaton to concatenate.\n        :param Nfa rhs: Second automaton to concatenate.\n        :param use_epsilon: Whether to concatenate over an epsilon symbol.\n        :return: Nfa: Concatenated automaton.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_102concatenate_with_result_state_maps = {"concatenate_with_result_state_maps", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_102concatenate_with_result_state_maps, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_101concatenate_with_result_state_maps};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_102concatenate_with_result_state_maps(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs = 0;
  bool __pyx_v_use_epsilon;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("concatenate_with_result_state_maps (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_rhs,&__pyx_n_s_use_epsilon,0};
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 743, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 743, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("concatenate_with_result_state_maps", 0, 2, 3, 1); __PYX_ERR(1, 743, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_use_epsilon);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 743, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "concatenate_with_result_state_maps") < 0)) __PYX_ERR(1, 743, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_rhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[1]);
    if (values[2]) {
      __pyx_v_use_epsilon = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_use_epsilon == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(1, 744, __pyx_L3_error)
    } else {

      /* "libmata/nfa/nfa.pyx":744
 * 
 *     @classmethod
 *     def concatenate_with_result_state_maps(cls, Nfa lhs, Nfa rhs, use_epsilon: bool = False) \             # <<<<<<<<<<<<<<
 *             -> tuple[Nfa, dict[str, str], dict[str, str]]:
 *         """Concatenate two NFAs.
 */
      __pyx_v_use_epsilon = ((bool)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("concatenate_with_result_state_maps", 0, 2, 3, __pyx_nargs); __PYX_ERR(1, 743, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.concatenate_with_result_state_maps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 744, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "rhs", 0))) __PYX_ERR(1, 744, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_101concatenate_with_result_state_maps(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_rhs, __pyx_v_use_epsilon);

  /* "libmata/nfa/nfa.pyx":743
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def concatenate_with_result_state_maps(cls, Nfa lhs, Nfa rhs, use_epsilon: bool = False) \
 *             -> tuple[Nfa, dict[str, str], dict[str, str]]:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_101concatenate_with_result_state_maps(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, bool __pyx_v_use_epsilon) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  Mata::Nfa::StateToStateMap __pyx_v_c_lhs_map;
  Mata::Nfa::StateToStateMap __pyx_v_c_rhs_map;
  PyObject *__pyx_v_lhs_map = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_rhs_map = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::unordered_map<Mata::Nfa::State,Mata::Nfa::State> ::iterator __pyx_t_2;
  std::pair<Mata::Nfa::State,Mata::Nfa::State>  __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("concatenate_with_result_state_maps", 0);

  /* "libmata/nfa/nfa.pyx":753
 *         :return: Nfa: Concatenated automaton.
 *         """
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         cdef StateToStateMap c_lhs_map
 *         cdef StateToStateMap c_rhs_map
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":756
 *         cdef StateToStateMap c_lhs_map
 *         cdef StateToStateMap c_rhs_map
 *         mata_nfa.concatenate(result.thisptr.get(), dereference(lhs.thisptr.get()), dereference(rhs.thisptr.get()),             # <<<<<<<<<<<<<<
 *                          use_epsilon, &c_lhs_map, &c_rhs_map)
 *         lhs_map = {}
 */
  Mata::Nfa::Plumbing::concatenate(__pyx_v_result->thisptr.get(), (*__pyx_v_lhs->thisptr.get()), (*__pyx_v_rhs->thisptr.get()), __pyx_v_use_epsilon, (&__pyx_v_c_lhs_map), (&__pyx_v_c_rhs_map));

  /* "libmata/nfa/nfa.pyx":758
 *         mata_nfa.concatenate(result.thisptr.get(), dereference(lhs.thisptr.get()), dereference(rhs.thisptr.get()),
 *                          use_epsilon, &c_lhs_map, &c_rhs_map)
 *         lhs_map = {}             # <<<<<<<<<<<<<<
 *         for key, value in c_lhs_map:
 *             lhs_map[key] = value
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_lhs_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":759
 *                          use_epsilon, &c_lhs_map, &c_rhs_map)
 *         lhs_map = {}
 *         for key, value in c_lhs_map:             # <<<<<<<<<<<<<<
 *             lhs_map[key] = value
 *         rhs_map = {}
 */
  __pyx_t_2 = __pyx_v_c_lhs_map.begin();
  for (;;) {
    if (!(__pyx_t_2 != __pyx_v_c_lhs_map.end())) break;
    __pyx_t_3 = *__pyx_t_2;
    ++__pyx_t_2;
    __pyx_t_1 = __pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 759, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 759, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 759, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 759, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 759, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
      index = 0; __pyx_t_4 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(1, 759, __pyx_L1_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 759, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "libmata/nfa/nfa.pyx":760
 *         lhs_map = {}
 *         for key, value in c_lhs_map:
 *             lhs_map[key] = value             # <<<<<<<<<<<<<<
 *         rhs_map = {}
 *         for key, value in c_rhs_map:
 */
    if (unlikely((PyDict_SetItem(__pyx_v_lhs_map, __pyx_v_key, __pyx_v_value) < 0))) __PYX_ERR(1, 760, __pyx_L1_error)

    /* "libmata/nfa/nfa.pyx":759
 *                          use_epsilon, &c_lhs_map, &c_rhs_map)
 *         lhs_map = {}
 *         for key, value in c_lhs_map:             # <<<<<<<<<<<<<<
 *             lhs_map[key] = value
 *         rhs_map = {}
 */
  }

  /* "libmata/nfa/nfa.pyx":761
 *         for key, value in c_lhs_map:
 *             lhs_map[key] = value
 *         rhs_map = {}             # <<<<<<<<<<<<<<
 *         for key, value in c_rhs_map:
 *             rhs_map[key] = value
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rhs_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":762
 *             lhs_map[key] = value
 *         rhs_map = {}
 *         for key, value in c_rhs_map:             # <<<<<<<<<<<<<<
 *             rhs_map[key] = value
 * 
 */
  __pyx_t_2 = __pyx_v_c_rhs_map.begin();
  for (;;) {
    if (!(__pyx_t_2 != __pyx_v_c_rhs_map.end())) break;
    __pyx_t_3 = *__pyx_t_2;
    ++__pyx_t_2;
    __pyx_t_1 = __pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 762, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
      index = 0; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_4 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(1, 762, __pyx_L1_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L11_unpacking_done;
      __pyx_L10_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 762, __pyx_L1_error)
      __pyx_L11_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "libmata/nfa/nfa.pyx":763
 *         rhs_map = {}
 *         for key, value in c_rhs_map:
 *             rhs_map[key] = value             # <<<<<<<<<<<<<<
 * 
 *         return result, lhs_map, rhs_map
 */
    if (unlikely((PyDict_SetItem(__pyx_v_rhs_map, __pyx_v_key, __pyx_v_value) < 0))) __PYX_ERR(1, 763, __pyx_L1_error)

    /* "libmata/nfa/nfa.pyx":762
 *             lhs_map[key] = value
 *         rhs_map = {}
 *         for key, value in c_rhs_map:             # <<<<<<<<<<<<<<
 *             rhs_map[key] = value
 * 
 */
  }

  /* "libmata/nfa/nfa.pyx":765
 *             rhs_map[key] = value
 * 
 *         return result, lhs_map, rhs_map             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __Pyx_GIVEREF((PyObject *)__pyx_v_result);
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_result));
  __Pyx_INCREF(__pyx_v_lhs_map);
  __Pyx_GIVEREF(__pyx_v_lhs_map);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_lhs_map);
  __Pyx_INCREF(__pyx_v_rhs_map);
  __Pyx_GIVEREF(__pyx_v_rhs_map);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_rhs_map);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":743
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def concatenate_with_result_state_maps(cls, Nfa lhs, Nfa rhs, use_epsilon: bool = False) \
 *             -> tuple[Nfa, dict[str, str], dict[str, str]]:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.concatenate_with_result_state_maps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_lhs_map);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_rhs_map);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":768
 * 
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def complement(cls, Nfa lhs, alph.Alphabet alphabet, params = None):
 *         """Performs complement of lhs
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_104complement(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_103complement, "Performs complement of lhs\n\n        :param Nfa lhs: complemented automaton\n        :param OnTheFlyAlphabet alphabet: alphabet of the lhs\n        :param dict params: additional params\n          - \"algorithm\": \"classical\" (classical algorithm determinizes the automaton, makes it complete and swaps final and non-final states);\n          - \"minimize\": \"true\"/\"false\" (whether to compute minimal deterministic automaton for classical algorithm);\n        :return: complemented automaton\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_104complement = {"complement", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_104complement, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_103complement};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_104complement(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet = 0;
  PyObject *__pyx_v_params = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("complement (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_alphabet,&__pyx_n_s_params,0};
    PyObject* values[3] = {0,0,0};

    /* "libmata/nfa/nfa.pyx":769
 * 
 *     @classmethod
 *     def complement(cls, Nfa lhs, alph.Alphabet alphabet, params = None):             # <<<<<<<<<<<<<<
 *         """Performs complement of lhs
 * 
 */
    values[2] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 768, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_alphabet)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 768, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("complement", 0, 2, 3, 1); __PYX_ERR(1, 768, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_params);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 768, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "complement") < 0)) __PYX_ERR(1, 768, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_alphabet = ((struct __pyx_obj_7libmata_9alphabets_Alphabet *)values[1]);
    __pyx_v_params = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("complement", 0, 2, 3, __pyx_nargs); __PYX_ERR(1, 768, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.complement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 769, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_alphabet), __pyx_ptype_7libmata_9alphabets_Alphabet, 1, "alphabet", 0))) __PYX_ERR(1, 769, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_103complement(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_alphabet, __pyx_v_params);

  /* "libmata/nfa/nfa.pyx":768
 * 
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def complement(cls, Nfa lhs, alph.Alphabet alphabet, params = None):
 *         """Performs complement of lhs
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_103complement(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet, PyObject *__pyx_v_params) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  PyObject *__pyx_9genexpr12__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr12__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Mata::Alphabet *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  Mata::Nfa::StringMap __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("complement", 0);
  __Pyx_INCREF(__pyx_v_params);

  /* "libmata/nfa/nfa.pyx":779
 *         :return: complemented automaton
 *         """
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         params = params or {'algorithm': 'classical', 'minimize': 'false'}
 *         mata_nfa.complement(
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":780
 *         """
 *         result = Nfa()
 *         params = params or {'algorithm': 'classical', 'minimize': 'false'}             # <<<<<<<<<<<<<<
 *         mata_nfa.complement(
 *             result.thisptr.get(),
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_params); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 780, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_params);
    __pyx_t_1 = __pyx_v_params;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_algorithm, __pyx_n_u_classical) < 0) __PYX_ERR(1, 780, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_minimize, __pyx_n_u_false) < 0) __PYX_ERR(1, 780, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF_SET(__pyx_v_params, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":784
 *             result.thisptr.get(),
 *             dereference(lhs.thisptr.get()),
 *             <CAlphabet&>dereference(alphabet.as_base()),             # <<<<<<<<<<<<<<
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_7libmata_9alphabets_Alphabet *)__pyx_v_alphabet->__pyx_vtab)->as_base(__pyx_v_alphabet); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 784, __pyx_L1_error)
  { /* enter inner scope */

    /* "libmata/nfa/nfa.pyx":785
 *             dereference(lhs.thisptr.get()),
 *             <CAlphabet&>dereference(alphabet.as_base()),
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()
 */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 785, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "libmata/nfa/nfa.pyx":787
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()             # <<<<<<<<<<<<<<
 *             },
 *         )
 */
    __pyx_t_5 = 0;
    if (unlikely(__pyx_v_params == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(1, 787, __pyx_L7_error)
    }
    __pyx_t_8 = __Pyx_dict_iterator(__pyx_v_params, 0, __pyx_n_s_items, (&__pyx_t_6), (&__pyx_t_7)); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 787, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_8;
    __pyx_t_8 = 0;
    while (1) {
      __pyx_t_10 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_6, &__pyx_t_5, &__pyx_t_8, &__pyx_t_9, NULL, __pyx_t_7);
      if (unlikely(__pyx_t_10 == 0)) break;
      if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(1, 787, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_k, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_v, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "libmata/nfa/nfa.pyx":786
 *             <CAlphabet&>dereference(alphabet.as_base()),
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v             # <<<<<<<<<<<<<<
 *                 for k, v in params.items()
 *             },
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr12__pyx_v_k, __pyx_n_s_encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 786, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_11 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_10 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_utf_8};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 786, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __pyx_t_2 = PyUnicode_Check(__pyx_9genexpr12__pyx_v_v); 
      if (__pyx_t_2) {
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr12__pyx_v_v, __pyx_n_s_encode); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 786, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_10 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_kp_u_utf_8};
          __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 786, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __pyx_t_8 = __pyx_t_11;
        __pyx_t_11 = 0;
      } else {
        __Pyx_INCREF(__pyx_9genexpr12__pyx_v_v);
        __pyx_t_8 = __pyx_9genexpr12__pyx_v_v;
      }
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_9, (PyObject*)__pyx_t_8))) __PYX_ERR(1, 786, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_k); __pyx_9genexpr12__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_v); __pyx_9genexpr12__pyx_v_v = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_k); __pyx_9genexpr12__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_v); __pyx_9genexpr12__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */

  /* "libmata/nfa/nfa.pyx":785
 *             dereference(lhs.thisptr.get()),
 *             <CAlphabet&>dereference(alphabet.as_base()),
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()
 */
  __pyx_t_14 = __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 785, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":781
 *         result = Nfa()
 *         params = params or {'algorithm': 'classical', 'minimize': 'false'}
 *         mata_nfa.complement(             # <<<<<<<<<<<<<<
 *             result.thisptr.get(),
 *             dereference(lhs.thisptr.get()),
 */
  try {
    Mata::Nfa::Plumbing::complement(__pyx_v_result->thisptr.get(), (*__pyx_v_lhs->thisptr.get()), ((Mata::Alphabet &)(*__pyx_t_4)), __pyx_t_14);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 781, __pyx_L1_error)
  }

  /* "libmata/nfa/nfa.pyx":790
 *             },
 *         )
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":768
 * 
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def complement(cls, Nfa lhs, alph.Alphabet alphabet, params = None):
 *         """Performs complement of lhs
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.complement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":792
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def make_complete(cls, Nfa lhs, State sink_state, alph.Alphabet alphabet):
 *         """Makes lhs complete
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_106make_complete(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_105make_complete, "Makes lhs complete\n\n        :param Nfa lhs: automaton that will be made complete\n        :param Symbol sink_state: sink state of the automaton\n        :param OnTheFlyAlphabet alphabet: alphabet of the\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_106make_complete = {"make_complete", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_106make_complete, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_105make_complete};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_106make_complete(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  Mata::Nfa::State __pyx_v_sink_state;
  struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_complete (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_sink_state,&__pyx_n_s_alphabet,0};
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 792, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sink_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 792, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("make_complete", 1, 3, 3, 1); __PYX_ERR(1, 792, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_alphabet)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 792, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("make_complete", 1, 3, 3, 2); __PYX_ERR(1, 792, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "make_complete") < 0)) __PYX_ERR(1, 792, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_sink_state = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_sink_state == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 793, __pyx_L3_error)
    __pyx_v_alphabet = ((struct __pyx_obj_7libmata_9alphabets_Alphabet *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("make_complete", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 792, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.make_complete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 793, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_alphabet), __pyx_ptype_7libmata_9alphabets_Alphabet, 1, "alphabet", 0))) __PYX_ERR(1, 793, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_105make_complete(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_sink_state, __pyx_v_alphabet);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_105make_complete(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, Mata::Nfa::State __pyx_v_sink_state, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Mata::Nfa::State __pyx_t_6;
  Mata::Alphabet *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("make_complete", 0);

  /* "libmata/nfa/nfa.pyx":800
 *         :param OnTheFlyAlphabet alphabet: alphabet of the
 *         """
 *         if not lhs.thisptr.get().is_state(sink_state):             # <<<<<<<<<<<<<<
 *             lhs.thisptr.get().add_state(lhs.size())
 *         mata_nfa.make_complete(lhs.thisptr.get(), <CAlphabet&>dereference(alphabet.as_base()), sink_state)
 */
  __pyx_t_1 = (!(__pyx_v_lhs->thisptr.get()->is_state(__pyx_v_sink_state) != 0));
  if (__pyx_t_1) {

    /* "libmata/nfa/nfa.pyx":801
 *         """
 *         if not lhs.thisptr.get().is_state(sink_state):
 *             lhs.thisptr.get().add_state(lhs.size())             # <<<<<<<<<<<<<<
 *         mata_nfa.make_complete(lhs.thisptr.get(), <CAlphabet&>dereference(alphabet.as_base()), sink_state)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_lhs), __pyx_n_s_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 801, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[1] = {__pyx_t_4, };
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 801, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_6 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_6 == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 801, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (void)(__pyx_v_lhs->thisptr.get()->add_state(__pyx_t_6));

    /* "libmata/nfa/nfa.pyx":800
 *         :param OnTheFlyAlphabet alphabet: alphabet of the
 *         """
 *         if not lhs.thisptr.get().is_state(sink_state):             # <<<<<<<<<<<<<<
 *             lhs.thisptr.get().add_state(lhs.size())
 *         mata_nfa.make_complete(lhs.thisptr.get(), <CAlphabet&>dereference(alphabet.as_base()), sink_state)
 */
  }

  /* "libmata/nfa/nfa.pyx":802
 *         if not lhs.thisptr.get().is_state(sink_state):
 *             lhs.thisptr.get().add_state(lhs.size())
 *         mata_nfa.make_complete(lhs.thisptr.get(), <CAlphabet&>dereference(alphabet.as_base()), sink_state)             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_7libmata_9alphabets_Alphabet *)__pyx_v_alphabet->__pyx_vtab)->as_base(__pyx_v_alphabet); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 802, __pyx_L1_error)
  try {
    Mata::Nfa::Plumbing::make_complete(__pyx_v_lhs->thisptr.get(), ((Mata::Alphabet &)(*__pyx_t_7)), __pyx_v_sink_state);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 802, __pyx_L1_error)
  }

  /* "libmata/nfa/nfa.pyx":792
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def make_complete(cls, Nfa lhs, State sink_state, alph.Alphabet alphabet):
 *         """Makes lhs complete
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.make_complete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":804
 *         mata_nfa.make_complete(lhs.thisptr.get(), <CAlphabet&>dereference(alphabet.as_base()), sink_state)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def revert(cls, Nfa lhs):
 *         """Reverses transitions in the lhs
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_108revert(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_107revert, "Reverses transitions in the lhs\n\n        :param Nfa lhs: source automaton\n        :return: automaton with reversed transitions\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_108revert = {"revert", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_108revert, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_107revert};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_108revert(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("revert (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 804, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "revert") < 0)) __PYX_ERR(1, 804, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("revert", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 804, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.revert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 805, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_107revert(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_107revert(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("revert", 0);

  /* "libmata/nfa/nfa.pyx":811
 *         :return: automaton with reversed transitions
 *         """
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         mata_nfa.revert(result.thisptr.get(), dereference(lhs.thisptr.get()))
 *         return result
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":812
 *         """
 *         result = Nfa()
 *         mata_nfa.revert(result.thisptr.get(), dereference(lhs.thisptr.get()))             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  Mata::Nfa::Plumbing::revert(__pyx_v_result->thisptr.get(), (*__pyx_v_lhs->thisptr.get()));

  /* "libmata/nfa/nfa.pyx":813
 *         result = Nfa()
 *         mata_nfa.revert(result.thisptr.get(), dereference(lhs.thisptr.get()))
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":804
 *         mata_nfa.make_complete(lhs.thisptr.get(), <CAlphabet&>dereference(alphabet.as_base()), sink_state)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def revert(cls, Nfa lhs):
 *         """Reverses transitions in the lhs
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.revert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":815
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def remove_epsilon(cls, Nfa lhs, Symbol epsilon = CEPSILON):
 *         """Removes transitions that contain epsilon symbol.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_110remove_epsilon(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_109remove_epsilon, "Removes transitions that contain epsilon symbol.\n\n        :param Nfa lhs: Automaton, where epsilon transitions will be removed.\n        :param Symbol epsilon: Symbol representing the epsilon.\n        :return: Nfa: Automaton with epsilon transitions removed.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_110remove_epsilon = {"remove_epsilon", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_110remove_epsilon, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_109remove_epsilon};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_110remove_epsilon(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  Mata::Symbol __pyx_v_epsilon;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_epsilon (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_epsilon,0};
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 815, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_epsilon);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 815, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "remove_epsilon") < 0)) __PYX_ERR(1, 815, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    if (values[1]) {
      __pyx_v_epsilon = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_epsilon == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 816, __pyx_L3_error)
    } else {
      __pyx_v_epsilon = __pyx_k__11;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove_epsilon", 0, 1, 2, __pyx_nargs); __PYX_ERR(1, 815, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.remove_epsilon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 816, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_109remove_epsilon(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_epsilon);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_109remove_epsilon(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, Mata::Symbol __pyx_v_epsilon) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove_epsilon", 0);

  /* "libmata/nfa/nfa.pyx":823
 *         :return: Nfa: Automaton with epsilon transitions removed.
 *         """
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         mata_nfa.remove_epsilon(result.thisptr.get(), dereference(lhs.thisptr.get()), epsilon)
 *         return result
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":824
 *         """
 *         result = Nfa()
 *         mata_nfa.remove_epsilon(result.thisptr.get(), dereference(lhs.thisptr.get()), epsilon)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  try {
    Mata::Nfa::Plumbing::remove_epsilon(__pyx_v_result->thisptr.get(), (*__pyx_v_lhs->thisptr.get()), __pyx_v_epsilon);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 824, __pyx_L1_error)
  }

  /* "libmata/nfa/nfa.pyx":825
 *         result = Nfa()
 *         mata_nfa.remove_epsilon(result.thisptr.get(), dereference(lhs.thisptr.get()), epsilon)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def remove_epsilon_inplace(self, Symbol epsilon = CEPSILON):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":815
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def remove_epsilon(cls, Nfa lhs, Symbol epsilon = CEPSILON):
 *         """Removes transitions that contain epsilon symbol.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.remove_epsilon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":827
 *         return result
 * 
 *     def remove_epsilon_inplace(self, Symbol epsilon = CEPSILON):             # <<<<<<<<<<<<<<
 *         """Removes transitions which contain epsilon symbol.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_112remove_epsilon_inplace(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_111remove_epsilon_inplace, "Removes transitions which contain epsilon symbol.\n\n        TODO: Possibly there may be issue with setting the size of the automaton beforehand?\n\n        :param Symbol epsilon: Symbol representing the epsilon symbol.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_112remove_epsilon_inplace = {"remove_epsilon_inplace", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_112remove_epsilon_inplace, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_111remove_epsilon_inplace};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_112remove_epsilon_inplace(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Symbol __pyx_v_epsilon;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_epsilon_inplace (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_epsilon,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_epsilon);
          if (value) { values[0] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 827, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "remove_epsilon_inplace") < 0)) __PYX_ERR(1, 827, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_epsilon = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_epsilon == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 827, __pyx_L3_error)
    } else {
      __pyx_v_epsilon = __pyx_k__12;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove_epsilon_inplace", 0, 0, 1, __pyx_nargs); __PYX_ERR(1, 827, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.remove_epsilon_inplace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_111remove_epsilon_inplace(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_epsilon);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_111remove_epsilon_inplace(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Symbol __pyx_v_epsilon) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove_epsilon_inplace", 0);

  /* "libmata/nfa/nfa.pyx":834
 *         :param Symbol epsilon: Symbol representing the epsilon symbol.
 *         """
 *         self.thisptr.get().remove_epsilon(epsilon)             # <<<<<<<<<<<<<<
 * 
 *     def get_epsilon_transitions(self, State state, Symbol epsilon = CEPSILON) -> Move | None:
 */
  try {
    __pyx_v_self->thisptr.get()->remove_epsilon(__pyx_v_epsilon);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 834, __pyx_L1_error)
  }

  /* "libmata/nfa/nfa.pyx":827
 *         return result
 * 
 *     def remove_epsilon_inplace(self, Symbol epsilon = CEPSILON):             # <<<<<<<<<<<<<<
 *         """Removes transitions which contain epsilon symbol.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.remove_epsilon_inplace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":836
 *         self.thisptr.get().remove_epsilon(epsilon)
 * 
 *     def get_epsilon_transitions(self, State state, Symbol epsilon = CEPSILON) -> Move | None:             # <<<<<<<<<<<<<<
 *         """Get epsilon transitions for a state.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_114get_epsilon_transitions(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_113get_epsilon_transitions, "Get epsilon transitions for a state.\n\n        :param state: State to get epsilon transitions for.\n        :param epsilon: Epsilon symbol.\n        :return: Epsilon transitions if there are any epsilon transitions for the passed state. None otherwise.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_114get_epsilon_transitions = {"get_epsilon_transitions", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_114get_epsilon_transitions, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_113get_epsilon_transitions};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_114get_epsilon_transitions(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  Mata::Nfa::State __pyx_v_state;
  Mata::Symbol __pyx_v_epsilon;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_epsilon_transitions (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_state,&__pyx_n_s_epsilon,0};
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 836, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_epsilon);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 836, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_epsilon_transitions") < 0)) __PYX_ERR(1, 836, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_state = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_state == ((Mata::Nfa::State)-1)) && PyErr_Occurred())) __PYX_ERR(1, 836, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_epsilon = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_epsilon == ((Mata::Symbol)-1)) && PyErr_Occurred())) __PYX_ERR(1, 836, __pyx_L3_error)
    } else {
      __pyx_v_epsilon = __pyx_k__13;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_epsilon_transitions", 0, 1, 2, __pyx_nargs); __PYX_ERR(1, 836, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_epsilon_transitions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_113get_epsilon_transitions(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v_state, __pyx_v_epsilon);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_113get_epsilon_transitions(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, Mata::Nfa::State __pyx_v_state, Mata::Symbol __pyx_v_epsilon) {
  Mata::Util::OrdVector<Mata::Nfa::Move> ::const_iterator __pyx_v_c_epsilon_transitions_iter;
  Mata::Nfa::Move __pyx_v_epsilon_transitions;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_epsilon_transitions", 0);

  /* "libmata/nfa/nfa.pyx":843
 *         :return: Epsilon transitions if there are any epsilon transitions for the passed state. None otherwise.
 *         """
 *         cdef COrdVector[CMove].const_iterator c_epsilon_transitions_iter = self.thisptr.get().get_epsilon_transitions(             # <<<<<<<<<<<<<<
 *             state, epsilon
 *         )
 */
  __pyx_v_c_epsilon_transitions_iter = __pyx_v_self->thisptr.get()->get_epsilon_transitions(__pyx_v_state, __pyx_v_epsilon);

  /* "libmata/nfa/nfa.pyx":846
 *             state, epsilon
 *         )
 *         if c_epsilon_transitions_iter == self.thisptr.get().get_moves_from(state).cend():             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
  __pyx_t_1 = (__pyx_v_c_epsilon_transitions_iter == __pyx_v_self->thisptr.get()->get_moves_from(__pyx_v_state).cend());
  if (__pyx_t_1) {

    /* "libmata/nfa/nfa.pyx":847
 *         )
 *         if c_epsilon_transitions_iter == self.thisptr.get().get_moves_from(state).cend():
 *             return None             # <<<<<<<<<<<<<<
 * 
 *         cdef CMove epsilon_transitions = dereference(c_epsilon_transitions_iter)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "libmata/nfa/nfa.pyx":846
 *             state, epsilon
 *         )
 *         if c_epsilon_transitions_iter == self.thisptr.get().get_moves_from(state).cend():             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
  }

  /* "libmata/nfa/nfa.pyx":849
 *             return None
 * 
 *         cdef CMove epsilon_transitions = dereference(c_epsilon_transitions_iter)             # <<<<<<<<<<<<<<
 *         return Move(epsilon_transitions.symbol, epsilon_transitions.targets.ToVector())
 * 
 */
  __pyx_v_epsilon_transitions = (*__pyx_v_c_epsilon_transitions_iter);

  /* "libmata/nfa/nfa.pyx":850
 * 
 *         cdef CMove epsilon_transitions = dereference(c_epsilon_transitions_iter)
 *         return Move(epsilon_transitions.symbol, epsilon_transitions.targets.ToVector())             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_epsilon_transitions.symbol); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State(__pyx_v_epsilon_transitions.targets.ToVector()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Move), __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":836
 *         self.thisptr.get().remove_epsilon(epsilon)
 * 
 *     def get_epsilon_transitions(self, State state, Symbol epsilon = CEPSILON) -> Move | None:             # <<<<<<<<<<<<<<
 *         """Get epsilon transitions for a state.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_epsilon_transitions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":853
 * 
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def minimize(cls, Nfa lhs):
 *         """Minimizes the automaton lhs
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_116minimize(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_115minimize, "Minimizes the automaton lhs\n\n        :param Nfa lhs: automaton to be minimized\n        :return: minimized automaton\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_116minimize = {"minimize", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_116minimize, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_115minimize};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_116minimize(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("minimize (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 853, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "minimize") < 0)) __PYX_ERR(1, 853, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("minimize", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 853, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.minimize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 854, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_115minimize(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_115minimize(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("minimize", 0);

  /* "libmata/nfa/nfa.pyx":860
 *         :return: minimized automaton
 *         """
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         mata_nfa.minimize(result.thisptr.get(), dereference(lhs.thisptr.get()))
 *         return result
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":861
 *         """
 *         result = Nfa()
 *         mata_nfa.minimize(result.thisptr.get(), dereference(lhs.thisptr.get()))             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  Mata::Nfa::Plumbing::minimize(__pyx_v_result->thisptr.get(), (*__pyx_v_lhs->thisptr.get()));

  /* "libmata/nfa/nfa.pyx":862
 *         result = Nfa()
 *         mata_nfa.minimize(result.thisptr.get(), dereference(lhs.thisptr.get()))
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":853
 * 
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def minimize(cls, Nfa lhs):
 *         """Minimizes the automaton lhs
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.minimize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":864
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def reduce_with_state_map(cls, Nfa aut, bool trim_input = True, params = None):
 *         """Reduce the automaton.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_118reduce_with_state_map(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_117reduce_with_state_map, "Reduce the automaton.\n\n        :param Nfa aut: Original automaton to reduce.\n        :param bool trim_input: Whether to trim the input automaton first or not.\n        :param Dict params: Additional parameters for the reduction algorithm:\n            - \"algorithm\": \"simulation\"\n        :return: (Reduced automaton, state map of original to new states)\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_118reduce_with_state_map = {"reduce_with_state_map", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_118reduce_with_state_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_117reduce_with_state_map};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_118reduce_with_state_map(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_aut = 0;
  bool __pyx_v_trim_input;
  PyObject *__pyx_v_params = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reduce_with_state_map (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_aut,&__pyx_n_s_trim_input,&__pyx_n_s_params,0};
    PyObject* values[3] = {0,0,0};

    /* "libmata/nfa/nfa.pyx":865
 * 
 *     @classmethod
 *     def reduce_with_state_map(cls, Nfa aut, bool trim_input = True, params = None):             # <<<<<<<<<<<<<<
 *         """Reduce the automaton.
 * 
 */
    values[2] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_aut)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 864, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_trim_input);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 864, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_params);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 864, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "reduce_with_state_map") < 0)) __PYX_ERR(1, 864, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_aut = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    if (values[1]) {
      __pyx_v_trim_input = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_trim_input == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(1, 865, __pyx_L3_error)
    } else {
      __pyx_v_trim_input = ((bool)1);
    }
    __pyx_v_params = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("reduce_with_state_map", 0, 1, 3, __pyx_nargs); __PYX_ERR(1, 864, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.reduce_with_state_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_aut), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "aut", 0))) __PYX_ERR(1, 865, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_117reduce_with_state_map(((PyTypeObject*)__pyx_v_cls), __pyx_v_aut, __pyx_v_trim_input, __pyx_v_params);

  /* "libmata/nfa/nfa.pyx":864
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def reduce_with_state_map(cls, Nfa aut, bool trim_input = True, params = None):
 *         """Reduce the automaton.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_117reduce_with_state_map(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_aut, bool __pyx_v_trim_input, PyObject *__pyx_v_params) {
  Mata::Nfa::StateToStateMap __pyx_v_state_map;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  PyObject *__pyx_9genexpr13__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr13__pyx_v_v = NULL;
  PyObject *__pyx_9genexpr14__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr14__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  Mata::Nfa::StringMap __pyx_t_12;
  std::unordered_map<Mata::Nfa::State,Mata::Nfa::State> ::iterator __pyx_t_13;
  std::pair<Mata::Nfa::State,Mata::Nfa::State>  __pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reduce_with_state_map", 0);
  __Pyx_INCREF(__pyx_v_params);

  /* "libmata/nfa/nfa.pyx":874
 *         :return: (Reduced automaton, state map of original to new states)
 *         """
 *         params = params or {"algorithm": "simulation"}             # <<<<<<<<<<<<<<
 *         cdef StateToStateMap state_map
 *         result = Nfa()
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_params); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 874, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_params);
    __pyx_t_1 = __pyx_v_params;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_algorithm, __pyx_n_u_simulation) < 0) __PYX_ERR(1, 874, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF_SET(__pyx_v_params, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":876
 *         params = params or {"algorithm": "simulation"}
 *         cdef StateToStateMap state_map
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, &state_map,
 *             {
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":877
 *         cdef StateToStateMap state_map
 *         result = Nfa()
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, &state_map,             # <<<<<<<<<<<<<<
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') for k, v in params.items()
 */
  { /* enter inner scope */

    /* "libmata/nfa/nfa.pyx":878
 *         result = Nfa()
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, &state_map,
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') for k, v in params.items()
 *             }
 */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 878, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "libmata/nfa/nfa.pyx":879
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, &state_map,
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') for k, v in params.items()             # <<<<<<<<<<<<<<
 *             }
 *         )
 */
    __pyx_t_4 = 0;
    if (unlikely(__pyx_v_params == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(1, 879, __pyx_L7_error)
    }
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_v_params, 0, __pyx_n_s_items, (&__pyx_t_5), (&__pyx_t_6)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 879, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_5, &__pyx_t_4, &__pyx_t_7, &__pyx_t_8, NULL, __pyx_t_6);
      if (unlikely(__pyx_t_9 == 0)) break;
      if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 879, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v_k, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v_v, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr13__pyx_v_k, __pyx_n_s_encode); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 879, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_10 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_kp_u_utf_8};
        __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 879, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr13__pyx_v_v, __pyx_n_s_encode); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 879, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_9 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_utf_8};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 879, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_8, (PyObject*)__pyx_t_7))) __PYX_ERR(1, 879, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_k); __pyx_9genexpr13__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_v); __pyx_9genexpr13__pyx_v_v = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_k); __pyx_9genexpr13__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_v); __pyx_9genexpr13__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */

  /* "libmata/nfa/nfa.pyx":878
 *         result = Nfa()
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, &state_map,
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') for k, v in params.items()
 *             }
 */
  __pyx_t_12 = __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 878, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":877
 *         cdef StateToStateMap state_map
 *         result = Nfa()
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, &state_map,             # <<<<<<<<<<<<<<
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') for k, v in params.items()
 */
  Mata::Nfa::Plumbing::reduce(__pyx_v_result->thisptr.get(), (*__pyx_v_aut->thisptr.get()), __pyx_v_trim_input, (&__pyx_v_state_map), __pyx_t_12);

  /* "libmata/nfa/nfa.pyx":883
 *         )
 * 
 *         return result, {k: v for k, v in state_map}             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 883, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_13 = __pyx_v_state_map.begin();
    for (;;) {
      if (!(__pyx_t_13 != __pyx_v_state_map.end())) break;
      __pyx_t_14 = *__pyx_t_13;
      ++__pyx_t_13;
      __pyx_t_3 = __pyx_convert_pair_to_py_Mata_3a__3a_Nfa_3a__3a_State____Mata_3a__3a_Nfa_3a__3a_State(__pyx_t_14); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 883, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_3);
      if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
        PyObject* sequence = __pyx_t_3;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(1, 883, __pyx_L13_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 883, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 883, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_10 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 883, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_15 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_10);
        index = 0; __pyx_t_7 = __pyx_t_15(__pyx_t_10); if (unlikely(!__pyx_t_7)) goto __pyx_L16_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        index = 1; __pyx_t_8 = __pyx_t_15(__pyx_t_10); if (unlikely(!__pyx_t_8)) goto __pyx_L16_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_10), 2) < 0) __PYX_ERR(1, 883, __pyx_L13_error)
        __pyx_t_15 = NULL;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L17_unpacking_done;
        __pyx_L16_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_15 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(1, 883, __pyx_L13_error)
        __pyx_L17_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr14__pyx_v_k, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr14__pyx_v_v, __pyx_t_8);
      __pyx_t_8 = 0;
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_9genexpr14__pyx_v_k, (PyObject*)__pyx_9genexpr14__pyx_v_v))) __PYX_ERR(1, 883, __pyx_L13_error)
    }
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_k); __pyx_9genexpr14__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_v); __pyx_9genexpr14__pyx_v_v = 0;
    goto __pyx_L19_exit_scope;
    __pyx_L13_error:;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_k); __pyx_9genexpr14__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_v); __pyx_9genexpr14__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L19_exit_scope:;
  } /* exit inner scope */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __Pyx_GIVEREF((PyObject *)__pyx_v_result);
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_result));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":864
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def reduce_with_state_map(cls, Nfa aut, bool trim_input = True, params = None):
 *         """Reduce the automaton.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.reduce_with_state_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_v);
  __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":885
 *         return result, {k: v for k, v in state_map}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def reduce(cls, Nfa aut, bool trim_input = True, params = None):
 *         """Reduce the automaton.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_120reduce(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_119reduce, "Reduce the automaton.\n\n        :param bool trim_input: Whether to trim the input automaton first or not.\n        :param Nfa aut: Original automaton to reduce.\n        :param Dict params: Additional parameters for the reduction algorithm:\n            - \"algorithm\": \"simulation\"\n        :return: Reduced automaton\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_120reduce = {"reduce", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_120reduce, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_119reduce};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_120reduce(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_aut = 0;
  bool __pyx_v_trim_input;
  PyObject *__pyx_v_params = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reduce (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_aut,&__pyx_n_s_trim_input,&__pyx_n_s_params,0};
    PyObject* values[3] = {0,0,0};

    /* "libmata/nfa/nfa.pyx":886
 * 
 *     @classmethod
 *     def reduce(cls, Nfa aut, bool trim_input = True, params = None):             # <<<<<<<<<<<<<<
 *         """Reduce the automaton.
 * 
 */
    values[2] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_aut)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 885, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_trim_input);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 885, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_params);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 885, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "reduce") < 0)) __PYX_ERR(1, 885, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_aut = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    if (values[1]) {
      __pyx_v_trim_input = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_trim_input == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(1, 886, __pyx_L3_error)
    } else {
      __pyx_v_trim_input = ((bool)1);
    }
    __pyx_v_params = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("reduce", 0, 1, 3, __pyx_nargs); __PYX_ERR(1, 885, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.reduce", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_aut), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "aut", 0))) __PYX_ERR(1, 886, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_119reduce(((PyTypeObject*)__pyx_v_cls), __pyx_v_aut, __pyx_v_trim_input, __pyx_v_params);

  /* "libmata/nfa/nfa.pyx":885
 *         return result, {k: v for k, v in state_map}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def reduce(cls, Nfa aut, bool trim_input = True, params = None):
 *         """Reduce the automaton.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_119reduce(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_aut, bool __pyx_v_trim_input, PyObject *__pyx_v_params) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_result = NULL;
  PyObject *__pyx_9genexpr15__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr15__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  Mata::Nfa::StringMap __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reduce", 0);
  __Pyx_INCREF(__pyx_v_params);

  /* "libmata/nfa/nfa.pyx":895
 *         :return: Reduced automaton
 *         """
 *         params = params or {"algorithm": "simulation"}             # <<<<<<<<<<<<<<
 *         result = Nfa()
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, NULL,
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_params); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 895, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_params);
    __pyx_t_1 = __pyx_v_params;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_algorithm, __pyx_n_u_simulation) < 0) __PYX_ERR(1, 895, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF_SET(__pyx_v_params, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":896
 *         """
 *         params = params or {"algorithm": "simulation"}
 *         result = Nfa()             # <<<<<<<<<<<<<<
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, NULL,
 *             {
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":897
 *         params = params or {"algorithm": "simulation"}
 *         result = Nfa()
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, NULL,             # <<<<<<<<<<<<<<
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') for k, v in params.items()
 */
  { /* enter inner scope */

    /* "libmata/nfa/nfa.pyx":898
 *         result = Nfa()
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, NULL,
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') for k, v in params.items()
 *             }
 */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 898, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "libmata/nfa/nfa.pyx":899
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, NULL,
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') for k, v in params.items()             # <<<<<<<<<<<<<<
 *             }
 *         )
 */
    __pyx_t_4 = 0;
    if (unlikely(__pyx_v_params == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(1, 899, __pyx_L7_error)
    }
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_v_params, 0, __pyx_n_s_items, (&__pyx_t_5), (&__pyx_t_6)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 899, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_5, &__pyx_t_4, &__pyx_t_7, &__pyx_t_8, NULL, __pyx_t_6);
      if (unlikely(__pyx_t_9 == 0)) break;
      if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 899, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_9genexpr15__pyx_v_k, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr15__pyx_v_v, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr15__pyx_v_k, __pyx_n_s_encode); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 899, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_10 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_kp_u_utf_8};
        __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 899, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr15__pyx_v_v, __pyx_n_s_encode); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 899, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_9 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_utf_8};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 899, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_8, (PyObject*)__pyx_t_7))) __PYX_ERR(1, 899, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_k); __pyx_9genexpr15__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_v); __pyx_9genexpr15__pyx_v_v = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_k); __pyx_9genexpr15__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_v); __pyx_9genexpr15__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */

  /* "libmata/nfa/nfa.pyx":898
 *         result = Nfa()
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, NULL,
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') for k, v in params.items()
 *             }
 */
  __pyx_t_12 = __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 898, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":897
 *         params = params or {"algorithm": "simulation"}
 *         result = Nfa()
 *         mata_nfa.reduce(result.thisptr.get(), dereference(aut.thisptr.get()), trim_input, NULL,             # <<<<<<<<<<<<<<
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') for k, v in params.items()
 */
  Mata::Nfa::Plumbing::reduce(__pyx_v_result->thisptr.get(), (*__pyx_v_aut->thisptr.get()), __pyx_v_trim_input, NULL, __pyx_t_12);

  /* "libmata/nfa/nfa.pyx":902
 *             }
 *         )
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":885
 *         return result, {k: v for k, v in state_map}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def reduce(cls, Nfa aut, bool trim_input = True, params = None):
 *         """Reduce the automaton.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.reduce", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":904
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def compute_relation(cls, Nfa lhs, params = None):
 *         """Computes the relation for the automaton
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_122compute_relation(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_121compute_relation, "Computes the relation for the automaton\n\n        :param Nfa lhs: automaton\n        :param Dict params: parameters of the computed relation\n        :return: computd relation\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_122compute_relation = {"compute_relation", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_122compute_relation, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_121compute_relation};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_122compute_relation(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  PyObject *__pyx_v_params = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_relation (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_params,0};
    PyObject* values[2] = {0,0};

    /* "libmata/nfa/nfa.pyx":905
 * 
 *     @classmethod
 *     def compute_relation(cls, Nfa lhs, params = None):             # <<<<<<<<<<<<<<
 *         """Computes the relation for the automaton
 * 
 */
    values[1] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 904, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_params);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 904, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "compute_relation") < 0)) __PYX_ERR(1, 904, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_params = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_relation", 0, 1, 2, __pyx_nargs); __PYX_ERR(1, 904, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.compute_relation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 905, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_121compute_relation(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_params);

  /* "libmata/nfa/nfa.pyx":904
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def compute_relation(cls, Nfa lhs, params = None):
 *         """Computes the relation for the automaton
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_121compute_relation(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, PyObject *__pyx_v_params) {
  Simlib::Util::BinaryRelation __pyx_v_relation;
  struct __pyx_obj_7libmata_5utils_BinaryRelation *__pyx_v_result = NULL;
  size_t __pyx_v_relation_size;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  bool __pyx_v_val;
  PyObject *__pyx_9genexpr16__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr16__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Mata::Nfa::StringMap __pyx_t_13;
  size_t __pyx_t_14;
  size_t __pyx_t_15;
  size_t __pyx_t_16;
  size_t __pyx_t_17;
  size_t __pyx_t_18;
  size_t __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_relation", 0);
  __Pyx_INCREF(__pyx_v_params);

  /* "libmata/nfa/nfa.pyx":912
 *         :return: computd relation
 *         """
 *         params = params or {'relation': 'simulation', 'direction': 'forward'}             # <<<<<<<<<<<<<<
 *         cdef CBinaryRelation relation = mata_nfa.compute_relation(
 *             dereference(lhs.thisptr.get()),
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_params); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 912, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_params);
    __pyx_t_1 = __pyx_v_params;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_relation, __pyx_n_u_simulation) < 0) __PYX_ERR(1, 912, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_direction, __pyx_n_u_forward) < 0) __PYX_ERR(1, 912, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF_SET(__pyx_v_params, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":914
 *         params = params or {'relation': 'simulation', 'direction': 'forward'}
 *         cdef CBinaryRelation relation = mata_nfa.compute_relation(
 *             dereference(lhs.thisptr.get()),             # <<<<<<<<<<<<<<
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 */
  { /* enter inner scope */

    /* "libmata/nfa/nfa.pyx":915
 *         cdef CBinaryRelation relation = mata_nfa.compute_relation(
 *             dereference(lhs.thisptr.get()),
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()
 */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 915, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "libmata/nfa/nfa.pyx":917
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()             # <<<<<<<<<<<<<<
 *             }
 *         )
 */
    __pyx_t_4 = 0;
    if (unlikely(__pyx_v_params == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(1, 917, __pyx_L7_error)
    }
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_v_params, 0, __pyx_n_s_items, (&__pyx_t_5), (&__pyx_t_6)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 917, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_5, &__pyx_t_4, &__pyx_t_7, &__pyx_t_8, NULL, __pyx_t_6);
      if (unlikely(__pyx_t_9 == 0)) break;
      if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 917, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_k, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_v, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "libmata/nfa/nfa.pyx":916
 *             dereference(lhs.thisptr.get()),
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v             # <<<<<<<<<<<<<<
 *                 for k, v in params.items()
 *             }
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr16__pyx_v_k, __pyx_n_s_encode); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 916, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_10 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_kp_u_utf_8};
        __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 916, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __pyx_t_2 = PyUnicode_Check(__pyx_9genexpr16__pyx_v_v); 
      if (__pyx_t_2) {
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr16__pyx_v_v, __pyx_n_s_encode); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 916, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_9 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_kp_u_utf_8};
          __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 916, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __pyx_t_7 = __pyx_t_10;
        __pyx_t_10 = 0;
      } else {
        __Pyx_INCREF(__pyx_9genexpr16__pyx_v_v);
        __pyx_t_7 = __pyx_9genexpr16__pyx_v_v;
      }
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_8, (PyObject*)__pyx_t_7))) __PYX_ERR(1, 916, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_k); __pyx_9genexpr16__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_v); __pyx_9genexpr16__pyx_v_v = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_k); __pyx_9genexpr16__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_v); __pyx_9genexpr16__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */

  /* "libmata/nfa/nfa.pyx":915
 *         cdef CBinaryRelation relation = mata_nfa.compute_relation(
 *             dereference(lhs.thisptr.get()),
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()
 */
  __pyx_t_13 = __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 915, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":913
 *         """
 *         params = params or {'relation': 'simulation', 'direction': 'forward'}
 *         cdef CBinaryRelation relation = mata_nfa.compute_relation(             # <<<<<<<<<<<<<<
 *             dereference(lhs.thisptr.get()),
 *             {
 */
  __pyx_v_relation = Mata::Nfa::Algorithms::compute_relation((*__pyx_v_lhs->thisptr.get()), __pyx_t_13);

  /* "libmata/nfa/nfa.pyx":920
 *             }
 *         )
 *         result = BinaryRelation()             # <<<<<<<<<<<<<<
 *         cdef size_t relation_size = relation.size()
 *         result.resize(relation_size)
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_5utils_BinaryRelation)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_7libmata_5utils_BinaryRelation *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":921
 *         )
 *         result = BinaryRelation()
 *         cdef size_t relation_size = relation.size()             # <<<<<<<<<<<<<<
 *         result.resize(relation_size)
 *         for i in range(0, relation_size):
 */
  __pyx_v_relation_size = __pyx_v_relation.size();

  /* "libmata/nfa/nfa.pyx":922
 *         result = BinaryRelation()
 *         cdef size_t relation_size = relation.size()
 *         result.resize(relation_size)             # <<<<<<<<<<<<<<
 *         for i in range(0, relation_size):
 *             for j in range(0, relation_size):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_result), __pyx_n_s_resize); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyInt_FromSize_t(__pyx_v_relation_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":923
 *         cdef size_t relation_size = relation.size()
 *         result.resize(relation_size)
 *         for i in range(0, relation_size):             # <<<<<<<<<<<<<<
 *             for j in range(0, relation_size):
 *                 val = relation.get(i, j)
 */
  __pyx_t_14 = __pyx_v_relation_size;
  __pyx_t_15 = __pyx_t_14;
  for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
    __pyx_v_i = __pyx_t_16;

    /* "libmata/nfa/nfa.pyx":924
 *         result.resize(relation_size)
 *         for i in range(0, relation_size):
 *             for j in range(0, relation_size):             # <<<<<<<<<<<<<<
 *                 val = relation.get(i, j)
 *                 result.set(i, j, val)
 */
    __pyx_t_17 = __pyx_v_relation_size;
    __pyx_t_18 = __pyx_t_17;
    for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
      __pyx_v_j = __pyx_t_19;

      /* "libmata/nfa/nfa.pyx":925
 *         for i in range(0, relation_size):
 *             for j in range(0, relation_size):
 *                 val = relation.get(i, j)             # <<<<<<<<<<<<<<
 *                 result.set(i, j, val)
 *         return result
 */
      __pyx_v_val = __pyx_v_relation.get(__pyx_v_i, __pyx_v_j);

      /* "libmata/nfa/nfa.pyx":926
 *             for j in range(0, relation_size):
 *                 val = relation.get(i, j)
 *                 result.set(i, j, val)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_result), __pyx_n_s_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 926, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyInt_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 926, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyInt_FromSize_t(__pyx_v_j); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 926, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_10 = __Pyx_PyBool_FromLong(__pyx_v_val); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 926, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_11, __pyx_t_7, __pyx_t_8, __pyx_t_10};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 926, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
  }

  /* "libmata/nfa/nfa.pyx":927
 *                 val = relation.get(i, j)
 *                 result.set(i, j, val)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     # Tests
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":904
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def compute_relation(cls, Nfa lhs, params = None):
 *         """Computes the relation for the automaton
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.compute_relation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":930
 * 
 *     # Tests
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_deterministic(cls, Nfa lhs):
 *         """Tests if the lhs is determinstic
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_124is_deterministic(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_123is_deterministic, "Tests if the lhs is determinstic\n\n        :param Nfa lhs: non-determinstic finite automaton\n        :return: true if the lhs is deterministic\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_124is_deterministic = {"is_deterministic", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_124is_deterministic, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_123is_deterministic};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_124is_deterministic(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_deterministic (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 930, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "is_deterministic") < 0)) __PYX_ERR(1, 930, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_deterministic", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 930, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_deterministic", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 931, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_123is_deterministic(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_123is_deterministic(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_deterministic", 0);

  /* "libmata/nfa/nfa.pyx":937
 *         :return: true if the lhs is deterministic
 *         """
 *         return mata_nfa.is_deterministic(dereference(lhs.thisptr.get()))             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(Mata::Nfa::is_deterministic((*__pyx_v_lhs->thisptr.get()))); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":930
 * 
 *     # Tests
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_deterministic(cls, Nfa lhs):
 *         """Tests if the lhs is determinstic
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_deterministic", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":939
 *         return mata_nfa.is_deterministic(dereference(lhs.thisptr.get()))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_lang_empty(cls, Nfa lhs, Run run = None):
 *         """Checks if language of automaton lhs is empty, if not, returns path of states as counter
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_126is_lang_empty(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_125is_lang_empty, "Checks if language of automaton lhs is empty, if not, returns path of states as counter\n        example.\n\n        :param Nfa lhs:\n        :return: true if the lhs is empty, counter example if lhs is not empty\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_126is_lang_empty = {"is_lang_empty", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_126is_lang_empty, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_125is_lang_empty};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_126is_lang_empty(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_run = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_lang_empty (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_run,0};
    PyObject* values[2] = {0,0};

    /* "libmata/nfa/nfa.pyx":940
 * 
 *     @classmethod
 *     def is_lang_empty(cls, Nfa lhs, Run run = None):             # <<<<<<<<<<<<<<
 *         """Checks if language of automaton lhs is empty, if not, returns path of states as counter
 *         example.
 */
    values[1] = (PyObject *)((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 939, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_run);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 939, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "is_lang_empty") < 0)) __PYX_ERR(1, 939, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_run = ((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_lang_empty", 0, 1, 2, __pyx_nargs); __PYX_ERR(1, 939, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_lang_empty", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 940, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_run), __pyx_ptype_7libmata_3nfa_3nfa_Run, 1, "run", 0))) __PYX_ERR(1, 940, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_125is_lang_empty(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_run);

  /* "libmata/nfa/nfa.pyx":939
 *         return mata_nfa.is_deterministic(dereference(lhs.thisptr.get()))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_lang_empty(cls, Nfa lhs, Run run = None):
 *         """Checks if language of automaton lhs is empty, if not, returns path of states as counter
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_125is_lang_empty(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_run) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_lang_empty", 0);

  /* "libmata/nfa/nfa.pyx":947
 *         :return: true if the lhs is empty, counter example if lhs is not empty
 *         """
 *         if run:             # <<<<<<<<<<<<<<
 *             return mata_nfa.is_lang_empty(dereference(lhs.thisptr.get()), run.thisptr)
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_run)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 947, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "libmata/nfa/nfa.pyx":948
 *         """
 *         if run:
 *             return mata_nfa.is_lang_empty(dereference(lhs.thisptr.get()), run.thisptr)             # <<<<<<<<<<<<<<
 *         else:
 *             return mata_nfa.is_lang_empty(dereference(lhs.thisptr.get()), NULL)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyBool_FromLong(Mata::Nfa::is_lang_empty((*__pyx_v_lhs->thisptr.get()), __pyx_v_run->thisptr)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "libmata/nfa/nfa.pyx":947
 *         :return: true if the lhs is empty, counter example if lhs is not empty
 *         """
 *         if run:             # <<<<<<<<<<<<<<
 *             return mata_nfa.is_lang_empty(dereference(lhs.thisptr.get()), run.thisptr)
 *         else:
 */
  }

  /* "libmata/nfa/nfa.pyx":950
 *             return mata_nfa.is_lang_empty(dereference(lhs.thisptr.get()), run.thisptr)
 *         else:
 *             return mata_nfa.is_lang_empty(dereference(lhs.thisptr.get()), NULL)             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyBool_FromLong(Mata::Nfa::is_lang_empty((*__pyx_v_lhs->thisptr.get()), NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "libmata/nfa/nfa.pyx":939
 *         return mata_nfa.is_deterministic(dereference(lhs.thisptr.get()))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_lang_empty(cls, Nfa lhs, Run run = None):
 *         """Checks if language of automaton lhs is empty, if not, returns path of states as counter
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_lang_empty", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":952
 *             return mata_nfa.is_lang_empty(dereference(lhs.thisptr.get()), NULL)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_universal(cls, Nfa lhs, alph.Alphabet alphabet, params = None):
 *         """Tests if lhs is universal wrt given alphabet
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_128is_universal(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_127is_universal, "Tests if lhs is universal wrt given alphabet\n\n        :param Nfa lhs: automaton tested for universality\n        :param OnTheFlyAlphabet alphabet: on the fly alphabet\n        :param dict params: additional params to the function, currently supports key 'algorithm',\n            which determines used universality test\n        :return: true if lhs is universal\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_128is_universal = {"is_universal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_128is_universal, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_127is_universal};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_128is_universal(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet = 0;
  PyObject *__pyx_v_params = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_universal (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_alphabet,&__pyx_n_s_params,0};
    PyObject* values[3] = {0,0,0};

    /* "libmata/nfa/nfa.pyx":953
 * 
 *     @classmethod
 *     def is_universal(cls, Nfa lhs, alph.Alphabet alphabet, params = None):             # <<<<<<<<<<<<<<
 *         """Tests if lhs is universal wrt given alphabet
 * 
 */
    values[2] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 952, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_alphabet)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 952, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("is_universal", 0, 2, 3, 1); __PYX_ERR(1, 952, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_params);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 952, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "is_universal") < 0)) __PYX_ERR(1, 952, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_alphabet = ((struct __pyx_obj_7libmata_9alphabets_Alphabet *)values[1]);
    __pyx_v_params = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_universal", 0, 2, 3, __pyx_nargs); __PYX_ERR(1, 952, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_universal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 953, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_alphabet), __pyx_ptype_7libmata_9alphabets_Alphabet, 1, "alphabet", 0))) __PYX_ERR(1, 953, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_127is_universal(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_alphabet, __pyx_v_params);

  /* "libmata/nfa/nfa.pyx":952
 *             return mata_nfa.is_lang_empty(dereference(lhs.thisptr.get()), NULL)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_universal(cls, Nfa lhs, alph.Alphabet alphabet, params = None):
 *         """Tests if lhs is universal wrt given alphabet
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_127is_universal(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet, PyObject *__pyx_v_params) {
  PyObject *__pyx_9genexpr17__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr17__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Mata::Alphabet *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  Mata::Nfa::StringMap __pyx_t_14;
  bool __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_universal", 0);
  __Pyx_INCREF(__pyx_v_params);

  /* "libmata/nfa/nfa.pyx":962
 *         :return: true if lhs is universal
 *         """
 *         params = params or {'algorithm': 'antichains'}             # <<<<<<<<<<<<<<
 *         return mata_nfa.is_universal(
 *             dereference(lhs.thisptr.get()),
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_params); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 962, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_params);
    __pyx_t_1 = __pyx_v_params;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_algorithm, __pyx_n_u_antichains) < 0) __PYX_ERR(1, 962, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF_SET(__pyx_v_params, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":963
 *         """
 *         params = params or {'algorithm': 'antichains'}
 *         return mata_nfa.is_universal(             # <<<<<<<<<<<<<<
 *             dereference(lhs.thisptr.get()),
 *             <CAlphabet&>dereference(alphabet.as_base()),
 */
  __Pyx_XDECREF(__pyx_r);

  /* "libmata/nfa/nfa.pyx":965
 *         return mata_nfa.is_universal(
 *             dereference(lhs.thisptr.get()),
 *             <CAlphabet&>dereference(alphabet.as_base()),             # <<<<<<<<<<<<<<
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_7libmata_9alphabets_Alphabet *)__pyx_v_alphabet->__pyx_vtab)->as_base(__pyx_v_alphabet); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 965, __pyx_L1_error)
  { /* enter inner scope */

    /* "libmata/nfa/nfa.pyx":966
 *             dereference(lhs.thisptr.get()),
 *             <CAlphabet&>dereference(alphabet.as_base()),
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()
 */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 966, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "libmata/nfa/nfa.pyx":968
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()             # <<<<<<<<<<<<<<
 *             }
 *         )
 */
    __pyx_t_5 = 0;
    if (unlikely(__pyx_v_params == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(1, 968, __pyx_L7_error)
    }
    __pyx_t_8 = __Pyx_dict_iterator(__pyx_v_params, 0, __pyx_n_s_items, (&__pyx_t_6), (&__pyx_t_7)); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 968, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_8;
    __pyx_t_8 = 0;
    while (1) {
      __pyx_t_10 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_6, &__pyx_t_5, &__pyx_t_8, &__pyx_t_9, NULL, __pyx_t_7);
      if (unlikely(__pyx_t_10 == 0)) break;
      if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(1, 968, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_9genexpr17__pyx_v_k, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr17__pyx_v_v, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "libmata/nfa/nfa.pyx":967
 *             <CAlphabet&>dereference(alphabet.as_base()),
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v             # <<<<<<<<<<<<<<
 *                 for k, v in params.items()
 *             }
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr17__pyx_v_k, __pyx_n_s_encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 967, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_11 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_10 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_utf_8};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 967, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __pyx_t_2 = PyUnicode_Check(__pyx_9genexpr17__pyx_v_v); 
      if (__pyx_t_2) {
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr17__pyx_v_v, __pyx_n_s_encode); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 967, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_10 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_kp_u_utf_8};
          __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 967, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __pyx_t_8 = __pyx_t_11;
        __pyx_t_11 = 0;
      } else {
        __Pyx_INCREF(__pyx_9genexpr17__pyx_v_v);
        __pyx_t_8 = __pyx_9genexpr17__pyx_v_v;
      }
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_9, (PyObject*)__pyx_t_8))) __PYX_ERR(1, 967, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_k); __pyx_9genexpr17__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_v); __pyx_9genexpr17__pyx_v_v = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_k); __pyx_9genexpr17__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_v); __pyx_9genexpr17__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */

  /* "libmata/nfa/nfa.pyx":966
 *             dereference(lhs.thisptr.get()),
 *             <CAlphabet&>dereference(alphabet.as_base()),
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()
 */
  __pyx_t_14 = __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 966, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":963
 *         """
 *         params = params or {'algorithm': 'antichains'}
 *         return mata_nfa.is_universal(             # <<<<<<<<<<<<<<
 *             dereference(lhs.thisptr.get()),
 *             <CAlphabet&>dereference(alphabet.as_base()),
 */
  try {
    __pyx_t_15 = Mata::Nfa::is_universal((*__pyx_v_lhs->thisptr.get()), ((Mata::Alphabet &)(*__pyx_t_4)), __pyx_t_14);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 963, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_15); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":952
 *             return mata_nfa.is_lang_empty(dereference(lhs.thisptr.get()), NULL)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_universal(cls, Nfa lhs, alph.Alphabet alphabet, params = None):
 *         """Tests if lhs is universal wrt given alphabet
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_universal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":972
 *         )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_included_with_cex(
 *             cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_130is_included_with_cex(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_129is_included_with_cex, "Test inclusion between two automata\n\n        :param Nfa lhs: smaller automaton\n        :param Nfa rhs: bigger automaton\n        :param alph.Alphabet alphabet: alpabet shared by two automata\n        :param dict params: additional params\n        :return: true if lhs is included by rhs, counter example word if not\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_130is_included_with_cex = {"is_included_with_cex", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_130is_included_with_cex, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_129is_included_with_cex};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_130is_included_with_cex(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs = 0;
  struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet = 0;
  PyObject *__pyx_v_params = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_included_with_cex (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_rhs,&__pyx_n_s_alphabet,&__pyx_n_s_params,0};
    PyObject* values[4] = {0,0,0,0};

    /* "libmata/nfa/nfa.pyx":974
 *     @classmethod
 *     def is_included_with_cex(
 *             cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None             # <<<<<<<<<<<<<<
 *     ):
 *         """Test inclusion between two automata
 */
    values[2] = (PyObject *)((struct __pyx_obj_7libmata_9alphabets_Alphabet *)Py_None);
    values[3] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 972, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 972, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("is_included_with_cex", 0, 2, 4, 1); __PYX_ERR(1, 972, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_alphabet);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 972, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_params);
          if (value) { values[3] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 972, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "is_included_with_cex") < 0)) __PYX_ERR(1, 972, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_rhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[1]);
    __pyx_v_alphabet = ((struct __pyx_obj_7libmata_9alphabets_Alphabet *)values[2]);
    __pyx_v_params = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_included_with_cex", 0, 2, 4, __pyx_nargs); __PYX_ERR(1, 972, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_included_with_cex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 974, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "rhs", 0))) __PYX_ERR(1, 974, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_alphabet), __pyx_ptype_7libmata_9alphabets_Alphabet, 1, "alphabet", 0))) __PYX_ERR(1, 974, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_129is_included_with_cex(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_rhs, __pyx_v_alphabet, __pyx_v_params);

  /* "libmata/nfa/nfa.pyx":972
 *         )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_included_with_cex(
 *             cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_129is_included_with_cex(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet, PyObject *__pyx_v_params) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_run = NULL;
  Mata::Alphabet *__pyx_v_c_alphabet;
  bool __pyx_v_result;
  PyObject *__pyx_9genexpr18__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr18__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Mata::Alphabet *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  Mata::Nfa::StringMap __pyx_t_14;
  bool __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_included_with_cex", 0);
  __Pyx_INCREF(__pyx_v_params);

  /* "libmata/nfa/nfa.pyx":984
 *         :return: true if lhs is included by rhs, counter example word if not
 *         """
 *         run = Run()             # <<<<<<<<<<<<<<
 *         cdef CAlphabet* c_alphabet = NULL
 *         if alphabet:
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Run)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_run = ((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":985
 *         """
 *         run = Run()
 *         cdef CAlphabet* c_alphabet = NULL             # <<<<<<<<<<<<<<
 *         if alphabet:
 *             c_alphabet = alphabet.as_base()
 */
  __pyx_v_c_alphabet = NULL;

  /* "libmata/nfa/nfa.pyx":986
 *         run = Run()
 *         cdef CAlphabet* c_alphabet = NULL
 *         if alphabet:             # <<<<<<<<<<<<<<
 *             c_alphabet = alphabet.as_base()
 *         params = params or {'algorithm': 'antichains'}
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_alphabet)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 986, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "libmata/nfa/nfa.pyx":987
 *         cdef CAlphabet* c_alphabet = NULL
 *         if alphabet:
 *             c_alphabet = alphabet.as_base()             # <<<<<<<<<<<<<<
 *         params = params or {'algorithm': 'antichains'}
 *         result = mata_nfa.is_included(
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_7libmata_9alphabets_Alphabet *)__pyx_v_alphabet->__pyx_vtab)->as_base(__pyx_v_alphabet); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 987, __pyx_L1_error)
    __pyx_v_c_alphabet = __pyx_t_3;

    /* "libmata/nfa/nfa.pyx":986
 *         run = Run()
 *         cdef CAlphabet* c_alphabet = NULL
 *         if alphabet:             # <<<<<<<<<<<<<<
 *             c_alphabet = alphabet.as_base()
 *         params = params or {'algorithm': 'antichains'}
 */
  }

  /* "libmata/nfa/nfa.pyx":988
 *         if alphabet:
 *             c_alphabet = alphabet.as_base()
 *         params = params or {'algorithm': 'antichains'}             # <<<<<<<<<<<<<<
 *         result = mata_nfa.is_included(
 *             dereference(lhs.thisptr.get()),
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_params); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 988, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_params);
    __pyx_t_1 = __pyx_v_params;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_algorithm, __pyx_n_u_antichains) < 0) __PYX_ERR(1, 988, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_1 = __pyx_t_4;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF_SET(__pyx_v_params, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":993
 *             dereference(rhs.thisptr.get()),
 *             run.thisptr,
 *             c_alphabet,             # <<<<<<<<<<<<<<
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 */
  { /* enter inner scope */

    /* "libmata/nfa/nfa.pyx":994
 *             run.thisptr,
 *             c_alphabet,
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()
 */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 994, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "libmata/nfa/nfa.pyx":996
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()             # <<<<<<<<<<<<<<
 *             }
 *         )
 */
    __pyx_t_5 = 0;
    if (unlikely(__pyx_v_params == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(1, 996, __pyx_L8_error)
    }
    __pyx_t_8 = __Pyx_dict_iterator(__pyx_v_params, 0, __pyx_n_s_items, (&__pyx_t_6), (&__pyx_t_7)); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 996, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_4);
    __pyx_t_4 = __pyx_t_8;
    __pyx_t_8 = 0;
    while (1) {
      __pyx_t_10 = __Pyx_dict_iter_next(__pyx_t_4, __pyx_t_6, &__pyx_t_5, &__pyx_t_8, &__pyx_t_9, NULL, __pyx_t_7);
      if (unlikely(__pyx_t_10 == 0)) break;
      if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(1, 996, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_9genexpr18__pyx_v_k, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr18__pyx_v_v, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "libmata/nfa/nfa.pyx":995
 *             c_alphabet,
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v             # <<<<<<<<<<<<<<
 *                 for k, v in params.items()
 *             }
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr18__pyx_v_k, __pyx_n_s_encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 995, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_11 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_10 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_utf_8};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 995, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __pyx_t_2 = PyUnicode_Check(__pyx_9genexpr18__pyx_v_v); 
      if (__pyx_t_2) {
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr18__pyx_v_v, __pyx_n_s_encode); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 995, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_10 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_kp_u_utf_8};
          __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 995, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __pyx_t_8 = __pyx_t_11;
        __pyx_t_11 = 0;
      } else {
        __Pyx_INCREF(__pyx_9genexpr18__pyx_v_v);
        __pyx_t_8 = __pyx_9genexpr18__pyx_v_v;
      }
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_9, (PyObject*)__pyx_t_8))) __PYX_ERR(1, 995, __pyx_L8_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_k); __pyx_9genexpr18__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_v); __pyx_9genexpr18__pyx_v_v = 0;
    goto __pyx_L11_exit_scope;
    __pyx_L8_error:;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_k); __pyx_9genexpr18__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_v); __pyx_9genexpr18__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L11_exit_scope:;
  } /* exit inner scope */

  /* "libmata/nfa/nfa.pyx":994
 *             run.thisptr,
 *             c_alphabet,
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()
 */
  __pyx_t_14 = __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 994, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":989
 *             c_alphabet = alphabet.as_base()
 *         params = params or {'algorithm': 'antichains'}
 *         result = mata_nfa.is_included(             # <<<<<<<<<<<<<<
 *             dereference(lhs.thisptr.get()),
 *             dereference(rhs.thisptr.get()),
 */
  try {
    __pyx_t_15 = Mata::Nfa::is_included((*__pyx_v_lhs->thisptr.get()), (*__pyx_v_rhs->thisptr.get()), __pyx_v_run->thisptr, __pyx_v_c_alphabet, __pyx_t_14);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 989, __pyx_L1_error)
  }
  __pyx_v_result = __pyx_t_15;

  /* "libmata/nfa/nfa.pyx":999
 *             }
 *         )
 *         return result, run             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_run);
  __Pyx_GIVEREF((PyObject *)__pyx_v_run);
  PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_run));
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":972
 *         )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_included_with_cex(
 *             cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_included_with_cex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_run);
  __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1001
 *         return result, run
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_included(
 *             cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_132is_included(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_131is_included, "Test inclusion between two automata\n\n        :param Nfa lhs: smaller automaton\n        :param Nfa rhs: bigger automaton\n        :param alph.Alphabet alphabet: alpabet shared by two automata\n        :param dict params: additional params\n        :return: true if lhs is included by rhs, counter example word if not\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_132is_included = {"is_included", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_132is_included, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_131is_included};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_132is_included(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs = 0;
  struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet = 0;
  PyObject *__pyx_v_params = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_included (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_rhs,&__pyx_n_s_alphabet,&__pyx_n_s_params,0};
    PyObject* values[4] = {0,0,0,0};

    /* "libmata/nfa/nfa.pyx":1003
 *     @classmethod
 *     def is_included(
 *             cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None             # <<<<<<<<<<<<<<
 *     ):
 *         """Test inclusion between two automata
 */
    values[2] = (PyObject *)((struct __pyx_obj_7libmata_9alphabets_Alphabet *)Py_None);
    values[3] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1001, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1001, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("is_included", 0, 2, 4, 1); __PYX_ERR(1, 1001, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_alphabet);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1001, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_params);
          if (value) { values[3] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1001, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "is_included") < 0)) __PYX_ERR(1, 1001, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_rhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[1]);
    __pyx_v_alphabet = ((struct __pyx_obj_7libmata_9alphabets_Alphabet *)values[2]);
    __pyx_v_params = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_included", 0, 2, 4, __pyx_nargs); __PYX_ERR(1, 1001, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_included", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 1003, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "rhs", 0))) __PYX_ERR(1, 1003, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_alphabet), __pyx_ptype_7libmata_9alphabets_Alphabet, 1, "alphabet", 0))) __PYX_ERR(1, 1003, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_131is_included(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_rhs, __pyx_v_alphabet, __pyx_v_params);

  /* "libmata/nfa/nfa.pyx":1001
 *         return result, run
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_included(
 *             cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_131is_included(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet, PyObject *__pyx_v_params) {
  Mata::Alphabet *__pyx_v_c_alphabet;
  bool __pyx_v_result;
  PyObject *__pyx_9genexpr19__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr19__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Mata::Alphabet *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  Mata::Nfa::StringMap __pyx_t_14;
  bool __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_included", 0);
  __Pyx_INCREF(__pyx_v_params);

  /* "libmata/nfa/nfa.pyx":1013
 *         :return: true if lhs is included by rhs, counter example word if not
 *         """
 *         cdef CAlphabet* c_alphabet = NULL             # <<<<<<<<<<<<<<
 *         if alphabet:
 *             c_alphabet = alphabet.as_base()
 */
  __pyx_v_c_alphabet = NULL;

  /* "libmata/nfa/nfa.pyx":1014
 *         """
 *         cdef CAlphabet* c_alphabet = NULL
 *         if alphabet:             # <<<<<<<<<<<<<<
 *             c_alphabet = alphabet.as_base()
 *         params = params or {'algorithm': 'antichains'}
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_alphabet)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 1014, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "libmata/nfa/nfa.pyx":1015
 *         cdef CAlphabet* c_alphabet = NULL
 *         if alphabet:
 *             c_alphabet = alphabet.as_base()             # <<<<<<<<<<<<<<
 *         params = params or {'algorithm': 'antichains'}
 *         result = mata_nfa.is_included(
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_7libmata_9alphabets_Alphabet *)__pyx_v_alphabet->__pyx_vtab)->as_base(__pyx_v_alphabet); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1015, __pyx_L1_error)
    __pyx_v_c_alphabet = __pyx_t_2;

    /* "libmata/nfa/nfa.pyx":1014
 *         """
 *         cdef CAlphabet* c_alphabet = NULL
 *         if alphabet:             # <<<<<<<<<<<<<<
 *             c_alphabet = alphabet.as_base()
 *         params = params or {'algorithm': 'antichains'}
 */
  }

  /* "libmata/nfa/nfa.pyx":1016
 *         if alphabet:
 *             c_alphabet = alphabet.as_base()
 *         params = params or {'algorithm': 'antichains'}             # <<<<<<<<<<<<<<
 *         result = mata_nfa.is_included(
 *             dereference(lhs.thisptr.get()),
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_params); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 1016, __pyx_L1_error)
  if (!__pyx_t_1) {
  } else {
    __Pyx_INCREF(__pyx_v_params);
    __pyx_t_3 = __pyx_v_params;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_algorithm, __pyx_n_u_antichains) < 0) __PYX_ERR(1, 1016, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF_SET(__pyx_v_params, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "libmata/nfa/nfa.pyx":1021
 *             dereference(rhs.thisptr.get()),
 *             NULL,
 *             c_alphabet,             # <<<<<<<<<<<<<<
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 */
  { /* enter inner scope */

    /* "libmata/nfa/nfa.pyx":1022
 *             NULL,
 *             c_alphabet,
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()
 */
    __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1022, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "libmata/nfa/nfa.pyx":1024
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()             # <<<<<<<<<<<<<<
 *             }
 *         )
 */
    __pyx_t_5 = 0;
    if (unlikely(__pyx_v_params == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(1, 1024, __pyx_L8_error)
    }
    __pyx_t_8 = __Pyx_dict_iterator(__pyx_v_params, 0, __pyx_n_s_items, (&__pyx_t_6), (&__pyx_t_7)); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1024, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_4);
    __pyx_t_4 = __pyx_t_8;
    __pyx_t_8 = 0;
    while (1) {
      __pyx_t_10 = __Pyx_dict_iter_next(__pyx_t_4, __pyx_t_6, &__pyx_t_5, &__pyx_t_8, &__pyx_t_9, NULL, __pyx_t_7);
      if (unlikely(__pyx_t_10 == 0)) break;
      if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(1, 1024, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_9genexpr19__pyx_v_k, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr19__pyx_v_v, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "libmata/nfa/nfa.pyx":1023
 *             c_alphabet,
 *             {
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v             # <<<<<<<<<<<<<<
 *                 for k, v in params.items()
 *             }
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr19__pyx_v_k, __pyx_n_s_encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1023, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_11 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_10 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_utf_8};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1023, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __pyx_t_1 = PyUnicode_Check(__pyx_9genexpr19__pyx_v_v); 
      if (__pyx_t_1) {
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr19__pyx_v_v, __pyx_n_s_encode); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1023, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_10 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_kp_u_utf_8};
          __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1023, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __pyx_t_8 = __pyx_t_11;
        __pyx_t_11 = 0;
      } else {
        __Pyx_INCREF(__pyx_9genexpr19__pyx_v_v);
        __pyx_t_8 = __pyx_9genexpr19__pyx_v_v;
      }
      if (unlikely(PyDict_SetItem(__pyx_t_3, (PyObject*)__pyx_t_9, (PyObject*)__pyx_t_8))) __PYX_ERR(1, 1023, __pyx_L8_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_k); __pyx_9genexpr19__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_v); __pyx_9genexpr19__pyx_v_v = 0;
    goto __pyx_L11_exit_scope;
    __pyx_L8_error:;
    __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_k); __pyx_9genexpr19__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_v); __pyx_9genexpr19__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L11_exit_scope:;
  } /* exit inner scope */

  /* "libmata/nfa/nfa.pyx":1022
 *             NULL,
 *             c_alphabet,
 *             {             # <<<<<<<<<<<<<<
 *                 k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                 for k, v in params.items()
 */
  __pyx_t_14 = __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(__pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1022, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "libmata/nfa/nfa.pyx":1017
 *             c_alphabet = alphabet.as_base()
 *         params = params or {'algorithm': 'antichains'}
 *         result = mata_nfa.is_included(             # <<<<<<<<<<<<<<
 *             dereference(lhs.thisptr.get()),
 *             dereference(rhs.thisptr.get()),
 */
  try {
    __pyx_t_15 = Mata::Nfa::is_included((*__pyx_v_lhs->thisptr.get()), (*__pyx_v_rhs->thisptr.get()), NULL, __pyx_v_c_alphabet, __pyx_t_14);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 1017, __pyx_L1_error)
  }
  __pyx_v_result = __pyx_t_15;

  /* "libmata/nfa/nfa.pyx":1027
 *             }
 *         )
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1001
 *         return result, run
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_included(
 *             cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_included", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1029
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def equivalence_check(cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None) -> bool:
 *         """Test equivalence of two automata.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_134equivalence_check(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_133equivalence_check, "Test equivalence of two automata.\n\n        :param Nfa lhs: Smaller automaton.\n        :param Nfa rhs: Bigger automaton.\n        :param alph.Alphabet alphabet: Alphabet shared by two automata.\n        :param dict params: Additional params:\n            - \"algorithm\": \"antichains\"\n        :return: True if lhs is equivalent to rhs, False otherwise.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_134equivalence_check = {"equivalence_check", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_134equivalence_check, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_133equivalence_check};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_134equivalence_check(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs = 0;
  struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet = 0;
  PyObject *__pyx_v_params = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("equivalence_check (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_rhs,&__pyx_n_s_alphabet,&__pyx_n_s_params,0};
    PyObject* values[4] = {0,0,0,0};

    /* "libmata/nfa/nfa.pyx":1030
 * 
 *     @classmethod
 *     def equivalence_check(cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None) -> bool:             # <<<<<<<<<<<<<<
 *         """Test equivalence of two automata.
 * 
 */
    values[2] = (PyObject *)((struct __pyx_obj_7libmata_9alphabets_Alphabet *)Py_None);
    values[3] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1029, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_rhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1029, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("equivalence_check", 0, 2, 4, 1); __PYX_ERR(1, 1029, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_alphabet);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1029, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_params);
          if (value) { values[3] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1029, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "equivalence_check") < 0)) __PYX_ERR(1, 1029, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_rhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[1]);
    __pyx_v_alphabet = ((struct __pyx_obj_7libmata_9alphabets_Alphabet *)values[2]);
    __pyx_v_params = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("equivalence_check", 0, 2, 4, __pyx_nargs); __PYX_ERR(1, 1029, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.equivalence_check", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 1030, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "rhs", 0))) __PYX_ERR(1, 1030, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_alphabet), __pyx_ptype_7libmata_9alphabets_Alphabet, 1, "alphabet", 0))) __PYX_ERR(1, 1030, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_133equivalence_check(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_rhs, __pyx_v_alphabet, __pyx_v_params);

  /* "libmata/nfa/nfa.pyx":1029
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def equivalence_check(cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None) -> bool:
 *         """Test equivalence of two automata.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_133equivalence_check(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_rhs, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet, PyObject *__pyx_v_params) {
  Mata::Alphabet *__pyx_v_c_alphabet;
  PyObject *__pyx_9genexpr20__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr20__pyx_v_v = NULL;
  PyObject *__pyx_9genexpr21__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr21__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Mata::Alphabet *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  Mata::Nfa::StringMap __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("equivalence_check", 0);
  __Pyx_INCREF(__pyx_v_params);

  /* "libmata/nfa/nfa.pyx":1040
 *         :return: True if lhs is equivalent to rhs, False otherwise.
 *         """
 *         params = params or {'algorithm': 'antichains'}             # <<<<<<<<<<<<<<
 *         cdef CAlphabet * c_alphabet = NULL
 *         if alphabet:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_params); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 1040, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_params);
    __pyx_t_1 = __pyx_v_params;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_algorithm, __pyx_n_u_antichains) < 0) __PYX_ERR(1, 1040, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF_SET(__pyx_v_params, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":1041
 *         """
 *         params = params or {'algorithm': 'antichains'}
 *         cdef CAlphabet * c_alphabet = NULL             # <<<<<<<<<<<<<<
 *         if alphabet:
 *             c_alphabet = alphabet.as_base()
 */
  __pyx_v_c_alphabet = NULL;

  /* "libmata/nfa/nfa.pyx":1042
 *         params = params or {'algorithm': 'antichains'}
 *         cdef CAlphabet * c_alphabet = NULL
 *         if alphabet:             # <<<<<<<<<<<<<<
 *             c_alphabet = alphabet.as_base()
 *             return mata_nfa.are_equivalent(
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_alphabet)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 1042, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "libmata/nfa/nfa.pyx":1043
 *         cdef CAlphabet * c_alphabet = NULL
 *         if alphabet:
 *             c_alphabet = alphabet.as_base()             # <<<<<<<<<<<<<<
 *             return mata_nfa.are_equivalent(
 *                 dereference(lhs.thisptr.get()),
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_7libmata_9alphabets_Alphabet *)__pyx_v_alphabet->__pyx_vtab)->as_base(__pyx_v_alphabet); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1043, __pyx_L1_error)
    __pyx_v_c_alphabet = __pyx_t_4;

    /* "libmata/nfa/nfa.pyx":1044
 *         if alphabet:
 *             c_alphabet = alphabet.as_base()
 *             return mata_nfa.are_equivalent(             # <<<<<<<<<<<<<<
 *                 dereference(lhs.thisptr.get()),
 *                 dereference(rhs.thisptr.get()),
 */
    __Pyx_XDECREF(__pyx_r);

    /* "libmata/nfa/nfa.pyx":1047
 *                 dereference(lhs.thisptr.get()),
 *                 dereference(rhs.thisptr.get()),
 *                 c_alphabet,             # <<<<<<<<<<<<<<
 *                 {
 *                     k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 */
    { /* enter inner scope */

      /* "libmata/nfa/nfa.pyx":1048
 *                 dereference(rhs.thisptr.get()),
 *                 c_alphabet,
 *                 {             # <<<<<<<<<<<<<<
 *                     k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                     for k, v in params.items()
 */
      __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1048, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "libmata/nfa/nfa.pyx":1050
 *                 {
 *                     k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                     for k, v in params.items()             # <<<<<<<<<<<<<<
 *                 }
 *             )
 */
      __pyx_t_5 = 0;
      if (unlikely(__pyx_v_params == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
        __PYX_ERR(1, 1050, __pyx_L8_error)
      }
      __pyx_t_8 = __Pyx_dict_iterator(__pyx_v_params, 0, __pyx_n_s_items, (&__pyx_t_6), (&__pyx_t_7)); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1050, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_3);
      __pyx_t_3 = __pyx_t_8;
      __pyx_t_8 = 0;
      while (1) {
        __pyx_t_10 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_6, &__pyx_t_5, &__pyx_t_8, &__pyx_t_9, NULL, __pyx_t_7);
        if (unlikely(__pyx_t_10 == 0)) break;
        if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(1, 1050, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_XDECREF_SET(__pyx_9genexpr20__pyx_v_k, __pyx_t_8);
        __pyx_t_8 = 0;
        __Pyx_XDECREF_SET(__pyx_9genexpr20__pyx_v_v, __pyx_t_9);
        __pyx_t_9 = 0;

        /* "libmata/nfa/nfa.pyx":1049
 *                 c_alphabet,
 *                 {
 *                     k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v             # <<<<<<<<<<<<<<
 *                     for k, v in params.items()
 *                 }
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr20__pyx_v_k, __pyx_n_s_encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1049, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_11 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_10 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_utf_8};
          __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1049, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __pyx_t_2 = PyUnicode_Check(__pyx_9genexpr20__pyx_v_v); 
        if (__pyx_t_2) {
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr20__pyx_v_v, __pyx_n_s_encode); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1049, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_13 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_10 = 1;
            }
          }
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_kp_u_utf_8};
            __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1049, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __pyx_t_8 = __pyx_t_11;
          __pyx_t_11 = 0;
        } else {
          __Pyx_INCREF(__pyx_9genexpr20__pyx_v_v);
          __pyx_t_8 = __pyx_9genexpr20__pyx_v_v;
        }
        if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_9, (PyObject*)__pyx_t_8))) __PYX_ERR(1, 1049, __pyx_L8_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_k); __pyx_9genexpr20__pyx_v_k = 0;
      __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_v); __pyx_9genexpr20__pyx_v_v = 0;
      goto __pyx_L11_exit_scope;
      __pyx_L8_error:;
      __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_k); __pyx_9genexpr20__pyx_v_k = 0;
      __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_v); __pyx_9genexpr20__pyx_v_v = 0;
      goto __pyx_L1_error;
      __pyx_L11_exit_scope:;
    } /* exit inner scope */

    /* "libmata/nfa/nfa.pyx":1048
 *                 dereference(rhs.thisptr.get()),
 *                 c_alphabet,
 *                 {             # <<<<<<<<<<<<<<
 *                     k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                     for k, v in params.items()
 */
    __pyx_t_14 = __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1048, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "libmata/nfa/nfa.pyx":1044
 *         if alphabet:
 *             c_alphabet = alphabet.as_base()
 *             return mata_nfa.are_equivalent(             # <<<<<<<<<<<<<<
 *                 dereference(lhs.thisptr.get()),
 *                 dereference(rhs.thisptr.get()),
 */
    __pyx_t_1 = __Pyx_PyBool_FromLong(Mata::Nfa::are_equivalent((*__pyx_v_lhs->thisptr.get()), (*__pyx_v_rhs->thisptr.get()), __pyx_v_c_alphabet, __pyx_t_14)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "libmata/nfa/nfa.pyx":1042
 *         params = params or {'algorithm': 'antichains'}
 *         cdef CAlphabet * c_alphabet = NULL
 *         if alphabet:             # <<<<<<<<<<<<<<
 *             c_alphabet = alphabet.as_base()
 *             return mata_nfa.are_equivalent(
 */
  }

  /* "libmata/nfa/nfa.pyx":1054
 *             )
 *         else:
 *             return mata_nfa.are_equivalent(             # <<<<<<<<<<<<<<
 *                 dereference(lhs.thisptr.get()),
 *                 dereference(rhs.thisptr.get()),
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);

    /* "libmata/nfa/nfa.pyx":1056
 *             return mata_nfa.are_equivalent(
 *                 dereference(lhs.thisptr.get()),
 *                 dereference(rhs.thisptr.get()),             # <<<<<<<<<<<<<<
 *                 {
 *                     k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 */
    { /* enter inner scope */

      /* "libmata/nfa/nfa.pyx":1057
 *                 dereference(lhs.thisptr.get()),
 *                 dereference(rhs.thisptr.get()),
 *                 {             # <<<<<<<<<<<<<<
 *                     k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                     for k, v in params.items()
 */
      __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1057, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "libmata/nfa/nfa.pyx":1059
 *                 {
 *                     k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                     for k, v in params.items()             # <<<<<<<<<<<<<<
 *                 }
 *             )
 */
      __pyx_t_6 = 0;
      if (unlikely(__pyx_v_params == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
        __PYX_ERR(1, 1059, __pyx_L14_error)
      }
      __pyx_t_8 = __Pyx_dict_iterator(__pyx_v_params, 0, __pyx_n_s_items, (&__pyx_t_5), (&__pyx_t_7)); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1059, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_3);
      __pyx_t_3 = __pyx_t_8;
      __pyx_t_8 = 0;
      while (1) {
        __pyx_t_10 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_5, &__pyx_t_6, &__pyx_t_8, &__pyx_t_9, NULL, __pyx_t_7);
        if (unlikely(__pyx_t_10 == 0)) break;
        if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(1, 1059, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_XDECREF_SET(__pyx_9genexpr21__pyx_v_k, __pyx_t_8);
        __pyx_t_8 = 0;
        __Pyx_XDECREF_SET(__pyx_9genexpr21__pyx_v_v, __pyx_t_9);
        __pyx_t_9 = 0;

        /* "libmata/nfa/nfa.pyx":1058
 *                 dereference(rhs.thisptr.get()),
 *                 {
 *                     k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v             # <<<<<<<<<<<<<<
 *                     for k, v in params.items()
 *                 }
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr21__pyx_v_k, __pyx_n_s_encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1058, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_11 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_10 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_utf_8};
          __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1058, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __pyx_t_2 = PyUnicode_Check(__pyx_9genexpr21__pyx_v_v); 
        if (__pyx_t_2) {
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr21__pyx_v_v, __pyx_n_s_encode); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 1058, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_13 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_10 = 1;
            }
          }
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_kp_u_utf_8};
            __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1058, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __pyx_t_8 = __pyx_t_11;
          __pyx_t_11 = 0;
        } else {
          __Pyx_INCREF(__pyx_9genexpr21__pyx_v_v);
          __pyx_t_8 = __pyx_9genexpr21__pyx_v_v;
        }
        if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_9, (PyObject*)__pyx_t_8))) __PYX_ERR(1, 1058, __pyx_L14_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_k); __pyx_9genexpr21__pyx_v_k = 0;
      __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_v); __pyx_9genexpr21__pyx_v_v = 0;
      goto __pyx_L17_exit_scope;
      __pyx_L14_error:;
      __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_k); __pyx_9genexpr21__pyx_v_k = 0;
      __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_v); __pyx_9genexpr21__pyx_v_v = 0;
      goto __pyx_L1_error;
      __pyx_L17_exit_scope:;
    } /* exit inner scope */

    /* "libmata/nfa/nfa.pyx":1057
 *                 dereference(lhs.thisptr.get()),
 *                 dereference(rhs.thisptr.get()),
 *                 {             # <<<<<<<<<<<<<<
 *                     k.encode('utf-8'): v.encode('utf-8') if isinstance(v, str) else v
 *                     for k, v in params.items()
 */
    __pyx_t_14 = __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_std_3a__3a_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1057, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "libmata/nfa/nfa.pyx":1054
 *             )
 *         else:
 *             return mata_nfa.are_equivalent(             # <<<<<<<<<<<<<<
 *                 dereference(lhs.thisptr.get()),
 *                 dereference(rhs.thisptr.get()),
 */
    __pyx_t_1 = __Pyx_PyBool_FromLong(Mata::Nfa::are_equivalent((*__pyx_v_lhs->thisptr.get()), (*__pyx_v_rhs->thisptr.get()), __pyx_t_14)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "libmata/nfa/nfa.pyx":1029
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def equivalence_check(cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None) -> bool:
 *         """Test equivalence of two automata.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.equivalence_check", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_v);
  __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1063
 *             )
 * 
 *     def get_symbols(self):             # <<<<<<<<<<<<<<
 *         """Return a set of symbols used on the transitions in NFA.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_136get_symbols(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_135get_symbols, "Return a set of symbols used on the transitions in NFA.\n\n        :return: Set of symbols.\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_136get_symbols = {"get_symbols", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_136get_symbols, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_135get_symbols};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_136get_symbols(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_symbols (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_symbols", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_symbols", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_135get_symbols(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_135get_symbols(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  Mata::Util::OrdVector<Mata::Symbol>  __pyx_v_symbols;
  Mata::Symbol __pyx_9genexpr22__pyx_v_s;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Mata::Util::OrdVector<Mata::Symbol> ::iterator __pyx_t_2;
  Mata::Symbol __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_symbols", 0);

  /* "libmata/nfa/nfa.pyx":1068
 *         :return: Set of symbols.
 *         """
 *         cdef COrdVector[Symbol] symbols = self.thisptr.get().get_used_symbols()             # <<<<<<<<<<<<<<
 *         return {s for s in symbols}
 * 
 */
  __pyx_v_symbols = __pyx_v_self->thisptr.get()->get_used_symbols();

  /* "libmata/nfa/nfa.pyx":1069
 *         """
 *         cdef COrdVector[Symbol] symbols = self.thisptr.get().get_used_symbols()
 *         return {s for s in symbols}             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PySet_New(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_symbols.begin();
    for (;;) {
      if (!(__pyx_t_2 != __pyx_v_symbols.end())) break;
      __pyx_t_3 = *__pyx_t_2;
      ++__pyx_t_2;
      __pyx_9genexpr22__pyx_v_s = __pyx_t_3;
      __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_9genexpr22__pyx_v_s); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1069, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(PySet_Add(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(1, 1069, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1063
 *             )
 * 
 *     def get_symbols(self):             # <<<<<<<<<<<<<<
 *         """Return a set of symbols used on the transitions in NFA.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_symbols", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1071
 *         return {s for s in symbols}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_complete(cls, Nfa lhs, alph.Alphabet alphabet):
 *         """Test if automaton is complete
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_138is_complete(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_137is_complete, "Test if automaton is complete\n\n        :param Nf lhs: tested automaton\n        :param OnTheFlyAlphabet alphabet: alphabet of the automaton\n        :return: true if the automaton is complete\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_138is_complete = {"is_complete", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_138is_complete, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_137is_complete};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_138is_complete(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_complete (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_alphabet,0};
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1071, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_alphabet)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1071, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("is_complete", 1, 2, 2, 1); __PYX_ERR(1, 1071, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "is_complete") < 0)) __PYX_ERR(1, 1071, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_alphabet = ((struct __pyx_obj_7libmata_9alphabets_Alphabet *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_complete", 1, 2, 2, __pyx_nargs); __PYX_ERR(1, 1071, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_complete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 1072, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_alphabet), __pyx_ptype_7libmata_9alphabets_Alphabet, 1, "alphabet", 0))) __PYX_ERR(1, 1072, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_137is_complete(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_alphabet);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_137is_complete(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, struct __pyx_obj_7libmata_9alphabets_Alphabet *__pyx_v_alphabet) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Mata::Alphabet *__pyx_t_1;
  bool __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_complete", 0);

  /* "libmata/nfa/nfa.pyx":1079
 *         :return: true if the automaton is complete
 *         """
 *         return mata_nfa.is_complete(             # <<<<<<<<<<<<<<
 *             dereference(lhs.thisptr.get()),
 *             <CAlphabet&>dereference(alphabet.as_base())
 */
  __Pyx_XDECREF(__pyx_r);

  /* "libmata/nfa/nfa.pyx":1081
 *         return mata_nfa.is_complete(
 *             dereference(lhs.thisptr.get()),
 *             <CAlphabet&>dereference(alphabet.as_base())             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7libmata_9alphabets_Alphabet *)__pyx_v_alphabet->__pyx_vtab)->as_base(__pyx_v_alphabet); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1081, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1079
 *         :return: true if the automaton is complete
 *         """
 *         return mata_nfa.is_complete(             # <<<<<<<<<<<<<<
 *             dereference(lhs.thisptr.get()),
 *             <CAlphabet&>dereference(alphabet.as_base())
 */
  try {
    __pyx_t_2 = Mata::Nfa::is_complete((*__pyx_v_lhs->thisptr.get()), ((Mata::Alphabet &)(*__pyx_t_1)));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 1079, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1071
 *         return {s for s in symbols}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_complete(cls, Nfa lhs, alph.Alphabet alphabet):
 *         """Test if automaton is complete
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_complete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1084
 *         )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_in_lang(cls, Nfa lhs, vector[Symbol] word):
 *         """Tests if word is in language
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_140is_in_lang(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_139is_in_lang, "Tests if word is in language\n\n        :param Nfa lhs: tested automaton\n        :param vector[Symbol] word: tested word\n        :return: true if word is in language of automaton lhs\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_140is_in_lang = {"is_in_lang", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_140is_in_lang, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_139is_in_lang};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_140is_in_lang(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  std::vector<Mata::Symbol>  __pyx_v_word;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_in_lang (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_word,0};
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1084, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_word)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1084, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("is_in_lang", 1, 2, 2, 1); __PYX_ERR(1, 1084, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "is_in_lang") < 0)) __PYX_ERR(1, 1084, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_word = __pyx_convert_vector_from_py_Mata_3a__3a_Symbol(values[1]); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1085, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_in_lang", 1, 2, 2, __pyx_nargs); __PYX_ERR(1, 1084, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_in_lang", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 1085, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_139is_in_lang(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __PYX_STD_MOVE_IF_SUPPORTED(__pyx_v_word));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_139is_in_lang(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, std::vector<Mata::Symbol>  __pyx_v_word) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_run = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_in_lang", 0);

  /* "libmata/nfa/nfa.pyx":1092
 *         :return: true if word is in language of automaton lhs
 *         """
 *         run = Run()             # <<<<<<<<<<<<<<
 *         run.thisptr.word = word
 *         return mata_nfa.is_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Run)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_run = ((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":1093
 *         """
 *         run = Run()
 *         run.thisptr.word = word             # <<<<<<<<<<<<<<
 *         return mata_nfa.is_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))
 * 
 */
  __pyx_v_run->thisptr->word = __pyx_v_word;

  /* "libmata/nfa/nfa.pyx":1094
 *         run = Run()
 *         run.thisptr.word = word
 *         return mata_nfa.is_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(Mata::Nfa::is_in_lang((*__pyx_v_lhs->thisptr.get()), (*__pyx_v_run->thisptr))); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1084
 *         )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_in_lang(cls, Nfa lhs, vector[Symbol] word):
 *         """Tests if word is in language
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_in_lang", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_run);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1096
 *         return mata_nfa.is_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_prefix_in_lang(cls, Nfa lhs, vector[Symbol] word):
 *         """Test if any prefix of the word is in the language
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_142is_prefix_in_lang(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_141is_prefix_in_lang, "Test if any prefix of the word is in the language\n\n        :param Nfa lhs: tested automaton\n        :param vector[Symbol] word: tested word\n        :return: true if any prefix of word is in language of automaton lhs\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_142is_prefix_in_lang = {"is_prefix_in_lang", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_142is_prefix_in_lang, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_141is_prefix_in_lang};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_142is_prefix_in_lang(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  std::vector<Mata::Symbol>  __pyx_v_word;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_prefix_in_lang (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_word,0};
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1096, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_word)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1096, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("is_prefix_in_lang", 1, 2, 2, 1); __PYX_ERR(1, 1096, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "is_prefix_in_lang") < 0)) __PYX_ERR(1, 1096, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_word = __pyx_convert_vector_from_py_Mata_3a__3a_Symbol(values[1]); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1097, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_prefix_in_lang", 1, 2, 2, __pyx_nargs); __PYX_ERR(1, 1096, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_prefix_in_lang", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 1097, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_141is_prefix_in_lang(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __PYX_STD_MOVE_IF_SUPPORTED(__pyx_v_word));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_141is_prefix_in_lang(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, std::vector<Mata::Symbol>  __pyx_v_word) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_run = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_prefix_in_lang", 0);

  /* "libmata/nfa/nfa.pyx":1104
 *         :return: true if any prefix of word is in language of automaton lhs
 *         """
 *         run = Run()             # <<<<<<<<<<<<<<
 *         run.thisptr.word = word
 *         return mata_nfa.is_prfx_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Run)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_run = ((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":1105
 *         """
 *         run = Run()
 *         run.thisptr.word = word             # <<<<<<<<<<<<<<
 *         return mata_nfa.is_prfx_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))
 * 
 */
  __pyx_v_run->thisptr->word = __pyx_v_word;

  /* "libmata/nfa/nfa.pyx":1106
 *         run = Run()
 *         run.thisptr.word = word
 *         return mata_nfa.is_prfx_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(Mata::Nfa::is_prfx_in_lang((*__pyx_v_lhs->thisptr.get()), (*__pyx_v_run->thisptr))); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1096
 *         return mata_nfa.is_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_prefix_in_lang(cls, Nfa lhs, vector[Symbol] word):
 *         """Test if any prefix of the word is in the language
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.is_prefix_in_lang", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_run);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1108
 *         return mata_nfa.is_prfx_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def accepts_epsilon(cls, Nfa lhs):
 *         """Tests if automaton accepts epsilon
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_144accepts_epsilon(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_143accepts_epsilon, "Tests if automaton accepts epsilon\n\n        :param Nfa lhs: tested automaton\n        :return: true if automaton accepts epsilon\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_144accepts_epsilon = {"accepts_epsilon", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_144accepts_epsilon, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_143accepts_epsilon};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_144accepts_epsilon(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accepts_epsilon (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1108, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "accepts_epsilon") < 0)) __PYX_ERR(1, 1108, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accepts_epsilon", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1108, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.accepts_epsilon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 1109, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_143accepts_epsilon(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_143accepts_epsilon(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs) {
  Mata::Nfa::State __pyx_v_state;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Mata::Util::SparseSet<Mata::Nfa::State> ::iterator __pyx_t_1;
  Mata::Util::SparseSet<Mata::Nfa::State>  __pyx_t_2;
  Mata::Nfa::State __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("accepts_epsilon", 0);

  /* "libmata/nfa/nfa.pyx":1115
 *         :return: true if automaton accepts epsilon
 *         """
 *         for state in lhs.thisptr.get().initial:             # <<<<<<<<<<<<<<
 *             if lhs.has_final_state(state):
 *                 return True
 */
  __pyx_t_2 = __pyx_v_lhs->thisptr.get()->initial;
  __pyx_t_1 = __pyx_t_2.begin();
  for (;;) {
    if (!(__pyx_t_1 != __pyx_t_2.end())) break;
    __pyx_t_3 = *__pyx_t_1;
    ++__pyx_t_1;
    __pyx_v_state = __pyx_t_3;

    /* "libmata/nfa/nfa.pyx":1116
 *         """
 *         for state in lhs.thisptr.get().initial:
 *             if lhs.has_final_state(state):             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_lhs), __pyx_n_s_has_final_state); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_v_state); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_6};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1116, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(1, 1116, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_9) {

      /* "libmata/nfa/nfa.pyx":1117
 *         for state in lhs.thisptr.get().initial:
 *             if lhs.has_final_state(state):
 *                 return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      goto __pyx_L0;

      /* "libmata/nfa/nfa.pyx":1116
 *         """
 *         for state in lhs.thisptr.get().initial:
 *             if lhs.has_final_state(state):             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
 */
    }

    /* "libmata/nfa/nfa.pyx":1115
 *         :return: true if automaton accepts epsilon
 *         """
 *         for state in lhs.thisptr.get().initial:             # <<<<<<<<<<<<<<
 *             if lhs.has_final_state(state):
 *                 return True
 */
  }

  /* "libmata/nfa/nfa.pyx":1118
 *             if lhs.has_final_state(state):
 *                 return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     # Helper functions
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1108
 *         return mata_nfa.is_prfx_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def accepts_epsilon(cls, Nfa lhs):
 *         """Tests if automaton accepts epsilon
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.accepts_epsilon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1122
 *     # Helper functions
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def get_word_for_path(cls, Nfa lhs, path):
 *         """For a given path (set of states) returns a corresponding word
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_146get_word_for_path(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_145get_word_for_path, "For a given path (set of states) returns a corresponding word\n\n        >>> mata_nfa.Nfa.get_word_for_path(lhs, [0, 1, 2])\n        ([1, 1], True)\n\n        :param Nfa lhs: source automaton\n        :param list path: list of states\n        :return: pair of word (list of symbols) and true or false, whether the search was successful\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_146get_word_for_path = {"get_word_for_path", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_146get_word_for_path, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_145get_word_for_path};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_146get_word_for_path(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = 0;
  PyObject *__pyx_v_path = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_word_for_path (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lhs,&__pyx_n_s_path,0};
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lhs)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1122, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_path)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1122, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("get_word_for_path", 1, 2, 2, 1); __PYX_ERR(1, 1122, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_word_for_path") < 0)) __PYX_ERR(1, 1122, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)values[0]);
    __pyx_v_path = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_word_for_path", 1, 2, 2, __pyx_nargs); __PYX_ERR(1, 1122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_word_for_path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lhs), __pyx_ptype_7libmata_3nfa_3nfa_Nfa, 1, "lhs", 0))) __PYX_ERR(1, 1123, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_145get_word_for_path(((PyTypeObject*)__pyx_v_cls), __pyx_v_lhs, __pyx_v_path);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_145get_word_for_path(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs, PyObject *__pyx_v_path) {
  std::pair<Mata::Nfa::Run,bool>  __pyx_v_result;
  struct __pyx_obj_7libmata_3nfa_3nfa_Run *__pyx_v_input = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_word_for_path", 0);

  /* "libmata/nfa/nfa.pyx":1134
 *         """
 *         cdef pair[CRun, bool] result
 *         input = Run()             # <<<<<<<<<<<<<<
 *         input.path = path
 *         result = mata_nfa.get_word_for_path(dereference(lhs.thisptr.get()), dereference(input.thisptr))
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Run)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_input = ((struct __pyx_obj_7libmata_3nfa_3nfa_Run *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":1135
 *         cdef pair[CRun, bool] result
 *         input = Run()
 *         input.path = path             # <<<<<<<<<<<<<<
 *         result = mata_nfa.get_word_for_path(dereference(lhs.thisptr.get()), dereference(input.thisptr))
 *         return result.first.word, result.second
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_input), __pyx_n_s_path, __pyx_v_path) < 0) __PYX_ERR(1, 1135, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1136
 *         input = Run()
 *         input.path = path
 *         result = mata_nfa.get_word_for_path(dereference(lhs.thisptr.get()), dereference(input.thisptr))             # <<<<<<<<<<<<<<
 *         return result.first.word, result.second
 * 
 */
  __pyx_v_result = Mata::Nfa::get_word_for_path((*__pyx_v_lhs->thisptr.get()), (*__pyx_v_input->thisptr));

  /* "libmata/nfa/nfa.pyx":1137
 *         input.path = path
 *         result = mata_nfa.get_word_for_path(dereference(lhs.thisptr.get()), dereference(input.thisptr))
 *         return result.first.word, result.second             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_Mata_3a__3a_Symbol(__pyx_v_result.first.word); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_result.second); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1122
 *     # Helper functions
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def get_word_for_path(cls, Nfa lhs, path):
 *         """For a given path (set of states) returns a corresponding word
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.get_word_for_path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_input);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1139
 *         return result.first.word, result.second
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def encode_word(cls, string_to_symbol, word):
 *         """Encodes word based on a string to symbol map
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_148encode_word(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_3Nfa_147encode_word, "Encodes word based on a string to symbol map\n\n        >>> mata_nfa.Nfa.encode_word({'a': 1, 'b': 2, \"c\": 0}, \"abca\")\n        [1, 2, 0, 1]\n\n        :param dict string_to_symbol: dictionary of strings to integers\n        :param word: list of strings representing a encoded word\n        :return:\n        ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_148encode_word = {"encode_word", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_148encode_word, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_147encode_word};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_148encode_word(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_string_to_symbol = 0;
  PyObject *__pyx_v_word = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("encode_word (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_string_to_symbol,&__pyx_n_s_word,0};
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_string_to_symbol)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1139, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_word)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1139, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("encode_word", 1, 2, 2, 1); __PYX_ERR(1, 1139, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "encode_word") < 0)) __PYX_ERR(1, 1139, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_string_to_symbol = values[0];
    __pyx_v_word = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("encode_word", 1, 2, 2, __pyx_nargs); __PYX_ERR(1, 1139, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.encode_word", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_147encode_word(((PyTypeObject*)__pyx_v_cls), __pyx_v_string_to_symbol, __pyx_v_word);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_147encode_word(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_string_to_symbol, PyObject *__pyx_v_word) {
  Mata::Nfa::Run __pyx_v_result;
  PyObject *__pyx_9genexpr23__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr23__pyx_v_v = NULL;
  PyObject *__pyx_9genexpr24__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  Mata::StringToSymbolMap __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  std::vector<std::string>  __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_word", 0);

  /* "libmata/nfa/nfa.pyx":1150
 *         :return:
 *         """
 *         result = mata_nfa.encode_word(             # <<<<<<<<<<<<<<
 *             {k.encode('utf-8'): v for (k, v) in string_to_symbol.items()},
 *             [s.encode('utf-8') for s in word]
 */
  { /* enter inner scope */

    /* "libmata/nfa/nfa.pyx":1151
 *         """
 *         result = mata_nfa.encode_word(
 *             {k.encode('utf-8'): v for (k, v) in string_to_symbol.items()},             # <<<<<<<<<<<<<<
 *             [s.encode('utf-8') for s in word]
 *         )
 */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1151, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_string_to_symbol == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(1, 1151, __pyx_L5_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_string_to_symbol, 0, __pyx_n_s_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1151, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(1, 1151, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_9genexpr23__pyx_v_k, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr23__pyx_v_v, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr23__pyx_v_k, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1151, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_kp_u_utf_8};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1151, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_7, (PyObject*)__pyx_9genexpr23__pyx_v_v))) __PYX_ERR(1, 1151, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_k); __pyx_9genexpr23__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_v); __pyx_9genexpr23__pyx_v_v = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_k); __pyx_9genexpr23__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_v); __pyx_9genexpr23__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_t_10 = __pyx_convert_unordered_map_from_py_std_3a__3a_string__and_Mata_3a__3a_Symbol(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1151, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  { /* enter inner scope */

    /* "libmata/nfa/nfa.pyx":1152
 *         result = mata_nfa.encode_word(
 *             {k.encode('utf-8'): v for (k, v) in string_to_symbol.items()},
 *             [s.encode('utf-8') for s in word]             # <<<<<<<<<<<<<<
 *         )
 *         return result.word
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1152, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_v_word)) || PyTuple_CheckExact(__pyx_v_word)) {
      __pyx_t_2 = __pyx_v_word; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_word); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1152, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1152, __pyx_L11_error)
    }
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(1, 1152, __pyx_L11_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1152, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(1, 1152, __pyx_L11_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1152, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_11(__pyx_t_2);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(1, 1152, __pyx_L11_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr24__pyx_v_s, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr24__pyx_v_s, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1152, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_5 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_kp_u_utf_8};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1152, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_7))) __PYX_ERR(1, 1152, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_s); __pyx_9genexpr24__pyx_v_s = 0;
    goto __pyx_L15_exit_scope;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_s); __pyx_9genexpr24__pyx_v_s = 0;
    goto __pyx_L1_error;
    __pyx_L15_exit_scope:;
  } /* exit inner scope */
  __pyx_t_12 = __pyx_convert_vector_from_py_std_3a__3a_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":1150
 *         :return:
 *         """
 *         result = mata_nfa.encode_word(             # <<<<<<<<<<<<<<
 *             {k.encode('utf-8'): v for (k, v) in string_to_symbol.items()},
 *             [s.encode('utf-8') for s in word]
 */
  __pyx_v_result = Mata::Nfa::encode_word(__pyx_t_10, __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_12));

  /* "libmata/nfa/nfa.pyx":1154
 *             [s.encode('utf-8') for s in word]
 *         )
 *         return result.word             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_Mata_3a__3a_Symbol(__pyx_v_result.word); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1139
 *         return result.first.word, result.second
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def encode_word(cls, string_to_symbol, word):
 *         """Encodes word based on a string to symbol map
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.encode_word", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_v);
  __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_150__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_150__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_150__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_150__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_149__reduce_cython__(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_149__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_152__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3Nfa_152__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_152__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3Nfa_152__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_3Nfa_151__setstate_cython__(((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_3Nfa_151__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.Nfa.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1158
 * 
 * 
 * cdef subset_map_to_dictionary(umap[StateSet, State] subset_map):             # <<<<<<<<<<<<<<
 *     """Helper function that translates the unordered map to dictionary
 * 
 */

static PyObject *__pyx_f_7libmata_3nfa_3nfa_subset_map_to_dictionary(std::unordered_map<Mata::Nfa::StateSet,Mata::Nfa::State>  __pyx_v_subset_map) {
  PyObject *__pyx_v_result = NULL;
  std::unordered_map<Mata::Nfa::StateSet,Mata::Nfa::State> ::iterator __pyx_v_it;
  std::vector<Mata::Nfa::State>  __pyx_v_key;
  Mata::Nfa::State __pyx_v_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Mata::Nfa::State __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("subset_map_to_dictionary", 0);

  /* "libmata/nfa/nfa.pyx":1164
 *     :return: subset_map as dictionary
 *     """
 *     result = {}             # <<<<<<<<<<<<<<
 *     cdef umap[StateSet, State].iterator it = subset_map.begin()
 *     while it != subset_map.end():
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":1165
 *     """
 *     result = {}
 *     cdef umap[StateSet, State].iterator it = subset_map.begin()             # <<<<<<<<<<<<<<
 *     while it != subset_map.end():
 *         key = dereference(it).first.ToVector()
 */
  __pyx_v_it = __pyx_v_subset_map.begin();

  /* "libmata/nfa/nfa.pyx":1166
 *     result = {}
 *     cdef umap[StateSet, State].iterator it = subset_map.begin()
 *     while it != subset_map.end():             # <<<<<<<<<<<<<<
 *         key = dereference(it).first.ToVector()
 *         value = dereference(it).second
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_it != __pyx_v_subset_map.end());
    if (!__pyx_t_2) break;

    /* "libmata/nfa/nfa.pyx":1167
 *     cdef umap[StateSet, State].iterator it = subset_map.begin()
 *     while it != subset_map.end():
 *         key = dereference(it).first.ToVector()             # <<<<<<<<<<<<<<
 *         value = dereference(it).second
 *         result[tuple(sorted(key))] = value
 */
    __pyx_v_key = (*__pyx_v_it).first.ToVector();

    /* "libmata/nfa/nfa.pyx":1168
 *     while it != subset_map.end():
 *         key = dereference(it).first.ToVector()
 *         value = dereference(it).second             # <<<<<<<<<<<<<<
 *         result[tuple(sorted(key))] = value
 *         postinc(it)
 */
    __pyx_t_3 = (*__pyx_v_it).second;
    __pyx_v_value = __pyx_t_3;

    /* "libmata/nfa/nfa.pyx":1169
 *         key = dereference(it).first.ToVector()
 *         value = dereference(it).second
 *         result[tuple(sorted(key))] = value             # <<<<<<<<<<<<<<
 *         postinc(it)
 *     return result
 */
    __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_convert_vector_to_py_Mata_3a__3a_Nfa_3a__3a_State(__pyx_v_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PySequence_List(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_4 = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_7 = PyList_Sort(__pyx_t_4); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 1169, __pyx_L1_error)
    __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely((PyDict_SetItem(__pyx_v_result, __pyx_t_6, __pyx_t_1) < 0))) __PYX_ERR(1, 1169, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "libmata/nfa/nfa.pyx":1170
 *         value = dereference(it).second
 *         result[tuple(sorted(key))] = value
 *         postinc(it)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
    (void)((__pyx_v_it++));
  }

  /* "libmata/nfa/nfa.pyx":1171
 *         result[tuple(sorted(key))] = value
 *         postinc(it)
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1158
 * 
 * 
 * cdef subset_map_to_dictionary(umap[StateSet, State] subset_map):             # <<<<<<<<<<<<<<
 *     """Helper function that translates the unordered map to dictionary
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("libmata.nfa.nfa.subset_map_to_dictionary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1175
 * 
 * # Temporary for testing
 * def divisible_by(k: int):             # <<<<<<<<<<<<<<
 *     """Constructs automaton accepting strings containing ones divisible by "k"."""
 *     assert k > 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3divisible_by(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_2divisible_by, "Constructs automaton accepting strings containing ones divisible by \"k\".");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_3divisible_by = {"divisible_by", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3divisible_by, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_2divisible_by};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_3divisible_by(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_k = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("divisible_by (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_k,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_k)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1175, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "divisible_by") < 0)) __PYX_ERR(1, 1175, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_k = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("divisible_by", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1175, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.divisible_by", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_k), (&PyInt_Type), 0, "k", 1))) __PYX_ERR(1, 1175, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_2divisible_by(__pyx_self, __pyx_v_k);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_2divisible_by(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_k) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *__pyx_v_lhs = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("divisible_by", 0);

  /* "libmata/nfa/nfa.pyx":1177
 * def divisible_by(k: int):
 *     """Constructs automaton accepting strings containing ones divisible by "k"."""
 *     assert k > 1             # <<<<<<<<<<<<<<
 *     lhs = Nfa(k+1)
 *     lhs.make_initial_state(0)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_k, __pyx_int_1, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1177, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 1177, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(1, 1177, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(1, 1177, __pyx_L1_error)
  #endif

  /* "libmata/nfa/nfa.pyx":1178
 *     """Constructs automaton accepting strings containing ones divisible by "k"."""
 *     assert k > 1
 *     lhs = Nfa(k+1)             # <<<<<<<<<<<<<<
 *     lhs.make_initial_state(0)
 *     lhs.add_transition(0, 0, 0)
 */
  __pyx_t_1 = PyNumber_Add(__pyx_v_k, __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_lhs = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "libmata/nfa/nfa.pyx":1179
 *     assert k > 1
 *     lhs = Nfa(k+1)
 *     lhs.make_initial_state(0)             # <<<<<<<<<<<<<<
 *     lhs.add_transition(0, 0, 0)
 *     for i in range(1, k + 1):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_lhs), __pyx_n_s_make_initial_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_int_0};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "libmata/nfa/nfa.pyx":1180
 *     lhs = Nfa(k+1)
 *     lhs.make_initial_state(0)
 *     lhs.add_transition(0, 0, 0)             # <<<<<<<<<<<<<<
 *     for i in range(1, k + 1):
 *         lhs.add_transition(i - 1, 1, i)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_lhs), __pyx_n_s_add_transition); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "libmata/nfa/nfa.pyx":1181
 *     lhs.make_initial_state(0)
 *     lhs.add_transition(0, 0, 0)
 *     for i in range(1, k + 1):             # <<<<<<<<<<<<<<
 *         lhs.add_transition(i - 1, 1, i)
 *         lhs.add_transition(i, 0, i)
 */
  __pyx_t_1 = PyNumber_Add(__pyx_v_k, __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1181, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(1, 1181, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1181, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(1, 1181, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1181, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_7(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 1181, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "libmata/nfa/nfa.pyx":1182
 *     lhs.add_transition(0, 0, 0)
 *     for i in range(1, k + 1):
 *         lhs.add_transition(i - 1, 1, i)             # <<<<<<<<<<<<<<
 *         lhs.add_transition(i, 0, i)
 *     lhs.add_transition(k, 1, 1)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_lhs), __pyx_n_s_add_transition); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = __Pyx_PyInt_SubtractObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_9, __pyx_t_8, __pyx_int_1, __pyx_v_i};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1182, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "libmata/nfa/nfa.pyx":1183
 *     for i in range(1, k + 1):
 *         lhs.add_transition(i - 1, 1, i)
 *         lhs.add_transition(i, 0, i)             # <<<<<<<<<<<<<<
 *     lhs.add_transition(k, 1, 1)
 *     lhs.make_final_state(k)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_lhs), __pyx_n_s_add_transition); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_8, __pyx_v_i, __pyx_int_0, __pyx_v_i};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1183, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "libmata/nfa/nfa.pyx":1181
 *     lhs.make_initial_state(0)
 *     lhs.add_transition(0, 0, 0)
 *     for i in range(1, k + 1):             # <<<<<<<<<<<<<<
 *         lhs.add_transition(i - 1, 1, i)
 *         lhs.add_transition(i, 0, i)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "libmata/nfa/nfa.pyx":1184
 *         lhs.add_transition(i - 1, 1, i)
 *         lhs.add_transition(i, 0, i)
 *     lhs.add_transition(k, 1, 1)             # <<<<<<<<<<<<<<
 *     lhs.make_final_state(k)
 *     return lhs
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_lhs), __pyx_n_s_add_transition); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_v_k, __pyx_int_1, __pyx_int_1};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "libmata/nfa/nfa.pyx":1185
 *         lhs.add_transition(i, 0, i)
 *     lhs.add_transition(k, 1, 1)
 *     lhs.make_final_state(k)             # <<<<<<<<<<<<<<
 *     return lhs
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_lhs), __pyx_n_s_make_final_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_k};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "libmata/nfa/nfa.pyx":1186
 *     lhs.add_transition(k, 1, 1)
 *     lhs.make_final_state(k)
 *     return lhs             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_lhs);
  __pyx_r = ((PyObject *)__pyx_v_lhs);
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1175
 * 
 * # Temporary for testing
 * def divisible_by(k: int):             # <<<<<<<<<<<<<<
 *     """Constructs automaton accepting strings containing ones divisible by "k"."""
 *     assert k > 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("libmata.nfa.nfa.divisible_by", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_lhs);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1189
 * 
 * 
 * def run_safely_external_command(cmd: str, check_results=True, quiet=True, timeout=None, **kwargs):             # <<<<<<<<<<<<<<
 *     """Safely runs the piped command, without executing of the shell
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5run_safely_external_command(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_4run_safely_external_command, "Safely runs the piped command, without executing of the shell\n\n    Courtesy of: https://blog.avinetworks.com/tech/python-best-practices\n\n    :param str cmd: string with command that we are executing\n    :param bool check_results: check correct command exit code and raise exception in case of fail\n    :param bool quiet: if set to False, then it will print the output of the command\n    :param int timeout: timeout of the command\n    :param dict kwargs: additional args to subprocess call\n    :return: returned standard output and error\n    :raises subprocess.CalledProcessError: when any of the piped commands fails\n    ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_5run_safely_external_command = {"run_safely_external_command", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_5run_safely_external_command, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_4run_safely_external_command};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_5run_safely_external_command(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_cmd = 0;
  PyObject *__pyx_v_check_results = 0;
  PyObject *__pyx_v_quiet = 0;
  PyObject *__pyx_v_timeout = 0;
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run_safely_external_command (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cmd,&__pyx_n_s_check_results,&__pyx_n_s_quiet,&__pyx_n_s_timeout,0};
    PyObject* values[4] = {0,0,0,0};
    values[1] = ((PyObject *)((PyObject *)Py_True));
    values[2] = ((PyObject *)((PyObject *)Py_True));
    values[3] = ((PyObject *)((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_cmd)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1189, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_check_results);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1189, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_quiet);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1189, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[3] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1189, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, kwd_pos_args, "run_safely_external_command") < 0)) __PYX_ERR(1, 1189, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cmd = ((PyObject*)values[0]);
    __pyx_v_check_results = values[1];
    __pyx_v_quiet = values[2];
    __pyx_v_timeout = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("run_safely_external_command", 0, 1, 4, __pyx_nargs); __PYX_ERR(1, 1189, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("libmata.nfa.nfa.run_safely_external_command", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_cmd), (&PyUnicode_Type), 0, "cmd", 1))) __PYX_ERR(1, 1189, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_4run_safely_external_command(__pyx_self, __pyx_v_cmd, __pyx_v_check_results, __pyx_v_quiet, __pyx_v_timeout, __pyx_v_kwargs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_4run_safely_external_command(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cmd, PyObject *__pyx_v_check_results, PyObject *__pyx_v_quiet, PyObject *__pyx_v_timeout, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_unpiped_commands = NULL;
  PyObject *__pyx_v_cmd_no = NULL;
  PyObject *__pyx_v_objects = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_executed_command = NULL;
  PyObject *__pyx_v_stdin = NULL;
  PyObject *__pyx_v_stderr = NULL;
  PyObject *__pyx_v_piped_command = NULL;
  PyObject *__pyx_v_cmdout = NULL;
  PyObject *__pyx_v_cmderr = NULL;
  PyObject *__pyx_v_process = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *(*__pyx_t_16)(PyObject *);
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_safely_external_command", 0);

  /* "libmata/nfa/nfa.pyx":1203
 *     """
 *     # Split
 *     unpiped_commands = list(map(str.strip, cmd.split(" | ")))             # <<<<<<<<<<<<<<
 *     cmd_no = len(unpiped_commands)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyUnicode_Type)), __pyx_n_s_strip); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyUnicode_Split(__pyx_v_cmd, __Pyx_NoneAsNull(__pyx_kp_u__15), -1L); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PySequence_ListKeepNew(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_unpiped_commands = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "libmata/nfa/nfa.pyx":1204
 *     # Split
 *     unpiped_commands = list(map(str.strip, cmd.split(" | ")))
 *     cmd_no = len(unpiped_commands)             # <<<<<<<<<<<<<<
 * 
 *     # Run the command through pipes
 */
  __pyx_t_4 = PyList_GET_SIZE(__pyx_v_unpiped_commands); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1204, __pyx_L1_error)
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_cmd_no = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "libmata/nfa/nfa.pyx":1207
 * 
 *     # Run the command through pipes
 *     objects: List[subprocess.Popen] = []             # <<<<<<<<<<<<<<
 *     for i in range(cmd_no):
 *         executed_command = shlex.split(unpiped_commands[i])
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_objects = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "libmata/nfa/nfa.pyx":1208
 *     # Run the command through pipes
 *     objects: List[subprocess.Popen] = []
 *     for i in range(cmd_no):             # <<<<<<<<<<<<<<
 *         executed_command = shlex.split(unpiped_commands[i])
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_cmd_no); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1208, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(1, 1208, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1208, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(1, 1208, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1208, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 1208, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "libmata/nfa/nfa.pyx":1209
 *     objects: List[subprocess.Popen] = []
 *     for i in range(cmd_no):
 *         executed_command = shlex.split(unpiped_commands[i])             # <<<<<<<<<<<<<<
 * 
 *         # set streams
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_shlex); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_split); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_unpiped_commands, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_1};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_executed_command, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "libmata/nfa/nfa.pyx":1212
 * 
 *         # set streams
 *         stdin = None if i == 0 else objects[i-1].stdout             # <<<<<<<<<<<<<<
 *         stderr = subprocess.STDOUT if i < (cmd_no - 1) else subprocess.PIPE
 * 
 */
    __pyx_t_9 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_i, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(1, 1212, __pyx_L1_error)
    if (__pyx_t_9) {
      __Pyx_INCREF(Py_None);
      __pyx_t_3 = Py_None;
    } else {
      __pyx_t_6 = __Pyx_PyInt_SubtractObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_objects, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_stdout); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_3 = __pyx_t_6;
      __pyx_t_6 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_stdin, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "libmata/nfa/nfa.pyx":1213
 *         # set streams
 *         stdin = None if i == 0 else objects[i-1].stdout
 *         stderr = subprocess.STDOUT if i < (cmd_no - 1) else subprocess.PIPE             # <<<<<<<<<<<<<<
 * 
 *         # run the piped command and close the previous one
 */
    __pyx_t_6 = __Pyx_PyInt_SubtractObjC(__pyx_v_cmd_no, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_i, __pyx_t_6, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1213, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(1, 1213, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_9) {
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_subprocess); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_STDOUT); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_3 = __pyx_t_6;
      __pyx_t_6 = 0;
    } else {
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_subprocess); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_PIPE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_3 = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_stderr, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "libmata/nfa/nfa.pyx":1216
 * 
 *         # run the piped command and close the previous one
 *         piped_command = subprocess.Popen(             # <<<<<<<<<<<<<<
 *             executed_command,
 *             shell=False, stdin=stdin, stdout=subprocess.PIPE, stderr=stderr, **kwargs
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_subprocess); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Popen); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "libmata/nfa/nfa.pyx":1217
 *         # run the piped command and close the previous one
 *         piped_command = subprocess.Popen(
 *             executed_command,             # <<<<<<<<<<<<<<
 *             shell=False, stdin=stdin, stdout=subprocess.PIPE, stderr=stderr, **kwargs
 *         )
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_executed_command);
    __Pyx_GIVEREF(__pyx_v_executed_command);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_executed_command);

    /* "libmata/nfa/nfa.pyx":1218
 *         piped_command = subprocess.Popen(
 *             executed_command,
 *             shell=False, stdin=stdin, stdout=subprocess.PIPE, stderr=stderr, **kwargs             # <<<<<<<<<<<<<<
 *         )
 *         if i != 0:
 */
    __pyx_t_7 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_shell, Py_False) < 0) __PYX_ERR(1, 1218, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_stdin, __pyx_v_stdin) < 0) __PYX_ERR(1, 1218, __pyx_L1_error)
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_subprocess); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_PIPE); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_stdout, __pyx_t_11) < 0) __PYX_ERR(1, 1218, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_stderr, __pyx_v_stderr) < 0) __PYX_ERR(1, 1218, __pyx_L1_error)
    __pyx_t_6 = __pyx_t_7;
    __pyx_t_7 = 0;
    if (__Pyx_MergeKeywords(__pyx_t_6, __pyx_v_kwargs) < 0) __PYX_ERR(1, 1218, __pyx_L1_error)

    /* "libmata/nfa/nfa.pyx":1216
 * 
 *         # run the piped command and close the previous one
 *         piped_command = subprocess.Popen(             # <<<<<<<<<<<<<<
 *             executed_command,
 *             shell=False, stdin=stdin, stdout=subprocess.PIPE, stderr=stderr, **kwargs
 */
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_piped_command, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "libmata/nfa/nfa.pyx":1220
 *             shell=False, stdin=stdin, stdout=subprocess.PIPE, stderr=stderr, **kwargs
 *         )
 *         if i != 0:             # <<<<<<<<<<<<<<
 *             objects[i-1].stdout.close()  # type: ignore
 *         objects.append(piped_command)
 */
    __pyx_t_9 = (__Pyx_PyInt_BoolNeObjC(__pyx_v_i, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(1, 1220, __pyx_L1_error)
    if (__pyx_t_9) {

      /* "libmata/nfa/nfa.pyx":1221
 *         )
 *         if i != 0:
 *             objects[i-1].stdout.close()  # type: ignore             # <<<<<<<<<<<<<<
 *         objects.append(piped_command)
 * 
 */
      __pyx_t_6 = __Pyx_PyInt_SubtractObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1221, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_objects, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1221, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_stdout); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1221, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_close); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1221, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_8 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[1] = {__pyx_t_6, };
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1221, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "libmata/nfa/nfa.pyx":1220
 *             shell=False, stdin=stdin, stdout=subprocess.PIPE, stderr=stderr, **kwargs
 *         )
 *         if i != 0:             # <<<<<<<<<<<<<<
 *             objects[i-1].stdout.close()  # type: ignore
 *         objects.append(piped_command)
 */
    }

    /* "libmata/nfa/nfa.pyx":1222
 *         if i != 0:
 *             objects[i-1].stdout.close()  # type: ignore
 *         objects.append(piped_command)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_objects, __pyx_v_piped_command); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(1, 1222, __pyx_L1_error)

    /* "libmata/nfa/nfa.pyx":1208
 *     # Run the command through pipes
 *     objects: List[subprocess.Popen] = []
 *     for i in range(cmd_no):             # <<<<<<<<<<<<<<
 *         executed_command = shlex.split(unpiped_commands[i])
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":1224
 *         objects.append(piped_command)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         # communicate with the last piped object
 *         cmdout, cmderr = objects[-1].communicate(timeout=timeout)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
    __Pyx_XGOTREF(__pyx_t_13);
    __Pyx_XGOTREF(__pyx_t_14);
    __Pyx_XGOTREF(__pyx_t_15);
    /*try:*/ {

      /* "libmata/nfa/nfa.pyx":1226
 *     try:
 *         # communicate with the last piped object
 *         cmdout, cmderr = objects[-1].communicate(timeout=timeout)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(len(objects) - 1):
 */
      __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_objects, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1226, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_communicate); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1226, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1226, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_timeout, __pyx_v_timeout) < 0) __PYX_ERR(1, 1226, __pyx_L7_error)
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1226, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
        PyObject* sequence = __pyx_t_3;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(1, 1226, __pyx_L7_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        #else
        __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1226, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1226, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1226, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_16 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
        index = 0; __pyx_t_2 = __pyx_t_16(__pyx_t_6); if (unlikely(!__pyx_t_2)) goto __pyx_L13_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        index = 1; __pyx_t_7 = __pyx_t_16(__pyx_t_6); if (unlikely(!__pyx_t_7)) goto __pyx_L13_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_6), 2) < 0) __PYX_ERR(1, 1226, __pyx_L7_error)
        __pyx_t_16 = NULL;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L14_unpacking_done;
        __pyx_L13_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_16 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(1, 1226, __pyx_L7_error)
        __pyx_L14_unpacking_done:;
      }
      __pyx_v_cmdout = __pyx_t_2;
      __pyx_t_2 = 0;
      __pyx_v_cmderr = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "libmata/nfa/nfa.pyx":1228
 *         cmdout, cmderr = objects[-1].communicate(timeout=timeout)
 * 
 *         for i in range(len(objects) - 1):             # <<<<<<<<<<<<<<
 *             objects[i].wait(timeout=timeout)
 * 
 */
      __pyx_t_4 = PyList_GET_SIZE(__pyx_v_objects); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1228, __pyx_L7_error)
      __pyx_t_3 = PyInt_FromSsize_t((__pyx_t_4 - 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1228, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1228, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
        __pyx_t_3 = __pyx_t_7; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
        __pyx_t_5 = NULL;
      } else {
        __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1228, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1228, __pyx_L7_error)
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      for (;;) {
        if (likely(!__pyx_t_5)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_7 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(1, 1228, __pyx_L7_error)
            #else
            __pyx_t_7 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1228, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
          } else {
            if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(1, 1228, __pyx_L7_error)
            #else
            __pyx_t_7 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1228, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
          }
        } else {
          __pyx_t_7 = __pyx_t_5(__pyx_t_3);
          if (unlikely(!__pyx_t_7)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(1, 1228, __pyx_L7_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_7);
        }
        __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "libmata/nfa/nfa.pyx":1229
 * 
 *         for i in range(len(objects) - 1):
 *             objects[i].wait(timeout=timeout)             # <<<<<<<<<<<<<<
 * 
 *     except subprocess.TimeoutExpired:
 */
        __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_objects, __pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1229, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_wait); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1229, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1229, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_timeout, __pyx_v_timeout) < 0) __PYX_ERR(1, 1229, __pyx_L7_error)
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1229, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "libmata/nfa/nfa.pyx":1228
 *         cmdout, cmderr = objects[-1].communicate(timeout=timeout)
 * 
 *         for i in range(len(objects) - 1):             # <<<<<<<<<<<<<<
 *             objects[i].wait(timeout=timeout)
 * 
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "libmata/nfa/nfa.pyx":1224
 *         objects.append(piped_command)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         # communicate with the last piped object
 *         cmdout, cmderr = objects[-1].communicate(timeout=timeout)
 */
    }
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    goto __pyx_L12_try_end;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "libmata/nfa/nfa.pyx":1231
 *             objects[i].wait(timeout=timeout)
 * 
 *     except subprocess.TimeoutExpired:             # <<<<<<<<<<<<<<
 *         for process in objects:
 *             process.terminate()
 */
    __Pyx_ErrFetch(&__pyx_t_3, &__pyx_t_6, &__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_subprocess); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1231, __pyx_L9_except_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_TimeoutExpired); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1231, __pyx_L9_except_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_ErrRestore(__pyx_t_3, __pyx_t_6, __pyx_t_7);
    __pyx_t_3 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("libmata.nfa.nfa.run_safely_external_command", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_3) < 0) __PYX_ERR(1, 1231, __pyx_L9_except_error)
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_3);

      /* "libmata/nfa/nfa.pyx":1232
 * 
 *     except subprocess.TimeoutExpired:
 *         for process in objects:             # <<<<<<<<<<<<<<
 *             process.terminate()
 *         raise
 */
      __pyx_t_1 = __pyx_v_objects; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
      for (;;) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(1, 1232, __pyx_L9_except_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1232, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
        __Pyx_XDECREF_SET(__pyx_v_process, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "libmata/nfa/nfa.pyx":1233
 *     except subprocess.TimeoutExpired:
 *         for process in objects:
 *             process.terminate()             # <<<<<<<<<<<<<<
 *         raise
 * 
 */
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_process, __pyx_n_s_terminate); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1233, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_10 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_8 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[1] = {__pyx_t_10, };
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1233, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "libmata/nfa/nfa.pyx":1232
 * 
 *     except subprocess.TimeoutExpired:
 *         for process in objects:             # <<<<<<<<<<<<<<
 *             process.terminate()
 *         raise
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "libmata/nfa/nfa.pyx":1234
 *         for process in objects:
 *             process.terminate()
 *         raise             # <<<<<<<<<<<<<<
 * 
 *     # collect the return codes
 */
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ErrRestoreWithState(__pyx_t_7, __pyx_t_6, __pyx_t_3);
      __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_3 = 0; 
      __PYX_ERR(1, 1234, __pyx_L9_except_error)
    }
    goto __pyx_L9_except_error;

    /* "libmata/nfa/nfa.pyx":1224
 *         objects.append(piped_command)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         # communicate with the last piped object
 *         cmdout, cmderr = objects[-1].communicate(timeout=timeout)
 */
    __pyx_L9_except_error:;
    __Pyx_XGIVEREF(__pyx_t_13);
    __Pyx_XGIVEREF(__pyx_t_14);
    __Pyx_XGIVEREF(__pyx_t_15);
    __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
    goto __pyx_L1_error;
    __pyx_L12_try_end:;
  }

  /* "libmata/nfa/nfa.pyx":1237
 * 
 *     # collect the return codes
 *     if check_results:             # <<<<<<<<<<<<<<
 *         for i in range(cmd_no):
 *             if objects[i].returncode:
 */
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_check_results); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(1, 1237, __pyx_L1_error)
  if (__pyx_t_9) {

    /* "libmata/nfa/nfa.pyx":1238
 *     # collect the return codes
 *     if check_results:
 *         for i in range(cmd_no):             # <<<<<<<<<<<<<<
 *             if objects[i].returncode:
 *                 if not quiet and (cmdout or cmderr):
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_cmd_no); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_6 = __pyx_t_3; __Pyx_INCREF(__pyx_t_6); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1238, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1238, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(1, 1238, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1238, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(1, 1238, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1238, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_5(__pyx_t_6);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(1, 1238, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "libmata/nfa/nfa.pyx":1239
 *     if check_results:
 *         for i in range(cmd_no):
 *             if objects[i].returncode:             # <<<<<<<<<<<<<<
 *                 if not quiet and (cmdout or cmderr):
 *                     print(f"captured stdout: {cmdout.decode('utf-8')}", 'red')
 */
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_objects, __pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1239, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_returncode); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1239, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(1, 1239, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_9) {

        /* "libmata/nfa/nfa.pyx":1240
 *         for i in range(cmd_no):
 *             if objects[i].returncode:
 *                 if not quiet and (cmdout or cmderr):             # <<<<<<<<<<<<<<
 *                     print(f"captured stdout: {cmdout.decode('utf-8')}", 'red')
 *                     print(f"captured stderr: {cmderr.decode('utf-8')}", 'red')
 */
        __pyx_t_17 = __Pyx_PyObject_IsTrue(__pyx_v_quiet); if (unlikely((__pyx_t_17 < 0))) __PYX_ERR(1, 1240, __pyx_L1_error)
        __pyx_t_18 = (!__pyx_t_17);
        if (__pyx_t_18) {
        } else {
          __pyx_t_9 = __pyx_t_18;
          goto __pyx_L28_bool_binop_done;
        }
        __pyx_t_18 = __Pyx_PyObject_IsTrue(__pyx_v_cmdout); if (unlikely((__pyx_t_18 < 0))) __PYX_ERR(1, 1240, __pyx_L1_error)
        if (!__pyx_t_18) {
        } else {
          __pyx_t_9 = __pyx_t_18;
          goto __pyx_L28_bool_binop_done;
        }
        __pyx_t_18 = __Pyx_PyObject_IsTrue(__pyx_v_cmderr); if (unlikely((__pyx_t_18 < 0))) __PYX_ERR(1, 1240, __pyx_L1_error)
        __pyx_t_9 = __pyx_t_18;
        __pyx_L28_bool_binop_done:;
        if (__pyx_t_9) {

          /* "libmata/nfa/nfa.pyx":1241
 *             if objects[i].returncode:
 *                 if not quiet and (cmdout or cmderr):
 *                     print(f"captured stdout: {cmdout.decode('utf-8')}", 'red')             # <<<<<<<<<<<<<<
 *                     print(f"captured stderr: {cmderr.decode('utf-8')}", 'red')
 *                 raise subprocess.CalledProcessError(
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_cmdout, __pyx_n_s_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1241, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = NULL;
          __pyx_t_8 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_8 = 1;
            }
          }
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_kp_u_utf_8};
            __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1241, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1241, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyUnicode_Concat(__pyx_kp_u_captured_stdout, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1241, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1241, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
          __Pyx_INCREF(__pyx_n_u_red);
          __Pyx_GIVEREF(__pyx_n_u_red);
          PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_n_u_red);
          __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1241, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "libmata/nfa/nfa.pyx":1242
 *                 if not quiet and (cmdout or cmderr):
 *                     print(f"captured stdout: {cmdout.decode('utf-8')}", 'red')
 *                     print(f"captured stderr: {cmderr.decode('utf-8')}", 'red')             # <<<<<<<<<<<<<<
 *                 raise subprocess.CalledProcessError(
 *                     objects[i].returncode, unpiped_commands[i]
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_cmderr, __pyx_n_s_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1242, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = NULL;
          __pyx_t_8 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_8 = 1;
            }
          }
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_kp_u_utf_8};
            __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1242, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1242, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyUnicode_Concat(__pyx_kp_u_captured_stderr, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1242, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1242, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
          __Pyx_INCREF(__pyx_n_u_red);
          __Pyx_GIVEREF(__pyx_n_u_red);
          PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_n_u_red);
          __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1242, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "libmata/nfa/nfa.pyx":1240
 *         for i in range(cmd_no):
 *             if objects[i].returncode:
 *                 if not quiet and (cmdout or cmderr):             # <<<<<<<<<<<<<<
 *                     print(f"captured stdout: {cmdout.decode('utf-8')}", 'red')
 *                     print(f"captured stderr: {cmderr.decode('utf-8')}", 'red')
 */
        }

        /* "libmata/nfa/nfa.pyx":1243
 *                     print(f"captured stdout: {cmdout.decode('utf-8')}", 'red')
 *                     print(f"captured stderr: {cmderr.decode('utf-8')}", 'red')
 *                 raise subprocess.CalledProcessError(             # <<<<<<<<<<<<<<
 *                     objects[i].returncode, unpiped_commands[i]
 *                 )
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_subprocess); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1243, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CalledProcessError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1243, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "libmata/nfa/nfa.pyx":1244
 *                     print(f"captured stderr: {cmderr.decode('utf-8')}", 'red')
 *                 raise subprocess.CalledProcessError(
 *                     objects[i].returncode, unpiped_commands[i]             # <<<<<<<<<<<<<<
 *                 )
 * 
 */
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_objects, __pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1244, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_returncode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1244, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_unpiped_commands, __pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1244, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_11 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_8 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_11, __pyx_t_2, __pyx_t_3};
          __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1243, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(1, 1243, __pyx_L1_error)

        /* "libmata/nfa/nfa.pyx":1239
 *     if check_results:
 *         for i in range(cmd_no):
 *             if objects[i].returncode:             # <<<<<<<<<<<<<<
 *                 if not quiet and (cmdout or cmderr):
 *                     print(f"captured stdout: {cmdout.decode('utf-8')}", 'red')
 */
      }

      /* "libmata/nfa/nfa.pyx":1238
 *     # collect the return codes
 *     if check_results:
 *         for i in range(cmd_no):             # <<<<<<<<<<<<<<
 *             if objects[i].returncode:
 *                 if not quiet and (cmdout or cmderr):
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "libmata/nfa/nfa.pyx":1237
 * 
 *     # collect the return codes
 *     if check_results:             # <<<<<<<<<<<<<<
 *         for i in range(cmd_no):
 *             if objects[i].returncode:
 */
  }

  /* "libmata/nfa/nfa.pyx":1247
 *                 )
 * 
 *     return cmdout, cmderr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_cmdout);
  __Pyx_GIVEREF(__pyx_v_cmdout);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_cmdout);
  __Pyx_INCREF(__pyx_v_cmderr);
  __Pyx_GIVEREF(__pyx_v_cmderr);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_cmderr);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1189
 * 
 * 
 * def run_safely_external_command(cmd: str, check_results=True, quiet=True, timeout=None, **kwargs):             # <<<<<<<<<<<<<<
 *     """Safely runs the piped command, without executing of the shell
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("libmata.nfa.nfa.run_safely_external_command", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_unpiped_commands);
  __Pyx_XDECREF(__pyx_v_cmd_no);
  __Pyx_XDECREF(__pyx_v_objects);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_executed_command);
  __Pyx_XDECREF(__pyx_v_stdin);
  __Pyx_XDECREF(__pyx_v_stderr);
  __Pyx_XDECREF(__pyx_v_piped_command);
  __Pyx_XDECREF(__pyx_v_cmdout);
  __Pyx_XDECREF(__pyx_v_cmderr);
  __Pyx_XDECREF(__pyx_v_process);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1250
 * 
 * 
 * def get_elements_from_bool_vec(bool_vec: list[int]):             # <<<<<<<<<<<<<<
 *     cdef CBoolVector c_bool_vec = CBoolVector(bool_vec)
 *     cdef StateSet c_states
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_7get_elements_from_bool_vec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_7get_elements_from_bool_vec = {"get_elements_from_bool_vec", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_7get_elements_from_bool_vec, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_7get_elements_from_bool_vec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_bool_vec = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_elements_from_bool_vec (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_bool_vec,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_bool_vec)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1250, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_elements_from_bool_vec") < 0)) __PYX_ERR(1, 1250, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_bool_vec = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_elements_from_bool_vec", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 1250, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("libmata.nfa.nfa.get_elements_from_bool_vec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bool_vec), (&PyList_Type), 0, "bool_vec", 1))) __PYX_ERR(1, 1250, __pyx_L1_error)
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_6get_elements_from_bool_vec(__pyx_self, __pyx_v_bool_vec);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_6get_elements_from_bool_vec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bool_vec) {
  Mata::BoolVector __pyx_v_c_bool_vec;
  Mata::Nfa::StateSet __pyx_v_c_states;
  Mata::Nfa::State __pyx_9genexpr25__pyx_v_state;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  std::vector<uint8_t>  __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Mata::Util::OrdVector<Mata::Nfa::State> ::iterator __pyx_t_3;
  Mata::Nfa::State __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_elements_from_bool_vec", 0);

  /* "libmata/nfa/nfa.pyx":1251
 * 
 * def get_elements_from_bool_vec(bool_vec: list[int]):
 *     cdef CBoolVector c_bool_vec = CBoolVector(bool_vec)             # <<<<<<<<<<<<<<
 *     cdef StateSet c_states
 *     mata_nfa.get_elements(&c_states, c_bool_vec)
 */
  __pyx_t_1 = __pyx_convert_vector_from_py_uint8_t(__pyx_v_bool_vec); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1251, __pyx_L1_error)
  __pyx_v_c_bool_vec = Mata::BoolVector(__pyx_t_1);

  /* "libmata/nfa/nfa.pyx":1253
 *     cdef CBoolVector c_bool_vec = CBoolVector(bool_vec)
 *     cdef StateSet c_states
 *     mata_nfa.get_elements(&c_states, c_bool_vec)             # <<<<<<<<<<<<<<
 *     return { state for state in c_states }
 * 
 */
  Mata::Nfa::Plumbing::get_elements((&__pyx_v_c_states), __pyx_v_c_bool_vec);

  /* "libmata/nfa/nfa.pyx":1254
 *     cdef StateSet c_states
 *     mata_nfa.get_elements(&c_states, c_bool_vec)
 *     return { state for state in c_states }             # <<<<<<<<<<<<<<
 * 
 * # On the fly configuration of the library
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PySet_New(NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_v_c_states.begin();
    for (;;) {
      if (!(__pyx_t_3 != __pyx_v_c_states.end())) break;
      __pyx_t_4 = *__pyx_t_3;
      ++__pyx_t_3;
      __pyx_9genexpr25__pyx_v_state = __pyx_t_4;
      __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_9genexpr25__pyx_v_state); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1254, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(PySet_Add(__pyx_t_2, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 1254, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1250
 * 
 * 
 * def get_elements_from_bool_vec(bool_vec: list[int]):             # <<<<<<<<<<<<<<
 *     cdef CBoolVector c_bool_vec = CBoolVector(bool_vec)
 *     cdef StateSet c_states
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("libmata.nfa.nfa.get_elements_from_bool_vec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "libmata/nfa/nfa.pyx":1277
 * 
 * 
 * def store():             # <<<<<<<<<<<<<<
 *     """
 *     Returns the configuration of the library
 */

/* Python wrapper */
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_9store(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_7libmata_3nfa_3nfa_8store, "\n    Returns the configuration of the library\n    ");
static PyMethodDef __pyx_mdef_7libmata_3nfa_3nfa_9store = {"store", (PyCFunction)__pyx_pw_7libmata_3nfa_3nfa_9store, METH_NOARGS, __pyx_doc_7libmata_3nfa_3nfa_8store};
static PyObject *__pyx_pw_7libmata_3nfa_3nfa_9store(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("store (wrapper)", 0);
  __pyx_r = __pyx_pf_7libmata_3nfa_3nfa_8store(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7libmata_3nfa_3nfa_8store(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("store", 0);

  /* "libmata/nfa/nfa.pyx":1281
 *     Returns the configuration of the library
 *     """
 *     return _store             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_store_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "libmata/nfa/nfa.pyx":1277
 * 
 * 
 * def store():             # <<<<<<<<<<<<<<
 *     """
 *     Returns the configuration of the library
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("libmata.nfa.nfa.store", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_7libmata_3nfa_3nfa_Nfa(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)o);
  new((void*)&(p->thisptr)) std::shared_ptr<Mata::Nfa::Nfa> ();
  p->label = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7libmata_3nfa_3nfa_3Nfa_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7libmata_3nfa_3nfa_Nfa(PyObject *o) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *p = (struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7libmata_3nfa_3nfa_Nfa) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->thisptr);
  Py_CLEAR(p->label);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_7libmata_3nfa_3nfa_Nfa(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *p = (struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)o;
  if (p->label) {
    e = (*v)(p->label, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7libmata_3nfa_3nfa_Nfa(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *p = (struct __pyx_obj_7libmata_3nfa_3nfa_Nfa *)o;
  tmp = ((PyObject*)p->label);
  p->label = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7libmata_3nfa_3nfa_3Nfa_label(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7libmata_3nfa_3nfa_3Nfa_5label_1__get__(o);
}

static int __pyx_setprop_7libmata_3nfa_3nfa_3Nfa_label(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7libmata_3nfa_3nfa_3Nfa_5label_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7libmata_3nfa_3nfa_3Nfa_initial_states(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7libmata_3nfa_3nfa_3Nfa_14initial_states_1__get__(o);
}

static int __pyx_setprop_7libmata_3nfa_3nfa_3Nfa_initial_states(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7libmata_3nfa_3nfa_3Nfa_14initial_states_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7libmata_3nfa_3nfa_3Nfa_final_states(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7libmata_3nfa_3nfa_3Nfa_12final_states_1__get__(o);
}

static int __pyx_setprop_7libmata_3nfa_3nfa_3Nfa_final_states(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7libmata_3nfa_3nfa_3Nfa_12final_states_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_7libmata_3nfa_3nfa_3Nfa_76__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7libmata_3nfa_3nfa_3Nfa_76__repr__(self);
}

static PyMethodDef __pyx_methods_7libmata_3nfa_3nfa_Nfa[] = {
  {"is_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_3is_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_2is_state},
  {"add_new_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_5add_new_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_4add_new_state},
  {"add_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_7add_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_6add_state},
  {"make_initial_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_9make_initial_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_8make_initial_state},
  {"make_initial_states", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_11make_initial_states, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_10make_initial_states},
  {"has_initial_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_13has_initial_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_12has_initial_state},
  {"remove_initial_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_15remove_initial_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_14remove_initial_state},
  {"clear_initial", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_17clear_initial, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_16clear_initial},
  {"make_final_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_19make_final_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_18make_final_state},
  {"make_final_states", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_21make_final_states, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_20make_final_states},
  {"has_final_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_23has_final_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_22has_final_state},
  {"remove_final_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_25remove_final_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_24remove_final_state},
  {"clear_final", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_27clear_final, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_26clear_final},
  {"unify_initial", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_29unify_initial, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_28unify_initial},
  {"unify_final", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_31unify_final, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_30unify_final},
  {"add_transition_object", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_33add_transition_object, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_32add_transition_object},
  {"add_transition", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_35add_transition, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_34add_transition},
  {"remove_trans", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_37remove_trans, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_36remove_trans},
  {"remove_trans_raw", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_39remove_trans_raw, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_38remove_trans_raw},
  {"has_transition", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_41has_transition, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_40has_transition},
  {"get_num_of_trans", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_43get_num_of_trans, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_42get_num_of_trans},
  {"clear", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_45clear, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_44clear},
  {"size", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_47size, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_46size},
  {"iterate", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_49iterate, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_48iterate},
  {"get_transitions_from_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_52get_transitions_from_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_51get_transitions_from_state},
  {"get_transitions_to_state", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_54get_transitions_to_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_53get_transitions_to_state},
  {"get_trans_as_sequence", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_56get_trans_as_sequence, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_55get_trans_as_sequence},
  {"get_trans_from_state_as_sequence", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_58get_trans_from_state_as_sequence, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_57get_trans_from_state_as_sequence},
  {"get_useful_states", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_60get_useful_states, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_59get_useful_states},
  {"get_reachable_states", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_62get_reachable_states, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_61get_reachable_states},
  {"get_terminating_states", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_64get_terminating_states, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_63get_terminating_states},
  {"trim", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_66trim, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_65trim},
  {"trim_with_state_map", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_68trim_with_state_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_67trim_with_state_map},
  {"get_one_letter_aut", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_70get_one_letter_aut, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_69get_one_letter_aut},
  {"is_epsilon", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_72is_epsilon, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_71is_epsilon},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7libmata_3nfa_3nfa_3Nfa_76__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"to_dot_file", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_78to_dot_file, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_77to_dot_file},
  {"to_dot_str", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_80to_dot_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_79to_dot_str},
  {"to_dataframe", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_82to_dataframe, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_81to_dataframe},
  {"to_networkx_graph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_84to_networkx_graph, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_83to_networkx_graph},
  {"post_map_of", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_86post_map_of, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_85post_map_of},
  {"post_of", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_88post_of, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_87post_of},
  {"determinize_with_subset_map", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_90determinize_with_subset_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_89determinize_with_subset_map},
  {"determinize", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_92determinize, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_91determinize},
  {"union", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_94union, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_93union},
  {"intersection", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_96intersection, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_95intersection},
  {"intersection_with_product_map", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_98intersection_with_product_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_97intersection_with_product_map},
  {"concatenate", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_100concatenate, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_99concatenate},
  {"concatenate_with_result_state_maps", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_102concatenate_with_result_state_maps, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_101concatenate_with_result_state_maps},
  {"complement", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_104complement, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_103complement},
  {"make_complete", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_106make_complete, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_105make_complete},
  {"revert", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_108revert, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_107revert},
  {"remove_epsilon", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_110remove_epsilon, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_109remove_epsilon},
  {"remove_epsilon_inplace", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_112remove_epsilon_inplace, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_111remove_epsilon_inplace},
  {"get_epsilon_transitions", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_114get_epsilon_transitions, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_113get_epsilon_transitions},
  {"minimize", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_116minimize, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_115minimize},
  {"reduce_with_state_map", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_118reduce_with_state_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_117reduce_with_state_map},
  {"reduce", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_120reduce, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_119reduce},
  {"compute_relation", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_122compute_relation, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_121compute_relation},
  {"is_deterministic", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_124is_deterministic, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_123is_deterministic},
  {"is_lang_empty", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_126is_lang_empty, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_125is_lang_empty},
  {"is_universal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_128is_universal, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_127is_universal},
  {"is_included_with_cex", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_130is_included_with_cex, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_129is_included_with_cex},
  {"is_included", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_132is_included, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_131is_included},
  {"equivalence_check", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_134equivalence_check, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_133equivalence_check},
  {"get_symbols", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_136get_symbols, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_135get_symbols},
  {"is_complete", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_138is_complete, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_137is_complete},
  {"is_in_lang", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_140is_in_lang, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_139is_in_lang},
  {"is_prefix_in_lang", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_142is_prefix_in_lang, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_141is_prefix_in_lang},
  {"accepts_epsilon", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_144accepts_epsilon, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_143accepts_epsilon},
  {"get_word_for_path", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_146get_word_for_path, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_145get_word_for_path},
  {"encode_word", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_148encode_word, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7libmata_3nfa_3nfa_3Nfa_147encode_word},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_150__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_152__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7libmata_3nfa_3nfa_Nfa[] = {
  {(char *)"label", __pyx_getprop_7libmata_3nfa_3nfa_3Nfa_label, __pyx_setprop_7libmata_3nfa_3nfa_3Nfa_label, (char *)0, 0},
  {(char *)"initial_states", __pyx_getprop_7libmata_3nfa_3nfa_3Nfa_initial_states, __pyx_setprop_7libmata_3nfa_3nfa_3Nfa_initial_states, (char *)0, 0},
  {(char *)"final_states", __pyx_getprop_7libmata_3nfa_3nfa_3Nfa_final_states, __pyx_setprop_7libmata_3nfa_3nfa_3Nfa_final_states, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7libmata_3nfa_3nfa_Nfa_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7libmata_3nfa_3nfa_Nfa},
  {Py_tp_repr, (void *)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_76__repr__},
  {Py_tp_str, (void *)__pyx_pw_7libmata_3nfa_3nfa_3Nfa_74__str__},
  {Py_tp_doc, (void *)PyDoc_STR("Wrapper over NFA\n\n    Note: In order to add more properties to Nfa, see `nfa.pxd`, where there is forward declaration\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7libmata_3nfa_3nfa_Nfa},
  {Py_tp_clear, (void *)__pyx_tp_clear_7libmata_3nfa_3nfa_Nfa},
  {Py_tp_methods, (void *)__pyx_methods_7libmata_3nfa_3nfa_Nfa},
  {Py_tp_getset, (void *)__pyx_getsets_7libmata_3nfa_3nfa_Nfa},
  {Py_tp_new, (void *)__pyx_tp_new_7libmata_3nfa_3nfa_Nfa},
  {0, 0},
};
static PyType_Spec __pyx_type_7libmata_3nfa_3nfa_Nfa_spec = {
  "libmata.nfa.nfa.Nfa",
  sizeof(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7libmata_3nfa_3nfa_Nfa_slots,
};
#else

static PyTypeObject __pyx_type_7libmata_3nfa_3nfa_Nfa = {
  PyVarObject_HEAD_INIT(0, 0)
  "libmata.nfa.nfa.""Nfa", /*tp_name*/
  sizeof(struct __pyx_obj_7libmata_3nfa_3nfa_Nfa), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7libmata_3nfa_3nfa_Nfa, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7libmata_3nfa_3nfa_3Nfa_76__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7libmata_3nfa_3nfa_3Nfa_74__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Wrapper over NFA\n\n    Note: In order to add more properties to Nfa, see `nfa.pxd`, where there is forward declaration\n    "), /*tp_doc*/
  __pyx_tp_traverse_7libmata_3nfa_3nfa_Nfa, /*tp_traverse*/
  __pyx_tp_clear_7libmata_3nfa_3nfa_Nfa, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7libmata_3nfa_3nfa_Nfa, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7libmata_3nfa_3nfa_Nfa, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7libmata_3nfa_3nfa_Nfa, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7libmata_3nfa_3nfa_Trans __pyx_vtable_7libmata_3nfa_3nfa_Trans;

static PyObject *__pyx_tp_new_7libmata_3nfa_3nfa_Trans(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7libmata_3nfa_3nfa_Trans *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7libmata_3nfa_3nfa_Trans *)o);
  p->__pyx_vtab = __pyx_vtabptr_7libmata_3nfa_3nfa_Trans;
  if (unlikely(__pyx_pw_7libmata_3nfa_3nfa_5Trans_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7libmata_3nfa_3nfa_Trans(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7libmata_3nfa_3nfa_Trans) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7libmata_3nfa_3nfa_5Trans_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_tp_richcompare_7libmata_3nfa_3nfa_Trans(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_7libmata_3nfa_3nfa_5Trans_5__eq__(o1, o2);
    }
    case Py_NE: {
      PyObject *ret;
      ret = __pyx_pw_7libmata_3nfa_3nfa_5Trans_5__eq__(o1, o2);
      if (likely(ret && ret != Py_NotImplemented)) {
        int b = __Pyx_PyObject_IsTrue(ret);
        Py_DECREF(ret);
        if (unlikely(b < 0)) return NULL;
        ret = (b) ? Py_False : Py_True;
        Py_INCREF(ret);
      }
      return ret;
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_7libmata_3nfa_3nfa_5Trans_src(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7libmata_3nfa_3nfa_5Trans_3src_1__get__(o);
}

static PyObject *__pyx_getprop_7libmata_3nfa_3nfa_5Trans_symb(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7libmata_3nfa_3nfa_5Trans_4symb_1__get__(o);
}

static PyObject *__pyx_getprop_7libmata_3nfa_3nfa_5Trans_tgt(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7libmata_3nfa_3nfa_5Trans_3tgt_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_7libmata_3nfa_3nfa_5Trans_11__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7libmata_3nfa_3nfa_5Trans_11__repr__(self);
}

static PyMethodDef __pyx_methods_7libmata_3nfa_3nfa_Trans[] = {
  {"__neq__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_5Trans_7__neq__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7libmata_3nfa_3nfa_5Trans_11__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_5Trans_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_5Trans_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7libmata_3nfa_3nfa_Trans[] = {
  {(char *)"src", __pyx_getprop_7libmata_3nfa_3nfa_5Trans_src, 0, (char *)PyDoc_STR("\n        :return: source state of the transition\n        "), 0},
  {(char *)"symb", __pyx_getprop_7libmata_3nfa_3nfa_5Trans_symb, 0, (char *)PyDoc_STR("\n        :return: symbol for the transition\n        "), 0},
  {(char *)"tgt", __pyx_getprop_7libmata_3nfa_3nfa_5Trans_tgt, 0, (char *)PyDoc_STR("\n        :return: target state of the transition\n        "), 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7libmata_3nfa_3nfa_Trans_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7libmata_3nfa_3nfa_Trans},
  {Py_tp_repr, (void *)__pyx_pw_7libmata_3nfa_3nfa_5Trans_11__repr__},
  {Py_tp_str, (void *)__pyx_pw_7libmata_3nfa_3nfa_5Trans_9__str__},
  {Py_tp_doc, (void *)PyDoc_STR("Wrapper over the transitions in NFA.")},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_7libmata_3nfa_3nfa_Trans},
  {Py_tp_methods, (void *)__pyx_methods_7libmata_3nfa_3nfa_Trans},
  {Py_tp_getset, (void *)__pyx_getsets_7libmata_3nfa_3nfa_Trans},
  {Py_tp_new, (void *)__pyx_tp_new_7libmata_3nfa_3nfa_Trans},
  {0, 0},
};
static PyType_Spec __pyx_type_7libmata_3nfa_3nfa_Trans_spec = {
  "libmata.nfa.nfa.Trans",
  sizeof(struct __pyx_obj_7libmata_3nfa_3nfa_Trans),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7libmata_3nfa_3nfa_Trans_slots,
};
#else

static PyTypeObject __pyx_type_7libmata_3nfa_3nfa_Trans = {
  PyVarObject_HEAD_INIT(0, 0)
  "libmata.nfa.nfa.""Trans", /*tp_name*/
  sizeof(struct __pyx_obj_7libmata_3nfa_3nfa_Trans), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7libmata_3nfa_3nfa_Trans, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7libmata_3nfa_3nfa_5Trans_11__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7libmata_3nfa_3nfa_5Trans_9__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("Wrapper over the transitions in NFA."), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_tp_richcompare_7libmata_3nfa_3nfa_Trans, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7libmata_3nfa_3nfa_Trans, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7libmata_3nfa_3nfa_Trans, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7libmata_3nfa_3nfa_Trans, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7libmata_3nfa_3nfa_Run(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  if (unlikely(__pyx_pw_7libmata_3nfa_3nfa_3Run_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7libmata_3nfa_3nfa_Run(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7libmata_3nfa_3nfa_Run) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7libmata_3nfa_3nfa_3Run_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_7libmata_3nfa_3nfa_3Run_word(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7libmata_3nfa_3nfa_3Run_4word_1__get__(o);
}

static int __pyx_setprop_7libmata_3nfa_3nfa_3Run_word(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7libmata_3nfa_3nfa_3Run_4word_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7libmata_3nfa_3nfa_3Run_path(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7libmata_3nfa_3nfa_3Run_4path_1__get__(o);
}

static int __pyx_setprop_7libmata_3nfa_3nfa_3Run_path(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7libmata_3nfa_3nfa_3Run_4path_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7libmata_3nfa_3nfa_Run[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Run_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_3Run_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7libmata_3nfa_3nfa_Run[] = {
  {(char *)"word", __pyx_getprop_7libmata_3nfa_3nfa_3Run_word, __pyx_setprop_7libmata_3nfa_3nfa_3Run_word, (char *)0, 0},
  {(char *)"path", __pyx_getprop_7libmata_3nfa_3nfa_3Run_path, __pyx_setprop_7libmata_3nfa_3nfa_3Run_path, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7libmata_3nfa_3nfa_Run_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7libmata_3nfa_3nfa_Run},
  {Py_tp_doc, (void *)PyDoc_STR("Wrapper over the run in NFA.")},
  {Py_tp_methods, (void *)__pyx_methods_7libmata_3nfa_3nfa_Run},
  {Py_tp_getset, (void *)__pyx_getsets_7libmata_3nfa_3nfa_Run},
  {Py_tp_new, (void *)__pyx_tp_new_7libmata_3nfa_3nfa_Run},
  {0, 0},
};
static PyType_Spec __pyx_type_7libmata_3nfa_3nfa_Run_spec = {
  "libmata.nfa.nfa.Run",
  sizeof(struct __pyx_obj_7libmata_3nfa_3nfa_Run),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7libmata_3nfa_3nfa_Run_slots,
};
#else

static PyTypeObject __pyx_type_7libmata_3nfa_3nfa_Run = {
  PyVarObject_HEAD_INIT(0, 0)
  "libmata.nfa.nfa.""Run", /*tp_name*/
  sizeof(struct __pyx_obj_7libmata_3nfa_3nfa_Run), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7libmata_3nfa_3nfa_Run, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("Wrapper over the run in NFA."), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7libmata_3nfa_3nfa_Run, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7libmata_3nfa_3nfa_Run, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7libmata_3nfa_3nfa_Run, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7libmata_3nfa_3nfa_Move(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  if (unlikely(__pyx_pw_7libmata_3nfa_3nfa_4Move_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7libmata_3nfa_3nfa_Move(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7libmata_3nfa_3nfa_Move) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7libmata_3nfa_3nfa_4Move_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_tp_richcompare_7libmata_3nfa_3nfa_Move(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_7libmata_3nfa_3nfa_4Move_13__eq__(o1, o2);
    }
    case Py_LT: {
      return __pyx_pw_7libmata_3nfa_3nfa_4Move_5__lt__(o1, o2);
    }
    case Py_GT: {
      return __pyx_pw_7libmata_3nfa_3nfa_4Move_7__gt__(o1, o2);
    }
    case Py_LE: {
      return __pyx_pw_7libmata_3nfa_3nfa_4Move_9__le__(o1, o2);
    }
    case Py_GE: {
      return __pyx_pw_7libmata_3nfa_3nfa_4Move_11__ge__(o1, o2);
    }
    case Py_NE: {
      PyObject *ret;
      ret = __pyx_pw_7libmata_3nfa_3nfa_4Move_13__eq__(o1, o2);
      if (likely(ret && ret != Py_NotImplemented)) {
        int b = __Pyx_PyObject_IsTrue(ret);
        Py_DECREF(ret);
        if (unlikely(b < 0)) return NULL;
        ret = (b) ? Py_False : Py_True;
        Py_INCREF(ret);
      }
      return ret;
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_7libmata_3nfa_3nfa_4Move_symbol(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7libmata_3nfa_3nfa_4Move_6symbol_1__get__(o);
}

static int __pyx_setprop_7libmata_3nfa_3nfa_4Move_symbol(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7libmata_3nfa_3nfa_4Move_6symbol_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7libmata_3nfa_3nfa_4Move_targets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7libmata_3nfa_3nfa_4Move_7targets_1__get__(o);
}

static int __pyx_setprop_7libmata_3nfa_3nfa_4Move_targets(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7libmata_3nfa_3nfa_4Move_7targets_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_7libmata_3nfa_3nfa_4Move_19__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_7libmata_3nfa_3nfa_4Move_19__repr__(self);
}

static PyMethodDef __pyx_methods_7libmata_3nfa_3nfa_Move[] = {
  {"__neq__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_4Move_15__neq__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_7libmata_3nfa_3nfa_4Move_19__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_4Move_21__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7libmata_3nfa_3nfa_4Move_23__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7libmata_3nfa_3nfa_Move[] = {
  {(char *)"symbol", __pyx_getprop_7libmata_3nfa_3nfa_4Move_symbol, __pyx_setprop_7libmata_3nfa_3nfa_4Move_symbol, (char *)0, 0},
  {(char *)"targets", __pyx_getprop_7libmata_3nfa_3nfa_4Move_targets, __pyx_setprop_7libmata_3nfa_3nfa_4Move_targets, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7libmata_3nfa_3nfa_Move_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7libmata_3nfa_3nfa_Move},
  {Py_tp_repr, (void *)__pyx_pw_7libmata_3nfa_3nfa_4Move_19__repr__},
  {Py_tp_str, (void *)__pyx_pw_7libmata_3nfa_3nfa_4Move_17__str__},
  {Py_tp_doc, (void *)PyDoc_STR("Wrapper over pair of symbol and states for transitions")},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_7libmata_3nfa_3nfa_Move},
  {Py_tp_methods, (void *)__pyx_methods_7libmata_3nfa_3nfa_Move},
  {Py_tp_getset, (void *)__pyx_getsets_7libmata_3nfa_3nfa_Move},
  {Py_tp_new, (void *)__pyx_tp_new_7libmata_3nfa_3nfa_Move},
  {0, 0},
};
static PyType_Spec __pyx_type_7libmata_3nfa_3nfa_Move_spec = {
  "libmata.nfa.nfa.Move",
  sizeof(struct __pyx_obj_7libmata_3nfa_3nfa_Move),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7libmata_3nfa_3nfa_Move_slots,
};
#else

static PyTypeObject __pyx_type_7libmata_3nfa_3nfa_Move = {
  PyVarObject_HEAD_INIT(0, 0)
  "libmata.nfa.nfa.""Move", /*tp_name*/
  sizeof(struct __pyx_obj_7libmata_3nfa_3nfa_Move), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7libmata_3nfa_3nfa_Move, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7libmata_3nfa_3nfa_4Move_19__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7libmata_3nfa_3nfa_4Move_17__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("Wrapper over pair of symbol and states for transitions"), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_tp_richcompare_7libmata_3nfa_3nfa_Move, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7libmata_3nfa_3nfa_Move, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7libmata_3nfa_3nfa_Move, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7libmata_3nfa_3nfa_Move, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *__pyx_freelist_7libmata_3nfa_3nfa___pyx_scope_struct__iterate[8];
static int __pyx_freecount_7libmata_3nfa_3nfa___pyx_scope_struct__iterate = 0;

static PyObject *__pyx_tp_new_7libmata_3nfa_3nfa___pyx_scope_struct__iterate(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((int)(__pyx_freecount_7libmata_3nfa_3nfa___pyx_scope_struct__iterate > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate)))) {
    o = (PyObject*)__pyx_freelist_7libmata_3nfa_3nfa___pyx_scope_struct__iterate[--__pyx_freecount_7libmata_3nfa_3nfa___pyx_scope_struct__iterate];
    memset(o, 0, sizeof(struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *)o);
  new((void*)&(p->__pyx_v_iterator)) Mata::Nfa::Nfa::const_iterator();
  new((void*)&(p->__pyx_v_lhs)) Mata::Nfa::Trans();
  return o;
}

static void __pyx_tp_dealloc_7libmata_3nfa_3nfa___pyx_scope_struct__iterate(PyObject *o) {
  struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *p = (struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7libmata_3nfa_3nfa___pyx_scope_struct__iterate) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->__pyx_v_iterator);
  __Pyx_call_destructor(p->__pyx_v_lhs);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_trans);
  if (CYTHON_COMPILING_IN_CPYTHON && ((int)(__pyx_freecount_7libmata_3nfa_3nfa___pyx_scope_struct__iterate < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate)))) {
    __pyx_freelist_7libmata_3nfa_3nfa___pyx_scope_struct__iterate[__pyx_freecount_7libmata_3nfa_3nfa___pyx_scope_struct__iterate++] = ((struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_7libmata_3nfa_3nfa___pyx_scope_struct__iterate(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *p = (struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_trans) {
    e = (*v)(((PyObject *)p->__pyx_v_trans), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7libmata_3nfa_3nfa___pyx_scope_struct__iterate_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7libmata_3nfa_3nfa___pyx_scope_struct__iterate},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7libmata_3nfa_3nfa___pyx_scope_struct__iterate},
  {Py_tp_new, (void *)__pyx_tp_new_7libmata_3nfa_3nfa___pyx_scope_struct__iterate},
  {0, 0},
};
static PyType_Spec __pyx_type_7libmata_3nfa_3nfa___pyx_scope_struct__iterate_spec = {
  "libmata.nfa.nfa.__pyx_scope_struct__iterate",
  sizeof(struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_7libmata_3nfa_3nfa___pyx_scope_struct__iterate_slots,
};
#else

static PyTypeObject __pyx_type_7libmata_3nfa_3nfa___pyx_scope_struct__iterate = {
  PyVarObject_HEAD_INIT(0, 0)
  "libmata.nfa.nfa.""__pyx_scope_struct__iterate", /*tp_name*/
  sizeof(struct __pyx_obj_7libmata_3nfa_3nfa___pyx_scope_struct__iterate), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7libmata_3nfa_3nfa___pyx_scope_struct__iterate, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7libmata_3nfa_3nfa___pyx_scope_struct__iterate, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7libmata_3nfa_3nfa___pyx_scope_struct__iterate, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
    {&__pyx_kp_u_0_3, __pyx_k_0_3, sizeof(__pyx_k_0_3), 0, 1, 0, 0},
    {&__pyx_kp_u_12, __pyx_k_12, sizeof(__pyx_k_12), 0, 1, 0, 0},
    {&__pyx_kp_u_1_5, __pyx_k_1_5, sizeof(__pyx_k_1_5), 0, 1, 0, 0},
    {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
    {&__pyx_n_s_CalledProcessError, __pyx_k_CalledProcessError, sizeof(__pyx_k_CalledProcessError), 0, 0, 1, 1},
    {&__pyx_kp_u_Cannot_translate_symbol, __pyx_k_Cannot_translate_symbol, sizeof(__pyx_k_Cannot_translate_symbol), 0, 1, 0, 0},
    {&__pyx_kp_u_Courier_Bold, __pyx_k_Courier_Bold, sizeof(__pyx_k_Courier_Bold), 0, 1, 0, 0},
    {&__pyx_n_s_DataFrame, __pyx_k_DataFrame, sizeof(__pyx_k_DataFrame), 0, 0, 1, 1},
    {&__pyx_n_s_DiGraph, __pyx_k_DiGraph, sizeof(__pyx_k_DiGraph), 0, 0, 1, 1},
    {&__pyx_kp_u_Encodes_word_based_on_a_string_t, __pyx_k_Encodes_word_based_on_a_string_t, sizeof(__pyx_k_Encodes_word_based_on_a_string_t), 0, 1, 0, 0},
    {&__pyx_kp_u_For_a_given_path_set_of_states_r, __pyx_k_For_a_given_path_set_of_states_r, sizeof(__pyx_k_For_a_given_path_set_of_states_r), 0, 1, 0, 0},
    {&__pyx_n_s_G, __pyx_k_G, sizeof(__pyx_k_G), 0, 0, 1, 1},
    {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
    {&__pyx_n_s_Move, __pyx_k_Move, sizeof(__pyx_k_Move), 0, 0, 1, 1},
    {&__pyx_kp_s_Move_None, __pyx_k_Move_None, sizeof(__pyx_k_Move_None), 0, 0, 1, 0},
    {&__pyx_n_s_Move___neq, __pyx_k_Move___neq, sizeof(__pyx_k_Move___neq), 0, 0, 1, 1},
    {&__pyx_n_s_Move___reduce_cython, __pyx_k_Move___reduce_cython, sizeof(__pyx_k_Move___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Move___setstate_cython, __pyx_k_Move___setstate_cython, sizeof(__pyx_k_Move___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa, __pyx_k_Nfa, sizeof(__pyx_k_Nfa), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa___reduce_cython, __pyx_k_Nfa___reduce_cython, sizeof(__pyx_k_Nfa___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa___setstate_cython, __pyx_k_Nfa___setstate_cython, sizeof(__pyx_k_Nfa___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_accepts_epsilon, __pyx_k_Nfa_accepts_epsilon, sizeof(__pyx_k_Nfa_accepts_epsilon), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_add_new_state, __pyx_k_Nfa_add_new_state, sizeof(__pyx_k_Nfa_add_new_state), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_add_state, __pyx_k_Nfa_add_state, sizeof(__pyx_k_Nfa_add_state), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_add_transition, __pyx_k_Nfa_add_transition, sizeof(__pyx_k_Nfa_add_transition), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_add_transition_object, __pyx_k_Nfa_add_transition_object, sizeof(__pyx_k_Nfa_add_transition_object), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_clear, __pyx_k_Nfa_clear, sizeof(__pyx_k_Nfa_clear), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_clear_final, __pyx_k_Nfa_clear_final, sizeof(__pyx_k_Nfa_clear_final), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_clear_initial, __pyx_k_Nfa_clear_initial, sizeof(__pyx_k_Nfa_clear_initial), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_complement, __pyx_k_Nfa_complement, sizeof(__pyx_k_Nfa_complement), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_compute_relation, __pyx_k_Nfa_compute_relation, sizeof(__pyx_k_Nfa_compute_relation), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_concatenate, __pyx_k_Nfa_concatenate, sizeof(__pyx_k_Nfa_concatenate), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_concatenate_with_result_stat, __pyx_k_Nfa_concatenate_with_result_stat, sizeof(__pyx_k_Nfa_concatenate_with_result_stat), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_determinize, __pyx_k_Nfa_determinize, sizeof(__pyx_k_Nfa_determinize), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_determinize_with_subset_map, __pyx_k_Nfa_determinize_with_subset_map, sizeof(__pyx_k_Nfa_determinize_with_subset_map), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_encode_word, __pyx_k_Nfa_encode_word, sizeof(__pyx_k_Nfa_encode_word), 0, 0, 1, 1},
    {&__pyx_kp_u_Nfa_encode_word_line_1139, __pyx_k_Nfa_encode_word_line_1139, sizeof(__pyx_k_Nfa_encode_word_line_1139), 0, 1, 0, 0},
    {&__pyx_n_s_Nfa_equivalence_check, __pyx_k_Nfa_equivalence_check, sizeof(__pyx_k_Nfa_equivalence_check), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_get_epsilon_transitions, __pyx_k_Nfa_get_epsilon_transitions, sizeof(__pyx_k_Nfa_get_epsilon_transitions), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_get_num_of_trans, __pyx_k_Nfa_get_num_of_trans, sizeof(__pyx_k_Nfa_get_num_of_trans), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_get_one_letter_aut, __pyx_k_Nfa_get_one_letter_aut, sizeof(__pyx_k_Nfa_get_one_letter_aut), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_get_reachable_states, __pyx_k_Nfa_get_reachable_states, sizeof(__pyx_k_Nfa_get_reachable_states), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_get_symbols, __pyx_k_Nfa_get_symbols, sizeof(__pyx_k_Nfa_get_symbols), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_get_terminating_states, __pyx_k_Nfa_get_terminating_states, sizeof(__pyx_k_Nfa_get_terminating_states), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_get_trans_as_sequence, __pyx_k_Nfa_get_trans_as_sequence, sizeof(__pyx_k_Nfa_get_trans_as_sequence), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_get_trans_from_state_as_sequ, __pyx_k_Nfa_get_trans_from_state_as_sequ, sizeof(__pyx_k_Nfa_get_trans_from_state_as_sequ), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_get_transitions_from_state, __pyx_k_Nfa_get_transitions_from_state, sizeof(__pyx_k_Nfa_get_transitions_from_state), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_get_transitions_to_state, __pyx_k_Nfa_get_transitions_to_state, sizeof(__pyx_k_Nfa_get_transitions_to_state), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_get_useful_states, __pyx_k_Nfa_get_useful_states, sizeof(__pyx_k_Nfa_get_useful_states), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_get_word_for_path, __pyx_k_Nfa_get_word_for_path, sizeof(__pyx_k_Nfa_get_word_for_path), 0, 0, 1, 1},
    {&__pyx_kp_u_Nfa_get_word_for_path_line_1122, __pyx_k_Nfa_get_word_for_path_line_1122, sizeof(__pyx_k_Nfa_get_word_for_path_line_1122), 0, 1, 0, 0},
    {&__pyx_n_s_Nfa_has_final_state, __pyx_k_Nfa_has_final_state, sizeof(__pyx_k_Nfa_has_final_state), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_has_initial_state, __pyx_k_Nfa_has_initial_state, sizeof(__pyx_k_Nfa_has_initial_state), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_has_transition, __pyx_k_Nfa_has_transition, sizeof(__pyx_k_Nfa_has_transition), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_intersection, __pyx_k_Nfa_intersection, sizeof(__pyx_k_Nfa_intersection), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_intersection_with_product_ma, __pyx_k_Nfa_intersection_with_product_ma, sizeof(__pyx_k_Nfa_intersection_with_product_ma), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_is_complete, __pyx_k_Nfa_is_complete, sizeof(__pyx_k_Nfa_is_complete), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_is_deterministic, __pyx_k_Nfa_is_deterministic, sizeof(__pyx_k_Nfa_is_deterministic), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_is_epsilon, __pyx_k_Nfa_is_epsilon, sizeof(__pyx_k_Nfa_is_epsilon), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_is_in_lang, __pyx_k_Nfa_is_in_lang, sizeof(__pyx_k_Nfa_is_in_lang), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_is_included, __pyx_k_Nfa_is_included, sizeof(__pyx_k_Nfa_is_included), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_is_included_with_cex, __pyx_k_Nfa_is_included_with_cex, sizeof(__pyx_k_Nfa_is_included_with_cex), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_is_lang_empty, __pyx_k_Nfa_is_lang_empty, sizeof(__pyx_k_Nfa_is_lang_empty), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_is_prefix_in_lang, __pyx_k_Nfa_is_prefix_in_lang, sizeof(__pyx_k_Nfa_is_prefix_in_lang), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_is_state, __pyx_k_Nfa_is_state, sizeof(__pyx_k_Nfa_is_state), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_is_universal, __pyx_k_Nfa_is_universal, sizeof(__pyx_k_Nfa_is_universal), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_iterate, __pyx_k_Nfa_iterate, sizeof(__pyx_k_Nfa_iterate), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_make_complete, __pyx_k_Nfa_make_complete, sizeof(__pyx_k_Nfa_make_complete), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_make_final_state, __pyx_k_Nfa_make_final_state, sizeof(__pyx_k_Nfa_make_final_state), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_make_final_states, __pyx_k_Nfa_make_final_states, sizeof(__pyx_k_Nfa_make_final_states), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_make_initial_state, __pyx_k_Nfa_make_initial_state, sizeof(__pyx_k_Nfa_make_initial_state), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_make_initial_states, __pyx_k_Nfa_make_initial_states, sizeof(__pyx_k_Nfa_make_initial_states), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_minimize, __pyx_k_Nfa_minimize, sizeof(__pyx_k_Nfa_minimize), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_post_map_of, __pyx_k_Nfa_post_map_of, sizeof(__pyx_k_Nfa_post_map_of), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_post_of, __pyx_k_Nfa_post_of, sizeof(__pyx_k_Nfa_post_of), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_reduce, __pyx_k_Nfa_reduce, sizeof(__pyx_k_Nfa_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_reduce_with_state_map, __pyx_k_Nfa_reduce_with_state_map, sizeof(__pyx_k_Nfa_reduce_with_state_map), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_remove_epsilon, __pyx_k_Nfa_remove_epsilon, sizeof(__pyx_k_Nfa_remove_epsilon), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_remove_epsilon_inplace, __pyx_k_Nfa_remove_epsilon_inplace, sizeof(__pyx_k_Nfa_remove_epsilon_inplace), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_remove_final_state, __pyx_k_Nfa_remove_final_state, sizeof(__pyx_k_Nfa_remove_final_state), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_remove_initial_state, __pyx_k_Nfa_remove_initial_state, sizeof(__pyx_k_Nfa_remove_initial_state), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_remove_trans, __pyx_k_Nfa_remove_trans, sizeof(__pyx_k_Nfa_remove_trans), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_remove_trans_raw, __pyx_k_Nfa_remove_trans_raw, sizeof(__pyx_k_Nfa_remove_trans_raw), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_revert, __pyx_k_Nfa_revert, sizeof(__pyx_k_Nfa_revert), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_size, __pyx_k_Nfa_size, sizeof(__pyx_k_Nfa_size), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_to_dataframe, __pyx_k_Nfa_to_dataframe, sizeof(__pyx_k_Nfa_to_dataframe), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_to_dot_file, __pyx_k_Nfa_to_dot_file, sizeof(__pyx_k_Nfa_to_dot_file), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_to_dot_str, __pyx_k_Nfa_to_dot_str, sizeof(__pyx_k_Nfa_to_dot_str), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_to_networkx_graph, __pyx_k_Nfa_to_networkx_graph, sizeof(__pyx_k_Nfa_to_networkx_graph), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_trim, __pyx_k_Nfa_trim, sizeof(__pyx_k_Nfa_trim), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_trim_with_state_map, __pyx_k_Nfa_trim_with_state_map, sizeof(__pyx_k_Nfa_trim_with_state_map), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_unify_final, __pyx_k_Nfa_unify_final, sizeof(__pyx_k_Nfa_unify_final), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_unify_initial, __pyx_k_Nfa_unify_initial, sizeof(__pyx_k_Nfa_unify_initial), 0, 0, 1, 1},
    {&__pyx_n_s_Nfa_union, __pyx_k_Nfa_union, sizeof(__pyx_k_Nfa_union), 0, 0, 1, 1},
    {&__pyx_n_s_PIPE, __pyx_k_PIPE, sizeof(__pyx_k_PIPE), 0, 0, 1, 1},
    {&__pyx_n_s_Popen, __pyx_k_Popen, sizeof(__pyx_k_Popen), 0, 0, 1, 1},
    {&__pyx_n_s_Run, __pyx_k_Run, sizeof(__pyx_k_Run), 0, 0, 1, 1},
    {&__pyx_n_s_Run___reduce_cython, __pyx_k_Run___reduce_cython, sizeof(__pyx_k_Run___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Run___setstate_cython, __pyx_k_Run___setstate_cython, sizeof(__pyx_k_Run___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_STDOUT, __pyx_k_STDOUT, sizeof(__pyx_k_STDOUT), 0, 0, 1, 1},
    {&__pyx_n_s_TimeoutExpired, __pyx_k_TimeoutExpired, sizeof(__pyx_k_TimeoutExpired), 0, 0, 1, 1},
    {&__pyx_n_s_Trans, __pyx_k_Trans, sizeof(__pyx_k_Trans), 0, 0, 1, 1},
    {&__pyx_n_s_Trans___neq, __pyx_k_Trans___neq, sizeof(__pyx_k_Trans___neq), 0, 0, 1, 1},
    {&__pyx_n_s_Trans___reduce_cython, __pyx_k_Trans___reduce_cython, sizeof(__pyx_k_Trans___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Trans___setstate_cython, __pyx_k_Trans___setstate_cython, sizeof(__pyx_k_Trans___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_kp_b__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 0, 0, 0},
    {&__pyx_kp_u__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 1, 0, 0},
    {&__pyx_n_s__16, __pyx_k__16, sizeof(__pyx_k__16), 0, 0, 1, 1},
    {&__pyx_n_s__165, __pyx_k__165, sizeof(__pyx_k__165), 0, 0, 1, 1},
    {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
    {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
    {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
    {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
    {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
    {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
    {&__pyx_n_s__80, __pyx_k__80, sizeof(__pyx_k__80), 0, 0, 1, 1},
    {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
    {&__pyx_n_s_accepts_epsilon, __pyx_k_accepts_epsilon, sizeof(__pyx_k_accepts_epsilon), 0, 0, 1, 1},
    {&__pyx_n_s_add_edge, __pyx_k_add_edge, sizeof(__pyx_k_add_edge), 0, 0, 1, 1},
    {&__pyx_n_s_add_new_state, __pyx_k_add_new_state, sizeof(__pyx_k_add_new_state), 0, 0, 1, 1},
    {&__pyx_n_s_add_state, __pyx_k_add_state, sizeof(__pyx_k_add_state), 0, 0, 1, 1},
    {&__pyx_n_s_add_transition, __pyx_k_add_transition, sizeof(__pyx_k_add_transition), 0, 0, 1, 1},
    {&__pyx_n_s_add_transition_object, __pyx_k_add_transition_object, sizeof(__pyx_k_add_transition_object), 0, 0, 1, 1},
    {&__pyx_n_u_algorithm, __pyx_k_algorithm, sizeof(__pyx_k_algorithm), 0, 1, 0, 1},
    {&__pyx_n_s_alphabet, __pyx_k_alphabet, sizeof(__pyx_k_alphabet), 0, 0, 1, 1},
    {&__pyx_n_u_alphabet, __pyx_k_alphabet, sizeof(__pyx_k_alphabet), 0, 1, 0, 1},
    {&__pyx_n_u_antichains, __pyx_k_antichains, sizeof(__pyx_k_antichains), 0, 1, 0, 1},
    {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_aut, __pyx_k_aut, sizeof(__pyx_k_aut), 0, 0, 1, 1},
    {&__pyx_kp_u_aut_dot, __pyx_k_aut_dot, sizeof(__pyx_k_aut_dot), 0, 1, 0, 0},
    {&__pyx_n_s_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 0, 1, 1},
    {&__pyx_n_s_bool_vec, __pyx_k_bool_vec, sizeof(__pyx_k_bool_vec), 0, 0, 1, 1},
    {&__pyx_n_s_c_alphabet, __pyx_k_c_alphabet, sizeof(__pyx_k_c_alphabet), 0, 0, 1, 1},
    {&__pyx_n_s_c_bool_vec, __pyx_k_c_bool_vec, sizeof(__pyx_k_c_bool_vec), 0, 0, 1, 1},
    {&__pyx_n_s_c_epsilon_transitions_iter, __pyx_k_c_epsilon_transitions_iter, sizeof(__pyx_k_c_epsilon_transitions_iter), 0, 0, 1, 1},
    {&__pyx_n_s_c_lhs_map, __pyx_k_c_lhs_map, sizeof(__pyx_k_c_lhs_map), 0, 0, 1, 1},
    {&__pyx_n_s_c_product_map, __pyx_k_c_product_map, sizeof(__pyx_k_c_product_map), 0, 0, 1, 1},
    {&__pyx_n_s_c_rhs_map, __pyx_k_c_rhs_map, sizeof(__pyx_k_c_rhs_map), 0, 0, 1, 1},
    {&__pyx_n_s_c_states, __pyx_k_c_states, sizeof(__pyx_k_c_states), 0, 0, 1, 1},
    {&__pyx_n_s_c_transition, __pyx_k_c_transition, sizeof(__pyx_k_c_transition), 0, 0, 1, 1},
    {&__pyx_n_s_c_transitions, __pyx_k_c_transitions, sizeof(__pyx_k_c_transitions), 0, 0, 1, 1},
    {&__pyx_kp_u_captured_stderr, __pyx_k_captured_stderr, sizeof(__pyx_k_captured_stderr), 0, 1, 0, 0},
    {&__pyx_kp_u_captured_stdout, __pyx_k_captured_stdout, sizeof(__pyx_k_captured_stdout), 0, 1, 0, 0},
    {&__pyx_n_s_check_results, __pyx_k_check_results, sizeof(__pyx_k_check_results), 0, 0, 1, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_u_classical, __pyx_k_classical, sizeof(__pyx_k_classical), 0, 1, 0, 1},
    {&__pyx_n_s_clear, __pyx_k_clear, sizeof(__pyx_k_clear), 0, 0, 1, 1},
    {&__pyx_n_s_clear_final, __pyx_k_clear_final, sizeof(__pyx_k_clear_final), 0, 0, 1, 1},
    {&__pyx_n_s_clear_initial, __pyx_k_clear_initial, sizeof(__pyx_k_clear_initial), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
    {&__pyx_n_s_cmd, __pyx_k_cmd, sizeof(__pyx_k_cmd), 0, 0, 1, 1},
    {&__pyx_n_s_cmd_no, __pyx_k_cmd_no, sizeof(__pyx_k_cmd_no), 0, 0, 1, 1},
    {&__pyx_n_s_cmderr, __pyx_k_cmderr, sizeof(__pyx_k_cmderr), 0, 0, 1, 1},
    {&__pyx_n_s_cmdout, __pyx_k_cmdout, sizeof(__pyx_k_cmdout), 0, 0, 1, 1},
    {&__pyx_n_u_color, __pyx_k_color, sizeof(__pyx_k_color), 0, 1, 0, 1},
    {&__pyx_n_s_columns, __pyx_k_columns, sizeof(__pyx_k_columns), 0, 0, 1, 1},
    {&__pyx_n_s_communicate, __pyx_k_communicate, sizeof(__pyx_k_communicate), 0, 0, 1, 1},
    {&__pyx_n_s_complement, __pyx_k_complement, sizeof(__pyx_k_complement), 0, 0, 1, 1},
    {&__pyx_n_s_compute_relation, __pyx_k_compute_relation, sizeof(__pyx_k_compute_relation), 0, 0, 1, 1},
    {&__pyx_n_s_concatenate, __pyx_k_concatenate, sizeof(__pyx_k_concatenate), 0, 0, 1, 1},
    {&__pyx_n_s_concatenate_with_result_state_ma, __pyx_k_concatenate_with_result_state_ma, sizeof(__pyx_k_concatenate_with_result_state_ma), 0, 0, 1, 1},
    {&__pyx_n_u_darkblue, __pyx_k_darkblue, sizeof(__pyx_k_darkblue), 0, 1, 0, 1},
    {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
    {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
    {&__pyx_n_s_determinize, __pyx_k_determinize, sizeof(__pyx_k_determinize), 0, 0, 1, 1},
    {&__pyx_n_s_determinize_with_subset_map, __pyx_k_determinize_with_subset_map, sizeof(__pyx_k_determinize_with_subset_map), 0, 0, 1, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_u_direction, __pyx_k_direction, sizeof(__pyx_k_direction), 0, 1, 0, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_s_divisible_by, __pyx_k_divisible_by, sizeof(__pyx_k_divisible_by), 0, 0, 1, 1},
    {&__pyx_kp_u_dot_O_T, __pyx_k_dot_O_T, sizeof(__pyx_k_dot_O_T), 0, 1, 0, 0},
    {&__pyx_n_u_edge_style, __pyx_k_edge_style, sizeof(__pyx_k_edge_style), 0, 1, 0, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
    {&__pyx_n_s_encode_word, __pyx_k_encode_word, sizeof(__pyx_k_encode_word), 0, 0, 1, 1},
    {&__pyx_n_s_encoding, __pyx_k_encoding, sizeof(__pyx_k_encoding), 0, 0, 1, 1},
    {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
    {&__pyx_n_s_epsilon, __pyx_k_epsilon, sizeof(__pyx_k_epsilon), 0, 0, 1, 1},
    {&__pyx_n_s_epsilon_transitions, __pyx_k_epsilon_transitions, sizeof(__pyx_k_epsilon_transitions), 0, 0, 1, 1},
    {&__pyx_n_s_equivalence_check, __pyx_k_equivalence_check, sizeof(__pyx_k_equivalence_check), 0, 0, 1, 1},
    {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
    {&__pyx_kp_u_error_while_dot_file, __pyx_k_error_while_dot_file, sizeof(__pyx_k_error_while_dot_file), 0, 1, 0, 0},
    {&__pyx_n_s_executed_command, __pyx_k_executed_command, sizeof(__pyx_k_executed_command), 0, 0, 1, 1},
    {&__pyx_n_u_false, __pyx_k_false, sizeof(__pyx_k_false), 0, 1, 0, 1},
    {&__pyx_n_u_fillcolor, __pyx_k_fillcolor, sizeof(__pyx_k_fillcolor), 0, 1, 0, 1},
    {&__pyx_n_u_filled, __pyx_k_filled, sizeof(__pyx_k_filled), 0, 1, 0, 1},
    {&__pyx_kp_u_final_states, __pyx_k_final_states, sizeof(__pyx_k_final_states), 0, 1, 0, 0},
    {&__pyx_n_u_fixedsize, __pyx_k_fixedsize, sizeof(__pyx_k_fixedsize), 0, 1, 0, 1},
    {&__pyx_n_u_fontname, __pyx_k_fontname, sizeof(__pyx_k_fontname), 0, 1, 0, 1},
    {&__pyx_n_u_fontsize, __pyx_k_fontsize, sizeof(__pyx_k_fontsize), 0, 1, 0, 1},
    {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
    {&__pyx_n_u_forward, __pyx_k_forward, sizeof(__pyx_k_forward), 0, 1, 0, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
    {&__pyx_n_s_get_alphabet_symbols, __pyx_k_get_alphabet_symbols, sizeof(__pyx_k_get_alphabet_symbols), 0, 0, 1, 1},
    {&__pyx_n_s_get_elements_from_bool_vec, __pyx_k_get_elements_from_bool_vec, sizeof(__pyx_k_get_elements_from_bool_vec), 0, 0, 1, 1},
    {&__pyx_n_s_get_epsilon_transitions, __pyx_k_get_epsilon_transitions, sizeof(__pyx_k_get_epsilon_transitions), 0, 0, 1, 1},
    {&__pyx_n_s_get_num_of_trans, __pyx_k_get_num_of_trans, sizeof(__pyx_k_get_num_of_trans), 0, 0, 1, 1},
    {&__pyx_n_s_get_one_letter_aut, __pyx_k_get_one_letter_aut, sizeof(__pyx_k_get_one_letter_aut), 0, 0, 1, 1},
    {&__pyx_n_s_get_reachable_states, __pyx_k_get_reachable_states, sizeof(__pyx_k_get_reachable_states), 0, 0, 1, 1},
    {&__pyx_n_s_get_symbols, __pyx_k_get_symbols, sizeof(__pyx_k_get_symbols), 0, 0, 1, 1},
    {&__pyx_n_s_get_terminating_states, __pyx_k_get_terminating_states, sizeof(__pyx_k_get_terminating_states), 0, 0, 1, 1},
    {&__pyx_n_s_get_trans_as_sequence, __pyx_k_get_trans_as_sequence, sizeof(__pyx_k_get_trans_as_sequence), 0, 0, 1, 1},
    {&__pyx_n_s_get_trans_from_state_as_sequence, __pyx_k_get_trans_from_state_as_sequence, sizeof(__pyx_k_get_trans_from_state_as_sequence), 0, 0, 1, 1},
    {&__pyx_n_s_get_transitions_from_state, __pyx_k_get_transitions_from_state, sizeof(__pyx_k_get_transitions_from_state), 0, 0, 1, 1},
    {&__pyx_n_s_get_transitions_to_state, __pyx_k_get_transitions_to_state, sizeof(__pyx_k_get_transitions_to_state), 0, 0, 1, 1},
    {&__pyx_n_s_get_useful_states, __pyx_k_get_useful_states, sizeof(__pyx_k_get_useful_states), 0, 0, 1, 1},
    {&__pyx_n_s_get_word_for_path, __pyx_k_get_word_for_path, sizeof(__pyx_k_get_word_for_path), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_s_graphviz_command, __pyx_k_graphviz_command, sizeof(__pyx_k_graphviz_command), 0, 0, 1, 1},
    {&__pyx_n_s_has_final_state, __pyx_k_has_final_state, sizeof(__pyx_k_has_final_state), 0, 0, 1, 1},
    {&__pyx_n_s_has_initial_state, __pyx_k_has_initial_state, sizeof(__pyx_k_has_initial_state), 0, 0, 1, 1},
    {&__pyx_n_s_has_transition, __pyx_k_has_transition, sizeof(__pyx_k_has_transition), 0, 0, 1, 1},
    {&__pyx_n_u_height, __pyx_k_height, sizeof(__pyx_k_height), 0, 1, 0, 1},
    {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_kp_u_initial_states, __pyx_k_initial_states, sizeof(__pyx_k_initial_states), 0, 1, 0, 0},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_n_s_input, __pyx_k_input, sizeof(__pyx_k_input), 0, 0, 1, 1},
    {&__pyx_n_s_input_states, __pyx_k_input_states, sizeof(__pyx_k_input_states), 0, 0, 1, 1},
    {&__pyx_n_s_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 0, 1, 1},
    {&__pyx_n_s_intersection, __pyx_k_intersection, sizeof(__pyx_k_intersection), 0, 0, 1, 1},
    {&__pyx_n_s_intersection_with_product_map, __pyx_k_intersection_with_product_map, sizeof(__pyx_k_intersection_with_product_map), 0, 0, 1, 1},
    {&__pyx_n_s_is_complete, __pyx_k_is_complete, sizeof(__pyx_k_is_complete), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_n_s_is_deterministic, __pyx_k_is_deterministic, sizeof(__pyx_k_is_deterministic), 0, 0, 1, 1},
    {&__pyx_n_s_is_epsilon, __pyx_k_is_epsilon, sizeof(__pyx_k_is_epsilon), 0, 0, 1, 1},
    {&__pyx_n_s_is_in_lang, __pyx_k_is_in_lang, sizeof(__pyx_k_is_in_lang), 0, 0, 1, 1},
    {&__pyx_n_s_is_included, __pyx_k_is_included, sizeof(__pyx_k_is_included), 0, 0, 1, 1},
    {&__pyx_n_s_is_included_with_cex, __pyx_k_is_included_with_cex, sizeof(__pyx_k_is_included_with_cex), 0, 0, 1, 1},
    {&__pyx_n_s_is_lang_empty, __pyx_k_is_lang_empty, sizeof(__pyx_k_is_lang_empty), 0, 0, 1, 1},
    {&__pyx_n_s_is_prefix_in_lang, __pyx_k_is_prefix_in_lang, sizeof(__pyx_k_is_prefix_in_lang), 0, 0, 1, 1},
    {&__pyx_n_s_is_state, __pyx_k_is_state, sizeof(__pyx_k_is_state), 0, 0, 1, 1},
    {&__pyx_n_s_is_universal, __pyx_k_is_universal, sizeof(__pyx_k_is_universal), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_it, __pyx_k_it, sizeof(__pyx_k_it), 0, 0, 1, 1},
    {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
    {&__pyx_n_s_iterate, __pyx_k_iterate, sizeof(__pyx_k_iterate), 0, 0, 1, 1},
    {&__pyx_n_s_iterator, __pyx_k_iterator, sizeof(__pyx_k_iterator), 0, 0, 1, 1},
    {&__pyx_n_s_iteritems, __pyx_k_iteritems, sizeof(__pyx_k_iteritems), 0, 0, 1, 1},
    {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
    {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
    {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
    {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
    {&__pyx_n_s_label, __pyx_k_label, sizeof(__pyx_k_label), 0, 0, 1, 1},
    {&__pyx_n_s_lhs, __pyx_k_lhs, sizeof(__pyx_k_lhs), 0, 0, 1, 1},
    {&__pyx_n_s_lhs_map, __pyx_k_lhs_map, sizeof(__pyx_k_lhs_map), 0, 0, 1, 1},
    {&__pyx_n_s_libmata_nfa_nfa, __pyx_k_libmata_nfa_nfa, sizeof(__pyx_k_libmata_nfa_nfa), 0, 0, 1, 1},
    {&__pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_k_libmata_nfa_nfa_pyx, sizeof(__pyx_k_libmata_nfa_nfa_pyx), 0, 0, 1, 0},
    {&__pyx_n_u_lightsteelblue, __pyx_k_lightsteelblue, sizeof(__pyx_k_lightsteelblue), 0, 1, 0, 1},
    {&__pyx_kp_s_list_int, __pyx_k_list_int, sizeof(__pyx_k_list_int), 0, 0, 1, 0},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_make_complete, __pyx_k_make_complete, sizeof(__pyx_k_make_complete), 0, 0, 1, 1},
    {&__pyx_n_s_make_final_state, __pyx_k_make_final_state, sizeof(__pyx_k_make_final_state), 0, 0, 1, 1},
    {&__pyx_n_s_make_final_states, __pyx_k_make_final_states, sizeof(__pyx_k_make_final_states), 0, 0, 1, 1},
    {&__pyx_n_s_make_initial_state, __pyx_k_make_initial_state, sizeof(__pyx_k_make_initial_state), 0, 0, 1, 1},
    {&__pyx_n_s_make_initial_states, __pyx_k_make_initial_states, sizeof(__pyx_k_make_initial_states), 0, 0, 1, 1},
    {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
    {&__pyx_n_u_midnightblue, __pyx_k_midnightblue, sizeof(__pyx_k_midnightblue), 0, 1, 0, 1},
    {&__pyx_n_s_minimize, __pyx_k_minimize, sizeof(__pyx_k_minimize), 0, 0, 1, 1},
    {&__pyx_n_u_minimize, __pyx_k_minimize, sizeof(__pyx_k_minimize), 0, 1, 0, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_neq, __pyx_k_neq, sizeof(__pyx_k_neq), 0, 0, 1, 1},
    {&__pyx_n_s_networkx, __pyx_k_networkx, sizeof(__pyx_k_networkx), 0, 0, 1, 1},
    {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
    {&__pyx_n_u_node_style, __pyx_k_node_style, sizeof(__pyx_k_node_style), 0, 1, 0, 1},
    {&__pyx_n_s_nx, __pyx_k_nx, sizeof(__pyx_k_nx), 0, 0, 1, 1},
    {&__pyx_kp_s_nx_Graph, __pyx_k_nx_Graph, sizeof(__pyx_k_nx_Graph), 0, 0, 1, 0},
    {&__pyx_n_s_objects, __pyx_k_objects, sizeof(__pyx_k_objects), 0, 0, 1, 1},
    {&__pyx_n_s_one_letter_aut, __pyx_k_one_letter_aut, sizeof(__pyx_k_one_letter_aut), 0, 0, 1, 1},
    {&__pyx_n_s_other, __pyx_k_other, sizeof(__pyx_k_other), 0, 0, 1, 1},
    {&__pyx_n_s_output, __pyx_k_output, sizeof(__pyx_k_output), 0, 0, 1, 1},
    {&__pyx_n_s_output_file, __pyx_k_output_file, sizeof(__pyx_k_output_file), 0, 0, 1, 1},
    {&__pyx_n_s_output_format, __pyx_k_output_format, sizeof(__pyx_k_output_format), 0, 0, 1, 1},
    {&__pyx_n_s_output_stream, __pyx_k_output_stream, sizeof(__pyx_k_output_stream), 0, 0, 1, 1},
    {&__pyx_n_s_pandas, __pyx_k_pandas, sizeof(__pyx_k_pandas), 0, 0, 1, 1},
    {&__pyx_kp_s_pandas_DataFrame, __pyx_k_pandas_DataFrame, sizeof(__pyx_k_pandas_DataFrame), 0, 0, 1, 0},
    {&__pyx_n_s_params, __pyx_k_params, sizeof(__pyx_k_params), 0, 0, 1, 1},
    {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
    {&__pyx_n_u_pdf, __pyx_k_pdf, sizeof(__pyx_k_pdf), 0, 1, 0, 1},
    {&__pyx_n_u_penwidth, __pyx_k_penwidth, sizeof(__pyx_k_penwidth), 0, 1, 0, 1},
    {&__pyx_n_s_piped_command, __pyx_k_piped_command, sizeof(__pyx_k_piped_command), 0, 0, 1, 1},
    {&__pyx_n_s_post_map_of, __pyx_k_post_map_of, sizeof(__pyx_k_post_map_of), 0, 0, 1, 1},
    {&__pyx_n_s_post_of, __pyx_k_post_of, sizeof(__pyx_k_post_of), 0, 0, 1, 1},
    {&__pyx_n_s_preserve_epsilon, __pyx_k_preserve_epsilon, sizeof(__pyx_k_preserve_epsilon), 0, 0, 1, 1},
    {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
    {&__pyx_n_s_process, __pyx_k_process, sizeof(__pyx_k_process), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_s_quiet, __pyx_k_quiet, sizeof(__pyx_k_quiet), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_u_red, __pyx_k_red, sizeof(__pyx_k_red), 0, 1, 0, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_2, __pyx_k_reduce_2, sizeof(__pyx_k_reduce_2), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_with_state_map, __pyx_k_reduce_with_state_map, sizeof(__pyx_k_reduce_with_state_map), 0, 0, 1, 1},
    {&__pyx_n_s_relation, __pyx_k_relation, sizeof(__pyx_k_relation), 0, 0, 1, 1},
    {&__pyx_n_u_relation, __pyx_k_relation, sizeof(__pyx_k_relation), 0, 1, 0, 1},
    {&__pyx_n_s_relation_size, __pyx_k_relation_size, sizeof(__pyx_k_relation_size), 0, 0, 1, 1},
    {&__pyx_n_s_remove_epsilon, __pyx_k_remove_epsilon, sizeof(__pyx_k_remove_epsilon), 0, 0, 1, 1},
    {&__pyx_n_s_remove_epsilon_inplace, __pyx_k_remove_epsilon_inplace, sizeof(__pyx_k_remove_epsilon_inplace), 0, 0, 1, 1},
    {&__pyx_n_s_remove_final_state, __pyx_k_remove_final_state, sizeof(__pyx_k_remove_final_state), 0, 0, 1, 1},
    {&__pyx_n_s_remove_initial_state, __pyx_k_remove_initial_state, sizeof(__pyx_k_remove_initial_state), 0, 0, 1, 1},
    {&__pyx_n_s_remove_trans, __pyx_k_remove_trans, sizeof(__pyx_k_remove_trans), 0, 0, 1, 1},
    {&__pyx_n_s_remove_trans_raw, __pyx_k_remove_trans_raw, sizeof(__pyx_k_remove_trans_raw), 0, 0, 1, 1},
    {&__pyx_n_s_resize, __pyx_k_resize, sizeof(__pyx_k_resize), 0, 0, 1, 1},
    {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
    {&__pyx_n_s_return, __pyx_k_return, sizeof(__pyx_k_return), 0, 0, 1, 1},
    {&__pyx_n_s_return_value, __pyx_k_return_value, sizeof(__pyx_k_return_value), 0, 0, 1, 1},
    {&__pyx_n_s_returncode, __pyx_k_returncode, sizeof(__pyx_k_returncode), 0, 0, 1, 1},
    {&__pyx_n_s_revert, __pyx_k_revert, sizeof(__pyx_k_revert), 0, 0, 1, 1},
    {&__pyx_n_s_rhs, __pyx_k_rhs, sizeof(__pyx_k_rhs), 0, 0, 1, 1},
    {&__pyx_n_s_rhs_map, __pyx_k_rhs_map, sizeof(__pyx_k_rhs_map), 0, 0, 1, 1},
    {&__pyx_n_s_run, __pyx_k_run, sizeof(__pyx_k_run), 0, 0, 1, 1},
    {&__pyx_n_s_run_safely_external_command, __pyx_k_run_safely_external_command, sizeof(__pyx_k_run_safely_external_command), 0, 0, 1, 1},
    {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
    {&__pyx_n_s_set, __pyx_k_set, sizeof(__pyx_k_set), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_shell, __pyx_k_shell, sizeof(__pyx_k_shell), 0, 0, 1, 1},
    {&__pyx_n_s_shlex, __pyx_k_shlex, sizeof(__pyx_k_shlex), 0, 0, 1, 1},
    {&__pyx_n_u_simulation, __pyx_k_simulation, sizeof(__pyx_k_simulation), 0, 1, 0, 1},
    {&__pyx_n_s_sink_state, __pyx_k_sink_state, sizeof(__pyx_k_sink_state), 0, 0, 1, 1},
    {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
    {&__pyx_n_u_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 1, 0, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
    {&__pyx_n_s_src, __pyx_k_src, sizeof(__pyx_k_src), 0, 0, 1, 1},
    {&__pyx_n_s_st, __pyx_k_st, sizeof(__pyx_k_st), 0, 0, 1, 1},
    {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
    {&__pyx_n_s_state_from, __pyx_k_state_from, sizeof(__pyx_k_state_from), 0, 0, 1, 1},
    {&__pyx_n_s_state_map, __pyx_k_state_map, sizeof(__pyx_k_state_map), 0, 0, 1, 1},
    {&__pyx_n_s_state_number, __pyx_k_state_number, sizeof(__pyx_k_state_number), 0, 0, 1, 1},
    {&__pyx_n_s_state_to, __pyx_k_state_to, sizeof(__pyx_k_state_to), 0, 0, 1, 1},
    {&__pyx_n_s_states, __pyx_k_states, sizeof(__pyx_k_states), 0, 0, 1, 1},
    {&__pyx_n_s_stderr, __pyx_k_stderr, sizeof(__pyx_k_stderr), 0, 0, 1, 1},
    {&__pyx_n_s_stdin, __pyx_k_stdin, sizeof(__pyx_k_stdin), 0, 0, 1, 1},
    {&__pyx_n_s_stdout, __pyx_k_stdout, sizeof(__pyx_k_stdout), 0, 0, 1, 1},
    {&__pyx_n_s_store, __pyx_k_store, sizeof(__pyx_k_store), 0, 0, 1, 1},
    {&__pyx_n_s_store_2, __pyx_k_store_2, sizeof(__pyx_k_store_2), 0, 0, 1, 1},
    {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
    {&__pyx_n_s_string_to_symbol, __pyx_k_string_to_symbol, sizeof(__pyx_k_string_to_symbol), 0, 0, 1, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
    {&__pyx_n_u_style, __pyx_k_style, sizeof(__pyx_k_style), 0, 1, 0, 1},
    {&__pyx_n_s_subprocess, __pyx_k_subprocess, sizeof(__pyx_k_subprocess), 0, 0, 1, 1},
    {&__pyx_n_s_subset_map, __pyx_k_subset_map, sizeof(__pyx_k_subset_map), 0, 0, 1, 1},
    {&__pyx_n_s_symb, __pyx_k_symb, sizeof(__pyx_k_symb), 0, 0, 1, 1},
    {&__pyx_n_s_symbol, __pyx_k_symbol, sizeof(__pyx_k_symbol), 0, 0, 1, 1},
    {&__pyx_n_u_symbol, __pyx_k_symbol, sizeof(__pyx_k_symbol), 0, 1, 0, 1},
    {&__pyx_n_s_symbol_map, __pyx_k_symbol_map, sizeof(__pyx_k_symbol_map), 0, 0, 1, 1},
    {&__pyx_n_s_symbols, __pyx_k_symbols, sizeof(__pyx_k_symbols), 0, 0, 1, 1},
    {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
    {&__pyx_n_u_target, __pyx_k_target, sizeof(__pyx_k_target), 0, 1, 0, 1},
    {&__pyx_n_s_targets, __pyx_k_targets, sizeof(__pyx_k_targets), 0, 0, 1, 1},
    {&__pyx_n_s_terminate, __pyx_k_terminate, sizeof(__pyx_k_terminate), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_tgt, __pyx_k_tgt, sizeof(__pyx_k_tgt), 0, 0, 1, 1},
    {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
    {&__pyx_n_s_timeout, __pyx_k_timeout, sizeof(__pyx_k_timeout), 0, 0, 1, 1},
    {&__pyx_n_s_to_dataframe, __pyx_k_to_dataframe, sizeof(__pyx_k_to_dataframe), 0, 0, 1, 1},
    {&__pyx_n_s_to_dot_file, __pyx_k_to_dot_file, sizeof(__pyx_k_to_dot_file), 0, 0, 1, 1},
    {&__pyx_n_s_to_dot_str, __pyx_k_to_dot_str, sizeof(__pyx_k_to_dot_str), 0, 0, 1, 1},
    {&__pyx_n_s_to_networkx_graph, __pyx_k_to_networkx_graph, sizeof(__pyx_k_to_networkx_graph), 0, 0, 1, 1},
    {&__pyx_n_s_tr, __pyx_k_tr, sizeof(__pyx_k_tr), 0, 0, 1, 1},
    {&__pyx_n_s_trans, __pyx_k_trans, sizeof(__pyx_k_trans), 0, 0, 1, 1},
    {&__pyx_kp_u_transitions, __pyx_k_transitions, sizeof(__pyx_k_transitions), 0, 1, 0, 0},
    {&__pyx_n_s_transitions_2, __pyx_k_transitions_2, sizeof(__pyx_k_transitions_2), 0, 0, 1, 1},
    {&__pyx_n_s_transitions_list, __pyx_k_transitions_list, sizeof(__pyx_k_transitions_list), 0, 0, 1, 1},
    {&__pyx_n_s_translate_symbol, __pyx_k_translate_symbol, sizeof(__pyx_k_translate_symbol), 0, 0, 1, 1},
    {&__pyx_n_s_transsymbols, __pyx_k_transsymbols, sizeof(__pyx_k_transsymbols), 0, 0, 1, 1},
    {&__pyx_n_s_trim, __pyx_k_trim, sizeof(__pyx_k_trim), 0, 0, 1, 1},
    {&__pyx_n_s_trim_input, __pyx_k_trim_input, sizeof(__pyx_k_trim_input), 0, 0, 1, 1},
    {&__pyx_n_s_trim_with_state_map, __pyx_k_trim_with_state_map, sizeof(__pyx_k_trim_with_state_map), 0, 0, 1, 1},
    {&__pyx_n_u_true, __pyx_k_true, sizeof(__pyx_k_true), 0, 1, 0, 1},
    {&__pyx_kp_s_tuple_Nfa_dict_str_str_dict_str, __pyx_k_tuple_Nfa_dict_str_str_dict_str, sizeof(__pyx_k_tuple_Nfa_dict_str_str_dict_str), 0, 0, 1, 0},
    {&__pyx_n_s_unify_final, __pyx_k_unify_final, sizeof(__pyx_k_unify_final), 0, 0, 1, 1},
    {&__pyx_n_s_unify_initial, __pyx_k_unify_initial, sizeof(__pyx_k_unify_initial), 0, 0, 1, 1},
    {&__pyx_n_s_union, __pyx_k_union, sizeof(__pyx_k_union), 0, 0, 1, 1},
    {&__pyx_n_s_unpiped_commands, __pyx_k_unpiped_commands, sizeof(__pyx_k_unpiped_commands), 0, 0, 1, 1},
    {&__pyx_n_s_use_epsilon, __pyx_k_use_epsilon, sizeof(__pyx_k_use_epsilon), 0, 0, 1, 1},
    {&__pyx_n_s_useful_states, __pyx_k_useful_states, sizeof(__pyx_k_useful_states), 0, 0, 1, 1},
    {&__pyx_n_s_useful_states_bool_vec, __pyx_k_useful_states_bool_vec, sizeof(__pyx_k_useful_states_bool_vec), 0, 0, 1, 1},
    {&__pyx_kp_u_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 1, 0, 0},
    {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
    {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
    {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
    {&__pyx_n_s_wait, __pyx_k_wait, sizeof(__pyx_k_wait), 0, 0, 1, 1},
    {&__pyx_n_u_width, __pyx_k_width, sizeof(__pyx_k_width), 0, 1, 0, 1},
    {&__pyx_kp_u_without_specified_alphabet, __pyx_k_without_specified_alphabet, sizeof(__pyx_k_without_specified_alphabet), 0, 1, 0, 0},
    {&__pyx_n_s_word, __pyx_k_word, sizeof(__pyx_k_word), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(1, 176, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(1, 566, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(1, 923, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(1, 1177, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 68, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "libmata/nfa/nfa.pyx":1180
 *     lhs = Nfa(k+1)
 *     lhs.make_initial_state(0)
 *     lhs.add_transition(0, 0, 0)             # <<<<<<<<<<<<<<
 *     for i in range(1, k + 1):
 *         lhs.add_transition(i - 1, 1, i)
 */
  __pyx_tuple__14 = PyTuple_Pack(3, __pyx_int_0, __pyx_int_0, __pyx_int_0); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(1, 1180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "libmata/nfa/nfa.pyx":30
 * cdef Symbol EPSILON = CEPSILON
 * 
 * def epsilon():             # <<<<<<<<<<<<<<
 *     return EPSILON
 * 
 */
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_epsilon, 30, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(1, 30, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_tuple__20 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":117
 *         return dereference(self.thisptr) == dereference(other.thisptr)
 * 
 *     def __neq__(self, Trans other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) != dereference(other.thisptr)
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_neq, 117, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(1, 117, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":172
 *         return self.symbol == other.symbol and self.targets == other.targets
 * 
 *     def __neq__(self, Move other):             # <<<<<<<<<<<<<<
 *         return self.symbol != other.symbol or self.targets != other.targets
 * 
 */
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_neq, 172, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(1, 172, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":232
 *             self.thisptr.get().final.insert(state)
 * 
 *     def is_state(self, state):             # <<<<<<<<<<<<<<
 *         """Tests if state is in the automaton
 * 
 */
  __pyx_tuple__29 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_state); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(1, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_is_state, 232, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(1, 232, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":240
 *         return self.thisptr.get().is_state(state)
 * 
 *     def add_new_state(self):             # <<<<<<<<<<<<<<
 *         """Adds new state to automaton
 * 
 */
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_add_new_state, 240, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(1, 240, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":247
 *         return self.thisptr.get().add_state()
 * 
 *     def add_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Adds passed state to the automaton.
 * 
 */
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_add_state, 247, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(1, 247, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":254
 *         return self.thisptr.get().add_state(state)
 * 
 *     def make_initial_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Makes specified state from the automaton initial.
 * 
 */
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_make_initial_state, 254, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(1, 254, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":261
 *         self.thisptr.get().initial.insert(state)
 * 
 *     def make_initial_states(self, vector[State] states):             # <<<<<<<<<<<<<<
 *         """Makes specified states from the automaton initial.
 * 
 */
  __pyx_tuple__34 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_states, __pyx_n_s_state); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(1, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_make_initial_states, 261, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(1, 261, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":269
 *             self.thisptr.get().initial.insert(state)
 * 
 *     def has_initial_state(self, State st):             # <<<<<<<<<<<<<<
 *         """Tests if automaton contains given state
 * 
 */
  __pyx_tuple__36 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_st); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(1, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_has_initial_state, 269, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(1, 269, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":277
 *         return self.thisptr.get().initial[st]
 * 
 *     def remove_initial_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Removes state from initial state set of the automaton.
 * 
 */
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_remove_initial_state, 277, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(1, 277, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":284
 *         self.thisptr.get().initial.erase(state)
 * 
 *     def clear_initial(self):             # <<<<<<<<<<<<<<
 *         """Clears initial state set of the automaton."""
 *         self.thisptr.get().initial.clear()
 */
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_clear_initial, 284, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(1, 284, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":288
 *         self.thisptr.get().initial.clear()
 * 
 *     def make_final_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Makes specified state from the automaton final.
 * 
 */
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_make_final_state, 288, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(1, 288, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":295
 *         self.thisptr.get().final.insert(state)
 * 
 *     def make_final_states(self, vector[State] states):             # <<<<<<<<<<<<<<
 *         """Makes specified states from the automaton final.
 * 
 */
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_make_final_states, 295, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(1, 295, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":303
 *             self.thisptr.get().final.insert(state)
 * 
 *     def has_final_state(self, State st):             # <<<<<<<<<<<<<<
 *         """Tests if automaton contains given state
 * 
 */
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_has_final_state, 303, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(1, 303, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":311
 *         return self.thisptr.get().final[st]
 * 
 *     def remove_final_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Removes state from final state set of the automaton.
 * 
 */
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_remove_final_state, 311, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(1, 311, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":318
 *         self.thisptr.get().final.erase(state)
 * 
 *     def clear_final(self):             # <<<<<<<<<<<<<<
 *         """Clears final state set of the automaton."""
 *         self.thisptr.get().final.clear()
 */
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_clear_final, 318, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(1, 318, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":322
 *         self.thisptr.get().final.clear()
 * 
 *     def unify_initial(self):             # <<<<<<<<<<<<<<
 *         """Unify initial states into a single new initial state."""
 *         self.thisptr.get().unify_initial()
 */
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_unify_initial, 322, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(1, 322, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":326
 *         self.thisptr.get().unify_initial()
 * 
 *     def unify_final(self):             # <<<<<<<<<<<<<<
 *         """Unify final states into a single new final state."""
 *         self.thisptr.get().unify_final()
 */
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_unify_final, 326, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(1, 326, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":330
 *         self.thisptr.get().unify_final()
 * 
 *     def add_transition_object(self, Trans tr):             # <<<<<<<<<<<<<<
 *         """Adds transition to automaton
 * 
 */
  __pyx_tuple__47 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_tr); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(1, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_add_transition_object, 330, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(1, 330, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":337
 *         self.thisptr.get().delta.add(dereference(tr.thisptr))
 * 
 *     def add_transition(self, State src, symb, State tgt, alph.Alphabet alphabet = None):             # <<<<<<<<<<<<<<
 *         """Constructs transition and adds it to automaton
 * 
 */
  __pyx_tuple__49 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_src, __pyx_n_s_symb, __pyx_n_s_tgt, __pyx_n_s_alphabet); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(1, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_add_transition, 337, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(1, 337, __pyx_L1_error)
  __pyx_tuple__51 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(1, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);

  /* "libmata/nfa/nfa.pyx":353
 *             self.thisptr.get().delta.add(src, symb, tgt)
 * 
 *     def remove_trans(self, Trans tr):             # <<<<<<<<<<<<<<
 *         """Removes transition from the automaton.
 * 
 */
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_remove_trans, 353, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(1, 353, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":360
 *         self.thisptr.get().delta.remove(dereference(tr.thisptr))
 * 
 *     def remove_trans_raw(self, State src, Symbol symb, State tgt):             # <<<<<<<<<<<<<<
 *         """Constructs transition and removes it from the automaton.
 * 
 */
  __pyx_tuple__53 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_src, __pyx_n_s_symb, __pyx_n_s_tgt); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(1, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_remove_trans_raw, 360, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(1, 360, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":369
 *         self.thisptr.get().delta.remove(src, symb, tgt)
 * 
 *     def has_transition(self, State src, Symbol symb, State tgt):             # <<<<<<<<<<<<<<
 *         """Tests if automaton contains transition
 * 
 */
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_has_transition, 369, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(1, 369, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":379
 *         return self.thisptr.get().delta.contains(src, symb, tgt)
 * 
 *     def get_num_of_trans(self):             # <<<<<<<<<<<<<<
 *         """Returns number of transitions in automaton
 * 
 */
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_num_of_trans, 379, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(1, 379, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":386
 *         return self.thisptr.get().get_num_of_trans()
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         """Clears all of the internals in the automaton"""
 *         self.thisptr.get().clear()
 */
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_clear, 386, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(1, 386, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":390
 *         self.thisptr.get().clear()
 * 
 *     def size(self) -> int:             # <<<<<<<<<<<<<<
 *         """Get the current number of states in the whole automaton.
 *         :return: The number of states.
 */
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_size, 390, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(1, 390, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":396
 *         return self.thisptr.get().size()
 * 
 *     def iterate(self):             # <<<<<<<<<<<<<<
 *         """Iterates over all transitions
 * 
 */
  __pyx_tuple__59 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_iterator, __pyx_n_s_trans, __pyx_n_s_lhs); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(1, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__59, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_iterate, 396, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(1, 396, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":409
 *             yield trans
 * 
 *     def get_transitions_from_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Returns list of Move for the given state
 * 
 */
  __pyx_tuple__60 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_transitions_2, __pyx_n_s_transitions_list, __pyx_n_s_it, __pyx_n_s_end, __pyx_n_s_transsymbols, __pyx_n_s_t); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(1, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_transitions_from_state, 409, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(1, 409, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":430
 *         return transsymbols
 * 
 *     def get_transitions_to_state(self, State state_to):             # <<<<<<<<<<<<<<
 *         """Get transitions leading to state_to (state_to is their target state).
 * 
 */
  __pyx_tuple__62 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_state_to, __pyx_n_s_c_transitions, __pyx_n_s_trans, __pyx_n_s_c_transition); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(1, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__62, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_transitions_to_state, 430, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(1, 430, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":441
 *         return trans
 * 
 *     def get_trans_as_sequence(self):             # <<<<<<<<<<<<<<
 *         """Get automaton transitions as a sequence.
 * 
 */
  __pyx_tuple__64 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_c_transitions, __pyx_n_s_transitions_2, __pyx_n_s_c_transition); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(1, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);
  __pyx_codeobj__65 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__64, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_trans_as_sequence, 441, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__65)) __PYX_ERR(1, 441, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":452
 *         return transitions
 * 
 *     def get_trans_from_state_as_sequence(self, State state_from):             # <<<<<<<<<<<<<<
 *         """Get automaton transitions from state_from as a sequence.
 * 
 */
  __pyx_tuple__66 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_state_from, __pyx_n_s_c_transitions, __pyx_n_s_transitions_2, __pyx_n_s_c_transition); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(1, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_trans_from_state_as_sequence, 452, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(1, 452, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":463
 *         return transitions
 * 
 *     def get_useful_states(self):             # <<<<<<<<<<<<<<
 *         """Get useful states (states which are reachable and terminating at the same time).
 * 
 */
  __pyx_tuple__68 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_useful_states_bool_vec, __pyx_n_s_useful_states, __pyx_n_s_state); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(1, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_useful_states, 463, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(1, 463, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":473
 *         return { state for state in useful_states }
 * 
 *     def get_reachable_states(self):             # <<<<<<<<<<<<<<
 *         """Get reachable states.
 * 
 */
  __pyx_tuple__70 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_return_value, __pyx_n_s_state); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(1, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);
  __pyx_codeobj__71 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__70, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_reachable_states, 473, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__71)) __PYX_ERR(1, 473, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":481
 *         return {state for state in return_value}
 * 
 *     def get_terminating_states(self):             # <<<<<<<<<<<<<<
 *         """Get terminating states (states with a path from them leading to any final state).
 * 
 */
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__70, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_terminating_states, 481, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(1, 481, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":489
 *         return {state for state in return_value}
 * 
 *     def trim(self):             # <<<<<<<<<<<<<<
 *         """Remove inaccessible (unreachable) and not co-accessible (non-terminating) states.
 * 
 */
  __pyx_codeobj__73 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_trim, 489, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__73)) __PYX_ERR(1, 489, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":498
 *         self.thisptr.get().trim(NULL)
 * 
 *     def trim_with_state_map(self):             # <<<<<<<<<<<<<<
 *         """Remove inaccessible (unreachable) and not co-accessible (non-terminating) states.
 * 
 */
  __pyx_tuple__74 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state_map, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(1, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__74, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_trim_with_state_map, 498, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(1, 498, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":511
 *         return {k: v for k, v in state_map}
 * 
 *     def get_one_letter_aut(self) -> Nfa:             # <<<<<<<<<<<<<<
 *         """Unify transitions to create a directed graph with at most a single transition between two states (using only
 *          one letter for all transitions).
 */
  __pyx_tuple__76 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_one_letter_aut); if (unlikely(!__pyx_tuple__76)) __PYX_ERR(1, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__76, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_one_letter_aut, 511, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(1, 511, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":521
 *         return one_letter_aut
 * 
 *     def is_epsilon(self, Symbol symbol) -> bool:             # <<<<<<<<<<<<<<
 *         """Check whether passed symbol is epsilon symbol or not.
 * 
 */
  __pyx_tuple__78 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_symbol); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(1, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);
  __pyx_codeobj__79 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__78, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_is_epsilon, 521, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__79)) __PYX_ERR(1, 521, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":546
 *         return str(self)
 * 
 *     def to_dot_file(self, output_file='aut.dot', output_format='pdf'):             # <<<<<<<<<<<<<<
 *         """Transforms the automaton to dot format.
 * 
 */
  __pyx_tuple__81 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_output_file, __pyx_n_s_output_format, __pyx_n_s_output, __pyx_n_s_graphviz_command, __pyx_n_s__80, __pyx_n_s_err); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(1, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);
  __pyx_codeobj__82 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__81, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_to_dot_file, 546, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__82)) __PYX_ERR(1, 546, __pyx_L1_error)
  __pyx_tuple__83 = PyTuple_Pack(2, __pyx_kp_u_aut_dot, __pyx_n_u_pdf); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(1, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);

  /* "libmata/nfa/nfa.pyx":568
 *             print(f"error while dot file: {err}")
 * 
 *     def to_dot_str(self, encoding='utf-8'):             # <<<<<<<<<<<<<<
 *         """Transforms the automaton to dot format string
 * 
 */
  __pyx_tuple__84 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_encoding, __pyx_n_s_output_stream, __pyx_n_s_result); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(1, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);
  __pyx_codeobj__85 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_to_dot_str, 568, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__85)) __PYX_ERR(1, 568, __pyx_L1_error)
  __pyx_tuple__86 = PyTuple_Pack(1, __pyx_kp_u_utf_8); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(1, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);

  /* "libmata/nfa/nfa.pyx":584
 *         return result.decode(encoding)
 * 
 *     def to_dataframe(self) -> pandas.DataFrame:             # <<<<<<<<<<<<<<
 *         """Transforms the automaton to DataFrame format.
 * 
 */
  __pyx_tuple__87 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_columns, __pyx_n_s_data, __pyx_n_s_trans); if (unlikely(!__pyx_tuple__87)) __PYX_ERR(1, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);
  __pyx_codeobj__88 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__87, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_to_dataframe, 584, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__88)) __PYX_ERR(1, 584, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":600
 *         return pandas.DataFrame(data, columns=columns)
 * 
 *     def to_networkx_graph(self) -> nx.Graph:             # <<<<<<<<<<<<<<
 *         """Transforms the automaton into networkx.Graph
 * 
 */
  __pyx_tuple__89 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_G, __pyx_n_s_trans); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(1, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);
  __pyx_codeobj__90 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__89, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_to_networkx_graph, 600, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__90)) __PYX_ERR(1, 600, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":616
 *         return G
 * 
 *     def post_map_of(self, State st, alph.Alphabet alphabet):             # <<<<<<<<<<<<<<
 *         """Returns mapping of symbols to set of states.
 * 
 */
  __pyx_tuple__91 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_st, __pyx_n_s_alphabet, __pyx_n_s_symbol_map, __pyx_n_s_symbol); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);
  __pyx_codeobj__92 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__91, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_post_map_of, 616, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__92)) __PYX_ERR(1, 616, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":628
 *         return symbol_map
 * 
 *     def post_of(self, vector[State] states, Symbol symbol):             # <<<<<<<<<<<<<<
 *         """Returns sets of reachable states from set of states through a symbol
 * 
 */
  __pyx_tuple__93 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_states, __pyx_n_s_symbol, __pyx_n_s_return_value, __pyx_n_s_input_states, __pyx_n_s_v); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(1, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);
  __pyx_codeobj__94 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_post_of, 628, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__94)) __PYX_ERR(1, 628, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":641
 * 
 *     # Operations
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def determinize_with_subset_map(cls, Nfa lhs):
 *         """Determinize the lhs automaton
 */
  __pyx_tuple__95 = PyTuple_Pack(4, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_result, __pyx_n_s_subset_map); if (unlikely(!__pyx_tuple__95)) __PYX_ERR(1, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);
  __pyx_codeobj__96 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_determinize_with_subset_map, 641, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__96)) __PYX_ERR(1, 641, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":653
 *         return result, subset_map_to_dictionary(subset_map)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def determinize(cls, Nfa lhs):
 *         """Determinize the lhs automaton
 */
  __pyx_tuple__97 = PyTuple_Pack(3, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_result); if (unlikely(!__pyx_tuple__97)) __PYX_ERR(1, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__97);
  __Pyx_GIVEREF(__pyx_tuple__97);
  __pyx_codeobj__98 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__97, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_determinize, 653, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__98)) __PYX_ERR(1, 653, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":664
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def union(cls, Nfa lhs, Nfa rhs):
 *         """Performs union of lhs and rhs
 */
  __pyx_tuple__99 = PyTuple_Pack(4, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_rhs, __pyx_n_s_result); if (unlikely(!__pyx_tuple__99)) __PYX_ERR(1, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);
  __pyx_codeobj__100 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__99, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_union, 664, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__100)) __PYX_ERR(1, 664, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":678
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def intersection(cls, Nfa lhs, Nfa rhs, preserve_epsilon: bool = False):
 *         """Performs intersection of lhs and rhs.
 */
  __pyx_tuple__101 = PyTuple_Pack(5, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_rhs, __pyx_n_s_preserve_epsilon, __pyx_n_s_result); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(1, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);
  __pyx_codeobj__102 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__101, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_intersection, 678, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__102)) __PYX_ERR(1, 678, __pyx_L1_error)
  __pyx_tuple__103 = PyTuple_Pack(1, Py_False); if (unlikely(!__pyx_tuple__103)) __PYX_ERR(1, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);

  /* "libmata/nfa/nfa.pyx":702
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def intersection_with_product_map(cls, Nfa lhs, Nfa rhs, preserve_epsilon: bool = False):
 *         """Performs intersection of lhs and rhs.
 */
  __pyx_tuple__104 = PyTuple_Pack(8, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_rhs, __pyx_n_s_preserve_epsilon, __pyx_n_s_result, __pyx_n_s_c_product_map, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(1, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);
  __pyx_codeobj__105 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__104, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_intersection_with_product_map, 702, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__105)) __PYX_ERR(1, 702, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":728
 *         return result, {tuple(k): v for k, v in c_product_map}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def concatenate(cls, Nfa lhs, Nfa rhs, use_epsilon: bool = False) -> Nfa:
 *         """Concatenate two NFAs.
 */
  __pyx_tuple__106 = PyTuple_Pack(5, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_rhs, __pyx_n_s_use_epsilon, __pyx_n_s_result); if (unlikely(!__pyx_tuple__106)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);
  __pyx_codeobj__107 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__106, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_concatenate, 728, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__107)) __PYX_ERR(1, 728, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":743
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def concatenate_with_result_state_maps(cls, Nfa lhs, Nfa rhs, use_epsilon: bool = False) \
 *             -> tuple[Nfa, dict[str, str], dict[str, str]]:
 */
  __pyx_tuple__108 = PyTuple_Pack(11, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_rhs, __pyx_n_s_use_epsilon, __pyx_n_s_result, __pyx_n_s_c_lhs_map, __pyx_n_s_c_rhs_map, __pyx_n_s_lhs_map, __pyx_n_s_key, __pyx_n_s_value, __pyx_n_s_rhs_map); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(1, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__108, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_concatenate_with_result_state_ma, 743, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(1, 743, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":768
 * 
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def complement(cls, Nfa lhs, alph.Alphabet alphabet, params = None):
 *         """Performs complement of lhs
 */
  __pyx_tuple__110 = PyTuple_Pack(7, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_alphabet, __pyx_n_s_params, __pyx_n_s_result, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);
  __pyx_codeobj__111 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__110, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_complement, 768, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__111)) __PYX_ERR(1, 768, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":792
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def make_complete(cls, Nfa lhs, State sink_state, alph.Alphabet alphabet):
 *         """Makes lhs complete
 */
  __pyx_tuple__112 = PyTuple_Pack(4, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_sink_state, __pyx_n_s_alphabet); if (unlikely(!__pyx_tuple__112)) __PYX_ERR(1, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);
  __pyx_codeobj__113 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__112, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_make_complete, 792, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__113)) __PYX_ERR(1, 792, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":804
 *         mata_nfa.make_complete(lhs.thisptr.get(), <CAlphabet&>dereference(alphabet.as_base()), sink_state)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def revert(cls, Nfa lhs):
 *         """Reverses transitions in the lhs
 */
  __pyx_codeobj__114 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__97, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_revert, 804, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__114)) __PYX_ERR(1, 804, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":815
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def remove_epsilon(cls, Nfa lhs, Symbol epsilon = CEPSILON):
 *         """Removes transitions that contain epsilon symbol.
 */
  __pyx_tuple__115 = PyTuple_Pack(4, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_epsilon, __pyx_n_s_result); if (unlikely(!__pyx_tuple__115)) __PYX_ERR(1, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__115);
  __Pyx_GIVEREF(__pyx_tuple__115);
  __pyx_codeobj__116 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__115, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_remove_epsilon, 815, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__116)) __PYX_ERR(1, 815, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":827
 *         return result
 * 
 *     def remove_epsilon_inplace(self, Symbol epsilon = CEPSILON):             # <<<<<<<<<<<<<<
 *         """Removes transitions which contain epsilon symbol.
 * 
 */
  __pyx_tuple__117 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_epsilon); if (unlikely(!__pyx_tuple__117)) __PYX_ERR(1, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);
  __pyx_codeobj__118 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__117, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_remove_epsilon_inplace, 827, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__118)) __PYX_ERR(1, 827, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":836
 *         self.thisptr.get().remove_epsilon(epsilon)
 * 
 *     def get_epsilon_transitions(self, State state, Symbol epsilon = CEPSILON) -> Move | None:             # <<<<<<<<<<<<<<
 *         """Get epsilon transitions for a state.
 * 
 */
  __pyx_tuple__119 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_epsilon, __pyx_n_s_c_epsilon_transitions_iter, __pyx_n_s_epsilon_transitions); if (unlikely(!__pyx_tuple__119)) __PYX_ERR(1, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__119);
  __Pyx_GIVEREF(__pyx_tuple__119);
  __pyx_codeobj__120 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_epsilon_transitions, 836, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__120)) __PYX_ERR(1, 836, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":853
 * 
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def minimize(cls, Nfa lhs):
 *         """Minimizes the automaton lhs
 */
  __pyx_codeobj__121 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__97, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_minimize, 853, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__121)) __PYX_ERR(1, 853, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":864
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def reduce_with_state_map(cls, Nfa aut, bool trim_input = True, params = None):
 *         """Reduce the automaton.
 */
  __pyx_tuple__122 = PyTuple_Pack(10, __pyx_n_s_cls, __pyx_n_s_aut, __pyx_n_s_trim_input, __pyx_n_s_params, __pyx_n_s_state_map, __pyx_n_s_result, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__122)) __PYX_ERR(1, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);
  __pyx_codeobj__123 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__122, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_reduce_with_state_map, 864, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__123)) __PYX_ERR(1, 864, __pyx_L1_error)
  __pyx_tuple__124 = PyTuple_Pack(2, Py_True, Py_None); if (unlikely(!__pyx_tuple__124)) __PYX_ERR(1, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);

  /* "libmata/nfa/nfa.pyx":885
 *         return result, {k: v for k, v in state_map}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def reduce(cls, Nfa aut, bool trim_input = True, params = None):
 *         """Reduce the automaton.
 */
  __pyx_tuple__125 = PyTuple_Pack(7, __pyx_n_s_cls, __pyx_n_s_aut, __pyx_n_s_trim_input, __pyx_n_s_params, __pyx_n_s_result, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__125)) __PYX_ERR(1, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);
  __pyx_codeobj__126 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__125, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_reduce_2, 885, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__126)) __PYX_ERR(1, 885, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":904
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def compute_relation(cls, Nfa lhs, params = None):
 *         """Computes the relation for the automaton
 */
  __pyx_tuple__127 = PyTuple_Pack(11, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_params, __pyx_n_s_relation, __pyx_n_s_result, __pyx_n_s_relation_size, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_val, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__127)) __PYX_ERR(1, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);
  __pyx_codeobj__128 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__127, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_compute_relation, 904, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__128)) __PYX_ERR(1, 904, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":930
 * 
 *     # Tests
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_deterministic(cls, Nfa lhs):
 *         """Tests if the lhs is determinstic
 */
  __pyx_tuple__129 = PyTuple_Pack(2, __pyx_n_s_cls, __pyx_n_s_lhs); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(1, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);
  __pyx_codeobj__130 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__129, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_is_deterministic, 930, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__130)) __PYX_ERR(1, 930, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":939
 *         return mata_nfa.is_deterministic(dereference(lhs.thisptr.get()))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_lang_empty(cls, Nfa lhs, Run run = None):
 *         """Checks if language of automaton lhs is empty, if not, returns path of states as counter
 */
  __pyx_tuple__131 = PyTuple_Pack(3, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_run); if (unlikely(!__pyx_tuple__131)) __PYX_ERR(1, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);
  __pyx_codeobj__132 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__131, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_is_lang_empty, 939, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__132)) __PYX_ERR(1, 939, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":952
 *             return mata_nfa.is_lang_empty(dereference(lhs.thisptr.get()), NULL)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_universal(cls, Nfa lhs, alph.Alphabet alphabet, params = None):
 *         """Tests if lhs is universal wrt given alphabet
 */
  __pyx_tuple__133 = PyTuple_Pack(6, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_alphabet, __pyx_n_s_params, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__133)) __PYX_ERR(1, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);
  __pyx_codeobj__134 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__133, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_is_universal, 952, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__134)) __PYX_ERR(1, 952, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":972
 *         )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_included_with_cex(
 *             cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None
 */
  __pyx_tuple__135 = PyTuple_Pack(10, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_rhs, __pyx_n_s_alphabet, __pyx_n_s_params, __pyx_n_s_run, __pyx_n_s_c_alphabet, __pyx_n_s_result, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__135)) __PYX_ERR(1, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);
  __pyx_codeobj__136 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_is_included_with_cex, 972, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__136)) __PYX_ERR(1, 972, __pyx_L1_error)
  __pyx_tuple__137 = PyTuple_Pack(2, Py_None, Py_None); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(1, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);

  /* "libmata/nfa/nfa.pyx":1001
 *         return result, run
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_included(
 *             cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None
 */
  __pyx_tuple__138 = PyTuple_Pack(9, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_rhs, __pyx_n_s_alphabet, __pyx_n_s_params, __pyx_n_s_c_alphabet, __pyx_n_s_result, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(1, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);
  __pyx_codeobj__139 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__138, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_is_included, 1001, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__139)) __PYX_ERR(1, 1001, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1029
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def equivalence_check(cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None) -> bool:
 *         """Test equivalence of two automata.
 */
  __pyx_tuple__140 = PyTuple_Pack(10, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_rhs, __pyx_n_s_alphabet, __pyx_n_s_params, __pyx_n_s_c_alphabet, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__140)) __PYX_ERR(1, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);
  __pyx_codeobj__141 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__140, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_equivalence_check, 1029, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__141)) __PYX_ERR(1, 1029, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1063
 *             )
 * 
 *     def get_symbols(self):             # <<<<<<<<<<<<<<
 *         """Return a set of symbols used on the transitions in NFA.
 * 
 */
  __pyx_tuple__142 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_symbols, __pyx_n_s_s); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(1, 1063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);
  __pyx_codeobj__143 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__142, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_symbols, 1063, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__143)) __PYX_ERR(1, 1063, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1071
 *         return {s for s in symbols}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_complete(cls, Nfa lhs, alph.Alphabet alphabet):
 *         """Test if automaton is complete
 */
  __pyx_tuple__144 = PyTuple_Pack(3, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_alphabet); if (unlikely(!__pyx_tuple__144)) __PYX_ERR(1, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);
  __pyx_codeobj__145 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__144, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_is_complete, 1071, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__145)) __PYX_ERR(1, 1071, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1084
 *         )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_in_lang(cls, Nfa lhs, vector[Symbol] word):
 *         """Tests if word is in language
 */
  __pyx_tuple__146 = PyTuple_Pack(4, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_word, __pyx_n_s_run); if (unlikely(!__pyx_tuple__146)) __PYX_ERR(1, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);
  __pyx_codeobj__147 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__146, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_is_in_lang, 1084, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__147)) __PYX_ERR(1, 1084, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1096
 *         return mata_nfa.is_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_prefix_in_lang(cls, Nfa lhs, vector[Symbol] word):
 *         """Test if any prefix of the word is in the language
 */
  __pyx_codeobj__148 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__146, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_is_prefix_in_lang, 1096, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__148)) __PYX_ERR(1, 1096, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1108
 *         return mata_nfa.is_prfx_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def accepts_epsilon(cls, Nfa lhs):
 *         """Tests if automaton accepts epsilon
 */
  __pyx_tuple__149 = PyTuple_Pack(3, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_state); if (unlikely(!__pyx_tuple__149)) __PYX_ERR(1, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);
  __pyx_codeobj__150 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_accepts_epsilon, 1108, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__150)) __PYX_ERR(1, 1108, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1122
 *     # Helper functions
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def get_word_for_path(cls, Nfa lhs, path):
 *         """For a given path (set of states) returns a corresponding word
 */
  __pyx_tuple__151 = PyTuple_Pack(5, __pyx_n_s_cls, __pyx_n_s_lhs, __pyx_n_s_path, __pyx_n_s_result, __pyx_n_s_input); if (unlikely(!__pyx_tuple__151)) __PYX_ERR(1, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__151);
  __Pyx_GIVEREF(__pyx_tuple__151);
  __pyx_codeobj__152 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__151, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_word_for_path, 1122, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__152)) __PYX_ERR(1, 1122, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1139
 *         return result.first.word, result.second
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def encode_word(cls, string_to_symbol, word):
 *         """Encodes word based on a string to symbol map
 */
  __pyx_tuple__153 = PyTuple_Pack(7, __pyx_n_s_cls, __pyx_n_s_string_to_symbol, __pyx_n_s_word, __pyx_n_s_result, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_s); if (unlikely(!__pyx_tuple__153)) __PYX_ERR(1, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__153);
  __Pyx_GIVEREF(__pyx_tuple__153);
  __pyx_codeobj__154 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__153, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_encode_word, 1139, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__154)) __PYX_ERR(1, 1139, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__155 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__155)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__156 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__156)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1175
 * 
 * # Temporary for testing
 * def divisible_by(k: int):             # <<<<<<<<<<<<<<
 *     """Constructs automaton accepting strings containing ones divisible by "k"."""
 *     assert k > 1
 */
  __pyx_tuple__157 = PyTuple_Pack(3, __pyx_n_s_k, __pyx_n_s_lhs, __pyx_n_s_i); if (unlikely(!__pyx_tuple__157)) __PYX_ERR(1, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__157);
  __Pyx_GIVEREF(__pyx_tuple__157);
  __pyx_codeobj__158 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__157, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_divisible_by, 1175, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__158)) __PYX_ERR(1, 1175, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1189
 * 
 * 
 * def run_safely_external_command(cmd: str, check_results=True, quiet=True, timeout=None, **kwargs):             # <<<<<<<<<<<<<<
 *     """Safely runs the piped command, without executing of the shell
 * 
 */
  __pyx_tuple__159 = PyTuple_Pack(16, __pyx_n_s_cmd, __pyx_n_s_check_results, __pyx_n_s_quiet, __pyx_n_s_timeout, __pyx_n_s_kwargs, __pyx_n_s_unpiped_commands, __pyx_n_s_cmd_no, __pyx_n_s_objects, __pyx_n_s_i, __pyx_n_s_executed_command, __pyx_n_s_stdin, __pyx_n_s_stderr, __pyx_n_s_piped_command, __pyx_n_s_cmdout, __pyx_n_s_cmderr, __pyx_n_s_process); if (unlikely(!__pyx_tuple__159)) __PYX_ERR(1, 1189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__159);
  __Pyx_GIVEREF(__pyx_tuple__159);
  __pyx_codeobj__160 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 16, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__159, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_run_safely_external_command, 1189, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__160)) __PYX_ERR(1, 1189, __pyx_L1_error)
  __pyx_tuple__161 = PyTuple_Pack(3, ((PyObject *)Py_True), ((PyObject *)Py_True), ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__161)) __PYX_ERR(1, 1189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__161);
  __Pyx_GIVEREF(__pyx_tuple__161);

  /* "libmata/nfa/nfa.pyx":1250
 * 
 * 
 * def get_elements_from_bool_vec(bool_vec: list[int]):             # <<<<<<<<<<<<<<
 *     cdef CBoolVector c_bool_vec = CBoolVector(bool_vec)
 *     cdef StateSet c_states
 */
  __pyx_tuple__162 = PyTuple_Pack(4, __pyx_n_s_bool_vec, __pyx_n_s_c_bool_vec, __pyx_n_s_c_states, __pyx_n_s_state); if (unlikely(!__pyx_tuple__162)) __PYX_ERR(1, 1250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__162);
  __Pyx_GIVEREF(__pyx_tuple__162);
  __pyx_codeobj__163 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__162, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_get_elements_from_bool_vec, 1250, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__163)) __PYX_ERR(1, 1250, __pyx_L1_error)

  /* "libmata/nfa/nfa.pyx":1277
 * 
 * 
 * def store():             # <<<<<<<<<<<<<<
 *     """
 *     Returns the configuration of the library
 */
  __pyx_codeobj__164 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_libmata_nfa_nfa_pyx, __pyx_n_s_store, 1277, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__164)) __PYX_ERR(1, 1277, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* AssertionsEnabled.init */
  __Pyx_init_assertions_enabled();

if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7libmata_3nfa_3nfa_Nfa = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7libmata_3nfa_3nfa_Nfa_spec, NULL); if (unlikely(!__pyx_ptype_7libmata_3nfa_3nfa_Nfa)) __PYX_ERR(1, 183, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7libmata_3nfa_3nfa_Nfa_spec, __pyx_ptype_7libmata_3nfa_3nfa_Nfa) < 0) __PYX_ERR(1, 183, __pyx_L1_error)
  #else
  __pyx_ptype_7libmata_3nfa_3nfa_Nfa = &__pyx_type_7libmata_3nfa_3nfa_Nfa;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7libmata_3nfa_3nfa_Nfa) < 0) __PYX_ERR(1, 183, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dictoffset && __pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, "__str__"); if (unlikely(!wrapper)) __PYX_ERR(1, 183, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7libmata_3nfa_3nfa_3Nfa_73__str__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7libmata_3nfa_3nfa_3Nfa_73__str__.doc = __pyx_doc_7libmata_3nfa_3nfa_3Nfa_73__str__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7libmata_3nfa_3nfa_3Nfa_73__str__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Nfa, (PyObject *) __pyx_ptype_7libmata_3nfa_3nfa_Nfa) < 0) __PYX_ERR(1, 183, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7libmata_3nfa_3nfa_Nfa) < 0) __PYX_ERR(1, 183, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7libmata_3nfa_3nfa_Trans = &__pyx_vtable_7libmata_3nfa_3nfa_Trans;
  __pyx_vtable_7libmata_3nfa_3nfa_Trans.copy_from = (PyObject *(*)(struct __pyx_obj_7libmata_3nfa_3nfa_Trans *, Mata::Nfa::Trans))__pyx_f_7libmata_3nfa_3nfa_5Trans_copy_from;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7libmata_3nfa_3nfa_Trans = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7libmata_3nfa_3nfa_Trans_spec, NULL); if (unlikely(!__pyx_ptype_7libmata_3nfa_3nfa_Trans)) __PYX_ERR(1, 67, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7libmata_3nfa_3nfa_Trans_spec, __pyx_ptype_7libmata_3nfa_3nfa_Trans) < 0) __PYX_ERR(1, 67, __pyx_L1_error)
  #else
  __pyx_ptype_7libmata_3nfa_3nfa_Trans = &__pyx_type_7libmata_3nfa_3nfa_Trans;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7libmata_3nfa_3nfa_Trans) < 0) __PYX_ERR(1, 67, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7libmata_3nfa_3nfa_Trans->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7libmata_3nfa_3nfa_Trans->tp_dictoffset && __pyx_ptype_7libmata_3nfa_3nfa_Trans->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7libmata_3nfa_3nfa_Trans->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7libmata_3nfa_3nfa_Trans, __pyx_vtabptr_7libmata_3nfa_3nfa_Trans) < 0) __PYX_ERR(1, 67, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7libmata_3nfa_3nfa_Trans) < 0) __PYX_ERR(1, 67, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Trans, (PyObject *) __pyx_ptype_7libmata_3nfa_3nfa_Trans) < 0) __PYX_ERR(1, 67, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7libmata_3nfa_3nfa_Trans) < 0) __PYX_ERR(1, 67, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7libmata_3nfa_3nfa_Run = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7libmata_3nfa_3nfa_Run_spec, NULL); if (unlikely(!__pyx_ptype_7libmata_3nfa_3nfa_Run)) __PYX_ERR(1, 33, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7libmata_3nfa_3nfa_Run_spec, __pyx_ptype_7libmata_3nfa_3nfa_Run) < 0) __PYX_ERR(1, 33, __pyx_L1_error)
  #else
  __pyx_ptype_7libmata_3nfa_3nfa_Run = &__pyx_type_7libmata_3nfa_3nfa_Run;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7libmata_3nfa_3nfa_Run) < 0) __PYX_ERR(1, 33, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7libmata_3nfa_3nfa_Run->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7libmata_3nfa_3nfa_Run->tp_dictoffset && __pyx_ptype_7libmata_3nfa_3nfa_Run->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7libmata_3nfa_3nfa_Run->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Run, (PyObject *) __pyx_ptype_7libmata_3nfa_3nfa_Run) < 0) __PYX_ERR(1, 33, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7libmata_3nfa_3nfa_Run) < 0) __PYX_ERR(1, 33, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7libmata_3nfa_3nfa_Move = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7libmata_3nfa_3nfa_Move_spec, NULL); if (unlikely(!__pyx_ptype_7libmata_3nfa_3nfa_Move)) __PYX_ERR(1, 127, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7libmata_3nfa_3nfa_Move_spec, __pyx_ptype_7libmata_3nfa_3nfa_Move) < 0) __PYX_ERR(1, 127, __pyx_L1_error)
  #else
  __pyx_ptype_7libmata_3nfa_3nfa_Move = &__pyx_type_7libmata_3nfa_3nfa_Move;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7libmata_3nfa_3nfa_Move) < 0) __PYX_ERR(1, 127, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7libmata_3nfa_3nfa_Move->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7libmata_3nfa_3nfa_Move->tp_dictoffset && __pyx_ptype_7libmata_3nfa_3nfa_Move->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7libmata_3nfa_3nfa_Move->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Move, (PyObject *) __pyx_ptype_7libmata_3nfa_3nfa_Move) < 0) __PYX_ERR(1, 127, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7libmata_3nfa_3nfa_Move) < 0) __PYX_ERR(1, 127, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7libmata_3nfa_3nfa___pyx_scope_struct__iterate_spec, NULL); if (unlikely(!__pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate)) __PYX_ERR(1, 396, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7libmata_3nfa_3nfa___pyx_scope_struct__iterate_spec, __pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate) < 0) __PYX_ERR(1, 396, __pyx_L1_error)
  #else
  __pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate = &__pyx_type_7libmata_3nfa_3nfa___pyx_scope_struct__iterate;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate) < 0) __PYX_ERR(1, 396, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate->tp_dictoffset && __pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7libmata_3nfa_3nfa___pyx_scope_struct__iterate->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("libmata.utils"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7libmata_5utils_BinaryRelation = __Pyx_ImportType_3_0_0(__pyx_t_1, "libmata.utils", "BinaryRelation", sizeof(struct __pyx_obj_7libmata_5utils_BinaryRelation), __PYX_GET_STRUCT_ALIGNMENT_3_0_0(struct __pyx_obj_7libmata_5utils_BinaryRelation),__Pyx_ImportType_CheckSize_Warn_3_0_0); if (!__pyx_ptype_7libmata_5utils_BinaryRelation) __PYX_ERR(2, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("libmata.alphabets"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7libmata_9alphabets_Alphabet = __Pyx_ImportType_3_0_0(__pyx_t_1, "libmata.alphabets", "Alphabet", sizeof(struct __pyx_obj_7libmata_9alphabets_Alphabet), __PYX_GET_STRUCT_ALIGNMENT_3_0_0(struct __pyx_obj_7libmata_9alphabets_Alphabet),__Pyx_ImportType_CheckSize_Warn_3_0_0); if (!__pyx_ptype_7libmata_9alphabets_Alphabet) __PYX_ERR(3, 33, __pyx_L1_error)
  __pyx_vtabptr_7libmata_9alphabets_Alphabet = (struct __pyx_vtabstruct_7libmata_9alphabets_Alphabet*)__Pyx_GetVtable(__pyx_ptype_7libmata_9alphabets_Alphabet); if (unlikely(!__pyx_vtabptr_7libmata_9alphabets_Alphabet)) __PYX_ERR(3, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_nfa(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_nfa},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "nfa",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initnfa(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initnfa(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_nfa(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_nfa(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_nfa(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'nfa' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("nfa", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to nfa pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_nfa(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_libmata__nfa__nfa) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(1, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "libmata.nfa.nfa")) {
      if (unlikely((PyDict_SetItemString(modules, "libmata.nfa.nfa", __pyx_m) < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif

  /* "libmata/nfa/nfa.pyx":1
 * import shlex             # <<<<<<<<<<<<<<
 * import subprocess
 * import pandas
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_shlex, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_shlex, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":2
 * import shlex
 * import subprocess             # <<<<<<<<<<<<<<
 * import pandas
 * import networkx as nx
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_subprocess, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_subprocess, __pyx_t_2) < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":3
 * import shlex
 * import subprocess
 * import pandas             # <<<<<<<<<<<<<<
 * import networkx as nx
 * 
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_pandas, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pandas, __pyx_t_2) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":4
 * import subprocess
 * import pandas
 * import networkx as nx             # <<<<<<<<<<<<<<
 * 
 * from libc.stdint cimport uint8_t
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_networkx, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nx, __pyx_t_2) < 0) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":28
 * from libmata.utils cimport COrdVector, CBinaryRelation, BinaryRelation
 * 
 * cdef Symbol EPSILON = CEPSILON             # <<<<<<<<<<<<<<
 * 
 * def epsilon():
 */
  __pyx_v_7libmata_3nfa_3nfa_EPSILON = Mata::Nfa::EPSILON;

  /* "libmata/nfa/nfa.pyx":30
 * cdef Symbol EPSILON = CEPSILON
 * 
 * def epsilon():             # <<<<<<<<<<<<<<
 *     return EPSILON
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_1epsilon, 0, __pyx_n_s_epsilon, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__17)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_epsilon, __pyx_t_2) < 0) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Run_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Run___reduce_cython, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__19)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Run_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Run___setstate_cython, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__21)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":117
 *         return dereference(self.thisptr) == dereference(other.thisptr)
 * 
 *     def __neq__(self, Trans other):             # <<<<<<<<<<<<<<
 *         return dereference(self.thisptr) != dereference(other.thisptr)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_5Trans_7__neq__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Trans___neq, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Trans->tp_dict, __pyx_n_s_neq, __pyx_t_2) < 0) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Trans);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_5Trans_13__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Trans___reduce_cython, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__24)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_5Trans_15__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Trans___setstate_cython, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":172
 *         return self.symbol == other.symbol and self.targets == other.targets
 * 
 *     def __neq__(self, Move other):             # <<<<<<<<<<<<<<
 *         return self.symbol != other.symbol or self.targets != other.targets
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_4Move_15__neq__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Move___neq, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__26)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Move->tp_dict, __pyx_n_s_neq, __pyx_t_2) < 0) __PYX_ERR(1, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Move);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_4Move_21__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Move___reduce_cython, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_4Move_23__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Move___setstate_cython, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":232
 *             self.thisptr.get().final.insert(state)
 * 
 *     def is_state(self, state):             # <<<<<<<<<<<<<<
 *         """Tests if state is in the automaton
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_3is_state, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_is_state, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_state, __pyx_t_2) < 0) __PYX_ERR(1, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":240
 *         return self.thisptr.get().is_state(state)
 * 
 *     def add_new_state(self):             # <<<<<<<<<<<<<<
 *         """Adds new state to automaton
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_5add_new_state, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_add_new_state, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_add_new_state, __pyx_t_2) < 0) __PYX_ERR(1, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":247
 *         return self.thisptr.get().add_state()
 * 
 *     def add_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Adds passed state to the automaton.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_7add_state, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_add_state, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_add_state, __pyx_t_2) < 0) __PYX_ERR(1, 247, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":254
 *         return self.thisptr.get().add_state(state)
 * 
 *     def make_initial_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Makes specified state from the automaton initial.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_9make_initial_state, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_make_initial_state, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_make_initial_state, __pyx_t_2) < 0) __PYX_ERR(1, 254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":261
 *         self.thisptr.get().initial.insert(state)
 * 
 *     def make_initial_states(self, vector[State] states):             # <<<<<<<<<<<<<<
 *         """Makes specified states from the automaton initial.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_11make_initial_states, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_make_initial_states, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_make_initial_states, __pyx_t_2) < 0) __PYX_ERR(1, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":269
 *             self.thisptr.get().initial.insert(state)
 * 
 *     def has_initial_state(self, State st):             # <<<<<<<<<<<<<<
 *         """Tests if automaton contains given state
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_13has_initial_state, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_has_initial_state, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_has_initial_state, __pyx_t_2) < 0) __PYX_ERR(1, 269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":277
 *         return self.thisptr.get().initial[st]
 * 
 *     def remove_initial_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Removes state from initial state set of the automaton.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_15remove_initial_state, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_remove_initial_state, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_remove_initial_state, __pyx_t_2) < 0) __PYX_ERR(1, 277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":284
 *         self.thisptr.get().initial.erase(state)
 * 
 *     def clear_initial(self):             # <<<<<<<<<<<<<<
 *         """Clears initial state set of the automaton."""
 *         self.thisptr.get().initial.clear()
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_17clear_initial, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_clear_initial, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_clear_initial, __pyx_t_2) < 0) __PYX_ERR(1, 284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":288
 *         self.thisptr.get().initial.clear()
 * 
 *     def make_final_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Makes specified state from the automaton final.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_19make_final_state, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_make_final_state, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_make_final_state, __pyx_t_2) < 0) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":295
 *         self.thisptr.get().final.insert(state)
 * 
 *     def make_final_states(self, vector[State] states):             # <<<<<<<<<<<<<<
 *         """Makes specified states from the automaton final.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_21make_final_states, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_make_final_states, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_make_final_states, __pyx_t_2) < 0) __PYX_ERR(1, 295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":303
 *             self.thisptr.get().final.insert(state)
 * 
 *     def has_final_state(self, State st):             # <<<<<<<<<<<<<<
 *         """Tests if automaton contains given state
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_23has_final_state, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_has_final_state, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_has_final_state, __pyx_t_2) < 0) __PYX_ERR(1, 303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":311
 *         return self.thisptr.get().final[st]
 * 
 *     def remove_final_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Removes state from final state set of the automaton.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_25remove_final_state, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_remove_final_state, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_remove_final_state, __pyx_t_2) < 0) __PYX_ERR(1, 311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":318
 *         self.thisptr.get().final.erase(state)
 * 
 *     def clear_final(self):             # <<<<<<<<<<<<<<
 *         """Clears final state set of the automaton."""
 *         self.thisptr.get().final.clear()
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_27clear_final, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_clear_final, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_clear_final, __pyx_t_2) < 0) __PYX_ERR(1, 318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":322
 *         self.thisptr.get().final.clear()
 * 
 *     def unify_initial(self):             # <<<<<<<<<<<<<<
 *         """Unify initial states into a single new initial state."""
 *         self.thisptr.get().unify_initial()
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_29unify_initial, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_unify_initial, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_unify_initial, __pyx_t_2) < 0) __PYX_ERR(1, 322, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":326
 *         self.thisptr.get().unify_initial()
 * 
 *     def unify_final(self):             # <<<<<<<<<<<<<<
 *         """Unify final states into a single new final state."""
 *         self.thisptr.get().unify_final()
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_31unify_final, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_unify_final, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_unify_final, __pyx_t_2) < 0) __PYX_ERR(1, 326, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":330
 *         self.thisptr.get().unify_final()
 * 
 *     def add_transition_object(self, Trans tr):             # <<<<<<<<<<<<<<
 *         """Adds transition to automaton
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_33add_transition_object, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_add_transition_object, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_add_transition_object, __pyx_t_2) < 0) __PYX_ERR(1, 330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":337
 *         self.thisptr.get().delta.add(dereference(tr.thisptr))
 * 
 *     def add_transition(self, State src, symb, State tgt, alph.Alphabet alphabet = None):             # <<<<<<<<<<<<<<
 *         """Constructs transition and adds it to automaton
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_35add_transition, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_add_transition, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__51);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_add_transition, __pyx_t_2) < 0) __PYX_ERR(1, 337, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":353
 *             self.thisptr.get().delta.add(src, symb, tgt)
 * 
 *     def remove_trans(self, Trans tr):             # <<<<<<<<<<<<<<
 *         """Removes transition from the automaton.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_37remove_trans, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_remove_trans, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_remove_trans, __pyx_t_2) < 0) __PYX_ERR(1, 353, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":360
 *         self.thisptr.get().delta.remove(dereference(tr.thisptr))
 * 
 *     def remove_trans_raw(self, State src, Symbol symb, State tgt):             # <<<<<<<<<<<<<<
 *         """Constructs transition and removes it from the automaton.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_39remove_trans_raw, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_remove_trans_raw, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_remove_trans_raw, __pyx_t_2) < 0) __PYX_ERR(1, 360, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":369
 *         self.thisptr.get().delta.remove(src, symb, tgt)
 * 
 *     def has_transition(self, State src, Symbol symb, State tgt):             # <<<<<<<<<<<<<<
 *         """Tests if automaton contains transition
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_41has_transition, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_has_transition, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_has_transition, __pyx_t_2) < 0) __PYX_ERR(1, 369, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":379
 *         return self.thisptr.get().delta.contains(src, symb, tgt)
 * 
 *     def get_num_of_trans(self):             # <<<<<<<<<<<<<<
 *         """Returns number of transitions in automaton
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_43get_num_of_trans, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_get_num_of_trans, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__56)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_num_of_trans, __pyx_t_2) < 0) __PYX_ERR(1, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":386
 *         return self.thisptr.get().get_num_of_trans()
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         """Clears all of the internals in the automaton"""
 *         self.thisptr.get().clear()
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_45clear, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_clear, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__57)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_clear, __pyx_t_2) < 0) __PYX_ERR(1, 386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":390
 *         self.thisptr.get().clear()
 * 
 *     def size(self) -> int:             # <<<<<<<<<<<<<<
 *         """Get the current number of states in the whole automaton.
 *         :return: The number of states.
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_int) < 0) __PYX_ERR(1, 390, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_47size, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_size, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_size, __pyx_t_3) < 0) __PYX_ERR(1, 390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":396
 *         return self.thisptr.get().size()
 * 
 *     def iterate(self):             # <<<<<<<<<<<<<<
 *         """Iterates over all transitions
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_49iterate, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_iterate, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__7)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_iterate, __pyx_t_3) < 0) __PYX_ERR(1, 396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":409
 *             yield trans
 * 
 *     def get_transitions_from_state(self, State state):             # <<<<<<<<<<<<<<
 *         """Returns list of Move for the given state
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_52get_transitions_from_state, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_get_transitions_from_state, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_transitions_from_state, __pyx_t_3) < 0) __PYX_ERR(1, 409, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":430
 *         return transsymbols
 * 
 *     def get_transitions_to_state(self, State state_to):             # <<<<<<<<<<<<<<
 *         """Get transitions leading to state_to (state_to is their target state).
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_54get_transitions_to_state, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_get_transitions_to_state, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__63)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_transitions_to_state, __pyx_t_3) < 0) __PYX_ERR(1, 430, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":441
 *         return trans
 * 
 *     def get_trans_as_sequence(self):             # <<<<<<<<<<<<<<
 *         """Get automaton transitions as a sequence.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_56get_trans_as_sequence, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_get_trans_as_sequence, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__65)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_trans_as_sequence, __pyx_t_3) < 0) __PYX_ERR(1, 441, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":452
 *         return transitions
 * 
 *     def get_trans_from_state_as_sequence(self, State state_from):             # <<<<<<<<<<<<<<
 *         """Get automaton transitions from state_from as a sequence.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_58get_trans_from_state_as_sequence, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_get_trans_from_state_as_sequ, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_trans_from_state_as_sequence, __pyx_t_3) < 0) __PYX_ERR(1, 452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":463
 *         return transitions
 * 
 *     def get_useful_states(self):             # <<<<<<<<<<<<<<
 *         """Get useful states (states which are reachable and terminating at the same time).
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_60get_useful_states, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_get_useful_states, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__69)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_useful_states, __pyx_t_3) < 0) __PYX_ERR(1, 463, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":473
 *         return { state for state in useful_states }
 * 
 *     def get_reachable_states(self):             # <<<<<<<<<<<<<<
 *         """Get reachable states.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_62get_reachable_states, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_get_reachable_states, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__71)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_reachable_states, __pyx_t_3) < 0) __PYX_ERR(1, 473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":481
 *         return {state for state in return_value}
 * 
 *     def get_terminating_states(self):             # <<<<<<<<<<<<<<
 *         """Get terminating states (states with a path from them leading to any final state).
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_64get_terminating_states, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_get_terminating_states, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_terminating_states, __pyx_t_3) < 0) __PYX_ERR(1, 481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":489
 *         return {state for state in return_value}
 * 
 *     def trim(self):             # <<<<<<<<<<<<<<
 *         """Remove inaccessible (unreachable) and not co-accessible (non-terminating) states.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_66trim, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_trim, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__73)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_trim, __pyx_t_3) < 0) __PYX_ERR(1, 489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":498
 *         self.thisptr.get().trim(NULL)
 * 
 *     def trim_with_state_map(self):             # <<<<<<<<<<<<<<
 *         """Remove inaccessible (unreachable) and not co-accessible (non-terminating) states.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_68trim_with_state_map, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_trim_with_state_map, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__75)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_trim_with_state_map, __pyx_t_3) < 0) __PYX_ERR(1, 498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":511
 *         return {k: v for k, v in state_map}
 * 
 *     def get_one_letter_aut(self) -> Nfa:             # <<<<<<<<<<<<<<
 *         """Unify transitions to create a directed graph with at most a single transition between two states (using only
 *          one letter for all transitions).
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_Nfa) < 0) __PYX_ERR(1, 511, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_70get_one_letter_aut, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_get_one_letter_aut, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__77)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_one_letter_aut, __pyx_t_2) < 0) __PYX_ERR(1, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":521
 *         return one_letter_aut
 * 
 *     def is_epsilon(self, Symbol symbol) -> bool:             # <<<<<<<<<<<<<<
 *         """Check whether passed symbol is epsilon symbol or not.
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(1, 521, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_72is_epsilon, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_is_epsilon, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_epsilon, __pyx_t_3) < 0) __PYX_ERR(1, 521, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":546
 *         return str(self)
 * 
 *     def to_dot_file(self, output_file='aut.dot', output_format='pdf'):             # <<<<<<<<<<<<<<
 *         """Transforms the automaton to dot format.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_78to_dot_file, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_to_dot_file, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__83);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_to_dot_file, __pyx_t_3) < 0) __PYX_ERR(1, 546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":568
 *             print(f"error while dot file: {err}")
 * 
 *     def to_dot_str(self, encoding='utf-8'):             # <<<<<<<<<<<<<<
 *         """Transforms the automaton to dot format string
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_80to_dot_str, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_to_dot_str, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__86);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_to_dot_str, __pyx_t_3) < 0) __PYX_ERR(1, 568, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":584
 *         return result.decode(encoding)
 * 
 *     def to_dataframe(self) -> pandas.DataFrame:             # <<<<<<<<<<<<<<
 *         """Transforms the automaton to DataFrame format.
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_kp_s_pandas_DataFrame) < 0) __PYX_ERR(1, 584, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_82to_dataframe, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_to_dataframe, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_to_dataframe, __pyx_t_2) < 0) __PYX_ERR(1, 584, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":600
 *         return pandas.DataFrame(data, columns=columns)
 * 
 *     def to_networkx_graph(self) -> nx.Graph:             # <<<<<<<<<<<<<<
 *         """Transforms the automaton into networkx.Graph
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_kp_s_nx_Graph) < 0) __PYX_ERR(1, 600, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_84to_networkx_graph, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_to_networkx_graph, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__90)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_to_networkx_graph, __pyx_t_3) < 0) __PYX_ERR(1, 600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":616
 *         return G
 * 
 *     def post_map_of(self, State st, alph.Alphabet alphabet):             # <<<<<<<<<<<<<<
 *         """Returns mapping of symbols to set of states.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_86post_map_of, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_post_map_of, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_post_map_of, __pyx_t_3) < 0) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":628
 *         return symbol_map
 * 
 *     def post_of(self, vector[State] states, Symbol symbol):             # <<<<<<<<<<<<<<
 *         """Returns sets of reachable states from set of states through a symbol
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_88post_of, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_post_of, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_post_of, __pyx_t_3) < 0) __PYX_ERR(1, 628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":641
 * 
 *     # Operations
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def determinize_with_subset_map(cls, Nfa lhs):
 *         """Determinize the lhs automaton
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_90determinize_with_subset_map, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_determinize_with_subset_map, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__96)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_determinize_with_subset_map, __pyx_t_3) < 0) __PYX_ERR(1, 641, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_determinize_with_subset_map); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_determinize_with_subset_map, __pyx_t_2) < 0) __PYX_ERR(1, 641, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":653
 *         return result, subset_map_to_dictionary(subset_map)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def determinize(cls, Nfa lhs):
 *         """Determinize the lhs automaton
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_92determinize, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_determinize, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__98)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_determinize, __pyx_t_2) < 0) __PYX_ERR(1, 653, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_determinize); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_determinize, __pyx_t_3) < 0) __PYX_ERR(1, 653, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":664
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def union(cls, Nfa lhs, Nfa rhs):
 *         """Performs union of lhs and rhs
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_94union, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_union, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_union, __pyx_t_3) < 0) __PYX_ERR(1, 664, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_union); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_union, __pyx_t_2) < 0) __PYX_ERR(1, 664, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":678
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def intersection(cls, Nfa lhs, Nfa rhs, preserve_epsilon: bool = False):
 *         """Performs intersection of lhs and rhs.
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_preserve_epsilon, __pyx_n_s_bool) < 0) __PYX_ERR(1, 678, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_96intersection, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_intersection, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__102)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__103);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_intersection, __pyx_t_3) < 0) __PYX_ERR(1, 678, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_intersection); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_intersection, __pyx_t_2) < 0) __PYX_ERR(1, 678, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":702
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def intersection_with_product_map(cls, Nfa lhs, Nfa rhs, preserve_epsilon: bool = False):
 *         """Performs intersection of lhs and rhs.
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_preserve_epsilon, __pyx_n_s_bool) < 0) __PYX_ERR(1, 702, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_98intersection_with_product_map, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_intersection_with_product_ma, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__105)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__103);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_intersection_with_product_map, __pyx_t_3) < 0) __PYX_ERR(1, 702, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_intersection_with_product_map); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_intersection_with_product_map, __pyx_t_2) < 0) __PYX_ERR(1, 702, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":728
 *         return result, {tuple(k): v for k, v in c_product_map}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def concatenate(cls, Nfa lhs, Nfa rhs, use_epsilon: bool = False) -> Nfa:
 *         """Concatenate two NFAs.
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_use_epsilon, __pyx_n_s_bool) < 0) __PYX_ERR(1, 728, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_Nfa) < 0) __PYX_ERR(1, 728, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_100concatenate, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_concatenate, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__107)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__103);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_concatenate, __pyx_t_3) < 0) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_concatenate); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_concatenate, __pyx_t_2) < 0) __PYX_ERR(1, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":743
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def concatenate_with_result_state_maps(cls, Nfa lhs, Nfa rhs, use_epsilon: bool = False) \
 *             -> tuple[Nfa, dict[str, str], dict[str, str]]:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_use_epsilon, __pyx_n_s_bool) < 0) __PYX_ERR(1, 743, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_kp_s_tuple_Nfa_dict_str_str_dict_str) < 0) __PYX_ERR(1, 743, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_102concatenate_with_result_state_maps, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_concatenate_with_result_stat, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__103);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_concatenate_with_result_state_ma, __pyx_t_3) < 0) __PYX_ERR(1, 743, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_concatenate_with_result_state_ma); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_concatenate_with_result_state_ma, __pyx_t_2) < 0) __PYX_ERR(1, 743, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":768
 * 
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def complement(cls, Nfa lhs, alph.Alphabet alphabet, params = None):
 *         """Performs complement of lhs
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_104complement, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_complement, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__111)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__51);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_complement, __pyx_t_2) < 0) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_complement); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_complement, __pyx_t_3) < 0) __PYX_ERR(1, 768, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":792
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def make_complete(cls, Nfa lhs, State sink_state, alph.Alphabet alphabet):
 *         """Makes lhs complete
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_106make_complete, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_make_complete, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__113)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_make_complete, __pyx_t_3) < 0) __PYX_ERR(1, 792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_make_complete); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_make_complete, __pyx_t_2) < 0) __PYX_ERR(1, 792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":804
 *         mata_nfa.make_complete(lhs.thisptr.get(), <CAlphabet&>dereference(alphabet.as_base()), sink_state)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def revert(cls, Nfa lhs):
 *         """Reverses transitions in the lhs
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_108revert, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_revert, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__114)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_revert, __pyx_t_2) < 0) __PYX_ERR(1, 804, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_revert); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_revert, __pyx_t_3) < 0) __PYX_ERR(1, 804, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":816
 * 
 *     @classmethod
 *     def remove_epsilon(cls, Nfa lhs, Symbol epsilon = CEPSILON):             # <<<<<<<<<<<<<<
 *         """Removes transitions that contain epsilon symbol.
 * 
 */
  __pyx_k__11 = Mata::Nfa::EPSILON;

  /* "libmata/nfa/nfa.pyx":815
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def remove_epsilon(cls, Nfa lhs, Symbol epsilon = CEPSILON):
 *         """Removes transitions that contain epsilon symbol.
 */
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(Mata::Nfa::EPSILON); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_110remove_epsilon, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_remove_epsilon, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__116)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_remove_epsilon, __pyx_t_3) < 0) __PYX_ERR(1, 815, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_remove_epsilon); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_remove_epsilon, __pyx_t_2) < 0) __PYX_ERR(1, 815, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":827
 *         return result
 * 
 *     def remove_epsilon_inplace(self, Symbol epsilon = CEPSILON):             # <<<<<<<<<<<<<<
 *         """Removes transitions which contain epsilon symbol.
 * 
 */
  __pyx_k__12 = Mata::Nfa::EPSILON;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(Mata::Nfa::EPSILON); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_112remove_epsilon_inplace, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_remove_epsilon_inplace, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__118)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_remove_epsilon_inplace, __pyx_t_2) < 0) __PYX_ERR(1, 827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":836
 *         self.thisptr.get().remove_epsilon(epsilon)
 * 
 *     def get_epsilon_transitions(self, State state, Symbol epsilon = CEPSILON) -> Move | None:             # <<<<<<<<<<<<<<
 *         """Get epsilon transitions for a state.
 * 
 */
  __pyx_k__13 = Mata::Nfa::EPSILON;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(Mata::Nfa::EPSILON); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_kp_s_Move_None) < 0) __PYX_ERR(1, 836, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_114get_epsilon_transitions, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_get_epsilon_transitions, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__120)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_epsilon_transitions, __pyx_t_4) < 0) __PYX_ERR(1, 836, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":853
 * 
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def minimize(cls, Nfa lhs):
 *         """Minimizes the automaton lhs
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_116minimize, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_minimize, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__121)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_minimize, __pyx_t_4) < 0) __PYX_ERR(1, 853, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_minimize); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_minimize, __pyx_t_2) < 0) __PYX_ERR(1, 853, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":864
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def reduce_with_state_map(cls, Nfa aut, bool trim_input = True, params = None):
 *         """Reduce the automaton.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_118reduce_with_state_map, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_reduce_with_state_map, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__123)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__124);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_reduce_with_state_map, __pyx_t_2) < 0) __PYX_ERR(1, 864, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_reduce_with_state_map); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_reduce_with_state_map, __pyx_t_4) < 0) __PYX_ERR(1, 864, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":885
 *         return result, {k: v for k, v in state_map}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def reduce(cls, Nfa aut, bool trim_input = True, params = None):
 *         """Reduce the automaton.
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_120reduce, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_reduce, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__126)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__124);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_reduce_2, __pyx_t_4) < 0) __PYX_ERR(1, 885, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_reduce_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_reduce_2, __pyx_t_2) < 0) __PYX_ERR(1, 885, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":904
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def compute_relation(cls, Nfa lhs, params = None):
 *         """Computes the relation for the automaton
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_122compute_relation, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_compute_relation, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__128)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__51);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_compute_relation, __pyx_t_2) < 0) __PYX_ERR(1, 904, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_compute_relation); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_compute_relation, __pyx_t_4) < 0) __PYX_ERR(1, 904, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":930
 * 
 *     # Tests
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_deterministic(cls, Nfa lhs):
 *         """Tests if the lhs is determinstic
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_124is_deterministic, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_is_deterministic, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__130)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_deterministic, __pyx_t_4) < 0) __PYX_ERR(1, 930, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_is_deterministic); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_deterministic, __pyx_t_2) < 0) __PYX_ERR(1, 930, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":939
 *         return mata_nfa.is_deterministic(dereference(lhs.thisptr.get()))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_lang_empty(cls, Nfa lhs, Run run = None):
 *         """Checks if language of automaton lhs is empty, if not, returns path of states as counter
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_126is_lang_empty, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_is_lang_empty, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__132)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__51);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_lang_empty, __pyx_t_2) < 0) __PYX_ERR(1, 939, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_is_lang_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_lang_empty, __pyx_t_4) < 0) __PYX_ERR(1, 939, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":952
 *             return mata_nfa.is_lang_empty(dereference(lhs.thisptr.get()), NULL)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_universal(cls, Nfa lhs, alph.Alphabet alphabet, params = None):
 *         """Tests if lhs is universal wrt given alphabet
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_128is_universal, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_is_universal, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__134)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__51);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_universal, __pyx_t_4) < 0) __PYX_ERR(1, 952, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_is_universal); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_universal, __pyx_t_2) < 0) __PYX_ERR(1, 952, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":972
 *         )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_included_with_cex(
 *             cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_130is_included_with_cex, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_is_included_with_cex, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__136)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__137);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_included_with_cex, __pyx_t_2) < 0) __PYX_ERR(1, 972, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_is_included_with_cex); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_included_with_cex, __pyx_t_4) < 0) __PYX_ERR(1, 972, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":1001
 *         return result, run
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_included(
 *             cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_132is_included, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_is_included, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__139)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__137);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_included, __pyx_t_4) < 0) __PYX_ERR(1, 1001, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_is_included); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_included, __pyx_t_2) < 0) __PYX_ERR(1, 1001, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":1029
 *         return result
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def equivalence_check(cls, Nfa lhs, Nfa rhs, alph.Alphabet alphabet = None, params = None) -> bool:
 *         """Test equivalence of two automata.
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(1, 1029, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_134equivalence_check, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_equivalence_check, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__141)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__137);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_equivalence_check, __pyx_t_4) < 0) __PYX_ERR(1, 1029, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_equivalence_check); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_equivalence_check, __pyx_t_2) < 0) __PYX_ERR(1, 1029, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":1063
 *             )
 * 
 *     def get_symbols(self):             # <<<<<<<<<<<<<<
 *         """Return a set of symbols used on the transitions in NFA.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_136get_symbols, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_get_symbols, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__143)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_symbols, __pyx_t_2) < 0) __PYX_ERR(1, 1063, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":1071
 *         return {s for s in symbols}
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_complete(cls, Nfa lhs, alph.Alphabet alphabet):
 *         """Test if automaton is complete
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_138is_complete, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_is_complete, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__145)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_complete, __pyx_t_2) < 0) __PYX_ERR(1, 1071, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_is_complete); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_complete, __pyx_t_4) < 0) __PYX_ERR(1, 1071, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":1084
 *         )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_in_lang(cls, Nfa lhs, vector[Symbol] word):
 *         """Tests if word is in language
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_140is_in_lang, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_is_in_lang, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__147)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_in_lang, __pyx_t_4) < 0) __PYX_ERR(1, 1084, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_is_in_lang); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_in_lang, __pyx_t_2) < 0) __PYX_ERR(1, 1084, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":1096
 *         return mata_nfa.is_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def is_prefix_in_lang(cls, Nfa lhs, vector[Symbol] word):
 *         """Test if any prefix of the word is in the language
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_142is_prefix_in_lang, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_is_prefix_in_lang, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__148)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_prefix_in_lang, __pyx_t_2) < 0) __PYX_ERR(1, 1096, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_is_prefix_in_lang); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_is_prefix_in_lang, __pyx_t_4) < 0) __PYX_ERR(1, 1096, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":1108
 *         return mata_nfa.is_prfx_in_lang(dereference(lhs.thisptr.get()), dereference(run.thisptr))
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def accepts_epsilon(cls, Nfa lhs):
 *         """Tests if automaton accepts epsilon
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_144accepts_epsilon, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_accepts_epsilon, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__150)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_accepts_epsilon, __pyx_t_4) < 0) __PYX_ERR(1, 1108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_accepts_epsilon); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_accepts_epsilon, __pyx_t_2) < 0) __PYX_ERR(1, 1108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":1122
 *     # Helper functions
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def get_word_for_path(cls, Nfa lhs, path):
 *         """For a given path (set of states) returns a corresponding word
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_146get_word_for_path, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_get_word_for_path, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__152)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_word_for_path, __pyx_t_2) < 0) __PYX_ERR(1, 1122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_get_word_for_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_get_word_for_path, __pyx_t_4) < 0) __PYX_ERR(1, 1122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "libmata/nfa/nfa.pyx":1139
 *         return result.first.word, result.second
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def encode_word(cls, string_to_symbol, word):
 *         """Encodes word based on a string to symbol map
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_148encode_word, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa_encode_word, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__154)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_encode_word, __pyx_t_4) < 0) __PYX_ERR(1, 1139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa, __pyx_n_s_encode_word); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_7libmata_3nfa_3nfa_Nfa->tp_dict, __pyx_n_s_encode_word, __pyx_t_2) < 0) __PYX_ERR(1, 1139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_7libmata_3nfa_3nfa_Nfa);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_150__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa___reduce_cython, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__155)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3Nfa_152__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Nfa___setstate_cython, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__156)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":1175
 * 
 * # Temporary for testing
 * def divisible_by(k: int):             # <<<<<<<<<<<<<<
 *     """Constructs automaton accepting strings containing ones divisible by "k"."""
 *     assert k > 1
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_k, __pyx_n_s_int) < 0) __PYX_ERR(1, 1175, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_3divisible_by, 0, __pyx_n_s_divisible_by, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__158)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_divisible_by, __pyx_t_4) < 0) __PYX_ERR(1, 1175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "libmata/nfa/nfa.pyx":1189
 * 
 * 
 * def run_safely_external_command(cmd: str, check_results=True, quiet=True, timeout=None, **kwargs):             # <<<<<<<<<<<<<<
 *     """Safely runs the piped command, without executing of the shell
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_cmd, __pyx_n_s_str) < 0) __PYX_ERR(1, 1189, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_5run_safely_external_command, 0, __pyx_n_s_run_safely_external_command, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__160)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__161);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_run_safely_external_command, __pyx_t_2) < 0) __PYX_ERR(1, 1189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":1250
 * 
 * 
 * def get_elements_from_bool_vec(bool_vec: list[int]):             # <<<<<<<<<<<<<<
 *     cdef CBoolVector c_bool_vec = CBoolVector(bool_vec)
 *     cdef StateSet c_states
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_bool_vec, __pyx_kp_s_list_int) < 0) __PYX_ERR(1, 1250, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_7get_elements_from_bool_vec, 0, __pyx_n_s_get_elements_from_bool_vec, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__163)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_elements_from_bool_vec, __pyx_t_4) < 0) __PYX_ERR(1, 1250, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "libmata/nfa/nfa.pyx":1258
 * # On the fly configuration of the library
 * _store = {
 *     'node_style': {             # <<<<<<<<<<<<<<
 *         "style": "filled",
 *         "color": "darkblue",
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "libmata/nfa/nfa.pyx":1259
 * _store = {
 *     'node_style': {
 *         "style": "filled",             # <<<<<<<<<<<<<<
 *         "color": "darkblue",
 *         "fillcolor": "lightsteelblue",
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(9); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_style, __pyx_n_u_filled) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_color, __pyx_n_u_darkblue) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_fillcolor, __pyx_n_u_lightsteelblue) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_fontname, __pyx_kp_u_Courier_Bold) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_width, __pyx_kp_u_0_3) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_height, __pyx_kp_u_0_3) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_fontsize, __pyx_kp_u_12) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_fixedsize, __pyx_n_u_true) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_penwidth, __pyx_kp_u_1_5) < 0) __PYX_ERR(1, 1259, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_node_style, __pyx_t_2) < 0) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":1270
 *     },
 *     'edge_style': {
 *         "penwidth": "1.5",             # <<<<<<<<<<<<<<
 *         "color": "midnightblue",
 *     },
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_penwidth, __pyx_kp_u_1_5) < 0) __PYX_ERR(1, 1270, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_color, __pyx_n_u_midnightblue) < 0) __PYX_ERR(1, 1270, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_edge_style, __pyx_t_2) < 0) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "libmata/nfa/nfa.pyx":1273
 *         "color": "midnightblue",
 *     },
 *     'alphabet': None,             # <<<<<<<<<<<<<<
 * }
 * 
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_alphabet, Py_None) < 0) __PYX_ERR(1, 1258, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_store_2, __pyx_t_4) < 0) __PYX_ERR(1, 1257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "libmata/nfa/nfa.pyx":1277
 * 
 * 
 * def store():             # <<<<<<<<<<<<<<
 *     """
 *     Returns the configuration of the library
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7libmata_3nfa_3nfa_9store, 0, __pyx_n_s_store, NULL, __pyx_n_s_libmata_nfa_nfa, __pyx_d, ((PyObject *)__pyx_codeobj__164)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_store, __pyx_t_4) < 0) __PYX_ERR(1, 1277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "libmata/nfa/nfa.pyx":1
 * import shlex             # <<<<<<<<<<<<<<
 * import subprocess
 * import pandas
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_kp_u_Nfa_get_word_for_path_line_1122, __pyx_kp_u_For_a_given_path_set_of_states_r) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_kp_u_Nfa_encode_word_line_1139, __pyx_kp_u_Encodes_word_based_on_a_string_t) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init libmata.nfa.nfa", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init libmata.nfa.nfa");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]);
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
    Py_DECREF(argstuple);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
        if (__Pyx_IsCyOrPyCFunction(func))
#else
        if (PyCFunction_Check(func))
#endif
        {
            if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
                return __Pyx_PyObject_CallMethO(func, NULL);
            }
        }
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (PyCFunction_Check(func))
        {
            if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
                return __Pyx_PyObject_CallMethO(func, args[0]);
            }
        }
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    #if CYTHON_VECTORCALL
    vectorcallfunc f = _PyVectorcall_Function(func);
    if (f) {
        return f(func, args, (size_t)nargs, kwargs);
    }
    #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
    if (__Pyx_CyFunction_CheckExact(func)) {
        __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
        if (f) return f(func, args, (size_t)nargs, kwargs);
    }
    #endif
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg = NULL;
    return __Pyx_PyObject_FastCall(func, (&arg)+1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;  // error
            return kwvalues[i];
        }
    }
    return NULL;  // not found (no exception set)
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        if (unlikely(PyTuple_GET_SIZE(kw) == 0))
            return 1;
        if (!kw_allowed) {
            key = PyTuple_GET_ITEM(kw, 0);
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < PyTuple_GET_SIZE(kw); pos++) {
            key = PyTuple_GET_ITEM(kw, pos);
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        if (kwds_is_tuple) {
            if (pos >= PyTuple_GET_SIZE(kwds)) break;
            key = PyTuple_GET_ITEM(kwds, pos);
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    CYTHON_MAYBE_UNUSED_VAR(max_char);
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    CYTHON_UNUSED_VAR(value_count);
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* UnicodeConcatInPlace */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
static int
__Pyx_unicode_modifiable(PyObject *unicode)
{
    if (Py_REFCNT(unicode) != 1)
        return 0;
    if (!PyUnicode_CheckExact(unicode))
        return 0;
    if (PyUnicode_CHECK_INTERNED(unicode))
        return 0;
    return 1;
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    ) {
    PyObject *left = *p_left;
    Py_ssize_t left_len, right_len, new_len;
    if (unlikely(__Pyx_PyUnicode_READY(left) == -1))
        return NULL;
    if (unlikely(__Pyx_PyUnicode_READY(right) == -1))
        return NULL;
    left_len = PyUnicode_GET_LENGTH(left);
    if (left_len == 0) {
        Py_INCREF(right);
        return right;
    }
    right_len = PyUnicode_GET_LENGTH(right);
    if (right_len == 0) {
        Py_INCREF(left);
        return left;
    }
    if (unlikely(left_len > PY_SSIZE_T_MAX - right_len)) {
        PyErr_SetString(PyExc_OverflowError,
                        "strings are too large to concat");
        return NULL;
    }
    new_len = left_len + right_len;
    if (__Pyx_unicode_modifiable(left)
            && PyUnicode_CheckExact(right)
            && PyUnicode_KIND(right) <= PyUnicode_KIND(left)
            && !(PyUnicode_IS_ASCII(left) && !PyUnicode_IS_ASCII(right))) {
        __Pyx_GIVEREF(*p_left);
        if (unlikely(PyUnicode_Resize(p_left, new_len) != 0)) {
            __Pyx_GOTREF(*p_left);
            return NULL;
        }
        __Pyx_INCREF(*p_left);
        _PyUnicode_FastCopyCharacters(*p_left, left_len, right, 0, right_len);
        return *p_left;
    } else {
        return __Pyx_PyUnicode_Concat(left, right);
    }
  }
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = (long)((unsigned long)a - (unsigned long)b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return PyLong_FromLong(-intval);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (meth) {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* PyIntCompare */
static CYTHON_INLINE int __Pyx_PyInt_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        return (a == b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* MergeKeywords */
static int __Pyx_MergeKeywords(PyObject *kwdict, PyObject *source_mapping) {
    PyObject *iter, *key = NULL, *value = NULL;
    int source_is_dict, result;
    Py_ssize_t orig_length, ppos = 0;
    iter = __Pyx_dict_iterator(source_mapping, 0, __pyx_n_s_items, &orig_length, &source_is_dict);
    if (unlikely(!iter)) {
        PyObject *args;
        if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError))) goto bad;
        PyErr_Clear();
        args = PyTuple_Pack(1, source_mapping);
        if (likely(args)) {
            PyObject *fallback = PyObject_Call((PyObject*)&PyDict_Type, args, NULL);
            Py_DECREF(args);
            if (likely(fallback)) {
                iter = __Pyx_dict_iterator(fallback, 1, __pyx_n_s_items, &orig_length, &source_is_dict);
                Py_DECREF(fallback);
            }
        }
        if (unlikely(!iter)) goto bad;
    }
    while (1) {
        result = __Pyx_dict_iter_next(iter, orig_length, &ppos, &key, &value, NULL, source_is_dict);
        if (unlikely(result < 0)) goto bad;
        if (!result) break;
        if (unlikely(PyDict_Contains(kwdict, key))) {
            __Pyx_RaiseDoubleKeywordsError("function", key);
            result = -1;
        } else {
            result = PyDict_SetItem(kwdict, key, value);
        }
        Py_DECREF(key);
        Py_DECREF(value);
        if (unlikely(result < 0)) goto bad;
    }
    Py_XDECREF(iter);
    return 0;
bad:
    Py_XDECREF(iter);
    return -1;
}

/* PyIntCompare */
static CYTHON_INLINE int __Pyx_PyInt_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 0;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        return (a != b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) != 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 1;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 1;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal != 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        return ((double)a != (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n = PyTuple_GET_SIZE(bases);
    for (i = 1; i < n; i++)
    {
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
            return -1;
        }
        if (dictoffset == 0 && b->tp_dictoffset)
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "extension type '%.200s' has no __dict__ slot, "
                "but base type '" __Pyx_FMT_TYPENAME "' has: "
                "either add 'cdef dict __dict__' to the extension type "
                "or add '__slots__ = [...]' to the base type",
                type_name, b_name);
            __Pyx_DECREF_TypeName(b_name);
            return -1;
        }
    }
    return 0;
}
#endif

/* PyType_Ready */
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetupReduce */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_0_0
#define __PYX_HAVE_RT_ImportType_3_0_0
static PyTypeObject *__Pyx_ImportType_3_0_0(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_0 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_0 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_0 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                #if CYTHON_COMPILING_IN_LIMITED_API
                module = PyImport_ImportModuleLevelObject(
                    name, empty_dict, empty_dict, from_list, 1);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                #endif
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            #if CYTHON_COMPILING_IN_LIMITED_API
            module = PyImport_ImportModuleLevelObject(
                name, empty_dict, empty_dict, from_list, level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportDottedModule */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__16;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    PyObject *abi_module = PyImport_AddModule((char*) __PYX_ABI_MODULE_NAME);
    if (unlikely(!abi_module)) return NULL;
    Py_INCREF(abi_module);
    return abi_module;
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
        PyList_SET_ITEM(fromlist, 0, marker);
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    cf->m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#ifdef _Py_TPFLAGS_HAVE_VECTORCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* ClassMethod */
static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    #if PY_MAJOR_VERSION == 2
    static PyTypeObject *methoddescr_type = NULL;
    if (unlikely(methoddescr_type == NULL)) {
       PyObject *meth = PyObject_GetAttrString((PyObject*)&PyList_Type, "append");
       if (unlikely(!meth)) return NULL;
       methoddescr_type = Py_TYPE(meth);
       Py_DECREF(meth);
    }
    #else
    PyTypeObject *methoddescr_type = &PyMethodDescr_Type;
    #endif
    if (__Pyx_TypeCheck(method, methoddescr_type))
#endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        #if PY_VERSION_HEX < 0x03020000
        PyTypeObject *d_type = descr->d_type;
        #else
        PyTypeObject *d_type = descr->d_common.d_type;
        #endif
        return PyDescr_NewClassMethod(d_type, descr->d_method);
    }
#endif
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else {
        return PyClassMethod_New(method);
    }
}

/* GetNameInClass */
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    PyObject *dict;
    assert(PyType_Check(nmspace));
#if CYTHON_USE_TYPE_SLOTS
    dict = ((PyTypeObject*)nmspace)->tp_dict;
    Py_XINCREF(dict);
#else
    dict = PyObject_GetAttr(nmspace, __pyx_n_s_dict);
#endif
    if (likely(dict)) {
        result = PyObject_GetItem(dict, name);
        Py_DECREF(dict);
        if (result) {
            return result;
        }
    }
    PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    _PyTraceback_Add(funcname, filename, py_line);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(size_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                                return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                                return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                                return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(size_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(size_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (size_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (size_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (size_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (size_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (size_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((size_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((size_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(unsigned long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(unsigned long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(unsigned long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(unsigned long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(unsigned long) >= 2 * PyLong_SHIFT)) {
                                return (unsigned long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(unsigned long) >= 3 * PyLong_SHIFT)) {
                                return (unsigned long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(unsigned long) >= 4 * PyLong_SHIFT)) {
                                return (unsigned long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(unsigned long) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(unsigned long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(unsigned long) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                                return (unsigned long) (((unsigned long)-1)*(((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(unsigned long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                                return (unsigned long) ((((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                                return (unsigned long) (((unsigned long)-1)*(((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                                return (unsigned long) ((((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT)) {
                                return (unsigned long) (((unsigned long)-1)*(((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT)) {
                                return (unsigned long) ((((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(unsigned long) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(unsigned long) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            unsigned long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (unsigned long) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (unsigned long) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (unsigned long) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (unsigned long) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (unsigned long) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(unsigned long) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((unsigned long) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(unsigned long) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((unsigned long) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((unsigned long) 1) << (sizeof(unsigned long) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (unsigned long) -1;
        }
    } else {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned long) -1;
        val = __Pyx_PyInt_As_unsigned_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned long");
    return (unsigned long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned long");
    return (unsigned long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint8_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint8_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT)) {
                                return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT)) {
                                return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT)) {
                                return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint8_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(uint8_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(uint8_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint8_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            uint8_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (uint8_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (uint8_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (uint8_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (uint8_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (uint8_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(uint8_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((uint8_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(uint8_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((uint8_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((uint8_t) 1) << (sizeof(uint8_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (uint8_t) -1;
        }
    } else {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyInt_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XSETREF(name, __Pyx_NewRef(__pyx_n_s__165));
    }
    return name;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                                return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                                return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                                return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(long) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                                return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                                return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                                return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(long) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (long) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (long) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (long) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (long) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (long) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((long) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((long) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                                return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                                return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                                return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(int) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                                return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                                return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                                return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(int) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (int) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (int) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (int) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (int) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (int) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((int) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((int) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_GeneratorType_slots
};
#else
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Generator_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_GeneratorType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
#endif
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compile time version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
